/*
 * @(#) ClassFile.java	1.5 00/04/19
 *
 * Copyright (c) 1997-1999 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.compiler;

import java.io.IOException;
import java.io.OutputStream;
import java.io.DataOutputStream;
import java.util.Vector;
import java.util.Enumeration;

/**
 * This class provides a way of making Java class file image.
 */
public class ClassFile implements Constants {

    private final static long MAGIC = 0xcafebabe0003002dL;
    private final static boolean DEBUG = false;

    private ConstantPool constantPool;
    private Vector methods = new Vector(); // MethodInfo
    private MethodInfo currentMethod;
    private Vector fields = new Vector(); // FieldInfo
    private Vector exceptionTable;
    private Vector interfaces = new Vector();
    private short thisClassIndex;
    private short superClassIndex;
    private short sourceFileNameIndex;
    private short accessFlags;
    private short maxStack;
    private short stackTop;
    private boolean locals[];
    public ClassFile parent;
    private String className;
    private int maxLocal;

    ByteBuffer codeBuffer;

    public ClassFile(String thisClass,
		     String superClass,
		     String sourceFile,
		     short accessFlags){

	constantPool = new ConstantPool();
        thisClassIndex = constantPool.addClass(thisClass);
        superClassIndex = constantPool.addClass(superClass);
        if (sourceFile != null){
            sourceFileNameIndex = constantPool.addUTF8(sourceFile);
	}
        this.accessFlags = accessFlags;
	this.className = thisClass;
    }

    public String getClassName(){
	return className;
    }

    public int codeSize(){
	return codeBuffer.size();
    }

    public void addInterface(String interfaceName) {
        interfaces.addElement(new Short(constantPool.addClass(interfaceName)));
    }

    public void addField(String fieldName, String type, short accessFlags) {
        fields.addElement(new FieldInfo(constantPool.addUTF8(fieldName),
					constantPool.addUTF8(type),
					accessFlags));
    }

    public short addConstant(String value){
	return constantPool.addConstant(value);
    }

    public void addConstant(String fieldName, String type, short flags, short valueIndex){
        short fieldNameIndex = constantPool.addUTF8(fieldName);
        short typeIndex = constantPool.addUTF8(type);
        short cvAttr[] = new short[4];
        cvAttr[0] = constantPool.addUTF8("ConstantValue");
        cvAttr[1] = 0;
        cvAttr[2] = 2;
        cvAttr[3] = valueIndex;
        fields.addElement(new FieldInfo(fieldNameIndex, typeIndex, flags, cvAttr));
    }

    public void addConstant(String fieldName, String type, short flags, int value){
	addConstant(fieldName, type, flags, constantPool.addConstant(value));
    }

    public void addConstant(String fieldName, String type, short flags, long value){
	addConstant(fieldName, type, flags, constantPool.addConstant(value));
    }

    public void addConstant(String fieldName, String type, short flags, double value){
	addConstant(fieldName, type, flags, constantPool.addConstant(value));
    }

    public void addConstant(String fieldName, String type, short flags, String value){
	addConstant(fieldName, type, flags, constantPool.addConstant(value));
    }

    public void openMethod(String methodName, String type, short flag){
	codeBuffer = new ByteBuffer();
	int sig = sizeOfParameters(type);
	int nlocals = sig & 0xff;
	if ((flag & ACC_STATIC) != ACC_STATIC){
	    nlocals++;
	}

	locals = new boolean[nlocals + 32];
	for (int i = 0; i < nlocals; i++){
	    this.locals[i] = true;
	}
	maxLocal = nlocals;

        currentMethod = new MethodInfo(constantPool.addUTF8(methodName),
				       constantPool.addUTF8(type),
				       flag);
        methods.addElement(currentMethod);
    }

    public void closeMethod(){
	int codeSize = codeSize();
	if (DEBUG){
	    System.out.println("codeSize = " + codeSize);
	}
	if (codeSize > 65535){
	    throw new ClassFormatError("code size is too large " + codeSize);
	}
	int exceptionSize = 0;
	if (exceptionTable != null){
	    exceptionSize = exceptionTable.size();
	}

        int attrLength = 2 +	// attribute_name_index
	    4 +    // attribute_length
	    2 +	// max_stack
	    2 +	// max_locals
	    4 +	// code_length
	    codeSize +
	    2 +	// exception_table_length
	    (exceptionSize * 8) +
	    2;	// attributes_count

	ByteBuffer codeAttr = new ByteBuffer(attrLength);
        int codeAttrIndex = constantPool.addUTF8("Code");
	codeAttr.add((short)codeAttrIndex);

	int attr_len = attrLength - 6;
	codeAttr.add(attr_len);
	codeAttr.add((short)maxStack);
	codeAttr.add((short)maxLocal);
	codeAttr.add(codeSize);
	codeAttr.append(codeBuffer);

        if (exceptionSize > 0) {
	    codeAttr.add((short)exceptionTable.size());
	    for (Enumeration e = exceptionTable.elements(); e.hasMoreElements(); ){
		ExceptionTableEntry entry = (ExceptionTableEntry)e.nextElement();
		codeAttr.add(entry.start.getPosition());
		codeAttr.add(entry.end.getPosition());
		codeAttr.add(entry.handler.getPosition());
		codeAttr.add(entry.catchType);
	    }
	} else {
	    codeAttr.add((short)0);
	}
	codeAttr.add((short)0);
        
        currentMethod.setCodeAttribute(codeAttr);

        exceptionTable = null;
        codeBuffer.setSize(0);
        currentMethod = null;
        maxStack = 0;
        stackTop = 0;
    }

    public void add(byte opcode){
	codeBuffer.add(opcode);
	stackTop += Opcode.stackGrowth[opcode & 0xff];
        if (stackTop > maxStack){
	    maxStack = stackTop;
	}
    }

    public Label getLabel(){
	return getLabel(false);
    }

    public Label getLabel(boolean fixed){
	Label label = new Label(this);
	if (fixed){
	    label.fix();
	}
	return label;
    }

    public int declareLocal(){
	if (maxLocal >= locals.length) {
	    boolean[] new_locals = new boolean[locals.length * 2];
	    System.arraycopy(locals, 0, new_locals, 0, locals.length);
	    locals = new_locals;
	}
	locals[maxLocal] = true;
	return maxLocal++;
    }

    public int getLocal(){
	for (int i = 0; i < maxLocal; i++){
	    if (!locals[i]){
		locals[i] = true;
		return i;
	    }
	}
	if (maxLocal >= locals.length) {
	    boolean[] new_locals = new boolean[locals.length * 2];
	    System.arraycopy(locals, 0, new_locals, 0, locals.length);
	    locals = new_locals;
	}
	locals[maxLocal] = true;
	return maxLocal++;
    }

    public void freeLocal(int index){
	locals[index] = false;
    }

    public void istoreLocal(int index){
	if (index >= 0 && index < 4){
	    add((byte)(Opcode.ISTORE_0 + index));
	} else {
	    add(Opcode.ISTORE, index);
	}
    }

    public void iloadLocal(int index){
	if (index >= 0 && index < 4) {
	    add((byte)(Opcode.ILOAD_0 + index));
	} else {
	    add(Opcode.ILOAD, index);
	}
    }
    
    public void lloadLocal(int index){
	if (index >= 0 && index < 4) {
	    add((byte)(Opcode.LLOAD_0 + index));
	} else {
	    add(Opcode.LLOAD, index);
	}
    }
    
    public void floadLocal(int index){
	if (index >= 0 && index < 4) {
	    add((byte)(Opcode.FLOAD_0 + index));
	} else {
	    add(Opcode.FLOAD, index);
	}
    }
    
    public void dloadLocal(int index){
	if (index >= 0 && index < 4) {
	    add((byte)(Opcode.DLOAD_0 + index));
	} else {
	    add(Opcode.DLOAD, index);
	}
    }
    
    public void loadLocal(int index){
	if (index >= 0 && index < 4) {
	    add((byte)(Opcode.ALOAD_0 + index));
	} else {
	    add(Opcode.ALOAD, index);
	}
    }
    
    public void storeLocal(int index){
	if (index >= 0 && index < 4){
	    add((byte)(Opcode.ASTORE_0 + index));
	} else {
	    add(Opcode.ASTORE, index);
	}
    }

    public void add(byte opcode, Label label) {
        switch (opcode) {
	case Opcode.GOTO: // fall-through
	case Opcode.IFEQ:
	case Opcode.IFNE:
	case Opcode.IFLT:
	case Opcode.IFGE:
	case Opcode.IFGT:
	case Opcode.IFLE:
	case Opcode.IF_ICMPEQ:
	case Opcode.IF_ICMPNE:
	case Opcode.IF_ICMPLT:
	case Opcode.IF_ICMPGE:
	case Opcode.IF_ICMPGT:
	case Opcode.IF_ICMPLE:
	case Opcode.IF_ACMPEQ:
	case Opcode.IF_ACMPNE:
	case Opcode.JSR:
	case Opcode.IFNULL:
	case Opcode.IFNONNULL:
	    int pc = codeBuffer.size();
	    codeBuffer.add(opcode);
	    label.register(pc, 2);
	    stackTop += Opcode.stackGrowth[opcode & 0xff];
	    break;
	}
    }

    void pushInteger(int number){
	if (number >= -1 && number <= 5){
	    add((byte)(Opcode.ICONST_0 + number));
	} else if (number > -0x80 && number < 0x80){
	    add(Opcode.BIPUSH, number);
	} else if (number > -0x8000 && number < 0x8000){
	    add(Opcode.SIPUSH, number);
	} else {
	    add(Opcode.LDC, constantPool.addConstant(number));
	}
    }

    void pushLong(long number){
	if (number == 0){
	    add(Opcode.LCONST_0);
	} else if (number == 1){
	    add(Opcode.LCONST_1);
	} else {
	    add(Opcode.LDC2_W, constantPool.addConstant(number));
	}
    }

    void pushFloat(float number){
	if (number == 0.0f){
	    add(Opcode.FCONST_0);
	} else if (number == 1.0f){
	    add(Opcode.FCONST_1);
	} else if (number == 2.0f){
	    add(Opcode.FCONST_2);
	} else {
	    add(Opcode.LDC_W, constantPool.addConstant(number));
	}
    }

    void pushDouble(double number){
	if (number == 0){
	    add(Opcode.DCONST_0);
	} else if (number == 1){
	    add(Opcode.DCONST_1);
	} else {
	    add(Opcode.LDC2_W, constantPool.addConstant(number));
	}
    }

    public void add(byte opcode, int operand) {
        stackTop += Opcode.stackGrowth[opcode & 0xFF];

        if (stackTop > maxStack){
	    maxStack = stackTop;
	}

        switch (opcode) {
	case Opcode.BIPUSH :
	    codeBuffer.add(opcode);
	    codeBuffer.add((byte)operand);
	    break;
	case Opcode.SIPUSH:
	    codeBuffer.add(opcode);
	    codeBuffer.add((short)operand);
	    break;
	case Opcode.NEWARRAY:
	    codeBuffer.add(opcode);
	    codeBuffer.add((byte)operand);
	    break;
	case Opcode.GETFIELD:
	case Opcode.PUTFIELD:
	    codeBuffer.add(opcode);
	    codeBuffer.add(operand);
	    break;
	case Opcode.LDC:
	case Opcode.LDC_W:
	case Opcode.LDC2_W:
	    if ((operand >= 256) ||
		(opcode == Opcode.LDC_W) ||
		(opcode == Opcode.LDC2_W)){

		if (opcode == Opcode.LDC){
		    codeBuffer.add(Opcode.LDC_W);
		} else {
		    codeBuffer.add(opcode);
		}
		codeBuffer.add((short)operand);
	    } else {
		codeBuffer.add(opcode);
		codeBuffer.add((byte)operand);
	    }
	    break;
	case Opcode.RET:
	case Opcode.ILOAD:
	case Opcode.LLOAD:
	case Opcode.FLOAD:
	case Opcode.DLOAD:
	case Opcode.ALOAD:
	case Opcode.ISTORE:
	case Opcode.LSTORE:
	case Opcode.FSTORE:
	case Opcode.DSTORE:
	case Opcode.ASTORE:
	    if (operand >= 256) {
		codeBuffer.add(Opcode.WIDE);
		codeBuffer.add(opcode);
		codeBuffer.add(operand);
	    } else {
		codeBuffer.add(opcode);
		codeBuffer.add((byte)operand);
	    }
	    break;
	default:
	    throw new ClassFileException("Unexpected opcode for 1 operand");
        }
    }

    public void add(byte opcode, int operand1, int operand2) {
        stackTop += Opcode.stackGrowth[opcode & 0xFF];

        if (stackTop > maxStack){
	    maxStack = stackTop;
	}
        if (opcode == Opcode.IINC) {
            if ((operand1 > 255) || (operand2 < -128) || (operand2 > 127)) {
                codeBuffer.add(Opcode.WIDE);
                codeBuffer.add(Opcode.IINC);
                codeBuffer.add((short)operand1);
                codeBuffer.add((short)operand2);
            } else {
                codeBuffer.add(Opcode.IINC);
                codeBuffer.add((byte)operand1);
                codeBuffer.add((byte)operand2);
            }
        } else {
            if (opcode == Opcode.MULTIANEWARRAY) {
                codeBuffer.add(Opcode.MULTIANEWARRAY);
                codeBuffer.add((short)operand1);
                codeBuffer.add((byte)operand2);
            } else {
                throw new ClassFileException("Unexpected opcode for 2 operands");
            }
        }
    }

    public void add(byte opcode, String className) {
        stackTop += Opcode.stackGrowth[opcode & 0xFF];

        switch (opcode) {
	case Opcode.NEW:
	case Opcode.ANEWARRAY:
	case Opcode.CHECKCAST:
	case Opcode.INSTANCEOF:
	    short classIndex = constantPool.addClass(className);
	    codeBuffer.add(opcode);
	    codeBuffer.add(classIndex);
	    break;
	default:
	    throw new ClassFileException("bad opcode for class reference:" + opcode);
        }
        if (stackTop > maxStack){
	    maxStack = stackTop;
	}
    }
    

    public void add(byte opcode, String className, String fieldName, String fieldType){
        stackTop += Opcode.stackGrowth[opcode & 0xFF];

        char fieldTypeChar = fieldType.charAt(0);
        int fieldSize = ((fieldTypeChar == 'J') || (fieldTypeChar == 'D')) ? 2 : 1;

        switch (opcode) {
	case Opcode.GETFIELD :
	case Opcode.GETSTATIC :
	    stackTop += fieldSize;
	    break;
	case Opcode.PUTSTATIC :
	case Opcode.PUTFIELD :
	    stackTop -= fieldSize;
	    break;
	default :
	    throw new ClassFileException("bad opcode for field reference:" + opcode);
        }
        short fieldRefIndex = constantPool.addFieldRef(className, fieldName, fieldType);
        codeBuffer.add(opcode);
        codeBuffer.add(fieldRefIndex);

        if (stackTop > maxStack){
	    maxStack = stackTop;
	}
    }
    
    public void add(byte opcode,
		    String className,
		    String methodName,
                    String parametersType,
		    String returnType) {
	int info = sizeOfParameters(parametersType);

        stackTop -= info & 0xffff;
        stackTop += Opcode.stackGrowth[opcode & 0xFF];

        if (stackTop > maxStack){
	    maxStack = stackTop;
	}

        switch (opcode) {
	case Opcode.INVOKEVIRTUAL:
	case Opcode.INVOKESPECIAL:
	case Opcode.INVOKESTATIC:
	case Opcode.INVOKEINTERFACE:
	    stackTop += sizeOfReturn(returnType);
	    codeBuffer.add(opcode);
	    if (opcode == Opcode.INVOKEINTERFACE) {
		int nargs = (info >> 16) & 0xffff;
		short ifMethodRefIndex =
		    constantPool.addInterfaceMethodRef(className, methodName, parametersType + returnType);
		codeBuffer.add(ifMethodRefIndex);
		codeBuffer.add((byte)(nargs + 1));
		codeBuffer.add((byte)0);
	    } else {
		short methodRefIndex =
		    constantPool.addMethodRef(className, methodName, parametersType + returnType);
		codeBuffer.add(methodRefIndex);
	    }
	    break;
	default :
	    throw new ClassFileException("bad opcode for method reference:" + opcode);
        }

        if (stackTop > maxStack){
	    maxStack = stackTop;
	}
    }

    public void push(){
	stackTop++;
        if (stackTop > maxStack){
	    maxStack = stackTop;
	}
    }

    public static final int sizeOfReturn(String sig){
	int size = 0;
	char c = sig.charAt(0);
	if (c != 'V'){
	    if ((c == 'J') || (c == 'D')){
		size += 2;
	    } else {
		size++;
	    }
	}
	return size;
    }

    public static final int sizeOfParameters(String sig){
	return sizeOfParameters(sig.toCharArray(), 1);
    }

    static int sizeOfParameters(char c[], int offset){
        int index = offset;
        int size = 0;
	int narg = 0;
    loop:
        while (index < c.length){
            switch (c[index]) {
	    case 'J':
	    case 'D':
		size += 2;
		index++;
		narg++;
		break;
	    case 'B':
	    case 'S':
	    case 'C':
	    case 'I':
	    case 'Z':
	    case 'F':
		size++;
		index++;
		narg++;
		break;
	    case '[':
		while (c[index] == '['){
		    index++;
		}
		if (c[index] != 'L') {
		    size++;
		    index++;
		    narg++;
		    break;
		}	// fall through
	    case 'L':
		size++;
		narg++;
		while (c[index++] != ';') {}
		break;
	    case ')':
		break loop;
	    default:
		throw new ClassFileException("Bad signature character:" + c[index]);
            }
        }
        return (narg << 16) | size;
    }

    public void addExceptionHandler(Label startLabel,
				    Label endLabel,
				    Label handlerLabel,
				    String catchClassName){
	short catch_type;
	if (catchClassName != null){
	    catch_type = constantPool.addClass(catchClassName);
	} else {
	    catch_type = 0;
	}
        ExceptionTableEntry newEntry =
	    new ExceptionTableEntry(startLabel, endLabel, handlerLabel, catch_type);

        if (exceptionTable == null) {
	    exceptionTable = new Vector();
	}
	exceptionTable.addElement(newEntry);
    }

    public void write(OutputStream stream) throws IOException {
	DataOutputStream out;
	if (stream instanceof DataOutputStream){
	    out = (DataOutputStream)stream;
	} else {
	    out = new DataOutputStream(stream);
	}
        short sourceFileIndex = 0;
        if (sourceFileNameIndex != 0){
            sourceFileIndex = constantPool.addUTF8("SourceFile");
	}
        out.writeLong(MAGIC);
        constantPool.write(out);
        out.writeShort(this.accessFlags);
        out.writeShort(thisClassIndex);
        out.writeShort(superClassIndex);
        out.writeShort(interfaces.size());
        for (int i = 0; i < interfaces.size(); i++) {
            out.writeShort(((Short)interfaces.elementAt(i)).shortValue());
        }
        out.writeShort(fields.size());
        for (int i = 0; i < fields.size(); i++) {
            ((FieldInfo)fields.elementAt(i)).write(out);
        }
        out.writeShort(methods.size());
        for (int i = 0; i < methods.size(); i++) {
            ((MethodInfo)methods.elementAt(i)).write(out);
        }
        if (sourceFileNameIndex != 0) {
            out.writeShort(1);      // attributes count
            out.writeShort(sourceFileIndex);
            out.writeInt(2);
            out.writeShort(sourceFileNameIndex);
        } else {
            out.writeShort(0);      // no attributes
	}
    }

    public String toString(){
	return getClass().getName() + "[" + className + "]";
    }


    static class MethodInfo {

	private short nameIndex;
	private short descriptorIndex;
	private short accessFlags;
	private ByteBuffer codeAttribute;

	MethodInfo(short nameIndex, short descriptorIndex, short accessFlags) {
	    this.nameIndex = nameIndex;
	    this.descriptorIndex = descriptorIndex;
	    this.accessFlags = accessFlags;
	}
	
	ByteBuffer getCodeAttribute(){
	    return this.codeAttribute;
	}

	void setCodeAttribute(ByteBuffer codeAttribute){
	    this.codeAttribute = codeAttribute;
	}
	
	void write(DataOutputStream out) throws IOException {
	    out.writeShort(accessFlags);
	    out.writeShort(nameIndex);
	    out.writeShort(descriptorIndex);
	    out.writeShort(1);  // attributes_count
	    codeAttribute.writeTo(out);
	}
    }

    static class FieldInfo {

	private short nameIndex;
	private short descriptorIndex;
	private short accessFlags;
	private short attributes[];

	FieldInfo(short nameIndex, short descriptorIndex, short accessFlags){
	    this.nameIndex = nameIndex;
	    this.descriptorIndex = descriptorIndex;
	    this.accessFlags = accessFlags;
	}

	FieldInfo(short nameIndex, short descriptorIndex, short accessFlags, short attributes[]){
	    this.nameIndex = nameIndex;
	    this.descriptorIndex = descriptorIndex;
	    this.accessFlags = accessFlags;
	    this.attributes = attributes;
	}
	
	void write(DataOutputStream out) throws IOException {
	    out.writeShort(accessFlags);
	    out.writeShort(nameIndex);
	    out.writeShort(descriptorIndex);
	    if (attributes == null){
		out.writeShort(0);              // no attributes
	    } else {
		out.writeShort(1);
		out.writeShort(attributes[0]);
		out.writeShort(attributes[1]);
		out.writeShort(attributes[2]);
		out.writeShort(attributes[3]);
	    }
	}
    }

    static class ExceptionTableEntry {
    
	Label start;
	Label end;
	Label handler;
	short catchType;

	ExceptionTableEntry(Label start, Label end, Label handler, short catchType){
	    this.start = start;
	    this.end = end;
	    this.handler = handler;
	    this.catchType = catchType;
	}
    }
}
