/*
 * @(#) CompileContext.java	1.6 00/07/26
 *
 * Copyright (c) 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.compiler;

import pnuts.lang.Package;
import pnuts.lang.Context;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.io.IOException;
import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.FileOutputStream;

/**
 * This class is used with pnuts.compiler.Compiler class to compile
 * Pnuts scripts.
 *
 * @see pnuts.compiler.Compiler
 * @see pnuts.lang.Context
 */
public class CompileContext extends Context {

    private final static boolean DEBUG = false;

    Frame env = new Frame();
    Symbol sym = new Symbol();
    Hashtable constants = new Hashtable();
    ClassFile cf;
    Vector classFiles = new Vector();
    String constClassName;
    Vector classes = new Vector();

    Class loadClasses(CodeLoader loader) throws IOException {
	ByteArrayOutputStream bout = new ByteArrayOutputStream();
	Class ret = load(cf, loader, bout);

	for (Enumeration e = classFiles.elements(); e.hasMoreElements(); ){
	    load((ClassFile)e.nextElement(), loader, bout);
	}
	resolve(loader);
	return ret;
    }

    void resolve(CodeLoader loader){
	for (Enumeration e = classes.elements(); e.hasMoreElements(); ){
	    Class c = (Class)e.nextElement();
	    loader.resolve(c);
	}
    }

    Class load(ClassFile file, CodeLoader loader) throws IOException {
	return load(file, loader, new ByteArrayOutputStream());
    }

    Class load(ClassFile file, CodeLoader loader, ByteArrayOutputStream bout)
           throws IOException
    {
	bout.reset();
	file.write(new DataOutputStream(bout));

	byte array[] = bout.toByteArray();
	Class ret = loader.define(file.getClassName(), array, 0, array.length);
	classes.addElement(ret);
	return ret;
    }

    /**
     * Get the primary class file
     */
    public ClassFile getClassFile(){
	return cf;
    }

    /**
     * Enumerate related class files.
     */
    public Enumeration getClassFiles(){
	return classFiles.elements();
    }

    public void write(DataOutputStream out) throws IOException {
	cf.write(out);
    }

    void debug(ClassFile file){
	try {
	    String fileName = "/tmp/" + file.getClassName() + ".class";
	    System.out.println(fileName);
	    FileOutputStream fout = new FileOutputStream(fileName);
	    DataOutputStream dout = new DataOutputStream(fout);
	    file.write(dout);
	    fout.close();
	} catch (IOException e){
	    e.printStackTrace();
	}
    }

    void debug(){
	debug(cf);
	for (Enumeration e = classFiles.elements(); e.hasMoreElements(); ){
	    ClassFile file = (ClassFile)e.nextElement();
	    debug(file);
	}
    }

    void _openFrame(String func, String locals[]){
	env = new Frame(locals, func, env);
    }

    void _closeFrame(){
	env = env.parent;
    }

    void openScope(String locals[]){
	env.openLocal();
	for (int i = 0; i < locals.length; i++){
	    _declare(locals[i]);
	}
    }

    void closeScope(){
	env.closeLocal();
    }

    void setReference(String symbol){
	String sym = symbol.intern();
	Frame f = env;
	f.setReference(sym);
    }

    Reference getReference(String symbol){
	String sym = symbol.intern();
	Frame f = env;
	Reference ref = f.getReference(sym);
	if (ref != null){
	    return ref;
	}
	f = f.parent;

	while (f != null){
	    Reference r = f.getReference(sym);
	    if (r != null){
		Frame f0 = env;
		Frame f1 = null;
		while (f0 != f){
		    f1 = f0;
		    if (!f0.imports.contains(sym)){
			if (DEBUG){
			    System.out.println(f0.fname + " imports " + sym);
			}
			f0.imports.addElement(sym);
		    }
		    f0 = f0.parent;
		}
		Reference ret = new Reference(sym, -1, ((r.offset >= 0) ? r.offset : 0));
		Reference _ref = f.getReference(sym);
		Vector vec = (Vector)f.exports.get(f1);
		if (vec == null){
		    vec = new Vector();
		    f.exports.put(f1, vec);
		}
		if (!vec.contains(_ref)){
		    vec.addElement(_ref);
		    if (DEBUG){
			System.out.println(f.fname + " exports " + _ref + " to " + f1.fname);
		    }
		}
		f0 = env;
		while (f0 != f1){
		    Frame p = f0.parent;
		    vec = (Vector)p.exports.get(f0);
		    if (vec == null){
			vec = new Vector();
			p.exports.put(f0, vec);
		    }
		    if (!vec.contains(ret)){
			vec.addElement(ret);
			if (DEBUG){
			    System.out.println(p.fname + " exports " + ret + " to " + f0.fname);
			}
		    }
		    f0 = p;
		}
		return ret;
	    }
	    f = f.parent;
	}
	return null;
    }

    int _declare(String symbol){
	int local = cf.getLocal();
	_declare(symbol, local);
	return local;
    }

    void _declare(String symbol, int local){
	_declare(symbol, local, 0);
    }

    void _declare(String symbol, int local, int idx){
	if (DEBUG){
	    System.out.println("_declare " + symbol + " " + local + " " + idx);
	}
	env._declare(symbol, local, idx);
    }

    int declare(String symbol){
	int local = cf.declareLocal();
	declare(symbol, local);
	return local;
    }

    void declare(String symbol, int local){
	declare(symbol, local, 0);
    }

    void declare(String symbol, int local, int idx){
	if (DEBUG){
	    System.out.println("declare " + symbol + " " + local + " " + idx);
	}
	env.declare(symbol, local, idx);
    }

    int contextIndex = 0;

    int getContextIndex(){
	return contextIndex;
    }

    void setContextIndex(int index){
	contextIndex = index;
    }


    Stack continueLabelStack = new Stack();
    Stack breakLabelStack = new Stack();

    Label getContinueLabel(){
	if (continueLabelStack.size() < 1){
	    return null;
	} else {
	    return (Label)continueLabelStack.peek();
	}
    }

    void pushContinueLabel(Label label){
	continueLabelStack.push(label);
    }

    Label popContinueLabel(){
	return (Label)continueLabelStack.pop();
    }

    Label getBreakLabel(){
	if (breakLabelStack.size() < 1){
	    return null;
	} else {
	    return (Label)breakLabelStack.peek();
	}
    }

    void pushBreakLabel(Label label){
	breakLabelStack.push(label);
    }

    Label popBreakLabel(){
	return (Label)breakLabelStack.pop();
    }

    public Object clone() {
	try {
	    return super.clone();
	} catch (Throwable t){
	    throw new InternalError();
	}
    }
}
