/*
 * @(#) Compiler.java	1.13 00/08/07
 *
 * Copyright (c) 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.compiler;

import pnuts.lang.Arithmetic;
import pnuts.lang.Runtime;
import pnuts.lang.Package;
import pnuts.lang.Pnuts;
import pnuts.lang.PnutsFunction;
import pnuts.lang.PnutsInterpreter;
import pnuts.lang.Function;
import pnuts.lang.Context;
import pnuts.lang.Visitor;
import pnuts.lang.SimpleNode;
import pnuts.lang.Jump;
import pnuts.lang.PnutsException;
import pnuts.lang.ParseException;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Vector;
import java.util.Enumeration;
import java.math.BigDecimal;
import java.math.BigInteger;
import pnuts.ext.UnparseVisitor;
import java.security.AccessController;
import java.security.PrivilegedAction;

/**
 * Pnuts to JVM bytecode compiler
 */
public class Compiler extends Runtime implements Visitor, Opcode, Constants {

    private final static boolean DEBUG = false;
    private final static boolean PROFILE = false;
    private long s;

    private static PnutsInterpreter interpreter; // fallback

    static PnutsInterpreter getInterpreter(){
	if (interpreter == null){
	    interpreter = new PnutsInterpreter();
	}
	return interpreter;
    }

    private static UnparseVisitor unparser;

    static UnparseVisitor getUnparseVisitor(){
	if (unparser == null){
	    unparser = new UnparseVisitor(false);
	}
	return unparser;
    }

    boolean _includeLineNo = false;
    boolean includeMainMethod = false;

    private long classCount = 0L;

    private static Integer ONE = new Integer(1);

    boolean automatic;

    String className;

    String runtimeClassName;
    boolean useDynamicProxy;

    final static boolean java2 = Pnuts.isJava2();
    final static CodeLoaderFactory codeLoaderFactory;
    final static CodeLoaderFactory privilegedCodeLoaderFactory;
    static {
	if (java2){
	    codeLoaderFactory = new Java2CodeLoaderFactory();
	    privilegedCodeLoaderFactory = new PrivilegedCodeLoaderFactory();
	} else {
	    codeLoaderFactory = new CodeLoaderFactory();
	    privilegedCodeLoaderFactory = null;
	}
    }

    public Compiler(){
	this(null);
    }

    public Compiler(String className){
	this(className, true);
    }

    public Compiler(String className, boolean automatic){
	this(className, automatic, true);
    }

    public Compiler(String className, boolean automatic, boolean useDynamicProxy){
	if (className == null){
	    this.className = "_pnuts_";
	} else {
	    this.className = className;
	}
	this.automatic = automatic;
	useDynamicProxy(useDynamicProxy);
    }

    public void includeLineNo(boolean flag){
	_includeLineNo = flag;
    }

    public void includeMainMethod(boolean flag){
	includeMainMethod = flag;
    }

    public void useDynamicProxy(boolean flag){
	this.useDynamicProxy = flag;
	if (flag){
	    runtimeClassName = "pnuts.compiler.DynamicRuntime";
	} else {
	    runtimeClassName = "pnuts.lang.Runtime";
	}
    }

    protected Object execute(CompileContext cc, Context context, boolean catchJump){
	if (DEBUG){
	    cc.debug();
	}
	try {
	    Runtime rt;
	    if (PROFILE){
		s = System.currentTimeMillis();
		CodeLoader loader = createCodeLoader(Pnuts.getClassLoader(context), true);
		Class cls = cc.loadClasses(loader);
		rt = (Runtime)cls.newInstance();
		System.out.println("load: " + (System.currentTimeMillis() - s));
	    } else {
		CodeLoader loader = createCodeLoader(Pnuts.getClassLoader(context), true);
		Class cls = cc.loadClasses(loader);
		rt = (Runtime)cls.newInstance();
	    }
	    return rt.execute(context);
	} catch (IOException io){
	    throw new PnutsException(io, context);
	} catch (InstantiationException ie){
	    throw new PnutsException(ie, context);
	} catch (IllegalAccessException iae){
	    throw new PnutsException(iae, context);
	}
    }

    public Object startSet(SimpleNode node, Context context){
	if (!_includeLineNo && Pnuts.isVerbose()){
	    _includeLineNo = true;
	}
	return start(node, context);
    }

    public Object start(SimpleNode node, Context context) {
	try {
	    return _start(node, context, true);
	} catch (ClassFormatError cfe){
	    if (DEBUG){
		System.out.println(cfe);
	    }
	    if (automatic){
		return node.accept(getInterpreter(), context);
	    } else {
		throw cfe;
	    }
	}
    }

    Object _start(SimpleNode node, Context context, boolean catchJump) {

	if (PROFILE){
	    s = System.currentTimeMillis();
	}

	CompileContext cc;
	if (automatic){
	    cc = new CompileContext();
	} else {
	    cc = (CompileContext)context;
	}
	cc.constClassName = className;
	cc.cf = new ClassFile(className, "pnuts.lang.Runtime", null, ACC_PUBLIC);
	ClassFile cf = cc.cf;
	cf.addInterface("pnuts.compiler.Compiled");

	/*
	 * constructor
	 */
	cf.openMethod("<init>", "()V", ACC_PUBLIC);
	cf.add(ALOAD_0);
	cf.add(INVOKESPECIAL, "pnuts.lang.Runtime", "<init>", "()", "V");
	cf.add(RETURN);
	cf.closeMethod();

	/*
	 * public static void main(String[])
	 */
	if (includeMainMethod){
	    cf.openMethod("main", "([Ljava/lang/String;)V", (short)(ACC_PUBLIC | ACC_STATIC));
	    cf.add(NEW, className);
	    cf.add(DUP);
	    cf.add(INVOKESPECIAL, className, "<init>", "()", "V");

	    Label catchStart = cf.getLabel(true);
	    cf.add(NEW, "pnuts.lang.Context");
	    cf.add(DUP);
	    cf.add(INVOKESPECIAL, "pnuts.lang.Context", "<init>", "()", "V");
	    cf.add(INVOKEVIRTUAL, "pnuts.lang.Runtime", "execute", "(Lpnuts/lang/Context;)", "Ljava/lang/Object;");
	    cf.add(POP);
	    cf.add(RETURN);

	    Label catchEnd = cf.getLabel(true);
	    Label catchTarget = cf.getLabel(true);
	    cf.push();
	    cf.add(INVOKEVIRTUAL, "pnuts.lang.Jump", "getValue", "()", "Ljava/lang/Object;");
	    cf.add(POP);
	    cf.add(RETURN);
	    cf.addExceptionHandler(catchStart, catchEnd, catchTarget, "pnuts.lang.Jump");

	    cf.closeMethod();
	}

	/*
	 * public Object execute(Context context);
	 */
	cf.openMethod("exec", "(Lpnuts/lang/Context;)Ljava/lang/Object;", ACC_PROTECTED);
	int ctx = 1;
	cc.setContextIndex(ctx);

	/*
	cf.add(ALOAD_0);
	cf.add(ALOAD_1);
	cf.add(INVOKESPECIAL, "pnuts.lang.Runtime", "execute", "(Lpnuts/lang/Context;)", "Ljava/lang/Object;");
	cf.add(POP);
	*/

	Label start = cf.getLabel(true);
	int n = node.jjtGetNumChildren();
	if (n > 0){
	    int m = n - 1;
	    for (int i = 0; i < m; i++){
		accept(node, i, cc);
		cf.add(POP);
	    }
	    accept(node, m, cc);
	} else {
	    cf.add(ACONST_NULL);
	}
	cf.add(ARETURN);
	cf.closeMethod();

	/*
	 * static {} clause
	 */
	cf.openMethod("<clinit>", "()V", ACC_STATIC);

	cf.addField("NO_PARAM", "[Ljava/lang/Object;", ACC_STATIC);
	cf.add(ICONST_0);
	cf.add(ANEWARRAY, "java.lang.Object");
	cf.add(PUTSTATIC, className, "NO_PARAM", "[Ljava/lang/Object;");

	for (Enumeration e = cc.constants.keys(); e.hasMoreElements();){
	    Object obj = e.nextElement();
	    String k = (String)cc.constants.get(obj);
	    if (DEBUG){
		System.out.println(k + " : " + obj);
	    }

	    if (obj instanceof Integer){
		int value = ((Integer)obj).intValue();
		cf.addField(k, "Ljava/lang/Integer;", ACC_STATIC);
		cf.add(NEW, "java.lang.Integer");
		cf.add(DUP);
		cf.pushInteger(value);
		cf.add(INVOKESPECIAL, "java.lang.Integer", "<init>", "(I)", "V");
		cf.add(PUTSTATIC, className, k, "Ljava/lang/Integer;");
	    } else if (obj instanceof Byte){
		byte value = ((Byte)obj).byteValue();
		cf.addField(k, "Ljava/lang/Byte;", ACC_STATIC);
		cf.add(NEW, "java.lang.Byte");
		cf.add(DUP);
		cf.pushInteger((int)value);
		cf.add(INVOKESPECIAL, "java.lang.Byte", "<init>", "(B)", "V");
		cf.add(PUTSTATIC, className, k, "Ljava/lang/Byte;");
	    } else if (obj instanceof Character){
		char value = ((Character)obj).charValue();
		cf.addField(k, "Ljava/lang/Character;", ACC_STATIC);
		cf.add(NEW, "java.lang.Character");
		cf.add(DUP);
 		cf.pushInteger((int)value);
		cf.add(INVOKESPECIAL, "java.lang.Character", "<init>", "(C)", "V");
		cf.add(PUTSTATIC, className, k, "Ljava/lang/Character;");
	    } else if (obj instanceof Long){
		long value = ((Long)obj).longValue();
		cf.addField(k, "Ljava/lang/Long;", ACC_STATIC);
		cf.add(NEW, "java.lang.Long");
		cf.add(DUP);
		cf.pushLong(value);
		cf.add(INVOKESPECIAL, "java.lang.Long", "<init>", "(J)", "V");
		cf.add(PUTSTATIC, className, k, "Ljava/lang/Long;");
	    } else if (obj instanceof Float){
		float value = ((Float)obj).floatValue();
		cf.addField(k, "Ljava/lang/Float;", ACC_STATIC);
		cf.add(NEW, "java.lang.Float");
		cf.add(DUP);
		cf.pushFloat(value);
		cf.add(INVOKESPECIAL, "java.lang.Float", "<init>", "(F)", "V");
		cf.add(PUTSTATIC, className, k, "Ljava/lang/Float;");
	    } else if (obj instanceof Double){
		double value = ((Double)obj).doubleValue();
		cf.addField(k, "Ljava/lang/Double;", ACC_STATIC);
		cf.add(NEW, "java.lang.Double");
		cf.add(DUP);
		cf.pushDouble(value);
		cf.add(INVOKESPECIAL, "java.lang.Double", "<init>", "(D)", "V");
		cf.add(PUTSTATIC, className, k, "Ljava/lang/Double;");
	    } else if (obj instanceof BigDecimal){
		cf.addField(k, "Ljava/math/BigDecimal;", ACC_STATIC);
		cf.add(NEW, "java.math.BigDecimal");
		cf.add(DUP);
		cf.add(LDC, cf.addConstant(obj.toString()));
		cf.add(INVOKESPECIAL, "java.math.BigDecimal", "<init>", "(Ljava/lang/String;)", "V");
		cf.add(PUTSTATIC, className, k, "Ljava/math/BigDecimal;");
	    } else if (obj instanceof BigInteger){
		cf.addField(k, "Ljava/math/BigInteger;", ACC_STATIC);
		cf.add(NEW, "java.math.BigInteger");
		cf.add(DUP);
		cf.add(LDC, cf.addConstant(obj.toString()));
		cf.add(INVOKESPECIAL, "java.math.BigInteger", "<init>", "(Ljava/lang/String;)", "V");
		cf.add(PUTSTATIC, className, k, "Ljava/math/BigInteger;");
	    }
	}
	cf.add(RETURN);
	cf.closeMethod();

	if (PROFILE){
	    System.out.println(className + " classFile: " + (System.currentTimeMillis() - s));
	}

	if (automatic){
	    return execute(cc, context, catchJump);
	} else {
	    return null;
	}
    }

    public Object expressionList(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	int n = node.jjtGetNumChildren();
	int m = n - 1;
	for (int i = 0; i < m; i++) {
	    accept(node, i, context);
	    cf.add(POP);
	}
	accept(node, m, context);
	return null;
    }

    public Object integerNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	String str = node.str;
	Object p[] = Runtime.parseInt(str);
	Number n = (Number)p[0];

	String assoc = (String)cc.constants.get(n);
	if (assoc == null){
	    assoc = gensym(context);
	    cc.constants.put(n, assoc);
	}

	if (n instanceof Integer){
	    cf.add(GETSTATIC, cc.constClassName, assoc, "Ljava/lang/Integer;");
	} else if (n instanceof Long){
	    cf.add(GETSTATIC, cc.constClassName, assoc, "Ljava/lang/Long;");
	} else if (n instanceof Byte){
	    cf.add(GETSTATIC, cc.constClassName, assoc, "Ljava/lang/Byte;");
	} else if (n instanceof BigInteger){
	    cf.add(GETSTATIC, cc.constClassName, assoc, "Ljava/math/BigInteger;");
	} else if (n instanceof BigDecimal){
	    cf.add(GETSTATIC, cc.constClassName, assoc, "Ljava/math/BigDecimal;");
	} else {
	    throw new PnutsException("compiler error");
	}

	if (p[1] != null){
	    int offset = ((int[])p[1])[0];
	    cf.add(CHECKCAST, "java.lang.Number");

	    cf.add(LDC, cf.addConstant(str.substring(0, offset)));
	    cf.add(LDC, cf.addConstant(str.substring(offset)));
	    cf.loadLocal(cc.getContextIndex());
	    cf.add(INVOKESTATIC,
		   "pnuts.lang.Runtime",
		   "quantity",
		   "(Ljava/lang/Number;Ljava/lang/String;Ljava/lang/String;Lpnuts/lang/Context;)",
		   "Ljava/lang/Object;");
	}
	return null;
    }

    public Object floatingNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	String str = node.str;
	Object p[] = Runtime.parseFloat(str);
	Number n = (Number)p[0];

	String assoc = (String)cc.constants.get(n);
	if (assoc == null){
	    assoc = gensym(context);
	    cc.constants.put(n, assoc);
	}
	if (n instanceof Float){
	    cf.add(GETSTATIC, cc.constClassName, assoc, "Ljava/lang/Float;");
	} else if (n instanceof Double){
	    cf.add(GETSTATIC, cc.constClassName, assoc, "Ljava/lang/Double;");
	} else if (n instanceof BigDecimal){
	    cf.add(GETSTATIC, cc.constClassName, assoc, "Ljava/math/BigDecimal;");
	} else {
	    throw new PnutsException("compiler error");
	}
	if (p[1] != null){
	    int offset = ((int[])p[1])[0];
	    cf.add(CHECKCAST, "java.lang.Number");

	    cf.add(LDC, cf.addConstant(str.substring(0, offset)));
	    cf.add(LDC, cf.addConstant(str.substring(offset)));
	    cf.loadLocal(cc.getContextIndex());
	    cf.add(INVOKESTATIC,
		   "pnuts.lang.Runtime",
		   "quantity",
		   "(Ljava/lang/Number;Ljava/lang/String;Ljava/lang/String;Lpnuts/lang/Context;)",
		   "Ljava/lang/Object;");
	}
	return null;
    }

    public Object stringNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	try {
	    String parsed;
	    if (node.info != null){
		parsed = (String)node.info;
	    } else {
		parsed = Runtime.parseString(node.str);
	    }
	    cf.add(LDC, cf.addConstant(parsed));
	    return null;
	} catch (PnutsException e){
	    throw new PnutsException(e.getThrowable(), context);
	}
    }

    public Object characterNode(SimpleNode node, Context context){
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	try {
	    Character ch = Runtime.parseChar(node.str);
	    String assoc = (String)cc.constants.get(ch);
	    if (assoc == null){
		assoc = gensym(context);
		cc.constants.put(ch, assoc);
	    }
	    cf.add(GETSTATIC, cc.constClassName, assoc, "Ljava/lang/Character;");
	    return null;
	} catch (IllegalArgumentException e){
	    throw new PnutsException("illegal.character", new Object[]{node.str}, context);
	}
    }

    public Object classNode(SimpleNode node, Context context){
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	int ctx = cc.getContextIndex();

	if (_includeLineNo){
	    cf.loadLocal(ctx);
	    cf.pushInteger(node.beginLine);
	    if (node.beginLine == node.endLine){
		cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "setLine", "(Lpnuts/lang/Context;I)", "V");
	    } else {
		cf.pushInteger(node.endLine);
		cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "setLine", "(Lpnuts/lang/Context;II)", "V");
	    }
	}

	int n = node.jjtGetNumChildren();

	String name = node.str;
	if (name == null){
	    cf.add(GETSTATIC, "pnuts.lang.PnutsFunction", "CLASS", "Lpnuts/lang/PnutsFunction;");
	    return null;
	}

	for (int i = 0; i < n; i++){
	    SimpleNode ch = node.jjtGetChild(i);
	    if (ch.id == JJTPACKAGE){
		name += ".";
		name += ch.str;
	    }
	}
	cf.add(LDC, cf.addConstant(name));
	cf.loadLocal(ctx);
	cf.add(INVOKESTATIC, "pnuts.lang.Pnuts", "loadClass", "(Ljava/lang/String;Lpnuts/lang/Context;)", "Ljava/lang/Class;");
	return null;
    }

    void booleanCheck(int id, ClassFile cf, Context context){
	int index = cf.getLocal();
	cf.storeLocal(index);
	cf.loadLocal(index);

	if (id != JJTTRUENODE &&
	    id != JJTFALSENODE &&
	    id != JJTLENODE &&
	    id != JJTLTNODE &&
	    id != JJTGENODE &&
	    id != JJTGTNODE &&
	    id != JJTEQUALNODE &&
	    id != JJTNOTEQNODE &&
	    id != JJTINSTANCEOFEXPRESSION &&
	    id != JJTORNODE &&
	    id != JJTANDNODE &&
	    id != JJTNOTNODE)
	{
	    cf.add(INSTANCEOF, "java.lang.Boolean");
	    Label l_next = cf.getLabel();
	    cf.add(IFNE, l_next);
	    error(cf, "boolean.expected", (CompileContext)context);
	    l_next.fix();
	    cf.loadLocal(index);
	}
	cf.freeLocal(index);
    }

    public Object andNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;

	accept(node, 0, context);
	booleanCheck(node.jjtGetChild(0).id, cf, context);
	cf.add(CHECKCAST, "java.lang.Boolean");
	int tgt1 = cf.getLocal();
	cf.storeLocal(tgt1);
	cf.loadLocal(tgt1);
	cf.add(INVOKEVIRTUAL, "java.lang.Boolean", "booleanValue", "()", "Z");
	Label l_else = cf.getLabel();
	cf.add(IFEQ, l_else);
	accept(node, 1, context);
	booleanCheck(node.jjtGetChild(1).id, cf, context);
	Label next = cf.getLabel();
	cf.add(GOTO, next);
	l_else.fix();
	cf.loadLocal(tgt1);
	next.fix();
	cf.freeLocal(tgt1);
	return null;
    }

    public Object orNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;

	accept(node, 0, context);
	booleanCheck(node.jjtGetChild(0).id, cf, context);
	cf.add(CHECKCAST, "java.lang.Boolean");
	int tgt1 = cf.getLocal();
	cf.storeLocal(tgt1);
	cf.loadLocal(tgt1);
	cf.add(INVOKEVIRTUAL, "java.lang.Boolean", "booleanValue", "()", "Z");
	Label l_else = cf.getLabel();
	cf.add(IFNE, l_else);
	accept(node, 1, context);
	booleanCheck(node.jjtGetChild(1).id, cf, context);
	Label next = cf.getLabel();
	cf.add(GOTO, next);
	l_else.fix();
	cf.loadLocal(tgt1);
	next.fix();
	cf.freeLocal(tgt1);
	return null;
    }

    public Object notNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	accept(node, 0, context);
	booleanCheck(node.jjtGetChild(0).id, cf, context);
	cf.add(CHECKCAST, "java.lang.Boolean");
	cf.add(INVOKEVIRTUAL, "java.lang.Boolean", "booleanValue", "()", "Z");
	Label t = cf.getLabel();
	Label next = cf.getLabel();
	cf.add(IFEQ, t);
	cf.add(GETSTATIC, "java.lang.Boolean", "FALSE", "Ljava/lang/Boolean;");
	cf.add(GOTO, next);
	t.fix();
	cf.add(GETSTATIC, "java.lang.Boolean", "TRUE", "Ljava/lang/Boolean;");
	next.fix();
	return null;
    }

    public Object shiftLeftNode(SimpleNode node, Context context) {
	binary(node, context, Arithmetic.SHIFT_L);
	return null;
    }

    public Object shiftRightNode(SimpleNode node, Context context) {
	binary(node, context, Arithmetic.SHIFT_R);
	return null;
    }

    public Object shiftArithmeticNode(SimpleNode node, Context context) {
	binary(node, context, Arithmetic.SHIFT_A);
	return null;
    }

    public Object addNode(SimpleNode node, Context context) {
	binary(node, context, Arithmetic.ADD);
	return null;
    }

    public Object subtractNode(SimpleNode node, Context context) {
	binary(node, context, Arithmetic.SUBTRACT);
	return null;
    }

    public Object multNode(SimpleNode node, Context context) {
	binary(node, context, Arithmetic.MULTIPLY);
	return null;
    }

    public Object divideNode(SimpleNode node, Context context) {
	binary(node, context, Arithmetic.DIVIDE);
	return null;
    }

    public Object negativeNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	int ctx = cc.getContextIndex();
	accept(node, 0, context);
	cf.pushInteger(Arithmetic.NEGATE);
	cf.loadLocal(ctx);
	cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "unary", "(Ljava/lang/Object;ILpnuts/lang/Context;)", "Ljava/lang/Object;");
	return null;
    }

    public Object modNode(SimpleNode node, Context context) {
	binary(node, context, Arithmetic.MOD);
	return null;
    }

    public Object logExorNode(SimpleNode node, Context context) {
	binary(node, context, Arithmetic.XOR);
	return null;
    }

    public Object logOrNode(SimpleNode node, Context context) {
	binary(node, context, Arithmetic.OR);
	return null;
    }

    public Object logAndNode(SimpleNode node, Context context) {
	binary(node, context, Arithmetic.AND);
	return null;
    }

    public Object logNotNode(SimpleNode node, Context context){
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	int ctx = cc.getContextIndex();
	accept(node, 0, context);
	cf.pushInteger(Arithmetic.NOT);
	cf.loadLocal(ctx);
	cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "unary", "(Ljava/lang/Object;ILpnuts/lang/Context;)", "Ljava/lang/Object;");
	return null;
    }

    static void getBoolean(ClassFile cf, boolean mode){
	byte inst = IF_ICMPEQ;
	if (mode){
	    inst = IF_ICMPNE;
	}
	Label deny = cf.getLabel();
	cf.add(inst, deny);
	cf.add(GETSTATIC, "java.lang.Boolean", "TRUE", "Ljava/lang/Boolean;");
	Label next = cf.getLabel();
	cf.add(GOTO, next);
	deny.fix();
	cf.add(GETSTATIC, "java.lang.Boolean", "FALSE", "Ljava/lang/Boolean;");
	next.fix();
    }

    public Object ltNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	compare(node, cf, cc, "RIGHT_IS_BIGGER");
	getBoolean(cf, true);
	return null;
    }

    public Object leNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	compare(node, cf, cc, "LEFT_IS_BIGGER");
	getBoolean(cf, false);
	return null;
    }

    public Object gtNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	compare(node, cf, cc, "LEFT_IS_BIGGER");
	getBoolean(cf, true);
	return null;
    }

    public Object geNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	compare(node, cf, cc, "RIGHT_IS_BIGGER");
	getBoolean(cf, false);
	return null;
    }

    public Object equalNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	compare(node, cf, cc, "EQUAL");
	getBoolean(cf, true);
	return null;
    }

    public Object notEqNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	compare(node, cf, cc, "EQUAL");
	getBoolean(cf, false);
	return null;
    }

    public Object ifStatement(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	int ctx = cc.getContextIndex();
	if (_includeLineNo){
	    cf.loadLocal(ctx);
	    cf.pushInteger(node.beginLine);
	    if (node.beginLine == node.endLine){
		cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "setLine", "(Lpnuts/lang/Context;I)", "V");
	    } else {
		cf.pushInteger(node.endLine);
		cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "setLine", "(Lpnuts/lang/Context;II)", "V");
	    }
	}
	accept(node, 0, context);
	booleanCheck(node.jjtGetChild(0).id, cf, context);
	cf.add(CHECKCAST, "java.lang.Boolean");
	cf.add(INVOKEVIRTUAL, "java.lang.Boolean", "booleanValue", "()", "Z");
	Label l_else = cf.getLabel();
	Label next = cf.getLabel();
	cf.add(IFEQ, l_else);
	accept(node, 1, context);
	cf.add(GOTO, next);
	l_else.fix();
	int n = node.jjtGetNumChildren();
	int j = 0;

	if (n < 3){
	    cf.add(ACONST_NULL);
	    cf.add(GOTO, next);
	}
	for (int i = 2; i < n; i++){
	    SimpleNode _node = node.jjtGetChild(i);
	    if (_node.id == JJTELSEIFNODE){
		Label next_else = cf.getLabel();
		accept(_node, 0, context);
		booleanCheck(_node.jjtGetChild(0).id, cf, context);
		cf.add(CHECKCAST, "java.lang.Boolean");
		cf.add(INVOKEVIRTUAL, "java.lang.Boolean", "booleanValue", "()", "Z");
		cf.add(IFEQ, next_else);
		accept(_node, 1, context);
		cf.add(GOTO, next);
		next_else.fix();
	    } else if (_node.id == JJTELSENODE){
		accept(_node, 0, context);
		cf.add(GOTO, next);
	    }
	}
	cf.add(ACONST_NULL);
	next.fix();
	return null;
    }

    public Object functionStatement(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	int ctx = cc.getContextIndex();
	ClassFile cf = cc.cf;

	if (_includeLineNo){
	    cf.loadLocal(ctx);
	    cf.pushInteger(node.beginLine);
	    if (node.beginLine == node.endLine){
		cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "setLine", "(Lpnuts/lang/Context;I)", "V");
	    } else {
		cf.pushInteger(node.endLine);
		cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "setLine", "(Lpnuts/lang/Context;II)", "V");
	    }
	}

	SimpleNode block = null;
	SimpleNode param = null;
	SimpleNode n = node.jjtGetChild(0);
	if (n.id == JJTPARAMLIST){
	    param = n;
	    block = node.jjtGetChild(1);
	} else if (n.id == JJTBLOCK){
	    block = n;
	}
	int nargs = param.jjtGetNumChildren();
	String[] locals = new String[nargs];
	SimpleNode n0 = null;
	if (nargs == 1 &&
	    (n0 = param.jjtGetChild(0)).id == JJTINDEXNODE){
	    nargs = -1;
	    locals[0] = n0.jjtGetChild(0).str;
	} else {
	    for (int j = 0; j < nargs; j++){
		locals[j] = param.jjtGetChild(j).str;
	    }
	}

	String cls = className + "$" + (classCount++ & 0x7fffffffffffffffL);
	cf = new ClassFile(cls, "pnuts.lang.Function", null, ACC_PUBLIC);
	cf.addInterface("pnuts.compiler.Compiled");
	cc.classFiles.addElement(cf);
	cf.parent = cc.cf;

	/*
	 * public PnutsFunction register(PnutsFunction pf);
	 */
	cf.openMethod("register", "(Lpnuts/lang/PnutsFunction;)Lpnuts/lang/PnutsFunction;", ACC_PUBLIC);

	cf.add(ALOAD_0);
	cf.add(ALOAD_1);
	cf.add(INVOKESPECIAL,
	       "pnuts.lang.Function",
	       "register",
	       "(Lpnuts/lang/PnutsFunction;)",
	       "Lpnuts/lang/PnutsFunction;");
	cf.add(ARETURN);
	cf.closeMethod();

	/*
	 * protected String unparse(Context context);
	 */
	cf.openMethod("unparse", "(Lpnuts/lang/Context;)Ljava/lang/String;", ACC_PROTECTED);
	cf.add(LDC, cf.addConstant((String)node.accept(getUnparseVisitor(), context)));
	cf.add(ARETURN);
	cf.closeMethod();

	String fname = node.str;

	/*
	 * protected Object exec(Object[] args, Context conext);
	 */
	cc._openFrame(fname, locals);
	cf.openMethod("exec", "([Ljava/lang/Object;Lpnuts/lang/Context;)Ljava/lang/Object;", ACC_PROTECTED);
	CompileContext cc2 = (CompileContext)cc.clone();
	cc2.cf = cf;
	cc2.setContextIndex(2);

	ByteBuffer code = cf.codeBuffer;
	ByteBuffer blockCode = new ByteBuffer();
	cf.codeBuffer = blockCode;

	if (fname != null){
	    int f = cf.getLocal();
	    cc2._declare(fname, f, -1);
	    cf.add(ALOAD_0);
	    cf.add(GETFIELD, "pnuts.lang.Function", "function", "Lpnuts/lang/PnutsFunction;");
	    cf.storeLocal(f);
	}

	int expt = cf.getLocal();
	cc2._declare("!exception", expt, -1);

	block.accept(this, cc2);

	cf.codeBuffer = code;

	String[] keys = cc.env.bottom.keys;
	LocalInfo[] infos = cc.env.bottom.info;

	int count = cc.env.bottom.count;

	for (int i = 0; i < cc.env.bottom.count; i++){
	    cf.add(ACONST_NULL);
	    cf.storeLocal(infos[i].map);
	}

	Reference eref = cc2.getReference("!exception");
	cf.add(ACONST_NULL);
	cf.storeLocal(eref.index);

	Label catchStart = cf.getLabel(true);

	blockCode.prepend(code);
	cf.codeBuffer = blockCode;

	cf.add(ARETURN);

	Label catchEnd = cf.getLabel(true);
	Label catchTarget = cf.getLabel(true);
	cf.push();

	int caught = cf.getLocal();
	cf.storeLocal(caught);
	cf.loadLocal(cc2.getContextIndex());
	cf.loadLocal(caught);
	cf.loadLocal(eref.index);

	cf.add(CHECKCAST, "java.util.Hashtable");
	cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "checkException", "(Lpnuts/lang/Context;Ljava/lang/Throwable;Ljava/util/Hashtable;)", "V");

	cf.add(ACONST_NULL);
	cf.add(ARETURN);

	cf.addExceptionHandler(catchStart, catchEnd, catchTarget, "java.lang.Throwable");

	StringBuffer arg =
	    new StringBuffer("(Ljava/lang/String;[Ljava/lang/String;ILpnuts/lang/SimpleNode;Lpnuts/lang/Package;Lpnuts/lang/Context;");
	int n_names = 0;
	if (cc.env.imports != null){
	    n_names = cc.env.imports.size();
	}
	String names[] = new String[n_names];
	int j = 0;
	if (n_names > 0){
	    for (Enumeration e = cc.env.imports.elements(); e.hasMoreElements(); ){
		String sym = (String)e.nextElement();
		if (DEBUG){
		    System.out.println(fname + ": " + sym);
		}
		names[j++] = sym;
		arg.append("[Ljava/lang/Object;");
	    }
	}

	if (DEBUG){
	    System.out.println(cf.getClassName() + ": " + arg);
	}

	cf.closeMethod();
	cc._closeFrame();

	arg.append(")V");
	/*
	 * constructor
	 */
	cf.openMethod("<init>", arg.toString(), (short)0);
	cf.add(ALOAD_0);
	cf.add(ALOAD_1); // String name
	cf.add(ALOAD_2); // String[] locals
	cf.add(ILOAD_3); // int nargs
	cf.add(ALOAD, 4);   // SimpleNode
	cf.add(ALOAD, 5);   // Package
	cf.add(ALOAD, 6);   // Context
	cf.add(INVOKESPECIAL,
	       "pnuts.lang.Function",
	       "<init>",
	       "(Ljava/lang/String;[Ljava/lang/String;ILpnuts/lang/SimpleNode;Lpnuts/lang/Package;Lpnuts/lang/Context;)",
	       "V");

	for (int i = 0; i < n_names; i++){
	    cf.addField(names[i], "[Ljava/lang/Object;", (short)0);
	    cf.add(ALOAD_0);
	    cf.add(ALOAD, 7 + i);
	    cf.add(PUTFIELD, cf.getClassName(), names[i], "[Ljava/lang/Object;");
	}

	cf.add(RETURN);
	cf.closeMethod();

	cf = cc.cf;
	cf.loadLocal(ctx);
	cf.add(INVOKEVIRTUAL, "pnuts.lang.Context", "getCurrentPackage", "()", "Lpnuts/lang/Package;");
	int pkg = cf.getLocal();
	cf.storeLocal(pkg);
	
	cf.add(NEW, cls);
	cf.add(DUP);
	int fnameIndex = 0;
	if (fname != null){
	    fnameIndex = cf.addConstant(fname);
	    cf.add(LDC, fnameIndex);
	} else {
	    cf.add(ACONST_NULL);
	}
	cf.pushInteger(locals.length);
	cf.add(ANEWARRAY, "java.lang.String");
	for (int i = 0; i < locals.length; i++){
	    cf.add(DUP);
	    cf.pushInteger(i);
	    cf.add(LDC, cf.addConstant(locals[i]));
	    cf.add(AASTORE);
	}
	cf.pushInteger(nargs);
	cf.add(ACONST_NULL);
	cf.loadLocal(pkg);
	cf.loadLocal(ctx);

	String arg2 = "(Ljava/lang/String;[Ljava/lang/String;ILpnuts/lang/SimpleNode;Lpnuts/lang/Package;Lpnuts/lang/Context;";

	Vector exports = (Vector)cc.env.exports.get(cc2.env);
	if (exports != null){
	    int size = exports.size();
	    for (int i = 0; i < size; i++){
		arg2 += "[Ljava/lang/Object;";
		Reference ref = (Reference)exports.elementAt(i);
		if (ref.index < 0){
		    cf.add(ALOAD_0);
		    cf.add(GETFIELD, cf.getClassName(), ref.symbol, "[Ljava/lang/Object;");
    		} else if (ref.offset < 0){
		    cf.add(ICONST_1);
		    cf.add(ANEWARRAY, "java.lang.Object");
		    cf.add(DUP);
		    cf.add(ICONST_0);
		    cf.loadLocal(ref.index);
		    cf.add(AASTORE);
		} else {
		    cf.loadLocal(ref.index);
		    cf.add(CHECKCAST, "[Ljava/lang/Object;");
		}
	    }
	}

	if (DEBUG){
	    System.out.println("arg2: " + cc2.env.fname + ": " + arg2);
	}

	cf.add(INVOKESPECIAL,
	       cls,
	       "<init>",
	       arg2 + ")",
	       "V");

	int func = cf.getLocal();
	cf.storeLocal(func);

	int pfunc = cf.getLocal();
	cf.add(ACONST_NULL);
	cf.storeLocal(pfunc);

	Reference ref = null;
	if (fname != null){
	    ref = cc.env.getReference(fname);
	    if (ref != null){
		ref.get(cc);
		cf.storeLocal(pfunc);
	    } else {
		cf.loadLocal(pkg);
		cf.loadLocal(func);
		cf.add(CHECKCAST, "pnuts.lang.Function");
		cf.add(INVOKEVIRTUAL, "pnuts.lang.Function", "getName", "()", "Ljava/lang/String;");
		int name = cf.getLocal();
		cf.storeLocal(name);
		cf.loadLocal(name);

		cf.add(INVOKEVIRTUAL, "pnuts.lang.Package", "defined", "(Ljava/lang/String;)", "Z");
		Label nodef = cf.getLabel();
		cf.add(IFEQ, nodef);
		cf.loadLocal(pkg);
		cf.loadLocal(name);
		cf.loadLocal(ctx);
		cf.add(INVOKEVIRTUAL,
		       "pnuts.lang.Package",
		       "get",
		       "(Ljava/lang/String;Lpnuts/lang/Context;)",
		       "Ljava/lang/Object;");
		cf.storeLocal(pfunc);
		cf.loadLocal(pfunc);
		cf.add(INSTANCEOF, "pnuts.lang.PnutsFunction");
		Label instance = cf.getLabel();
		cf.add(IFNE, instance);
		cf.add(ACONST_NULL);
		cf.storeLocal(pfunc);

		Label match = cf.getLabel();
		cf.add(GOTO, match);

		instance.fix();
		cf.loadLocal(pfunc);
		cf.add(CHECKCAST, "pnuts.lang.PnutsFunction");
		cf.add(INVOKEVIRTUAL, "pnuts.lang.PnutsFunction", "getName", "()", "Ljava/lang/String;");
		cf.loadLocal(name);
		cf.add(INVOKEVIRTUAL, "java.lang.Object", "equals", "(Ljava/lang/Object;)", "Z");

		cf.add(IFNE, match);
		cf.add(ACONST_NULL);
		cf.storeLocal(pfunc);

		match.fix();
		nodef.fix();
	    }
	}

	cf.loadLocal(func);
	cf.loadLocal(pfunc);
	cf.add(CHECKCAST, "pnuts.lang.PnutsFunction");
	cf.add(INVOKEVIRTUAL, cls, "register", "(Lpnuts/lang/PnutsFunction;)", "Lpnuts/lang/PnutsFunction;");
	cf.storeLocal(pfunc);

	if (fname == null){
	    cf.loadLocal(pfunc);
	    return null;
	}

	if (cc.env.parent != null){
	    if (ref == null){
		cc.declare(fname);
		ref = cc.getReference(fname);
	    }
	    ref.set(cf, pfunc);
	    cf.loadLocal(pfunc);
	} else {
	    cf.loadLocal(pkg);
	    cf.add(LDC, fnameIndex);
	    cf.loadLocal(pfunc);
	    cf.loadLocal(ctx);
	    cf.add(INVOKEVIRTUAL,
		   "pnuts.lang.Package",
		   "set",
		   "(Ljava/lang/String;Ljava/lang/Object;Lpnuts/lang/Context;)", "V");
	    cf.loadLocal(pfunc);
	}
	cf.freeLocal(func);
	cf.freeLocal(pfunc);
	cf.freeLocal(pkg);
	return null;
    }

    public Object applicationNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;

	cf.loadLocal(cc.getContextIndex());
	accept(node, 1, context);
	int argIdx = cf.getLocal();
	cf.storeLocal(argIdx);
	
	accept(node, 0, context);
	cf.loadLocal(argIdx);

	SimpleNode argNode = node.jjtGetChild(1);
	int nargs = argNode.jjtGetNumChildren();

	boolean types_created = false;
	for (int i = 0; i < nargs; i++){
	    SimpleNode n = argNode.jjtGetChild(i);
	    if (n.id == JJTCASTEXPRESSION){
		if (!types_created){
		    cf.pushInteger(nargs);
		    cf.add(ANEWARRAY, "java.lang.Class");
		    types_created = true;
		}
		cf.add(DUP);
		cf.pushInteger(i);
		accept(n, 0, context);
		cf.add(AASTORE);
	    }
	}
	if (!types_created){
	    cf.add(ACONST_NULL);
	}

	cf.add(INVOKESTATIC,
	       "pnuts.lang.Runtime",
	       "call",
	       "(Lpnuts/lang/Context;Ljava/lang/Object;[Ljava/lang/Object;[Ljava/lang/Class;)",
	       "Ljava/lang/Object;");

	cf.freeLocal(argIdx);
	return null;
    }

    public Object blockNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	int n = node.jjtGetNumChildren();

	if (n > 0){
	    int m = n - 1;
	    for (int i = 0; i < m; i++) {
		accept(node, i, context);
		cf.add(POP);
	    }
	    accept(node, m, context);
	} else {
	    cf.add(ACONST_NULL);
	}
	return null;
    }

    public Object trueNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	cc.cf.add(GETSTATIC, "java.lang.Boolean", "TRUE", "Ljava/lang/Boolean;");
	return null;
    }

    public Object falseNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	cc.cf.add(GETSTATIC, "java.lang.Boolean", "FALSE", "Ljava/lang/Boolean;");
	return null;
    }

    public Object nullNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	cc.cf.add(ACONST_NULL);
	return null;
    }

    public Object idNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	String symbol = node.str;
	int ctx = cc.getContextIndex();

	if (_includeLineNo){
	    cf.loadLocal(ctx);
	    cf.pushInteger(node.beginLine);
	    if (node.beginLine == node.endLine){
		cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "setLine", "(Lpnuts/lang/Context;I)", "V");
	    } else {
		cf.pushInteger(node.endLine);
		cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "setLine", "(Lpnuts/lang/Context;II)", "V");
	    }
	}

	Reference ref = cc.getReference(symbol);
	if (ref != null){
	    ref.get(cc);
	} else {
	    cf.loadLocal(ctx);
	    cf.add(LDC, cf.addConstant(symbol));
	    cf.add(INVOKEVIRTUAL, "pnuts.lang.Context", "getId", "(Ljava/lang/String;)", "Ljava/lang/Object;");
	}
	return null;
    }

    public Object global(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	cf.add(INVOKESTATIC, "pnuts.lang.Package", "getGlobalPackage", "()", "Lpnuts/lang/Package;");
	int gbl = cf.getLocal();
	cf.storeLocal(gbl);
	cf.loadLocal(gbl);
	int symbol = cf.addConstant(node.str);
	cf.add(LDC, symbol);
	cf.add(INVOKEVIRTUAL, "pnuts.lang.Package", "defined", "(Ljava/lang/String;)", "Z");
	Label else_ = cf.getLabel();
	cf.add(IFEQ, else_);
	cf.loadLocal(gbl);
	cf.add(LDC, symbol);
	int ctx = cc.getContextIndex();
	cf.loadLocal(ctx);
	cf.add(INVOKEVIRTUAL,
	       "pnuts.lang.Package",
	       "get",
	       "(Ljava/lang/String;Lpnuts/lang/Context;)",
	       "Ljava/lang/Object;");
	Label next = cf.getLabel();
	cf.add(GOTO, next);
	else_.fix();
	cf.add(NEW, "pnuts.lang.PnutsException");
	cf.add(DUP);
	cf.add(LDC, cf.addConstant("not.defined"));
	cf.add(ICONST_1);
	cf.add(ANEWARRAY, "java.lang.Object");
	cf.add(DUP);
	cf.add(ICONST_0);
	cf.add(LDC, symbol);
	cf.add(AASTORE);
	cf.loadLocal(ctx);
	cf.add(INVOKESPECIAL,
	       "pnuts.lang.PnutsException",
	       "<init>",
	       "(Ljava/lang/String;[Ljava/lang/Object;Lpnuts/lang/Context;)",
	       "V");
	cf.add(ATHROW);
	next.fix();

	cf.freeLocal(gbl);
	return null;
    }

    public Object listElements(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	int n = node.jjtGetNumChildren();
	if (n == 0){
	    cf.add(GETSTATIC, cc.constClassName, "NO_PARAM", "[Ljava/lang/Object;");
	} else {
	    cf.pushInteger(n);
	    cf.add(ANEWARRAY, "java.lang.Object");
	    for (int i = 0; i < n; i++){
		cf.add(DUP);
		cf.pushInteger(i);
		accept(node, i, context);
		cf.add(AASTORE);
	    }
	}
	return null;
    }

    public Object castExpression(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	int ctx = cc.getContextIndex();
	cf.loadLocal(ctx);
	accept(node, 0, context);
	cf.add(CHECKCAST, "java.lang.Class");
	accept(node, 1, context);
	cf.add(ICONST_1);
	cf.add(INVOKESTATIC,
	       "pnuts.lang.Runtime",
	       "cast",
	       "(Lpnuts/lang/Context;Ljava/lang/Class;Ljava/lang/Object;Z)",
	       "Ljava/lang/Object;");
	return null;
    }

    public Object memberNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	accept(node, 0, context);
	int tgt = cf.getLocal();
	cf.storeLocal(tgt);
	cf.loadLocal(tgt);
	Label else_if = cf.getLabel();
	Label else_end = cf.getLabel();
	cf.add(IFNONNULL, else_if);
	cf.add(NEW, "pnuts.lang.PnutsException");
	cf.add(DUP);
	cf.add(NEW, "java.lang.NullPointerException");
	cf.add(DUP);
	cf.add(INVOKESPECIAL, "java.lang.NullPointerException", "<init>", "()", "V");
	int ctx = cc.getContextIndex();
	cf.loadLocal(ctx);
	cf.add(INVOKESPECIAL,
	       "pnuts.lang.PnutsException",
	       "<init>",
	       "(Ljava/lang/Throwable;Lpnuts/lang/Context;)",
	       "V");
	cf.add(ATHROW);
	else_if.fix();
	if ("length".equals(node.str)){
	    cf.loadLocal(tgt);
	    cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "isArray", "(Ljava/lang/Object;)", "Z");
	    Label else_ = cf.getLabel();
	    cf.add(IFEQ, else_);
	    cf.add(NEW, "java.lang.Integer");
	    cf.add(DUP);
	    cf.loadLocal(tgt);
	    cf.add(INVOKESTATIC, "java.lang.reflect.Array", "getLength", "(Ljava/lang/Object;)", "I");
	    cf.add(INVOKESPECIAL, "java.lang.Integer", "<init>", "(I)", "V");
	    cf.add(GOTO, else_end);
	    else_.fix();
	}
	cf.loadLocal(ctx);
	cf.loadLocal(tgt);
	cf.add(LDC, cf.addConstant(node.str));
	cf.add(INVOKESTATIC,
	       "pnuts.lang.Runtime",
	       "getField",
	       "(Lpnuts/lang/Context;Ljava/lang/Object;Ljava/lang/String;)",
	       "Ljava/lang/Object;");
	else_end.fix();
	cf.freeLocal(tgt);
	return null;
    }

    public Object methodNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	int ctx = cc.getContextIndex();
	if (_includeLineNo){
	    cf.loadLocal(ctx);
	    cf.pushInteger(node.beginLine);
	    if (node.beginLine == node.endLine){
		cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "setLine", "(Lpnuts/lang/Context;I)", "V");
	    } else {
		cf.pushInteger(node.endLine);
		cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "setLine", "(Lpnuts/lang/Context;II)", "V");
	    }
	}
	accept(node, 0, context);
	int tgt = cf.getLocal();	
	cf.storeLocal(tgt);
	/**
	cf.loadLocal(tgt);
	Label nonnull = cf.getLabel();
	cf.add(IFNONNULL, nonnull);
	cf.add(NEW, "pnuts.lang.PnutsException");
	cf.add(DUP);
	cf.add(NEW, "java.lang.NullPointerException");
	cf.add(DUP);
	cf.add(INVOKESPECIAL, "java.lang.NullPointerException", "<init>", "()", "V");
	cf.loadLocal(ctx);
	cf.add(INVOKESPECIAL,
	       "pnuts.lang.PnutsException",
	       "<init>",
	       "(Ljava/lang/Throwable;Lpnuts/lang/Context;)",
	       "V");
	cf.add(ATHROW);
	nonnull.fix();
	**/
	SimpleNode argNode = node.jjtGetChild(1);
	int nargs = argNode.jjtGetNumChildren();

	boolean types_created = false;
	for (int i = 0; i < nargs; i++){
	    SimpleNode n = argNode.jjtGetChild(i);
	    if (n.id == JJTCASTEXPRESSION){
		if (!types_created){
		    cf.pushInteger(nargs);
		    cf.add(ANEWARRAY, "java.lang.Class");
		    types_created = true;
		}
		cf.add(DUP);
		cf.pushInteger(i);
		accept(n, 0, context);
		cf.add(AASTORE);
	    }
	}
	int types = cf.getLocal();
	if (types_created){
	    cf.storeLocal(types);
	}

	cf.loadLocal(ctx);
	cf.loadLocal(tgt);
	cf.add(INVOKEVIRTUAL, "java.lang.Object", "getClass", "()", "Ljava/lang/Class;");
	cf.add(LDC, cf.addConstant(node.str));
	accept(node, 1, context);
	if (types_created){
	    cf.loadLocal(types);
	} else {
	    cf.add(ACONST_NULL);
	}
	cf.loadLocal(tgt);
	cf.freeLocal(tgt);

	cf.add(INVOKESTATIC,
	       runtimeClassName,
	       "callMethod",
	       "(Lpnuts/lang/Context;Ljava/lang/Class;Ljava/lang/String;[Ljava/lang/Object;[Ljava/lang/Class;Ljava/lang/Object;)",
	       "Ljava/lang/Object;");
	return null;
    }

    public Object arrayType(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;

	SimpleNode n = node;
	int count = 0;
	while (n != null && n.id == JJTARRAYTYPE){
	    count++;
	    n = n.jjtGetChild(0);
	}
	if (n.id == JJTINDEXNODE){
	    Label clserr = cf.getLabel();
	    Label numerr = cf.getLabel();

	    Object[] idx = parseIndex(n);
	    ((SimpleNode)idx[0]).accept(this, context);
	    cf.add(DUP);
	    cf.add(INSTANCEOF, "java.lang.Class");

	    cf.add(IFEQ, clserr);
	    cf.add(CHECKCAST, "java.lang.Class");

	    Object[] dim = (Object[])idx[1];
	    cf.pushInteger(dim.length);
	    cf.add(NEWARRAY, T_INT);

	    for (int i = 0; i < dim.length; i++){
		cf.add(DUP);
		cf.pushInteger(i);
		((SimpleNode)dim[i]).accept(this, context);
		cf.add(DUP);
		cf.add(INSTANCEOF, "java.lang.Number");
		cf.add(IFEQ, numerr);

		cf.add(CHECKCAST, "java.lang.Number");
		cf.add(INVOKEVIRTUAL, "java.lang.Number", "intValue", "()", "I");
		cf.add(IASTORE);
	    }

	    cf.add(INVOKESTATIC,
		   "java.lang.reflect.Array",
		   "newInstance",
		   "(Ljava/lang/Class;[I)",
		   "Ljava/lang/Object;");
	    Label next = cf.getLabel();
	    cf.add(GOTO, next);
	    numerr.fix();
	    error(cf, "number.expected", cc);
	    cf.add(ACONST_NULL);
	    cf.add(GOTO, next);
	    clserr.fix();
	    error(cf, "classOrArray.expected", cc);
	    next.fix();
	    return null;
	} else {
	    n.accept(this, context);
	    int type = cf.getLocal();
	    cf.storeLocal(type);
	    cf.loadLocal(type);
	    cf.add(INSTANCEOF, "java.lang.Class");
	    Label err = cf.getLabel();
	    cf.add(IFEQ, err);
	    cf.loadLocal(type);
	    cf.freeLocal(type);
	    cf.add(CHECKCAST, "java.lang.Class");
	    cf.pushInteger(count);
	    cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "arrayType", "(Ljava/lang/Class;I)", "Ljava/lang/Class;");
	    Label next = cf.getLabel();
	    cf.add(GOTO, next);
	    err.fix();
	    error(cf, "classOrArray.expected", cc);
	    next.fix();
	}
	return null;
    }

    /**
     * @return [base_component_node, [idx_node_0, idx_node_1, ...]]
     */
    static Object[] parseIndex(SimpleNode node){
	SimpleNode c0 = node.jjtGetChild(0);
	SimpleNode c1 = node.jjtGetChild(1);
	if (c0.id != JJTINDEXNODE){
	    return new Object[]{c0, new Object[]{c1}};
	} else {
	    Object[] r = parseIndex(c0);
	    Object[] d = (Object[])r[1];
	    Object[] a = new Object[d.length + 1];
	    System.arraycopy(d, 0, a, 0, d.length);
	    a[d.length] = c1;
	    return new Object[]{r[0], a};
	}
    }

    public Object indexNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	Object[] idx = parseIndex(node);
	((SimpleNode)idx[0]).accept(this, context);
	int tgt = cf.getLocal();
	cf.storeLocal(tgt);
	cf.loadLocal(tgt);
	Object[] dim = (Object[])idx[1];
	cf.add(INSTANCEOF, "java.lang.Class");
	Label else_ = cf.getLabel();
	Label else_end = cf.getLabel();
	cf.add(IFEQ, else_);
	cf.loadLocal(tgt);
	cf.add(CHECKCAST, "java.lang.Class");
	cf.pushInteger(dim.length);
	cf.add(NEWARRAY, T_INT);
	int ctx = cc.getContextIndex();
	for (int i = 0; i < dim.length; i++){
	    cf.add(DUP);
	    cf.pushInteger(i);
	    ((SimpleNode)dim[i]).accept(this, context);
	    cf.add(DUP);
	    cf.add(INSTANCEOF, "java.lang.Number");
	    Label ok = cf.getLabel();
	    cf.add(IFNE, ok);
	    error(cf, "number.expected", cc);
	    ok.fix();
	    cf.add(CHECKCAST, "java.lang.Number");
	    cf.add(INVOKEVIRTUAL, "java.lang.Number", "intValue", "()", "I");
	    cf.add(IASTORE);
	}
	cf.add(INVOKESTATIC, "java.lang.reflect.Array", "newInstance", "(Ljava/lang/Class;[I)", "Ljava/lang/Object;");
	cf.add(GOTO, else_end);
	else_.fix();
	Label err = cf.getLabel();
	for (int i = 0; i < dim.length; i++){
	    ((SimpleNode)dim[i]).accept(this, context);
	    int _idx = cf.getLocal();
	    cf.storeLocal(_idx);
	    cf.loadLocal(_idx);
	    cf.add(INSTANCEOF, "java.lang.Number");
	    cf.add(IFEQ, err);
	    cf.loadLocal(tgt);
	    cf.loadLocal(_idx);
	    cf.add(CHECKCAST, "java.lang.Number");
	    cf.add(INVOKEVIRTUAL, "java.lang.Number", "intValue", "()", "I");
	    cf.loadLocal(ctx);
	    cf.add(INVOKESTATIC,
		   "pnuts.lang.Runtime",
		   "getIndex",
		   "(Ljava/lang/Object;ILpnuts/lang/Context;)",
		   "Ljava/lang/Object;");
	    cf.storeLocal(tgt);
	}
	cf.loadLocal(tgt);
	cf.add(GOTO, else_end);
	err.fix();

	cf.add(NEW, "pnuts.lang.PnutsException");
	cf.add(DUP);
	cf.add(LDC, cf.addConstant("number.expected"));
	cf.add(GETSTATIC, cc.constClassName, "NO_PARAM", "[Ljava/lang/Object;");
	cf.loadLocal(ctx);
	cf.add(INVOKESPECIAL,
	       "pnuts.lang.PnutsException",
	       "<init>",
	       "(Ljava/lang/String;[Ljava/lang/Object;Lpnuts/lang/Context;)",
	       "V");
	cf.add(ATHROW);
	
	else_end.fix();
	cf.freeLocal(tgt);
	return null;
    }

    public Object instanceofExpression(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	int ctx = cc.getContextIndex();

	accept(node, 1, context);
	int param = cf.getLocal();
	cf.storeLocal(param);

	Label next = cf.getLabel();
	Label err = cf.getLabel();
	Label l_false = cf.getLabel();

	cf.loadLocal(param);
	cf.add(INSTANCEOF, "java.lang.Class");
	cf.add(IFEQ, err);

	accept(node, 0, context);
	int tgt = cf.getLocal();
	cf.storeLocal(tgt);
	cf.loadLocal(tgt);
	cf.add(IFNULL, l_false);

	cf.loadLocal(param);
	cf.freeLocal(param);
	cf.add(CHECKCAST, "java.lang.Class");
	cf.loadLocal(tgt);
	cf.freeLocal(tgt);

	cf.add(INVOKEVIRTUAL, "java.lang.Class", "isInstance", "(Ljava/lang/Object;)", "Z");
	cf.add(IFEQ, l_false);
	cf.add(GETSTATIC, "java.lang.Boolean", "TRUE", "Ljava/lang/Boolean;");
	cf.add(GOTO, next);
	err.fix();
	cf.add(NEW, "pnuts.lang.PnutsException");
	cf.add(DUP);
	cf.add(LDC, cf.addConstant("class.expected"));
	cf.add(GETSTATIC, cc.constClassName, "NO_PARAM", "[Ljava/lang/Object;");
	cf.loadLocal(ctx);
	cf.add(INVOKESPECIAL,
	       "pnuts.lang.PnutsException",
	       "<init>",
	       "(Ljava/lang/String;[Ljava/lang/Object;Lpnuts/lang/Context;)",
	       "V");
	cf.add(ATHROW);
	l_false.fix();
	cf.add(GETSTATIC, "java.lang.Boolean", "FALSE", "Ljava/lang/Boolean;");
	next.fix();
	return null;
    }

    public Object assignment(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENT);
    }

    public Object assignmentTA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTTA);
    }

    public Object assignmentMA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTMA);
    }

    public Object assignmentDA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTDA);
    }

    public Object assignmentPA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTPA);
    }

    public Object assignmentSA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTSA);
    }

    public Object assignmentLA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTLA);
    }

    public Object assignmentRA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTRA);
    }

    public Object assignmentRAA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTRAA);
    }

    public Object assignmentAA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTAA);
    }

    public Object assignmentEA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTEA);
    }

    public Object assignmentOA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTOA);
    }

    public Object preIncrNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	String assoc = (String)cc.constants.get(ONE);
	if (assoc == null){
	    assoc = gensym(context);
	    cc.constants.put(ONE, assoc);
	}
	cf.add(GETSTATIC, cc.constClassName, assoc, "Ljava/lang/Integer;");
	_assign(node.jjtGetChild(0), context, JJTASSIGNMENTPA, false);
	return null;
    }

    public Object preDecrNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	String assoc = (String)cc.constants.get(ONE);
	if (assoc == null){
	    assoc = gensym(context);
	    cc.constants.put(ONE, assoc);
	}
	cf.add(GETSTATIC, cc.constClassName, assoc, "Ljava/lang/Integer;");
	_assign(node.jjtGetChild(0), context, JJTASSIGNMENTSA, false);
	return null;
    }

    public Object postIncrNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	String assoc = (String)cc.constants.get(ONE);
	if (assoc == null){
	    assoc = gensym(context);
	    cc.constants.put(ONE, assoc);
	}
	accept(node, 0, context);
	int ret = cf.getLocal();
	cf.storeLocal(ret);

	cf.add(NEW, "java.lang.Integer");
	cf.add(DUP);
	cf.add(ICONST_1);
	cf.add(INVOKESPECIAL, "java.lang.Integer", "<init>", "(I)", "V");

	_assign(node.jjtGetChild(0), context, JJTASSIGNMENTPA, false);
	cf.add(POP);
	cf.loadLocal(ret);
	cf.freeLocal(ret);
	return null;
    }

    public Object postDecrNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	String assoc = (String)cc.constants.get(ONE);
	if (assoc == null){
	    assoc = gensym(context);
	    cc.constants.put(ONE, assoc);
	}
	accept(node, 0, context);
	int ret = cf.getLocal();
	cf.storeLocal(ret);
	cf.add(GETSTATIC, cc.constClassName, assoc, "Ljava/lang/Integer;");
	_assign(node.jjtGetChild(0), context, JJTASSIGNMENTSA, false);
	cf.add(POP);
	cf.loadLocal(ret);
	cf.freeLocal(ret);
	return null;
    }

    public Object staticMethodNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	int ctx = cc.getContextIndex();
	if (_includeLineNo){
	    cf.loadLocal(ctx);
	    cf.pushInteger(node.beginLine);
	    if (node.beginLine == node.endLine){
		cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "setLine", "(Lpnuts/lang/Context;I)", "V");
	    } else {
		cf.pushInteger(node.endLine);
		cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "setLine", "(Lpnuts/lang/Context;II)", "V");
	    }
	}
	accept(node, 1, context);
	int args = cf.getLocal();
	cf.storeLocal(args);

	SimpleNode c1 = node.jjtGetChild(0);
	String pkgName = getPackageName(c1);

	SimpleNode argNode = node.jjtGetChild(1);
	int nargs = argNode.jjtGetNumChildren();

	cf.add(LDC, cf.addConstant(pkgName));
	cf.add(INVOKESTATIC, "pnuts.lang.Package", "find", "(Ljava/lang/String;)", "Lpnuts/lang/Package;");
	int pkg = cf.getLocal();
	cf.storeLocal(pkg);
	cf.loadLocal(pkg);
	Label l_null = cf.getLabel();
	cf.add(IFNULL, l_null);
	cf.loadLocal(pkg);
	cf.add(LDC, cf.addConstant(node.str));
	cf.loadLocal(ctx);
	cf.add(INVOKEVIRTUAL,
	       "pnuts.lang.Package",
	       "get",
	       "(Ljava/lang/String;Lpnuts/lang/Context;)",
	       "Ljava/lang/Object;");
	int got = cf.getLocal();
	cf.storeLocal(got);
	cf.loadLocal(got);
	cf.add(INSTANCEOF, "pnuts.lang.PnutsFunction");
	Label l_else = cf.getLabel();
	cf.add(IFEQ, l_else);
	cf.loadLocal(ctx);
	cf.loadLocal(got);
	cf.add(CHECKCAST, "pnuts.lang.PnutsFunction");
	cf.loadLocal(args);
	cf.add(INVOKESTATIC,
	       "pnuts.lang.Runtime",
	       "callFunction",
	       "(Lpnuts/lang/Context;Lpnuts/lang/PnutsFunction;[Ljava/lang/Object;)",
	       "Ljava/lang/Object;");
	Label next = cf.getLabel();
	cf.add(GOTO, next);
	l_else.fix();
	cf.loadLocal(got);
	cf.add(INSTANCEOF, "java.lang.Class");
	cf.add(IFEQ, l_null);

	boolean types_created = false;
	for (int i = 0; i < nargs; i++){
	    SimpleNode n = argNode.jjtGetChild(i);
	    if (n.id == JJTCASTEXPRESSION){
		if (!types_created){
		    cf.pushInteger(nargs);
		    cf.add(ANEWARRAY, "java.lang.Class");
		    types_created = true;
		}
		cf.add(DUP);
		cf.pushInteger(i);
		accept(n, 0, context);
		cf.add(AASTORE);
	    }
	}
	int types = cf.getLocal();
	if (types_created){
	    cf.storeLocal(types);
	}

	cf.loadLocal(ctx);
	cf.loadLocal(got);
	cf.add(CHECKCAST, "java.lang.Class");
	cf.loadLocal(args);
	if (types_created){
	    cf.loadLocal(types);
	} else {
	    cf.add(ACONST_NULL);
	}
	cf.add(INVOKESTATIC,
	       runtimeClassName,
	       "newInstance",
	       "(Lpnuts/lang/Context;Ljava/lang/Class;[Ljava/lang/Object;[Ljava/lang/Class;)",
	       "Ljava/lang/Object;");

	cf.add(GOTO, next);

	l_null.fix();
	accept(node, 0, context);
	int tgt = cf.getLocal();
	cf.storeLocal(tgt);
	cf.loadLocal(tgt);
	cf.add(INSTANCEOF, "java.lang.Class");
	Label ok = cf.getLabel();
	cf.add(IFNE, ok);
	cf.add(NEW, "pnuts.lang.PnutsException");
	cf.add(DUP);
	cf.add(LDC, cf.addConstant("illegal.staticCall"));
	cf.add(ICONST_3);
	cf.add(ANEWARRAY, "java.lang.Object");
	cf.add(DUP);
	cf.add(ICONST_0);
	cf.add(LDC, cf.addConstant(pkgName));
	cf.add(AASTORE);
	cf.add(DUP);
	cf.add(ICONST_1);
	cf.add(LDC, cf.addConstant(node.str));
	cf.add(AASTORE);
	cf.add(DUP);
	cf.add(ICONST_2);
	cf.add(NEW, "java.lang.Integer");
	cf.add(DUP);
	cf.pushInteger(nargs);
	cf.add(INVOKESPECIAL, "java.lang.Integer", "<init>", "(I)", "V");
	cf.add(AASTORE);
	cf.loadLocal(ctx);
	cf.add(INVOKESPECIAL,
	       "pnuts.lang.PnutsException",
	       "<init>",
	       "(Ljava/lang/String;[Ljava/lang/Object;Lpnuts/lang/Context;)",
	       "V");
	cf.add(ATHROW);
	ok.fix();

	types_created = false;
	for (int i = 0; i < nargs; i++){
	    SimpleNode n = argNode.jjtGetChild(i);
	    if (n.id == JJTCASTEXPRESSION){
		if (!types_created){
		    cf.pushInteger(nargs);
		    cf.add(ANEWARRAY, "java.lang.Class");
		    types_created = true;
		}
		cf.add(DUP);
		cf.pushInteger(i);
		accept(n, 0, context);
		cf.add(AASTORE);
	    }
	}
	types = cf.getLocal();
	if (types_created){
	    cf.storeLocal(types);
	}
	cf.loadLocal(ctx);
	cf.loadLocal(tgt);
	cf.add(CHECKCAST, "java.lang.Class");
	cf.add(LDC, cf.addConstant(node.str));
	cf.loadLocal(args);
	if (types_created){
	    cf.loadLocal(types);
	} else {
	    cf.add(ACONST_NULL);
	}
	cf.freeLocal(types);
	cf.add(ACONST_NULL);

	cf.add(INVOKESTATIC,
	       runtimeClassName,
	       "callMethod",
	       "(Lpnuts/lang/Context;Ljava/lang/Class;Ljava/lang/String;[Ljava/lang/Object;[Ljava/lang/Class;Ljava/lang/Object;)",
	       "Ljava/lang/Object;");
	next.fix();
	return null;
    }

    public Object staticMemberNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	int ctx = cc.getContextIndex();
	if (_includeLineNo){
	    cf.loadLocal(ctx);
	    cf.pushInteger(node.beginLine);
	    if (node.beginLine == node.endLine){
		cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "setLine", "(Lpnuts/lang/Context;I)", "V");
	    } else {
		cf.pushInteger(node.endLine);
		cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "setLine", "(Lpnuts/lang/Context;II)", "V");
	    }
	}
	SimpleNode c1 = node.jjtGetChild(0);
	Label next = cf.getLabel();

	String pkgName = getPackageName(c1);

	cf.add(LDC, cf.addConstant(pkgName));
	cf.add(INVOKESTATIC, "pnuts.lang.Package", "find", "(Ljava/lang/String;)", "Lpnuts/lang/Package;");
	int pkg = cf.getLocal();
	cf.storeLocal(pkg);
	cf.loadLocal(pkg);
	Label l_null = cf.getLabel();
	cf.add(IFNULL, l_null);
	cf.loadLocal(pkg);
	cf.add(LDC, cf.addConstant(node.str));
	cf.loadLocal(ctx);
	cf.add(INVOKEVIRTUAL,
	       "pnuts.lang.Package",
	       "defined",
	       "(Ljava/lang/String;Lpnuts/lang/Context;)",
	       "Z");
	Label undef = cf.getLabel();
	cf.add(IFEQ, undef);
	cf.loadLocal(pkg);
	cf.freeLocal(pkg);
	cf.add(LDC, cf.addConstant(node.str));
	cf.loadLocal(ctx);
	cf.add(INVOKEVIRTUAL,
	       "pnuts.lang.Package",
	       "get",
	       "(Ljava/lang/String;Lpnuts/lang/Context;)",
	       "Ljava/lang/Object;");
	cf.add(GOTO, next);

	undef.fix();
	cf.add(NEW, "pnuts.lang.PnutsException");
	cf.add(DUP);
	cf.add(LDC, cf.addConstant("not.defined"));
	cf.add(ICONST_1);
	cf.add(ANEWARRAY, "java.lang.Object");
	cf.add(DUP);
	cf.add(ICONST_0);
	cf.add(LDC, cf.addConstant(node.str));
	cf.add(AASTORE);
	cf.loadLocal(ctx);
	cf.add(INVOKESPECIAL,
	       "pnuts.lang.PnutsException",
	       "<init>",
	       "(Ljava/lang/String;[Ljava/lang/Object;Lpnuts/lang/Context;)",
	       "V");
	cf.add(ATHROW);

	l_null.fix();

	c1.accept(this, context);
	int tgt = cf.getLocal();
	cf.storeLocal(tgt);
	cf.loadLocal(tgt);
	cf.add(INSTANCEOF, "java.lang.Class");
	Label err = cf.getLabel();
	cf.add(IFEQ, err);

	cf.loadLocal(ctx);
	cf.loadLocal(tgt);
	cf.freeLocal(tgt);
	cf.add(CHECKCAST, "java.lang.Class");
	cf.add(LDC, cf.addConstant(node.str));
	cf.add(INVOKESTATIC,
	       "pnuts.lang.Runtime",
	       "getStaticField",
	       "(Lpnuts/lang/Context;Ljava/lang/Class;Ljava/lang/String;)",
	       "Ljava/lang/Object;");
	cf.add(GOTO, next);

	err.fix();
	cf.add(NEW, "pnuts.lang.PnutsException");
	cf.add(DUP);
	cf.add(LDC, cf.addConstant("packageOrClass.expected"));
	cf.add(GETSTATIC, cc.constClassName, "NO_PARAM", "[Ljava/lang/Object;");
	cf.loadLocal(ctx);
	cf.add(INVOKESPECIAL,
	       "pnuts.lang.PnutsException",
	       "<init>",
	       "(Ljava/lang/String;[Ljava/lang/Object;Lpnuts/lang/Context;)",
	       "V");
	cf.add(ATHROW);
	next.fix();
	return null;
    }

    public Object rangeNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	int ctx = cc.getContextIndex();

	accept(node, 0, context);
	int tgt = cf.getLocal();
	cf.storeLocal(tgt);

	accept(node, 1, context);
	int idx1 = cf.getLocal();
	cf.storeLocal(idx1);

	int idx2 = cf.getLocal();
	if (node.jjtGetNumChildren() >= 3){
	    accept(node, 2, context);
	} else {
	    cf.add(ACONST_NULL);
	}
	cf.storeLocal(idx2);

	cf.loadLocal(idx1);
	cf.add(INSTANCEOF, "java.lang.Number");
	Label err = cf.getLabel();
	Label ok = cf.getLabel();
	Label next = cf.getLabel();
	cf.add(IFEQ, err);

	cf.loadLocal(idx2);
	cf.add(IFNULL, ok);
	cf.loadLocal(idx2);
	cf.add(INSTANCEOF, "java.lang.Number");
	cf.add(IFEQ, err);

	ok.fix();
	cf.loadLocal(tgt);
	cf.freeLocal(tgt);
	cf.loadLocal(idx1);
	cf.freeLocal(idx1);
	cf.add(CHECKCAST, "java.lang.Number");
	cf.loadLocal(idx2);
	cf.freeLocal(idx2);
	cf.add(CHECKCAST, "java.lang.Number");
	cf.add(INVOKESTATIC,
	       "pnuts.lang.Runtime",
	       "range",
	       "(Ljava/lang/Object;Ljava/lang/Number;Ljava/lang/Number;)",
	       "Ljava/lang/Object;");
	cf.add(GOTO, next);

	err.fix();
	cf.add(NEW, "pnuts.lang.PnutsException");
	cf.add(DUP);
	cf.add(LDC, cf.addConstant("number.expected"));
	cf.add(GETSTATIC, cc.constClassName, "NO_PARAM", "[Ljava/lang/Object;");
	cf.loadLocal(ctx);
	cf.add(INVOKESPECIAL,
	       "pnuts.lang.PnutsException",
	       "<init>",
	       "(Ljava/lang/String;[Ljava/lang/Object;Lpnuts/lang/Context;)",
	       "V");
	cf.add(ATHROW);

	next.fix();
	return null;
    }

    public Object forStatement(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	int ctx = cc.getContextIndex();

	if (_includeLineNo){
	    cf.loadLocal(ctx);
	    cf.pushInteger(node.beginLine);
	    if (node.beginLine == node.endLine){
		cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "setLine", "(Lpnuts/lang/Context;I)", "V");
	    } else {
		cf.pushInteger(node.endLine);
		cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "setLine", "(Lpnuts/lang/Context;II)", "V");
	    }
	}

	SimpleNode initNode = null;
	SimpleNode condNode = null;
	SimpleNode updateNode = null;
	SimpleNode blockNode = null;

	int j = 0;
	SimpleNode n = node.jjtGetChild(j);
	if (n.id == JJTFORINIT){
	    initNode = n;
	    j++;
	}
	n = node.jjtGetChild(j);
	if (n.id != JJTFORUPDATE && n.id != JJTBLOCK){
	    condNode = n;
	    j++;
	}
	n = node.jjtGetChild(j);
	if (n.id == JJTFORUPDATE){
	    updateNode = n;
	    j++;
	}

	blockNode = node.jjtGetChild(j);

	int last = cf.getLocal();
	cf.add(ACONST_NULL);
	cf.storeLocal(last);

	if (initNode != null){
	    int num = initNode.jjtGetNumChildren();
	    String[] env = new String[num];
	    for (int i = 0; i < env.length; i++){
		SimpleNode sn = initNode.jjtGetChild(i);
		env[i] = sn.str;
	    }
	    cc.openScope(env);
	    for (int i = 0; i < env.length; i++){

		cf.add(ICONST_1);
		cf.add(ANEWARRAY, "java.lang.Object");
		cf.add(DUP);
		cf.add(ICONST_0);

		SimpleNode sn = initNode.jjtGetChild(i);
		accept(sn, 0, context);
		Reference ref = cc.getReference(env[i]);

		cf.add(AASTORE);
		cf.storeLocal(ref.index);
	    }
	} else {
	    cc.openScope(new String[]{});
	}

	Label start = cf.getLabel(true);
	Label cont = cf.getLabel();
	Label next = cf.getLabel();
	Label broke = cf.getLabel();
	cc.pushContinueLabel(cont);
	cc.pushBreakLabel(broke);

	if (condNode != null){
	    condNode.accept(this, context);
	    booleanCheck(condNode.id, cf, context);
	    cf.add(CHECKCAST, "java.lang.Boolean");
	    cf.add(INVOKEVIRTUAL, "java.lang.Boolean", "booleanValue", "()", "Z");
	} else {
	    cf.add(ICONST_1);
	}
	cf.add(IFEQ, next);

	blockNode.accept(this, context);
	cont.fix();
	cf.storeLocal(last);

	if (updateNode != null){
	    int num = updateNode.jjtGetNumChildren();
	    for (int i = 0; i < num; i++){
		accept(updateNode, i, context);
		cf.add(POP);
	    }
	}

	cf.add(GOTO, start);
	next.fix();
	cf.loadLocal(last);
	cf.freeLocal(last);
	cc.closeScope();
	broke.fix();
	cc.popContinueLabel();
	cc.popBreakLabel();
	return null;
    }

    public Object breakNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;

	if (node.jjtGetNumChildren() > 0){
	    accept(node, 0, context);
	} else {
	    cf.add(ACONST_NULL);
	}

	Label brk = cc.getBreakLabel();
	if (brk != null){
	    cf.add(GOTO, brk);
	}
	return null;
    }

    public Object continueNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	Label cnt = cc.getContinueLabel();
	cf.add(ACONST_NULL);
	if (cnt != null){
	    cf.add(GOTO, cnt);
	}
	return null;
    }

    public Object returnNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	if (node.jjtGetNumChildren() > 0){
	    accept(node, 0, context);
	} else {
	    cf.add(ACONST_NULL);
	}
	if (cc.env.parent != null){
	    cf.add(ARETURN);
	} else {
	    cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "jump", "(Ljava/lang/Object;)", "V");
	    cf.add(ACONST_NULL);
	}
	return null;
    }

    public Object catchNode(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	int ctx = cc.getContextIndex();

	if (node.jjtGetNumChildren() == 0){
	    cf.add(ACONST_NULL);
	    return null;
	}

	Label next = cf.getLabel();

	accept(node, 0, context);
	cf.add(CHECKCAST, "[Ljava/lang/Object;");
	int args = cf.getLocal();
	cf.storeLocal(args);

	Reference eref = cc.getReference("!exception");
	if (cc.env.parent != null){
	    cf.loadLocal(eref.index);
	    Label nonnull = cf.getLabel();
	    cf.add(IFNONNULL, nonnull);
	    cf.add(NEW, "java.util.Hashtable");
	    cf.add(DUP);
	    cf.pushInteger(10);
	    cf.add(INVOKESPECIAL, "java.util.Hashtable", "<init>", "(I)", "V");
	    cf.storeLocal(eref.index);
	    nonnull.fix();
	    cf.loadLocal(args);
	    cf.add(ICONST_1);
	    cf.add(AALOAD);
	    Label l_null = cf.getLabel();
	    cf.add(IFNULL, l_null);
	    cf.loadLocal(eref.index);
	    cf.add(CHECKCAST, "java.util.Hashtable");
	    cf.loadLocal(args);
	    cf.add(DUP);
	    cf.add(ICONST_0);
	    cf.add(AALOAD);
	    cf.add(SWAP);
	    cf.add(ICONST_1);
	    cf.add(AALOAD);
	    cf.add(INVOKEVIRTUAL,
		   "java.util.Hashtable",
		   "put",
		   "(Ljava/lang/Object;Ljava/lang/Object;)",
		   "Ljava/lang/Object;");
	    cf.add(POP);
	    cf.add(ACONST_NULL);
	    cf.add(GOTO, next);
	    l_null.fix();
	    cf.loadLocal(eref.index);
	    cf.add(CHECKCAST, "java.util.Hashtable");
	    cf.loadLocal(args);
	    cf.add(ICONST_0);
	    cf.add(AALOAD);
	    cf.add(INVOKEVIRTUAL, "java.util.Hashtable", "remove", "(Ljava/lang/Object;)", "Ljava/lang/Object;");
	    cf.add(POP);
	    cf.add(ACONST_NULL);
	    next.fix();
	} else {
	    cf.loadLocal(args);
	    cf.add(ICONST_0);
	    cf.add(AALOAD);
	    cf.add(CHECKCAST, "java.lang.Class");
	    cf.loadLocal(args);
	    cf.add(ICONST_1);
	    cf.add(AALOAD);
	    cf.add(CHECKCAST, "pnuts.lang.PnutsFunction");
	    cf.loadLocal(ctx);
	    cf.add(INVOKESTATIC,
		   "pnuts.lang.Runtime",
		   "catchException",
		   "(Ljava/lang/Class;Lpnuts/lang/PnutsFunction;Lpnuts/lang/Context;)",
		   "V");
	    cf.add(ACONST_NULL);
	}
	return null;
    }

    public Object whileStatement(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	String symbol = node.str;
	int ctx = cc.getContextIndex();

	if (_includeLineNo){
	    cf.loadLocal(ctx);
	    cf.pushInteger(node.beginLine);
	    if (node.beginLine == node.endLine){
		cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "setLine", "(Lpnuts/lang/Context;I)", "V");
	    } else {
		cf.pushInteger(node.endLine);
		cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "setLine", "(Lpnuts/lang/Context;II)", "V");
	    }
	}
	cf.add(ACONST_NULL);
	int last = cf.getLocal();
	cf.storeLocal(last);

	Label start = cf.getLabel(true);
	Label next = cf.getLabel();
	Label broke = cf.getLabel();
	Label cont = cf.getLabel();
	cc.pushContinueLabel(cont);
	cc.pushBreakLabel(broke);

	accept(node, 0, context);
	int tgt = cf.getLocal();
	cf.storeLocal(tgt);
	cf.loadLocal(tgt);

	cf.add(INSTANCEOF, "java.lang.Boolean");
	Label ok = cf.getLabel();
	cf.add(IFNE, ok);
	error(cf, "boolean.expected", cc);

	ok.fix();
	cf.loadLocal(tgt);
	cf.add(CHECKCAST, "java.lang.Boolean");
	cf.add(INVOKEVIRTUAL, "java.lang.Boolean", "booleanValue", "()", "Z");
	cf.add(IFEQ, next);
	accept(node, 1, context);
	cont.fix();
	cf.storeLocal(last);
	cf.add(GOTO, start);
	next.fix();
	cc.popContinueLabel();
	cc.popBreakLabel();
	cf.loadLocal(last);
	cf.freeLocal(last);
	cf.freeLocal(tgt);
	broke.fix();
	return null;
    }

    public Object foreachStatement(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	int ctx = cc.getContextIndex();

	if (_includeLineNo){
	    cf.loadLocal(ctx);
	    cf.pushInteger(node.beginLine);
	    if (node.beginLine == node.endLine){
		cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "setLine", "(Lpnuts/lang/Context;I)", "V");
	    } else {
		cf.pushInteger(node.endLine);
		cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "setLine", "(Lpnuts/lang/Context;II)", "V");
	    }
	}
	cc.openScope(new String[]{node.str});
	Reference count = cc.getReference(node.str);
	cf.add(ICONST_1);
	cf.add(ANEWARRAY, "java.lang.Object");
	cf.storeLocal(count.index);

	Label next = cf.getLabel();
	Label broke = cf.getLabel();
	cc.pushBreakLabel(broke);
	int last = cf.getLocal();
	cf.add(ACONST_NULL);
	cf.storeLocal(last);

	accept(node, 0, context);
	int tgt = cf.getLocal();
	cf.storeLocal(tgt);

	cf.loadLocal(tgt);
	Label nonnull = cf.getLabel();

	cf.add(IFNONNULL, nonnull);
	cf.add(GOTO, next);
	nonnull.fix();

	cf.loadLocal(tgt);
	cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "isArray", "(Ljava/lang/Object;)", "Z");
	Label checkEnum = cf.getLabel();
	cf.add(IFEQ, checkEnum);

	cf.add(ICONST_0);
	int counter = cf.getLocal();
	cf.istoreLocal(counter);

	int max = cf.getLocal();
	cf.loadLocal(tgt);
	cf.add(INVOKESTATIC, "java.lang.reflect.Array", "getLength", "(Ljava/lang/Object;)", "I");
	cf.istoreLocal(max);

	Label loop1 = cf.getLabel();
	cf.add(GOTO, loop1);
	Label body = cf.getLabel(true);
	Label cont = cf.getLabel();
	cc.pushContinueLabel(cont);

	cf.loadLocal(tgt);
	cf.iloadLocal(counter);
	cf.add(INVOKESTATIC, "java.lang.reflect.Array", "get", "(Ljava/lang/Object;I)", "Ljava/lang/Object;");

	int tmp = cf.getLocal();
	cf.storeLocal(tmp);
	count.set(cf, tmp);

	accept(node, 1, context);
	cont.fix();
	cf.storeLocal(last);

	cf.add(IINC, counter, 1);
	loop1.fix();
	cf.iloadLocal(counter);
	cf.iloadLocal(max);
	cf.add(IF_ICMPLT, body);
	cf.add(GOTO, next);
	
	checkEnum.fix();
	cf.loadLocal(tgt);
	cf.add(INSTANCEOF, "java.util.Enumeration");
	Label checkIterator = cf.getLabel();
	cf.add(IFEQ, checkIterator);

	Label loop2 = cf.getLabel();
	cf.add(GOTO, loop2);
	Label body2 = cf.getLabel(true);

	cf.loadLocal(tgt);
	cf.add(CHECKCAST, "java.util.Enumeration");
	cf.add(INVOKEINTERFACE, "java.util.Enumeration", "nextElement", "()", "Ljava/lang/Object;");

	int tmp2 = cf.getLocal();
	cf.storeLocal(tmp2);
	count.set(cf, tmp2);

	Label cont2 = cf.getLabel();
	cc.pushContinueLabel(cont2);
	accept(node, 1, context);
	cont2.fix();
	cf.storeLocal(last);

	loop2.fix();
	cf.loadLocal(tgt);
	cf.add(CHECKCAST, "java.util.Enumeration");
	cf.add(INVOKEINTERFACE, "java.util.Enumeration", "hasMoreElements", "()", "Z");
	cf.add(IFEQ, next);
	cf.add(GOTO, body2);

	checkIterator.fix();

	cf.loadLocal(tgt);
	cf.add(INSTANCEOF, "java.util.Iterator");
	cf.add(IFEQ, next);

	Label loop3 = cf.getLabel();
	cf.add(GOTO, loop3);
	Label body3 = cf.getLabel(true);

	cf.loadLocal(tgt);
	cf.add(CHECKCAST, "java.util.Iterator");
	cf.add(INVOKEINTERFACE, "java.util.Iterator", "next", "()", "Ljava/lang/Object;");

	int tmp3 = cf.getLocal();
	cf.storeLocal(tmp3);
	count.set(cf, tmp3);

	Label cont3 = cf.getLabel();
	cc.pushContinueLabel(cont3);

	accept(node, 1, context);
	cont3.fix();
	cf.storeLocal(last);

	loop3.fix();
	cf.loadLocal(tgt);
	cf.add(CHECKCAST, "java.util.Iterator");
	cf.add(INVOKEINTERFACE, "java.util.Iterator", "hasNext", "()", "Z");
	cf.add(IFEQ, next);
	cf.add(GOTO, body3);

	next.fix();
	cf.loadLocal(last);
	cf.freeLocal(last);
	cf.freeLocal(tmp);
	cf.freeLocal(tgt);
	cf.freeLocal(counter);
	cf.freeLocal(max);
	cc.popContinueLabel();
	cc.popContinueLabel();
	cc.popContinueLabel();
	cc.popBreakLabel();
	cc.closeScope();
	broke.fix();
	return null;
    }

    public Object switchStatement(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	int ctx = cc.getContextIndex();

	if (_includeLineNo){
	    cf.loadLocal(ctx);
	    cf.pushInteger(node.beginLine);
	    if (node.beginLine == node.endLine){
		cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "setLine", "(Lpnuts/lang/Context;I)", "V");
	    } else {
		cf.pushInteger(node.endLine);
		cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "setLine", "(Lpnuts/lang/Context;II)", "V");
	    }
	}

	int n = node.jjtGetNumChildren();
	accept(node, 0, context);
	int tgt = cf.getLocal();
	cf.storeLocal(tgt);

	cc.openScope(new String[]{});
	int match = cf.getLocal();
	cf.add(ICONST_0);
	cf.istoreLocal(match);
	int last = cf.getLocal();
	cf.add(ACONST_NULL);
	cf.storeLocal(last);

	Label start = cf.getLabel(true);
	Label next = cf.getLabel();
	cc.pushBreakLabel(next);

	for (int i = 1; i < n; i++){
	    SimpleNode _node = node.jjtGetChild(i);
	    if (_node.jjtGetNumChildren() == 1){ // case
		cf.iloadLocal(match);
		Label hit = cf.getLabel();
		cf.add(IFNE, hit);
		i++;
		accept(_node, 0, context);
		cf.loadLocal(tgt);
		cf.loadLocal(ctx);
		cf.add(INVOKESTATIC,
		       "pnuts.lang.Runtime",
		       "compareTo",
		       "(Ljava/lang/Object;Ljava/lang/Object;Lpnuts/lang/Context;)",
		       "I");
		cf.add(GETSTATIC, "pnuts.lang.Arithmetic", "EQUAL", "I");
		Label cont = cf.getLabel();
		cf.add(IF_ICMPNE, cont);
		cf.add(ICONST_1); // match = 1
		cf.istoreLocal(match);
		hit.fix();
		accept(node, i, context);
		cf.storeLocal(last);
		cont.fix();
	    } else { // default
		cf.add(ICONST_1);
		cf.istoreLocal(match);
		accept(node, ++i, context);
		cf.storeLocal(last);
	    }
	}
	cf.freeLocal(tgt);
	cf.freeLocal(match);
	cf.loadLocal(last);
	next.fix();
	cc.closeScope();
	cc.popBreakLabel();
	return null;
    }

    public Object switchBlock(SimpleNode node, Context context) {
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	int n = node.jjtGetNumChildren();
	if (n > 0){
	    int m = n - 1;
	    for (int i = 0; i < m; i++) {
		accept(node, i, context);
		cf.add(POP);
	    }
	    accept(node, m, context);
	} else {
	    cf.add(ACONST_NULL);
	}
	return null;
    }

    void binary(SimpleNode node, Context context, int operator){
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	int ctx = cc.getContextIndex();

	accept(node, 0, context);
	accept(node, 1, context);

	if (_includeLineNo){
	    cf.loadLocal(ctx);
	    cf.pushInteger(node.beginLine);
	    if (node.beginLine == node.endLine){
		cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "setLine", "(Lpnuts/lang/Context;I)", "V");
	    } else {
		cf.pushInteger(node.endLine);
		cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "setLine", "(Lpnuts/lang/Context;II)", "V");
	    }
	}

	cf.pushInteger(operator);
	cf.loadLocal(ctx);
	cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "binary", "(Ljava/lang/Object;Ljava/lang/Object;ILpnuts/lang/Context;)", "Ljava/lang/Object;");
    }

    private void compare(SimpleNode node, ClassFile cf, CompileContext cc, String mode){
	accept(node, 0, cc);
	accept(node, 1, cc);
	cf.loadLocal(cc.getContextIndex());
	cf.add(INVOKESTATIC, "pnuts.lang.Runtime", "compareTo", "(Ljava/lang/Object;Ljava/lang/Object;Lpnuts/lang/Context;)", "I");
	if (!"EQUAL".equals(mode)){
	    int t = cf.getLocal();
	    cf.istoreLocal(t);
	    cf.iloadLocal(t);
	    cf.add(ICONST_2); // Arithmetic.NOT_EQUAL
	    Label cont = cf.getLabel();
	    cf.add(IF_ICMPNE, cont);
	    error(cf, "illegal.operandType", cc);
	    cf.add(ATHROW);
	    cont.fix();
	    cf.iloadLocal(t);
	    cf.freeLocal(t);
	}
	cf.add(GETSTATIC, "pnuts.lang.Arithmetic", mode, "I");
    }

    static String getPackageName(SimpleNode node){
	if (node.jjtGetNumChildren() > 0){
	    SimpleNode c1 = node.jjtGetChild(0);
	    return getPackageName(c1) + "::" + node.str;
	} else {
	    return node.str;
	}
    }

    static String gensym(Context context){
	return ((CompileContext)context).sym.gen();
    }

    protected Object accept(SimpleNode node, int idx, Context context){
	return node.jjtGetChild(idx).accept(this, context);
    }


    private Object assign(SimpleNode node, Context context, int id){
	accept(node, 1, context);
	_assign(node.jjtGetChild(0), context, id, false);
	return null;
    }

    private Object _assign(SimpleNode lhs, Context context, int id, boolean mutable){
	CompileContext cc = (CompileContext)context;
	ClassFile cf = cc.cf;
	int ctx = cc.getContextIndex();
	int rhs = cf.getLocal();
	cf.storeLocal(rhs);

	if (lhs.id == JJTIDNODE){
	    assignId(cf, id, ctx, rhs, lhs, cc, mutable);
	} else if (lhs.id == JJTGLOBAL){
	    assignGlobal(cf, id, ctx, rhs, lhs, cc);
	} else if (lhs.id == JJTINDEXNODE){
	    assignIndex(cf, id, ctx, rhs, lhs, cc);
	} else if (lhs.id == JJTSTATICMEMBERNODE){
	    assignStaticMember(cf, id, ctx, rhs, lhs, cc);
	} else if (lhs.id == JJTMEMBERNODE){
	    assignMember(cf, id, ctx, rhs, lhs, cc);
	} else if (lhs.id == JJTRANGENODE){
	    assignRange(cf, id, ctx, rhs, lhs, cc);
	} else {
	    throw new PnutsException("illegal.assign", new Object[]{}, context);
	}
	cf.freeLocal(rhs);
	return null;
    }

    boolean inControl(SimpleNode node){
	while (node != null){
	    if (node.id == JJTFUNCTIONSTATEMENT){
		return false;
	    } else if (node.id == JJTSWITCHSTATEMENT ||
		       node.id == JJTFOREACHSTATEMENT ||
		       node.id == JJTFORSTATEMENT ||
		       node.id == JJTWHILESTATEMENT ||
		       node.id == JJTIFSTATEMENT){
		return true;
	    }
	    node = node.jjtGetParent();
	}
	return false;
    }

    void assignId(ClassFile cf, int id, int ctx, int rhs, SimpleNode lhs, CompileContext cc, boolean mutable){
	if (cc.env.parent != null && !mutable){
	    Reference ref = cc.getReference(lhs.str);
	    int idx = 0;
	    if (ref == null){
		cc.declare(lhs.str);
		ref = cc.getReference(lhs.str);
	    }
	    if (id != JJTASSIGNMENT){
		ref.get(cc);
		cf.loadLocal(rhs);
		compute(cf, id, ctx);
		cf.storeLocal(rhs);
	    }
	    ref.set(cf, rhs);
	    if (!inControl(lhs)){
		cc.setReference(lhs.str);
	    }
	    cf.loadLocal(rhs);
	} else {
	    Reference ref = cc.getReference(lhs.str);
	    if (ref != null){
		if (id != JJTASSIGNMENT){
		    ref.get(cc);
		    cf.loadLocal(rhs);
		    compute(cf, id, ctx);
		    cf.storeLocal(rhs);
		}
		ref.set(cf, rhs);
		cf.loadLocal(rhs);
	    } else {
		if (id != JJTASSIGNMENT){
		    cf.loadLocal(ctx);
		    cf.add(INVOKEVIRTUAL, "pnuts.lang.Context", "getCurrentPackage", "()", "Lpnuts/lang/Package;");
		    cf.add(LDC, cf.addConstant(lhs.str));
		    cf.add(INVOKEVIRTUAL, "pnuts.lang.Package", "get", "(Ljava/lang/String;)", "Ljava/lang/Object;");
		    cf.loadLocal(rhs);
		    compute(cf, id, ctx);
		    cf.storeLocal(rhs);
		}
		cf.loadLocal(ctx);
		cf.add(INVOKEVIRTUAL, "pnuts.lang.Context", "getCurrentPackage", "()", "Lpnuts/lang/Package;");
		cf.add(LDC, cf.addConstant(lhs.str));
		cf.loadLocal(rhs);
		cf.loadLocal(ctx);
		cf.add(INVOKEVIRTUAL,
		       "pnuts.lang.Package",
		       "set",
		       "(Ljava/lang/String;Ljava/lang/Object;Lpnuts/lang/Context;)",
		       "V");
		cf.loadLocal(rhs);
	    }
	}
    }

    void assignMember(ClassFile cf, int id, int ctx, int rhs, SimpleNode lhs, CompileContext cc){
	accept(lhs, 0, cc);
	int tgt = cf.getLocal();
	cf.storeLocal(tgt);
	if (id != JJTASSIGNMENT){
	    cf.loadLocal(ctx);
	    cf.loadLocal(tgt);
	    cf.add(LDC, cf.addConstant(lhs.str));
	    cf.add(INVOKESTATIC,
		   "pnuts.lang.Runtime",
		   "getField",
		   "(Lpnuts/lang/Context;Ljava/lang/Object;Ljava/lang/String;)",
		   "Ljava/lang/Object;");
	    cf.loadLocal(rhs);
	    compute(cf, id, ctx);
	    cf.storeLocal(rhs);
	}

	cf.loadLocal(ctx);
	cf.loadLocal(tgt);
	cf.freeLocal(tgt);
	cf.add(LDC, cf.addConstant(lhs.str));
	cf.loadLocal(rhs);
	cf.add(INVOKESTATIC,
	       "pnuts.lang.Runtime",
	       "putField",
	       "(Lpnuts/lang/Context;Ljava/lang/Object;Ljava/lang/String;Ljava/lang/Object;)",
	       "V");
	cf.loadLocal(rhs);
    }

    void assignIndex(ClassFile cf, int id, int ctx, int rhs, SimpleNode lhs, CompileContext cc){
	accept(lhs, 0, cc);
	int tgt = cf.getLocal();
	cf.storeLocal(tgt);

	accept(lhs, 1, cc);
	int idx = cf.getLocal();
	int num = cf.getLocal();
	cf.storeLocal(idx);

	cf.loadLocal(idx);
	cf.add(INSTANCEOF, "java.lang.Number");
	Label ok = cf.getLabel();
	cf.add(IFNE, ok);
	error(cf, "number.expected", cc);
	ok.fix();
	cf.loadLocal(idx);
	cf.add(CHECKCAST, "java.lang.Number");
	cf.add(INVOKEVIRTUAL, "java.lang.Number", "intValue", "()", "I");
	cf.istoreLocal(num);

	cf.loadLocal(tgt);
	cf.add(INSTANCEOF, "pnuts.lang.Indexed");
	Label l1 = cf.getLabel();
	cf.add(IFEQ, l1);

	Label next = cf.getLabel();
	if (id == JJTASSIGNMENT){
	    cf.loadLocal(tgt);
	    cf.add(CHECKCAST, "pnuts.lang.Indexed");
	    cf.iloadLocal(num);
	    cf.loadLocal(rhs);
	    cf.add(INVOKEINTERFACE, "pnuts.lang.Indexed", "set", "(ILjava/lang/Object;)", "V");
	    cf.loadLocal(rhs);
	    cf.add(GOTO, next);
	} else {
	    cf.loadLocal(tgt);
	    cf.add(CHECKCAST, "pnuts.lang.Indexed");
	    cf.iloadLocal(num);
	    cf.add(INVOKEINTERFACE, "pnuts.lang.Indexed", "get", "(I)", "Ljava/lang/Object;");
	    cf.loadLocal(rhs);
	    compute(cf, id, ctx);
	    cf.storeLocal(rhs);
	    cf.loadLocal(tgt);
	    cf.iloadLocal(num);
	    cf.loadLocal(rhs);
	    cf.add(INVOKEINTERFACE, "pnuts.lang.Indexed", "set", "(ILjava/lang/Object;)", "V");
	    cf.loadLocal(rhs);
	    cf.add(GOTO, next);
	}

	l1.fix();
	cf.loadLocal(tgt);
	cf.add(INSTANCEOF, "java.lang.String");
	Label l2 = cf.getLabel();
	cf.add(IFEQ, l2);
	if (id != JJTASSIGNMENT){
	    error(cf, "illegal.assign", cc);
	} else {
	    cf.loadLocal(rhs);
	    cf.add(INSTANCEOF, "java.lang.Character");
	    Label ok2 = cf.getLabel();
	    cf.add(IFNE, ok2);
	    error(cf, "illegal.assign", cc);
	    ok2.fix();
	    cf.loadLocal(tgt);
	    cf.loadLocal(idx);
	    cf.add(CHECKCAST, "java.lang.Number");
	    cf.add(DUP);
	    cf.loadLocal(rhs);
	    cf.add(INVOKESTATIC,
		   "pnuts.lang.Runtime",
		   "assignRange",
		   "(Ljava/lang/Object;Ljava/lang/Number;Ljava/lang/Number;Ljava/lang/Object;)",
		   "Ljava/lang/Object;");
	    SimpleNode ch = lhs.jjtGetChild(0);
	    if (ch.id == JJTIDNODE ||
		ch.id == JJTGLOBAL ||
		ch.id == JJTINDEXNODE ||
		ch.id == JJTMEMBERNODE ||
		ch.id == JJTSTATICMEMBERNODE){
		_assign(ch, cc, id, true);
	    }
	    cf.add(GOTO, next);
	}
	cf.freeLocal(idx);

	l2.fix();
	if (id == JJTASSIGNMENT){
	    cf.loadLocal(tgt);
	    cf.iloadLocal(num);
	    cf.loadLocal(rhs);
	    cf.add(INVOKESTATIC, "java.lang.reflect.Array", "set", "(Ljava/lang/Object;ILjava/lang/Object;)", "V");
	    cf.loadLocal(rhs);
	} else {
	    cf.loadLocal(tgt);
	    cf.iloadLocal(num);
	    cf.add(INVOKESTATIC, "java.lang.reflect.Array", "get", "(Ljava/lang/Object;I)", "Ljava/lang/Object;");
	    cf.loadLocal(rhs);
	    compute(cf, id, ctx);
	    cf.storeLocal(rhs);
	    cf.loadLocal(tgt);
	    cf.iloadLocal(num);
	    cf.loadLocal(rhs);
	    cf.add(INVOKESTATIC, "java.lang.reflect.Array", "set", "(Ljava/lang/Object;ILjava/lang/Object;)", "V");
	    cf.loadLocal(rhs);
	}
	cf.freeLocal(tgt);
	cf.freeLocal(num);
	next.fix();
    }

    void assignRange(ClassFile cf, int id, int ctx, int rhs, SimpleNode lhs, CompileContext cc){
	accept(lhs, 0, cc);
	int tgt = cf.getLocal();
	cf.storeLocal(tgt);

	accept(lhs, 1, cc);
	int idx1 = cf.getLocal();
	cf.storeLocal(idx1);

	int idx2 = cf.getLocal();
	if (lhs.jjtGetNumChildren() >= 3){
	    accept(lhs, 2, cc);
	} else {
	    cf.add(ACONST_NULL);
	}
	cf.storeLocal(idx2);

	Label err = cf.getLabel();
	Label ok = cf.getLabel();
	cf.loadLocal(idx1);
	cf.add(INSTANCEOF, "java.lang.Number");
	cf.add(IFEQ, err);

	cf.loadLocal(idx2);
	cf.add(IFNULL, ok);
	cf.loadLocal(idx2);
	cf.add(INSTANCEOF, "java.lang.Number");
	cf.add(IFNE, ok);

	err.fix();
	error(cf, "number.expected", cc);

	ok.fix();
	cf.loadLocal(tgt);
	cf.loadLocal(idx1);
	cf.freeLocal(idx1);
	cf.add(CHECKCAST, "java.lang.Number");
	cf.loadLocal(idx2);
	cf.freeLocal(idx2);
	cf.add(CHECKCAST, "java.lang.Number");
	cf.loadLocal(rhs);

	cf.add(INVOKESTATIC,
	       "pnuts.lang.Runtime",
	       "assignRange",
	       "(Ljava/lang/Object;Ljava/lang/Number;Ljava/lang/Number;Ljava/lang/Object;)",
	       "Ljava/lang/Object;");

	SimpleNode ch = lhs.jjtGetChild(0);
	if (ch.id == JJTIDNODE ||
	    ch.id == JJTGLOBAL ||
	    ch.id == JJTINDEXNODE ||
	    ch.id == JJTMEMBERNODE ||
	    ch.id == JJTSTATICMEMBERNODE)
	{
	    int assigned = cf.getLocal();
	    cf.storeLocal(assigned);
	    cf.loadLocal(tgt);
	    cf.freeLocal(tgt);
	    cf.add(INSTANCEOF, "java.lang.String");
	    Label no = cf.getLabel();
	    cf.add(IFEQ, no);
	    cf.loadLocal(assigned);
	    _assign(ch, cc, id, true);
	    cf.add(POP);
	    no.fix();
	    cf.loadLocal(assigned);
	    cf.freeLocal(assigned);
	}
    }

    void assignStaticMember(ClassFile cf, int id, int ctx, int rhs, SimpleNode lhs, CompileContext cc){
	SimpleNode c1 = lhs.jjtGetChild(0);
	String pkgName = getPackageName(c1);
	cf.add(LDC, cf.addConstant(pkgName));
	cf.add(INVOKESTATIC, "pnuts.lang.Package", "find", "(Ljava/lang/String;)", "Lpnuts/lang/Package;");
	int tgt = cf.getLocal();
	cf.storeLocal(tgt);
	cf.loadLocal(tgt);
	Label l_null = cf.getLabel();
	cf.add(IFNULL, l_null);
	cf.loadLocal(tgt);
	cf.add(LDC, cf.addConstant(lhs.str));
	cf.loadLocal(rhs);
	cf.loadLocal(ctx);
	cf.add(INVOKEVIRTUAL,
	       "pnuts.lang.Package",
	       "set",
	       "(Ljava/lang/String;Ljava/lang/Object;Lpnuts/lang/Context;)",
	       "V");
	cf.loadLocal(rhs);
	Label next = cf.getLabel();
	cf.add(GOTO, next);
	l_null.fix();
	c1.accept(this, cc);
	cf.storeLocal(tgt);
	cf.loadLocal(tgt);
	cf.add(INSTANCEOF, "java.lang.Class");
	Label err = cf.getLabel();
	cf.add(IFEQ, err);
	if (id != JJTASSIGNMENT){
	    cf.loadLocal(ctx);
	    cf.loadLocal(tgt);
	    cf.add(CHECKCAST, "java.lang.Class");
	    cf.add(LDC, cf.addConstant(lhs.str));
	    cf.add(INVOKESTATIC,
		   "pnuts.lang.Runtime",
		   "getStaticField",
		   "(Lpnuts/lang/Context;Ljava/lang/Class;Ljava/lang/String;)",
		   "Ljava/lang/Object;");
	    cf.loadLocal(rhs);
	    compute(cf, id, ctx);
	    cf.storeLocal(rhs);
	}
	cf.loadLocal(ctx);
	cf.loadLocal(tgt);
	cf.freeLocal(tgt);
	cf.add(CHECKCAST, "java.lang.Class");
	cf.add(LDC, cf.addConstant(lhs.str));
	cf.loadLocal(rhs);
	cf.add(INVOKESTATIC,
	       "pnuts.lang.Runtime",
	       "putStaticField",
	       "(Lpnuts/lang/Context;Ljava/lang/Class;Ljava/lang/String;Ljava/lang/Object;)",
	       "V");
	cf.loadLocal(rhs);
	cf.add(GOTO, next);
	err.fix();
	error(cf, "package.notFound", cc);
	next.fix();
    }

    void assignGlobal(ClassFile cf, int id, int ctx, int rhs, SimpleNode lhs, CompileContext cc){
	cf.add(INVOKESTATIC, "pnuts.lang.Package", "getGlobalPackage", "()", "Lpnuts/lang/Package;");
	if (id != JJTASSIGNMENT){
	    int gbl = cf.getLocal();
	    cf.storeLocal(gbl);
	    cf.loadLocal(gbl);
	    cf.add(LDC, cf.addConstant(lhs.str));
	    cf.loadLocal(ctx);
	    cf.add(INVOKEVIRTUAL,
		   "pnuts.lang.Package",
		   "get",
		   "(Ljava/lang/String;Lpnuts/lang/Context;)",
		   "Ljava/lang/Object;");
	    cf.loadLocal(rhs);
	    compute(cf, id, ctx);
	    cf.storeLocal(rhs);
	    cf.loadLocal(gbl);
	    cf.freeLocal(gbl);
	}
	cf.add(LDC, cf.addConstant(lhs.str));
	cf.loadLocal(rhs);
	cf.loadLocal(ctx);
	cf.add(INVOKEVIRTUAL,
	       "pnuts.lang.Package",
	       "set",
	       "(Ljava/lang/String;Ljava/lang/Object;Lpnuts/lang/Context;)",
	       "V");
	cf.loadLocal(rhs);
    }

    void error(ClassFile cf, String keyword, CompileContext cc){
	cf.add(NEW, "pnuts.lang.PnutsException");
	cf.add(DUP);
	cf.add(LDC, cf.addConstant(keyword));
	cf.add(GETSTATIC, cc.constClassName, "NO_PARAM", "[Ljava/lang/Object;");

	int ctx = cc.getContextIndex();
	cf.loadLocal(ctx);
	cf.add(INVOKESPECIAL,
	       "pnuts.lang.PnutsException",
	       "<init>",
	       "(Ljava/lang/String;[Ljava/lang/Object;Lpnuts/lang/Context;)",
	       "V");
	cf.add(ATHROW);
    }

    void compute(ClassFile cf, int id, int ctx){
	int operator = 0;
	switch (id){
	case JJTASSIGNMENTTA:
	    operator = Arithmetic.MULTIPLY;
	    break;
	case JJTASSIGNMENTMA:
	    operator = Arithmetic.MOD;
	    break;
	case JJTASSIGNMENTDA:
	    operator = Arithmetic.DIVIDE;
	    break;
	case JJTASSIGNMENTPA:
	    operator = Arithmetic.ADD;
	    break;
	case JJTASSIGNMENTSA:
	    operator = Arithmetic.SUBTRACT;
	    break;
	case JJTASSIGNMENTLA:
	    operator = Arithmetic.SHIFT_L;
	    break;
	case JJTASSIGNMENTRA:
	    operator = Arithmetic.SHIFT_R;
	    break;
	case JJTASSIGNMENTRAA:
	    operator = Arithmetic.SHIFT_A;
	    break;
	case JJTASSIGNMENTAA:
	    operator = Arithmetic.AND;
	    break;
	case JJTASSIGNMENTEA:
	    operator = Arithmetic.XOR;
	    break;
	case JJTASSIGNMENTOA:
	    operator = Arithmetic.OR;
	    break;
	default:
	    throw new RuntimeException("never happen");
	}
	cf.pushInteger(operator);
	cf.loadLocal(ctx);
	cf.add(INVOKESTATIC,
	       "pnuts.lang.Runtime",
	       "binary",
	       "(Ljava/lang/Object;Ljava/lang/Object;ILpnuts/lang/Context;)",
	       "Ljava/lang/Object;");
    }

    /**
     * Create a class loader to load compiled code.
     */
    static CodeLoader createCodeLoader(ClassLoader loader, boolean privileged){
	if (java2 && privileged){
	    return privilegedCodeLoaderFactory.create(loader);
	} else {
	    return codeLoaderFactory.create(loader);
	}
    }

    public static CodeLoader createCodeLoader(ClassLoader loader){
	return createCodeLoader(loader, false);
    }


    static class CodeLoaderFactory {
	public CodeLoader create(ClassLoader parent){
	    return new CodeLoader0(parent);
	}
    }

    static class Java2CodeLoaderFactory extends CodeLoaderFactory {
	public CodeLoader create(final ClassLoader parent){
	    return new CodeLoader(parent);
	}
    }

    static class PrivilegedCodeLoaderFactory extends CodeLoaderFactory {
	public CodeLoader create(final ClassLoader parent){
	    return (CodeLoader)AccessController.doPrivileged(new PrivilegedAction(){
		public Object run(){
		    return new CodeLoader(parent);
		} 
	    });
	}
    }

    /**
     * Compile a parsed expression.
     * The resulting bytecode is handled by a ClassFileHandler.
     */
    public Object compile(Pnuts pnuts, ClassFileHandler handler) {
	automatic = false;
	CompileContext cc = new CompileContext();
	pnuts.accept(this, cc);
	handler.handle(cc, cc.getClassFile());
	for (Enumeration e = cc.getClassFiles(); e.hasMoreElements(); ){
	    ClassFile cf = (ClassFile)e.nextElement();
	    handler.handle(cc, cf);
	}
	return handler.getResult(cc);
    }

    static class LoadHandler implements ClassFileHandler {
	CodeLoader loader;
	Class clazz = null;

	LoadHandler(CodeLoader loader){
	    this.loader = loader;
	}

	public void handle(CompileContext cc, ClassFile cf) {
	    try {
		Class c = cc.load(cf, loader);
		if (clazz == null){
		    clazz = c;
		}
	    } catch (IOException e){
		throw new PnutsException(e);
	    }
	}

	public Object getResult(CompileContext cc){
	    cc.resolve(loader);
	    return clazz;
	}
    }

    /**
     * Compile a parsed expression.
     *
     * @param pnuts	a parsed expression to be compiled
     * @param context	a context in which the expression is compiled.  
     * @return		a Pnuts object
     */
    public Pnuts compile(Pnuts pnuts, Context context){
	Throwable caught = null;
	try {
	    final CodeLoader loader = createCodeLoader(Pnuts.getClassLoader(context), true);
	    CompileContext cc = new CompileContext();
	    Class cls = (Class)compile(pnuts, new LoadHandler(loader));
	    final Runtime rt = (Runtime)cls.newInstance();

	    class CompiledPnuts extends Pnuts implements Compiled {
		public Object accept(Visitor visitor, Context context){
		    if (visitor instanceof Compiler){
			Compiler c = (Compiler)visitor;
			if (c.automatic){
			    return run(context);
			} else {
			    return this;
			}
		    } else {
			throw new RuntimeException("unsupported operation for compiled code");
		    }
		}

		public Object run(Context context){
		    return rt.execute(context);
		}
	    }
	    return new CompiledPnuts();
	} catch (Throwable t){
	    caught = t;
	}
	Runtime.checkException(context, caught);
	throw new PnutsException(caught, context);
    }

    /**
     * Compile an expression.
     *
     * @param expression	an expression to be compiled.
     * @return 			a Pnuts object.
     */
    public Pnuts compile(String expression){
	Context context = new Context();
	try {
	    return compile(Pnuts.parse(expression), context);
	} catch (ParseException e){
	    Runtime.checkException(context, e);
	    throw new PnutsException(e, context);
	}
    }

    /**
     * Compile an expression.
     * @param expression	an expression to be compiled
     * @param context		a context in which the expression is compiled.
     * @return			a Pnuts object.
     */
    public Pnuts compile(String expression, Context context){
	try {
	    return compile(Pnuts.parse(expression), context);
	} catch (ParseException e){
	    Runtime.checkException(context, e);
	    throw new PnutsException(e, context);
	}
    }

    /**
     * Compile a function group.
     * The resulting bytecode is handled by a ClassFileHandler.
     *
     * @param pf	a function group.
     * @param handler   the resulting bytecode is handle by this object.
     * @return 		the result of handler.getResult() method.
     */
    public Object compile(PnutsFunction pf, ClassFileHandler handler) {
	PnutsFunction ret = null;
	StringBuffer buf = new StringBuffer();
	for (Enumeration e = getFunctions(pf); e.hasMoreElements(); ){
	    Function f = (Function)e.nextElement();
	    int nargs = f.getNumberOfParameter();
	    String im[] = f.getImportEnv();
	    buf.append("import(null);");
	    for (int i = 0; i < im.length; i++){
		buf.append("import(\"" + im[i] + "\");");
	    }
	    buf.append(pf.unparse(nargs));
	    buf.append('\n');
	}
	try {
	    String expr = buf.toString();
	    if (DEBUG){
		System.out.println(expr);
	    }
	    return compile(Pnuts.parse(expr), handler);
	} catch (ParseException pe){
	    throw new PnutsException(pe);
	}
    }

    /**
     * Compile a function group
     *
     * @param   pf	a function group to be compiled.
     * @return	a compiled version of the function group
     */
    public PnutsFunction compile(PnutsFunction pf){
	return compile(pf, new Context());
    }

    /**
     * Compile a function group
     *
     * @param pf	a function group to be compiled.
     * @param context	a context in which the function is compiled.
     * @return	a compiled version of the function group
     */
    public PnutsFunction compile(PnutsFunction pf, Context context){
        StringBuffer buf = new StringBuffer();
	for (Enumeration e = getFunctions(pf); e.hasMoreElements(); ){
	    Function f = (Function)e.nextElement();
	    int nargs = f.getNumberOfParameter();
	    String im[] = f.getImportEnv();
	    buf.append("import(null);");
	    for (int i = 0; i < im.length; i++){
		buf.append("import(\"" + im[i] + "\");");
	    }
	    buf.append(pf.unparse(nargs));
	    buf.append('\n');
	}
	Context c = (Context)context.clone();
	setPackage(pf.getPackage(), c);
	return (PnutsFunction)compile(buf.toString(), context).run(c);
    }

    /**
     * Check if an object is compiled
     *
     * @param obj an object to be checked
     * @return true if the object is compiled
     */
    public static boolean isCompiled(Object obj){
	if (obj instanceof Compiled){
	    return true;
	}
	if (obj instanceof PnutsFunction){
	    for (Enumeration e = getFunctions((PnutsFunction)obj); e.hasMoreElements(); ){
		Function f = (Function)e.nextElement();
		if (!(f instanceof Compiled)){
		    return false;
		}
	    }
	    return true;
	}
	return false;
    }
}
