/*
 * @(#) JavaAdapter.java	1.8 00/08/01
 *
 * Copyright (c) 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.compiler;

import pnuts.lang.Pnuts;
import pnuts.lang.PnutsException;
import pnuts.lang.PnutsFunction;
import pnuts.lang.Function;
import pnuts.lang.Context;
import pnuts.lang.Package;
import pnuts.lang.Runtime;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.io.ByteArrayOutputStream;
import java.io.OutputStream;
import java.io.IOException;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.security.AccessController;
import java.security.PrivilegedAction;

/**
 * This class provides a way to extends a Java class in Pnuts.
 */
public class JavaAdapter extends Runtime implements Constants, Opcode {

    private final static boolean DEBUG = false;

    protected JavaAdapter(){
    }

    /**
     * Generates a class that extends the superClass and implements the interfaces.
     * Functions in the pkg are mapped to non-final/static public/protected methods
     * with the same name and the same number of parameters. Note that this method may
     * throw ClassFormatError.
     *
     * @param superClass super class that the generated class extends
     * @param interfaces list of interfaces that the generated class implements
     * @param pkg a Package that includes functions to be mapped to methods.
     * @param context a Context in which functions are extracted from pkg
     * @return the generated class
     */
    public static Class generateSubtype(Class superClass, Class[] interfaces, Package pkg, Context context)
	throws ClassNotFoundException
    {
	try {
	    ByteArrayOutputStream bout = new ByteArrayOutputStream();
	    if (superClass == null){
		superClass = Object.class;
	    }
	    String superClassName = superClass.getName();
	    String className = superClassName.replace('.', '_') + "__adapter";
	    ClassFile cf = getClassFileForSubtype(className, superClass, interfaces, pkg, context);
	    cf.write(bout);
	    return defineClass(className, bout.toByteArray(), superClass.getClassLoader());
	} catch (IOException io){
	    return null;
	}
    }

    /**
     * Generate an adapter class for bean events.
     * The generated class implements the the specified <em>listenerType</em>.
     * The constructor has two arguments (PnutsFunction, Context).  The method <em>methodName</em>
     * calls a function which is specified as a parameter of the constructor.
     *
     * @param listenerType a Class object of a EventListener subclass
     * @param methodName a method name of the listener type.
     * @return a Class object of the generated class
     */
    public static Class generateEventAdapter(Class listenerType, String methodName){
	try {
	    ByteArrayOutputStream bout = new ByteArrayOutputStream();
	    ClassFile cf = getClassFileForEventAdapter("__event_adapter__", listenerType, methodName);
	    cf.write(bout);
	    return defineClass(cf.getClassName(), bout.toByteArray(), null);
	} catch (IOException io){
	    return null;
	}
    }

    /**
     * Creates a class file image of a subtype of superClass (or some interfaces) and writes
     * to the output stream.
     *
     * @param superClass super class that the generated class extends
     * @param interfaces list of interfaces that the generated class implements
     * @param pkg a Package that includes functions to be mapped to methods.
     * @param context a Context in which functions are extracted from pkg
     */
    public static ClassFile getClassFileForSubtype(String className,
						   Class superClass,
						   Class[] interfaces,
						   Package pkg,
						   Context context)
	throws ClassNotFoundException
    {
	if (superClass == null){
	    superClass = Object.class;
	}
	ClassFile cf = new ClassFile(className,
				     superClass.getName(),
				     null,
				     ACC_PUBLIC);
	if (interfaces != null){
	    for (int i = 0; i < interfaces.length; i++){
		cf.addInterface(interfaces[i].getName());
	    }
	}
	Hashtable functions = new Hashtable(8);
	Hashtable methodNames = new Hashtable(8);
	Hashtable duplicates = new Hashtable(4);

	for (Enumeration e = pkg.keys(); e.hasMoreElements(); ){
	    String name = (String)e.nextElement();

	    Object value = pkg.get(name, context);
	    if (!(value instanceof PnutsFunction)){
		continue;
	    }
	    functions.put(value, name);

	    String methodName = toMethodName(name);
	    int idx = methodName.indexOf('(');
	    if (idx >= 0){
		methodName = methodName.substring(0, idx);
	    }
	    if (methodNames.get(methodName) != null){
		duplicates.put(methodName, methodName);
	    } else {
		methodNames.put(methodName, name);
	    }
	}

	constructor(cf, superClass, functions, duplicates);

	for (Enumeration e = functions.keys(); e.hasMoreElements(); ){
	    PnutsFunction func = (PnutsFunction)e.nextElement();
	    String sig = (String)functions.get(func);
	    String name = toMethodName(sig);

	    int idx = sig.indexOf('(');
	    Vector types = null;
	    if (idx >= 0){
		name = sig.substring(0, idx);
		types = new Vector();
		parseParameterSignature(sig.substring(idx), types, context);
	    }
	    for (Enumeration ee = getFunctions(func); ee.hasMoreElements(); ){
		Function f = (Function)ee.nextElement();
		int narg = f.getNumberOfParameter();
		Method m = findMethod(name, narg, superClass, interfaces, types);
		boolean dup = (duplicates.get(name) != null);
		if (m != null){
//		    define(cf, m.getParameterTypes(), m.getReturnType(), m.getModifiers(), name, func, sig, dup, context);
		    define(cf, m.getParameterTypes(), m.getReturnType(), Modifier.PUBLIC, name, func, sig, dup, context);
		} else {
		    Class[] parameterTypes = new Class[narg];
		    if (types == null){
			for (int i = 0; i < narg; i++){
			    parameterTypes[i] = Object.class;
			}
		    } else {
			for (int i = 0; i < narg; i++){
			    parameterTypes[i] = (Class)types.elementAt(i);
			}
		    }
		    define(cf, parameterTypes, Object.class, Modifier.PUBLIC, name, func, sig, dup, context);
		}
	    }
	}
	return cf;
    }

    static String toMethodName(String name){
	int idx = name.lastIndexOf('.');
	if (idx > 0){
	    name = name.substring(idx + 1);
	}
	return name;
    }

    private static void constructor(ClassFile cf, Class superClass, Hashtable functions, Hashtable duplicates){
	String className = cf.getClassName();

	cf.openMethod("<init>", "()V", ACC_PUBLIC);
	cf.add(ALOAD_0);
	cf.add(INVOKESPECIAL, superClass.getName(), "<init>", "()", "V");
	cf.add(RETURN);
	cf.closeMethod();

	cf.openMethod("<init>", "(Lpnuts/lang/Context;Lpnuts/lang/Package;)V", ACC_PUBLIC);
	cf.add(ALOAD_0);
	cf.add(INVOKESPECIAL, superClass.getName(), "<init>", "()", "V");

	cf.addField("_context", "Lpnuts/lang/Context;", (short)(ACC_PRIVATE | ACC_STATIC));
	cf.add(ALOAD_1);
	cf.add(PUTSTATIC, className, "_context", "Lpnuts/lang/Context;");

	cf.addField("_package", "Lpnuts/lang/Package;", (short)(ACC_PRIVATE | ACC_STATIC));
	cf.add(ALOAD_2);
	cf.add(PUTSTATIC, className, "_package", "Lpnuts/lang/Package;");

	for (Enumeration e = functions.keys(); e.hasMoreElements(); ){
	    PnutsFunction func = (PnutsFunction)e.nextElement();
	    String sig = (String)functions.get(func);
	    String methodName = toMethodName(sig);
	    int idx = sig.indexOf('(');
	    if (idx >= 0){
		methodName = sig.substring(0, idx);
	    }
	    if (duplicates.get(methodName) == null){
		cf.addField(methodName, "Lpnuts/lang/PnutsFunction;", (short)(ACC_PRIVATE | ACC_STATIC));
	    }
	}

	cf.add(RETURN);
	cf.closeMethod();
    }

    private static Method findMethod(String name, int narg, Class superClass, Class[] interfaces, Vector types)
	throws ClassNotFoundException
    {
	if (DEBUG){
	    System.out.println("findMethod(" + name + ", " + narg + ", " + superClass + ", " + interfaces);
	}

	if (interfaces != null){
	loop1:
	    for (int i = 0; i < interfaces.length; i++){
		Class _interface = interfaces[i];
		Method[] methods = _interface.getMethods(); 
		for (int j = 0; j < methods.length; j++){
		    Method m = methods[j];
		    int modifiers = m.getModifiers();
		    if (!Modifier.isPublic(modifiers) && !Modifier.isProtected(modifiers) ||
			Modifier.isStatic(modifiers) || 
			Modifier.isFinal(modifiers))
		    {
			continue;
		    }
		    Class[] parameterTypes = m.getParameterTypes();
		    if (m.getName().equals(name) && parameterTypes.length == narg){
			if (types == null){
			    return m;
			} else {
			    for (int k = 0; k < narg; k++){
				if (!types.elementAt(k).equals(parameterTypes[k])){
				    continue loop1;
				}
			    }
			    return m;
			}
		    }
		}
	    }
	}
	if (superClass == null){
	    superClass = Object.class;
	}
	while (superClass != null){
	    Method[] methods = superClass.getDeclaredMethods();
	loop2:
	    for (int j = 0; j < methods.length; j++){
		Method m = methods[j];
		int modifiers = m.getModifiers();
		if (!Modifier.isPublic(modifiers) && !Modifier.isProtected(modifiers) ||
		    Modifier.isStatic(modifiers) || 
		    Modifier.isFinal(modifiers))
		{
		    continue;
		}
		Class[] parameterTypes = m.getParameterTypes();
		if (m.getName().equals(name) && parameterTypes.length == narg){
		    if (types == null){
			return m;
		    } else {
			for (int k = 0; k < narg; k++){
			    if (!types.elementAt(k).equals(parameterTypes[k])){
				break loop2;
			    }
			}
			return m;
		    }
		}
	    }
	    superClass = superClass.getSuperclass();
	}
	return null;
    }

    private static void define(ClassFile cf,
			       Class[] parameterTypes,
			       Class returnType,
			       int modifiers,
			       String methodName,
			       PnutsFunction func,
			       String signature,
			       boolean dup,
			       Context context)
    {
	String className = cf.getClassName();
	cf.openMethod(methodName, makeSignature(parameterTypes, returnType), (short)(modifiers & (ACC_PUBLIC | ACC_PROTECTED)));

	int f = 0;
	Label next = null;
	if (!dup){
	    cf.add(GETSTATIC, className, methodName, "Lpnuts/lang/PnutsFunction;");
	    f = cf.getLocal();
	    cf.storeLocal(f);
	    next = cf.getLabel();
	    cf.loadLocal(f);
	    cf.add(IFNONNULL, next);
	}
	cf.add(GETSTATIC, className, "_package", "Lpnuts/lang/Package;");
	cf.add(LDC, cf.addConstant(signature));
	cf.add(GETSTATIC, className, "_context", "Lpnuts/lang/Context;");
	cf.add(INVOKEVIRTUAL, "pnuts.lang.Package", "get", "(Ljava/lang/String;Lpnuts/lang/Context;)", "Ljava/lang/Object;");
	cf.add(CHECKCAST, "pnuts.lang.PnutsFunction");

	if (!dup){
	    cf.storeLocal(f);
	    cf.loadLocal(f);
	    cf.add(PUTSTATIC, className, methodName, "Lpnuts/lang/PnutsFunction;");
	    next.fix();
	    cf.loadLocal(f);
	}

	int nargs = parameterTypes.length;
	if (nargs <= -0x8000 || nargs >= 0x8000){
	    throw new PnutsException("too many parameters", context);
	}
	pushInteger(cf, nargs);
	cf.add(ANEWARRAY, "java.lang.Object");
	for (int i = 0; i < nargs; i++){
	    cf.add(DUP);
	    pushInteger(cf, i);
	    if (parameterTypes[i].isPrimitive()){
		loadPrimitive(cf, parameterTypes[i], i + 1);
	    } else {
		cf.loadLocal(i + 1);
	    }
	    cf.add(AASTORE);
	}
	cf.add(GETSTATIC, className, "_context", "Lpnuts/lang/Context;");
	cf.add(INVOKEVIRTUAL, "pnuts.lang.PnutsFunction", "call", "([Ljava/lang/Object;Lpnuts/lang/Context;)", "Ljava/lang/Object;");
	if (returnType == void.class){
	    cf.add(POP);
	    cf.add(RETURN);
	} else if (returnType.isPrimitive()){
	    returnPrimitive(cf, returnType);
	} else {
	    if (returnType != Object.class){
		cf.add(CHECKCAST, returnType.getName());
	    }
	    cf.add(ARETURN);
	}
	cf.closeMethod();
    }

    private static void loadPrimitive(ClassFile cf, Class primitive, int index){
	if (primitive == int.class){
	    cf.add(NEW, "java.lang.Integer");
	    cf.add(DUP);
	    cf.iloadLocal(index);
	    cf.add(INVOKESPECIAL, "java.lang.Integer", "<init>", "(I)", "V");
	} else if (primitive == byte.class){
	    cf.add(NEW, "java.lang.Byte");
	    cf.add(DUP);
	    cf.iloadLocal(index);
	    cf.add(INVOKESPECIAL, "java.lang.Byte", "<init>", "(B)", "V");
	} else if (primitive == short.class){
	    cf.add(NEW, "java.lang.Short");
	    cf.add(DUP);
	    cf.iloadLocal(index);
	    cf.add(INVOKESPECIAL, "java.lang.Short", "<init>", "(S)", "V");
	} else if (primitive == char.class){
	    cf.add(NEW, "java.lang.Character");
	    cf.add(DUP);
	    cf.iloadLocal(index);
	    cf.add(INVOKESPECIAL, "java.lang.Character", "<init>", "(C)", "V");
	} else if (primitive == long.class){
	    cf.add(NEW, "java.lang.Long");
	    cf.add(DUP);
	    cf.lloadLocal(index);
	    cf.add(INVOKESPECIAL, "java.lang.Long", "<init>", "(J)", "V");
	} else if (primitive == float.class){
	    cf.add(NEW, "java.lang.Float");
	    cf.add(DUP);
	    cf.floadLocal(index);
	    cf.add(INVOKESPECIAL, "java.lang.Float", "<init>", "(F)", "V");
	} else if (primitive == double.class){
	    cf.add(NEW, "java.lang.Double");
	    cf.add(DUP);
	    cf.dloadLocal(index);
	    cf.add(INVOKESPECIAL, "java.lang.Double", "<init>", "(D)", "V");
	} else if (primitive == boolean.class){
	    cf.add(NEW, "java.lang.Boolean");
	    cf.add(DUP);
	    cf.iloadLocal(index);
	    cf.add(INVOKESPECIAL, "java.lang.Boolean", "<init>", "(Z)", "V");
	}
    }

    private static void returnPrimitive(ClassFile cf, Class type){
	if (type == int.class){
	    cf.add(CHECKCAST, "java.lang.Integer");
	    cf.add(INVOKEVIRTUAL, "java.lang.Integer", "intValue", "()", "I");
	    cf.add(IRETURN);
	} else if (type == byte.class){
	    cf.add(CHECKCAST, "java.lang.Byte");
	    cf.add(INVOKEVIRTUAL, "java.lang.Byte", "byteValue", "()", "B");
	    cf.add(IRETURN);
	} else if (type == short.class){
	    cf.add(CHECKCAST, "java.lang.Short");
	    cf.add(INVOKEVIRTUAL, "java.lang.Short", "shortValue", "()", "S");
	    cf.add(IRETURN);
	} else if (type == char.class){
	    cf.add(CHECKCAST, "java.lang.Character");
	    cf.add(INVOKEVIRTUAL, "java.lang.Character", "charValue", "()", "C");
	    cf.add(IRETURN);
	} else if (type == long.class){
	    cf.add(CHECKCAST, "java.lang.Long");
	    cf.add(INVOKEVIRTUAL, "java.lang.Long", "longValue", "()", "L");
	    cf.add(LRETURN);
	} else if (type == float.class){
	    cf.add(CHECKCAST, "java.lang.Float");
	    cf.add(INVOKEVIRTUAL, "java.lang.Float", "floatValue", "()", "F");
	    cf.add(FRETURN);
	} else if (type == double.class){
	    cf.add(CHECKCAST, "java.lang.Double");
	    cf.add(INVOKEVIRTUAL, "java.lang.Double", "doubleValue", "()", "D");
	    cf.add(DRETURN);
	} else if (type == boolean.class){
	    cf.add(CHECKCAST, "java.lang.Boolean");
	    cf.add(INVOKEVIRTUAL, "java.lang.Boolean", "booleanValue", "()", "Z");
	    cf.add(IRETURN);
	} else {
	    throw new InternalError();
	}
    }

    private static void parseParameterSignature(String signature, Vector types, Context context)
	throws ClassNotFoundException
    {
	char[] c = signature.toCharArray();
	int index = 1;
	int dim = 0;
	boolean returnPart = false;
    loop:
        while (index < c.length){
            switch (c[index]) {
	    case 'V':
		types.addElement(void.class);
		break loop;
	    case 'J':
		types.addElement(Runtime.arrayType(Long.TYPE, dim));
		dim = 0;
		index++;
		break;
	    case 'D':
		types.addElement(Runtime.arrayType(Double.TYPE, dim));
		dim = 0;
		index++;
		break;
	    case 'B':
		types.addElement(Runtime.arrayType(Byte.TYPE, dim));
		dim = 0;
		index++;
		break;
	    case 'S':
		types.addElement(Runtime.arrayType(Short.TYPE, dim));
		dim = 0;
		index++;
		break;
	    case 'C':
		types.addElement(Runtime.arrayType(Character.TYPE, dim));
		dim = 0;
		index++;
		break;
	    case 'I':
		types.addElement(Runtime.arrayType(Integer.TYPE, dim));
		dim = 0;
		index++;
		break;
	    case 'Z':
		types.addElement(Runtime.arrayType(Boolean.TYPE, dim));
		dim = 0;
		index++;
		break;
	    case 'F':
		types.addElement(Runtime.arrayType(Float.TYPE, dim));
		dim = 0;
		index++;
		break;
	    case '[':
		while (c[index] == '['){
		    dim++;
		    index++;
		}
		break;
	    case 'L':
	        int start = index + 1;
		while (c[index++] != ';') {}
		String cn = new String(c, start, index - start - 1);
		types.addElement(Runtime.arrayType(Pnuts.loadClass(cn.replace('/', '.'), context) , dim));
		dim = 0;
		break;
	    case ')':
		returnPart = true;
		index++;
		continue loop;
	    default:
		throw new PnutsException("illegal method signature", context);
            }
	    if (returnPart){
		break loop;
	    }
        }
    }

    private static String makeSignature(Class[] parameterTypes, Class returnType){
	StringBuffer buf = new StringBuffer("(");
	for (int i = 0; i < parameterTypes.length; i++){
	    Class type = parameterTypes[i];
	    buf.append(makeSignature(type));
	}
	buf.append(")");
	buf.append(makeSignature(returnType));
	return buf.toString();
    }

    private static String makeSignature(Class type){
	if (type == int.class){
	    return "I";
	} else if (type == byte.class){
	    return "B";
	} else if (type == long.class){
	    return "J";
	} else if (type == char.class){
	    return "C";
	} else if (type == short.class){
	    return "S";
	} else if (type == float.class){
	    return "F";
	} else if (type == double.class){
	    return "D";
	} else if (type == boolean.class){
	    return "Z";
	} else if (type == void.class){
	    return "V";
	} else {
	    if (type.isArray()){
		return "[" + makeSignature(type.getComponentType());
	    } else {
		return "L" + type.getName().replace('.', '/') + ";";
	    }
	}
    }

    private static void pushInteger(ClassFile cf, int number){
	if (number >= -1 && number <= 5){
	    cf.add((byte)(ICONST_0 + number));
	} else if (number > -0x80 && number < 0x80){
	    cf.add(BIPUSH, number);
	} else {
	    cf.add(SIPUSH, number);
	}
    }

    /**
     * Creates a class file of a event adapter that implements the specified listenerType.
     * The constructor of the created class takes PnutsFunction and Context as the parameters.
     * When the corresponding event is observed, the PnutsFunction is called in the Context.
     *
     * @param className the class name of the event adapter
     * @param listenerType a Class object of a EventListener subclass
     * @param methodName a method name of the listener type.
     */
    public static ClassFile getClassFileForEventAdapter(String className,
							Class listenerType,
							String methodName)
    {
	/*
	if (!EventListener.class.isAssignableFrom(listenerType)){
	    throw new RuntimeException("EventListener type is required:" + listenerType);
	}
	*/
	ClassFile cf = new ClassFile(className, "java.lang.Object", null, ACC_PUBLIC);
	cf.addInterface(listenerType.getName());
	cf.addField("context", "Lpnuts/lang/Context;", ACC_PRIVATE);
	cf.addField("function", "Lpnuts/lang/PnutsFunction;", ACC_PRIVATE);

	cf.openMethod("<init>", "(Lpnuts/lang/PnutsFunction;Lpnuts/lang/Context;)V", ACC_PUBLIC);
	cf.add(ALOAD_0);
	cf.add(INVOKESPECIAL, "java.lang.Object", "<init>", "()", "V");
	cf.add(ALOAD_0);
	cf.add(ALOAD_1);
	cf.add(PUTFIELD, cf.getClassName(), "function", "Lpnuts/lang/PnutsFunction;");
	cf.add(ALOAD_0);
	cf.add(ALOAD_2);
	cf.add(PUTFIELD, cf.getClassName(), "context", "Lpnuts/lang/Context;");
	cf.add(RETURN);
	cf.closeMethod();

	Method[] methods = listenerType.getMethods();
	for (int i = 0; i < methods.length; i++){
	    Method m = methods[i];
	    Class[] types = m.getParameterTypes();
	    if (m.getName().equals(methodName)){
		cf.openMethod(methodName, makeSignature(types, void.class), ACC_PUBLIC);
		cf.add(ALOAD_0);
		cf.add(GETFIELD, cf.getClassName(), "function", "Lpnuts/lang/PnutsFunction;");
		cf.add(ICONST_1);
		cf.add(ANEWARRAY, "java.lang.Object");
		cf.add(DUP);
		cf.add(ICONST_0);
		cf.add(ALOAD_1);
		cf.add(AASTORE);
		cf.add(ALOAD_0);
		cf.add(GETFIELD, cf.getClassName(), "context", "Lpnuts/lang/Context;");
		cf.add(INVOKEVIRTUAL, "pnuts.lang.PnutsFunction", "call", "([Ljava/lang/Object;Lpnuts/lang/Context;)", "Ljava/lang/Object;");
		cf.add(POP);
		cf.add(RETURN);
		cf.closeMethod();
	    } else {
		cf.openMethod(m.getName(), makeSignature(types, void.class), ACC_PUBLIC);
		cf.add(RETURN);
		cf.closeMethod();
	    }
	}
	return cf;
    }

    private static Class defineClass(String name, byte[] array, final ClassLoader parent){
	CodeLoader loader = null;
	if (Pnuts.isJava2()){
	    loader = (CodeLoader)AccessController.doPrivileged(new PrivilegedAction(){
		public Object run(){
		    return Compiler.createCodeLoader(parent);
		}
	    });
	} else {
	    loader = Compiler.createCodeLoader(parent);
	}
	Class cls = loader.define(name, array, 0, array.length);
	loader.resolve(cls);
	return cls;
    }
}
