/*
 * @(#) CachedPnutsImpl.java	1.1 00/08/01
 *
 * Copyright (c) 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.ext;

import java.io.File;
import java.io.FileInputStream;
import java.net.URL;
import java.net.URLConnection;
import pnuts.lang.Cache;
import pnuts.lang.Context;
import pnuts.lang.Pnuts;
import pnuts.lang.Runtime;
import pnuts.compiler.Compiler;
import pnuts.compiler.CompilerPnutsImpl;

/**
 * PnutsImpl which caches parsed (compiled) scripts and reuse them.
 * This class is useful when same scripts are executed over and over,
 * e.g. servlet scripts.
 *
 * @see pnuts.lang.PnutsImpl
 * @see pnuts.compiler.CompilerPnutsImpl
 */
public class CachedPnutsImpl extends CompilerPnutsImpl {
    private final static boolean DEBUG = false;
    private final static boolean java2 = Pnuts.isJava2();

    private boolean useCompiler;
    private Cache expr_cache;
    private Cache file_cache;

    public CachedPnutsImpl(){
	this(true);
    }

    /**
     * @param useCompiler true if compiler is used (default)
     */
    public CachedPnutsImpl(boolean useCompiler){
	this.useCompiler = useCompiler;
	reset();
    }

    /**
     * Reset the cache entries
     */
    public void reset(){
	if (java2){
	    expr_cache = new RefCache();
	    file_cache = new RefCache();
	} else {
	    expr_cache = new Cache(64, false);
	    file_cache = new Cache(64, false);
	}
    }

    /**
     * Evaluate an expression
     *
     * @param expr the expression to be evaluated
     * @param context the context in which the expression is evaluated
     * @see pnuts.lang.Pnuts#eval(String, pnuts.lang.Context)
     */
    public Object eval(String expr, Context context){
	Pnuts parsed = (Pnuts)expr_cache.get(expr);
	try {
	    if (parsed == null){
		parsed = Pnuts.parse(expr);
		if (useCompiler){
		    Compiler compiler = new Compiler(null, false, true);
		    if (DEBUG){
			System.out.println("compiling");
		    }
		    try {
			parsed = compiler.compile(parsed, context);
		    } catch (ClassFormatError cfe){
			/* ignore */
		    }
		}
		expr_cache.put(expr, parsed);
	    }
	    return parsed.run(context);
	} catch (Throwable t){
	    checkException(context, t);
	    return null;
	}
    }

    /**
     * Load a script file from a URL
     *
     * @param scriptURL the URL of the script
     * @param context the context in which the script is executed
     */    
    public Object load(URL scriptURL, Context context) {
	String protocol = scriptURL.getProtocol();
	Pnuts parsed = null;
	CacheEntry entry = null;

	try {
	    if ("file".equals(protocol)){
		String fileName = scriptURL.getFile();
		File file = new File(fileName);
		String canon = file.getCanonicalPath();
		long lastModified = file.lastModified();
		entry = (CacheEntry)file_cache.get(canon);
		if (entry == null ||
		    lastModified == 0 ||   // coundn't get the timestamp
		    entry.lastModified < lastModified)
		{
		    parsed = Pnuts.parse(new FileInputStream(file));
		    if (useCompiler){
			Compiler compiler = new Compiler(null, false, true);
			if (DEBUG){
			    System.out.println("compiling " + file);
			}
			try {
			    parsed = compiler.compile(parsed, context);
			} catch (ClassFormatError cfe){
			    /* ignore */
			}
		    }
		    if (lastModified != 0){
			entry = new CacheEntry();
			entry.lastModified = lastModified;
			entry.parsedExpression = parsed;
			file_cache.put(canon, entry);
		    }
		} else {
		    parsed = entry.parsedExpression;
		}
	    } else {
		URLConnection conn = scriptURL.openConnection();
		long lastModified = conn.getLastModified();
		entry = (CacheEntry)file_cache.get(scriptURL);
		if (entry == null ||
		    lastModified == 0 ||   // unknown
		    entry.lastModified < lastModified)
		{
		    parsed = Pnuts.parse(conn.getInputStream());
		    if (useCompiler){
			Compiler compiler = new Compiler(null, false, true);
			if (DEBUG){
			    System.out.println("compiling " + scriptURL);
			}
			try {
			    parsed = compiler.compile(parsed, context);
			} catch (ClassFormatError cfe){
				/* ignore */
			}
		    }
		    if (lastModified != 0){
			entry = new CacheEntry();
			entry.lastModified = lastModified;
			entry.parsedExpression = parsed;
			file_cache.put(scriptURL, entry);
		    }
		} else {
		    parsed = entry.parsedExpression;
		}
	    }
	    return parsed.run(context);
	} catch (Throwable t){
	    checkException(context, t);
	    return null;
	}
    }

    static class CacheEntry {
	long lastModified;
	Pnuts parsedExpression;
    }
}

