/*
 * @(#) ScriptPackage.java	1.5 00/08/05
 *
 * Copyright (c) 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.ext;

import pnuts.lang.Pnuts;
import pnuts.lang.Runtime;
import pnuts.lang.Context;
import pnuts.lang.Package;
import pnuts.lang.AbstractData;
import pnuts.lang.PnutsFunction;
import pnuts.lang.PnutsException;
import java.util.Enumeration;

/**
 * Package that is used in user's scripts as an hashtable.
 * Unlike pnuts.lang.Package, this class implements AbstractData.
 */
public class ScriptPackage extends Package implements AbstractData {

    public ScriptPackage(){
    }

    public ScriptPackage(String name){
	super(name);
    }

    public ScriptPackage(String name, Package parent){
	super(name, parent);
    }

    protected void init(Context context){
	super.init(context);
    }

    public Object invoke(String name, Object args[], Context context){
	Object o = get(name, context);
	if (o instanceof PnutsFunction){
	    return ((PnutsFunction)o).call(args, context);
	} else if (o instanceof Class){
	    return Runtime.callConstructor(context, (Class)o, args, null);
	} else {
	    throw new PnutsException("funcOrType.expected", new Object[]{Pnuts.format(o)}, context);
	}
    }

    /**
     * Defines a special function to create <a href="../../../doc/script_package.html">script packages</a>.
     */
    public static class Function extends PnutsFunction implements AbstractData {

	private String name;

	public Function(){
	    this("$");
	}

	public Function(String name){
	    this.name = name;
	}

	public String getName(){
	    return name;
	}

	/**
	 * @return null
	 */
	public Object get(String name, Context context){
	    return null;
	}

	/**
	 * Do nothing.
	 */
	public void set(String name, Object value, Context context){
	}

	/**
	 * Provides these methods:
	 *<pre>
	 * $.set(pkg, "name", value)
	 * $.get(pkg, "name")
	 * $.defined(pkg, "name")
	 * $.clear(pkg, "name")
	 * $.keys(pkg)
	 *</pre>
	 */
	public Object invoke(String name, Object[] args, Context context){
	    if ("set".equals(name)){
		if (args.length == 3){
		    Package pkg = (Package)args[0];
		    String sym = (String)args[1];
		    Object value = args[2];
		    pkg.set(sym.intern(), value, context);
		    return null;
		}
	    } else if ("get".equals(name)){
		if (args.length == 2){
		    Package pkg = (Package)args[0];
		    String sym = (String)args[1];
		    return pkg.get(sym.intern(), context);
		} 
	    } else if ("clear".equals(name)){
		if (args.length == 2){
		    Package pkg = (Package)args[0];
		    String sym = (String)args[1];
		    pkg.clear(sym.intern(), context);
		    return null;
		}
	    } else if ("defined".equals(name)){
		if (args.length == 2){
		    Package pkg = (Package)args[0];
		    String sym = (String)args[1];
		    boolean b = pkg.defined(sym.intern(), context);
		    return b ? Boolean.TRUE : Boolean.FALSE;
		}
	    } else if ("keys".equals(name)){
		if (args.length == 1){
		    Package pkg = (Package)args[0];
		    return pkg.keys();
		}
	    }
	    throw new PnutsException("method.notFound", new Object[]{name}, context);
	}

	/**
	 * Creates a <a href="../../../doc/script_package.html">script package</a>.
	 */
	protected Object exec(Object[] args, Context context){
	    if (args.length == 0){
		return new ScriptPackage();
	    } else if (args.length == 1 && args[0] instanceof String){
		Context c = (Context)context.clone();
		ScriptPackage p = new ScriptPackage();
		p.init(c);
		Pnuts.eval((String)args[0], c);
		return p;
	    } else if (args.length == 1 && args[0] instanceof Object[]){
		Object[] list = (Object[])args[0];
		ScriptPackage sp = new ScriptPackage();
		for (int i = 0; i < list.length; i++){
		    Object[] pair = (Object[])list[i];
		    String name = (String)pair[0];
		    sp.set(name.intern(), pair[1], context);
		}
		return sp;
	    } else {
		ScriptPackage sp = new ScriptPackage();
		for (int i = 0; i < args.length; i++){
		    Object arg = args[i];
		    if (arg instanceof PnutsFunction){
			String name = ((PnutsFunction)arg).getName();
			int idx = name.lastIndexOf('.');
			if (idx >= 0){
			    name = name.substring(idx + 1);
			}
			sp.set(name.intern(), arg, context);
		    }
		}
		return sp;
	    }
	}
    }

    public String toString(){
	return "<script package>";
    }
}
