/*
 * @(#) SecurePnutsImpl.java	1.2 00/08/03
 *
 * Copyright (c) 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.ext;

import java.security.CodeSource;
import java.security.Policy;
import java.security.Permission;
import java.security.Permissions;
import java.security.PermissionCollection;
import java.security.AccessControlContext;
import java.security.PrivilegedAction;
import java.security.AccessController;
import java.security.ProtectionDomain;
import java.io.File;
import java.io.FilePermission;
import java.net.URL;
import java.net.SocketPermission;
import pnuts.lang.SimpleNode;
import pnuts.lang.Context;
import pnuts.lang.PnutsImpl;
import pnuts.lang.PackagePermission;

/**
 * A PnutsImpl subclass that execute scripts in an access control context in 
 * Java2 Security.
 *
 * <pre>e.g.
 *  context.setPnutsImpl(new SecurePnutsImpl(new CompilerPnutsImpl()));
 * </pre>
 */
public class SecurePnutsImpl extends PnutsImpl {

    private CodeSource codeSource;
    private PnutsImpl impl;

    /**
     * A Constructor
     *
     * @param impl a PnutsImpl object
     */
    public SecurePnutsImpl(PnutsImpl impl){
	this(impl, null);
    }

    /**
     * A Constructor
     *
     * @param impl
     * @param codeSource a CodeSource object which indicates the source of the expression
     *        execute by eval(String, Context).
     */
    public SecurePnutsImpl(PnutsImpl impl, CodeSource codeSource){
	this.impl = impl;
	this.codeSource = codeSource;
    }

    /**
     * Evaluate an expreesion
     *
     * @param expr the expression to be evaluated
     * @param context the context in which the expression is evaluated
     * @return the result of the evaluation
     */
    public Object eval(final String expr, final Context context) {
	CodeSource cs = codeSource;
	if (cs != null){
	    AccessControlContext acc = getAccessControlContext(codeSource);
	    return AccessController.doPrivileged(new PrivilegedAction() {
		    public Object run() {
			return impl.eval(expr, context);
		    }
		}, acc);
	} else {
	    return impl.eval(expr, context);
	}
    }

    /**
     * Evaluate a parsed script
     *
     * @param node the parsed script
     * @param context the context in which the script is evaluated
     * @return the result of the evaluation
     */
    public Object accept(final SimpleNode node, final Context context){
	CodeSource cs = codeSource;
	if (cs != null){
	    AccessControlContext acc = getAccessControlContext(codeSource);
	    return AccessController.doPrivileged(new PrivilegedAction() {
		    public Object run() {
			return impl.accept(node, context);
		    }
		}, acc);
	} else {
	    return impl.accept(node, context);
	}
    }

    /**
     * Load a script file from a URL
     *
     * @param scriptURL the URL of the script
     * @param context the context in which the script is executed
     */
    public Object load(final URL scriptURL, final Context context) {
	CodeSource cs = codeSource;
	if (cs == null){
	    cs = new CodeSource(scriptURL, null);
	}
	AccessControlContext acc = getAccessControlContext(cs);
	return AccessController.doPrivileged(new PrivilegedAction() {
	    public Object run() {
		return impl.load(scriptURL, context);
	    }
	}, acc);
    }

    static AccessControlContext getAccessControlContext(final CodeSource codesource) {
	final URL codebase = codesource.getLocation();
	PermissionCollection perms = (PermissionCollection)
	    AccessController.doPrivileged(new PrivilegedAction() {
		public Object run() {
		    Policy p = Policy.getPolicy();
		    if (p != null) {
			return p.getPermissions(codesource);
		    } else {
			return null;
		    }
		}
	    });

	if (perms == null) {
	    perms = new Permissions();
	}

	Permission p;
	try {
	    p = codebase.openConnection().getPermission();
	} catch (java.io.IOException ioe) {
	    p = null;
	}

	if (p instanceof FilePermission) {
	    String path = p.getName();
	    int endIndex = path.lastIndexOf(File.separatorChar);
	    perms.add(p);
	    perms.add(new PackagePermission("*", "write, add, remove"));
	
	    if (endIndex != -1) {
		path = path.substring(0, endIndex+1);
		if (path.endsWith(File.separator)) {
		    path += "-";
		}
		perms.add(new FilePermission(path, "read"));
	    }
	} else {
	    String host = codebase.getHost();
	    if (host == null){
		host = "localhost";
	    }
	    perms.add(new SocketPermission(host, "connect, accept"));
	    if (p != null){
		perms.add(p);
	    }
	}
	ProtectionDomain domain = new ProtectionDomain(codesource, perms);
	return new AccessControlContext(new ProtectionDomain[] {domain});
    }

    public String toString(){
	return getClass().getName() + "[" + impl + ", " + codeSource + "]";
    }
}

