/*
 * @(#) Arithmetic.java	1.12 00/04/19
 *
 * Copyright (c) 1997-1999 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.lang;

import java.math.BigInteger;
import java.math.BigDecimal;
import java.lang.reflect.Array;
import java.lang.reflect.Method;

/**
 * The Arithmetic class is geared for transparent multi-precision arithmetic operations.
 */
public final class Arithmetic {
    private static final boolean DEBUG = false;
    /*
     * operators
     */
    public static final int ADD = 0;
    public static final int SUBTRACT = 1;
    public static final int MULTIPLY = 2;
    public static final int DIVIDE = 3;
    public static final int REMAINDER = 4;
    public static final int GCD = 5;
    public static final int MOD = 6;
    public static final int COMPARE = 7;
    public static final int NEGATE = 10;
    public static final int ABS = 11;
    public static final int AND = 12;
    public static final int OR = 13;
    public static final int XOR = 14;
    public static final int NOT = 15;
    public static final int SHIFT_R = 16;
    public static final int SHIFT_L = 17;
    public static final int SHIFT_A = 18;
    public static final int POW = 19;

    private static Integer zero = new Integer(0);

    private static BigInteger maxLong = BigInteger.valueOf(Long.MAX_VALUE);
    private static BigInteger minLong = BigInteger.valueOf(Long.MIN_VALUE);

    public final static int NOT_EQUAL = Numeric.NOT_EQUAL;
    public final static int LEFT_IS_BIGGER = Numeric.LEFT_IS_BIGGER;
    public final static int RIGHT_IS_BIGGER = Numeric.RIGHT_IS_BIGGER;
    public final static int EQUAL = Numeric.EQUAL;

    public final static Integer LEFT = new Integer(LEFT_IS_BIGGER);
    public final static Integer RIGHT = new Integer(RIGHT_IS_BIGGER);
    public final static Integer EQ = new Integer(EQUAL);
    public final static Integer NE = new Integer(NOT_EQUAL);

    final static Integer smallInt[] = new Integer[32];

    static {
	for (int i = 0; i < smallInt.length; i++){
	    smallInt[i] = new Integer(i);
	}
    }

    static Object numeric_object(int operator, Numeric n1, Object n2){
	switch (operator){
	case ADD:
	    return n1.add(n2);
	case SUBTRACT:
	    return n1.subtract(n2);
	case MULTIPLY:
	    return n1.multiply(n2);
	case DIVIDE:
	    return n1.divide(n2);
	case COMPARE:
	    return new Integer(n1.compareTo(n2));
	}
	throw new IllegalArgumentException("" + operator);
    }

    static Object object_numeric(int operator, Object n1, Numeric n2){
	switch (operator){
	case ADD:
	    return n2.add(n1);
	case SUBTRACT:
	    return binary(ADD, n2.negate(), n1);
	case MULTIPLY:
	    return n2.multiply(n1);
	case DIVIDE:
	    return binary(MULTIPLY, n2.inverse(), n1);
	case COMPARE:
	    int c = ((Numeric)n2).compareTo(n1);
	    if (c == LEFT_IS_BIGGER){
		return RIGHT;
 	    } else if (c == RIGHT_IS_BIGGER){
		return LEFT;
	    } else if (c == EQUAL){
		return EQ;
	    } else {
		return NE;
	    }
	}
	throw new IllegalArgumentException("" + operator);
    }

    static Object unary(int operator, Object n1){
	if (n1 instanceof BigDecimal){
	    switch (operator){
	    case NEGATE: return ((BigDecimal)n1).negate();
	    case ABS:    return ((BigDecimal)n1).abs();
	    default:   	 throw new IllegalArgumentException("" + operator);

	    }
	}
	if (n1 instanceof BigInteger){
	    switch (operator){
	      case NEGATE: return ((BigInteger)n1).negate();
	      case ABS:    return ((BigInteger)n1).abs();
	      case NOT:    return ((BigInteger)n1).not();
	      default:     throw new IllegalArgumentException("" + operator);
	    }
	}
	if (n1 instanceof Double){
	    double d = Double.NaN;
	    switch (operator){
	      case NEGATE: d = -((Double)n1).doubleValue(); break;
	      case ABS:    d = Math.abs(((Double)n1).doubleValue()); break;
	      default:     throw new IllegalArgumentException("" + operator);
	    }
	    return new Double(d);
	}
	if (n1 instanceof Float){
	    float d = Float.NaN;
	    switch (operator){
	      case NEGATE: d = -((Float)n1).floatValue(); break;
	      case ABS:    d = Math.abs(((Float)n1).floatValue()); break;
	      default:     throw new IllegalArgumentException("" + operator);
	    }
	    return new Float(d);
	}
	if (n1 instanceof Long ||
	    n1 instanceof Integer ||
	    n1 instanceof Short ||
	    n1 instanceof Byte){
	    long d = 0;
	    switch (operator){
	      case NEGATE: d = -((Number)n1).longValue(); break;
	      case ABS:    d = Math.abs(((Number)n1).longValue()); break;
	      case NOT:    d = ~((Number)n1).longValue(); break;
	      default:     throw new IllegalArgumentException("" + operator);
	    }
	    return longOrInt(d);
	}
	if (n1 instanceof Numeric){
	    switch (operator){
	    case NEGATE:
		if (((Numeric)n1).compareTo(zero) == RIGHT_IS_BIGGER){
		    return ((Numeric)n1).negate();
		} else {
		    return n1;
		}
	    }
	}
	throw new IllegalArgumentException("(" + n1 + ")");
    }

    static Object binary(int operator, Object n1, Object n2){
	int t1;
	int t2;
	if (n1 instanceof Integer){
	    t1 = 1;
	} else if (n1 instanceof Character){
	    t1 = 2;
	} else if (n1 instanceof Byte){
	    t1 = 4;
	} else if (n1 instanceof Short){
	    t1 = 8;
	} else if (n1 instanceof Long){
	    t1 = 16;
	} else if (n1 instanceof Float){
	    t1 = 32;
	} else if (n1 instanceof Double){
	    t1 = 64;
	} else if (n1 instanceof String){
	    if (operator == ADD){
		return concat(n1, n2);
	    }
	    t1 = 128;
	} else if (n1 instanceof BigDecimal){
	    t1 = 256;
	} else if (n1 instanceof BigInteger){
	    t1 = 512;
	} else if (n1 instanceof Numeric){
	    return numeric_object(operator, (Numeric)n1, n2);
	} else if (n1 == null){
	    if (operator == COMPARE){
		return n2 == null ? EQ : NE;
	    }
	    t1 = 0;
	} else {
	    t1 = 0;
	}
	if (n2 instanceof Integer){
	    t2 = 1;
	} else if (n2 instanceof Character){
	    t2 = 2;
	} else if (n2 instanceof Byte){
	    t2 = 4;
	} else if (n2 instanceof Short){
	    t2 = 8;
	} else if (n2 instanceof Long){
	    t2 = 16;
	} else if (n2 instanceof Float){
	    t2 = 32;
	} else if (n2 instanceof Double){
	    t2 = 64;
	} else if (n2 instanceof String){
	    if (operator == ADD){
		return concat(n1, n2);
	    }
	    t2 = 128;
	} else if (n2 instanceof BigDecimal){
	    t2 = 256;
	} else if (n2 instanceof BigInteger){
	    t2 = 512;
	} else if (n2 instanceof Numeric){
	    return object_numeric(operator, n1, (Numeric)n2);
	} else if (n2 == null && operator == COMPARE){
	    return NE;
	} else {
	    t2 = 0;
	}

	if (DEBUG){
	    System.out.println("t1 = " + t1 + ", t2 = " + t2);
	}
	switch ((t1 << 16) | t2){

	case (1 << 16) + 4: // int, byte
	case (1 << 16) + 8: // int, short
	case (4 << 16) + 1: // byte, int
	case (4 << 16) + 4: // byte, byte
	case (4 << 16) + 8: // byte, short
	case (8 << 16) + 1: // short, int
	case (8 << 16) + 4: // short, byte
	case (8 << 16) + 8: // short, short
		return op2_int(operator, ((Number)n1).intValue(), ((Number)n2).intValue());

	case (1 << 16) + 1: // int, int
	    if (operator == COMPARE){
		int i1 = ((Integer)n1).intValue();
		int i2 = ((Integer)n2).intValue();
		if (i1 < i2){
		    return RIGHT;
		} else if (i1 > i2){
		    return LEFT;
		} else {
		    return EQ;
		}
	    } else {
		return op2_int(operator, ((Integer)n1).intValue(), ((Integer)n2).intValue());
	    }
	case (4 << 16) + 2: // byte, char
	case (8 << 16) + 2: // short, char
	    return op2_int(operator, ((Number)n1).intValue(), (int)((Character)n2).charValue());

	case (1 << 16) + 2: // int, char
	    return op2_int(operator, ((Integer)n1).intValue(), (int)((Character)n2).charValue());

	case (1 << 16) + 16: // int, long
	case (16 << 16) + 1: // long, int
	case (16 << 16) + 16: // long, long
	    return op2_long(operator, ((Number)n1).longValue(), ((Number)n2).longValue());

	case (2 << 16) + 4: // char, byte
	case (2 << 16) + 8: // char, short
	    return op2_int(operator, (int)((Character)n1).charValue(), ((Number)n2).intValue());

	case (2 << 16) + 1: // char, int
	    return op2_int(operator, (int)((Character)n1).charValue(), ((Integer)n2).intValue());

	case (2 << 16) + 2: // char, char
	    if (operator == COMPARE){
		char c1 = ((Character)n1).charValue();
		char c2 = ((Character)n2).charValue();
		if (c1 < c2){
		    return RIGHT;
		} else if (c1 > c2){
		    return LEFT;
		} else {
		    return EQ;
		}
	    } else {
		return op2_int(operator,
			       (int)((Character)n1).charValue(),
			       (int)((Character)n2).charValue());
	    }
	case (2 << 16) + 16: // char, long
	    return op2_long(operator, (long)((Character)n1).charValue(), ((Long)n2).longValue());
	case (16 << 16) + 2: // long, char
	    return op2_long(operator, ((Long)n1).longValue(), (long)((Character)n2).charValue());
	case (1 << 16) + 32: // int,float
	case (16 << 16) + 32: // long,float
	case (32 << 16) + 1: // float, int
        case (32 << 16) + 16: // float, long
        case (32 << 16) + 32: // float, float
	    return op2_float(operator, ((Number)n1).floatValue(), ((Number)n2).floatValue());
	case (1 << 16) + 64: // int, double
	case (16 << 16) + 64: // long, double
	case (32 << 16) + 64: // float, double
	case (64 << 16) + 1: // double , int
	case (64 << 16) + 16: // double, long
	case (64 << 16) + 32: // double, float
	case (64 << 16) + 64: // double, double
	    return op2_double(operator, ((Number)n1).doubleValue(), ((Number)n2).doubleValue());
	case (2 << 16) + 32: // char, float
	    return op2_float(operator, (float)((Character)n1).charValue(), ((Float)n2).floatValue()); 
	case (32 << 16) + 2: // float, char
	    return op2_float(operator, ((Float)n1).floatValue(), (float)((Character)n2).charValue());
	case (2 << 16) + 64: // char, double
	    return op2_double(operator, (double)((Character)n1).charValue(), ((Double)n2).doubleValue());
	case (64 << 16) + 2: // double, char
	    return op2_double(operator, ((Double)n1).doubleValue(), (double)((Character)n2).charValue());

	case (1 << 16) + 512: // int, bigint
	case (4 << 16) + 512: // byte, bigint
	case (8 << 16) + 512: // short, bigint
	case (16 << 16) + 512: // long, bigint
	    return op2_bint(operator, BigInteger.valueOf(((Number)n1).longValue()), (BigInteger)n2);
	case (2 << 16) + 512: // char, bigint
	    return op2_bint(operator, BigInteger.valueOf((long)((Character)n1).charValue()), (BigInteger)n2);
	case (512 << 16) + 1: // bigint, int
	case (512 << 16) + 4: // bigint, byte
	case (512 << 16) + 8: // bigint, short
	case (512 << 16) + 16: // bigint, long
	    return op2_bint(operator, (BigInteger)n1, BigInteger.valueOf(((Number)n2).longValue()));
	case (512 << 16) + 2: // bigint, char
	    return op2_bint(operator, (BigInteger)n1, BigInteger.valueOf((long)((Character)n2).charValue()));
	case (512 << 16) + 512: // bigint, bigint
	    return op2_bint(operator, (BigInteger)n1, (BigInteger)n2);

	case (512 << 16) + 32: // bigint, float
	    {
		float ff = ((Float)n2).floatValue();
		if (Float.isInfinite(ff) || Float.isNaN(ff)){
		    return n2;
		}
		return op2_bdec(operator, new BigDecimal((BigInteger)n1), doubleToDecimal(ff));
	    }
	case (512 << 16) + 64: // bigint, double
	    {
		double dd = ((Double)n2).doubleValue();
		if (Double.isInfinite(dd) || Double.isNaN(dd)){
		    return n2;
		}
		return op2_bdec(operator, new BigDecimal((BigInteger)n1), doubleToDecimal(dd));
	    }
	case (32 << 16) + 512: // float, bigint
	    {
		float ff = ((Float)n1).floatValue();
		if (Float.isInfinite(ff) || Float.isNaN(ff)){
		    return n1;
		}
		return op2_bdec(operator, doubleToDecimal(ff), new BigDecimal((BigInteger)n2));
	    }
	case (64 << 16) + 512: // double, bigint
	    {
		double dd = ((Double)n1).doubleValue();
		if (Double.isInfinite(dd) || Double.isNaN(dd)){
		    return n1;
		}
		return op2_bdec(operator,  doubleToDecimal(dd), new BigDecimal((BigInteger)n2));
	    }
	case (1 << 16) + 256: // int, decimal
	case (4 << 16) + 256: // byte, decimal
	case (8 << 16) + 256: // short, decimal
	case (16 << 16) + 256: // long, decimal
	    return op2_bdec(operator, new BigDecimal(((Number)n1).longValue()), (BigDecimal)n2);
	case (2 << 16) + 256: // char, decimal
	    return op2_bdec(operator, new BigDecimal((long)((Character)n1).charValue()), (BigDecimal)n2);
	case (32 << 16) + 256: // float, decimal
	    {
		float ff = ((Float)n1).floatValue();
		if (Float.isInfinite(ff) || Float.isNaN(ff)){
		    return n1;
		}
		return op2_bdec(operator, doubleToDecimal(ff), (BigDecimal)n2);
	    }
	case (64 << 16) + 256: // double, decimal
	    {
		double dd = ((Double)n1).doubleValue();
		if (Double.isInfinite(dd) || Double.isNaN(dd)){
		    return n1;
		}
		return op2_bdec(operator, doubleToDecimal(dd), (BigDecimal)n2);
	    }
	case (256 << 16) + 1: // decimal, int
	case (256 << 16) + 4: // decimal, byte
	case (256 << 16) + 8: // decimal, short
	case (256 << 16) + 16: // decimal, long
	    return op2_bdec(operator, (BigDecimal)n1, new BigDecimal(((Number)n2).longValue()));
	case (256 << 16) + 2: // decimal, char
	    return op2_bdec(operator, (BigDecimal)n1, new BigDecimal((long)((Character)n2).charValue()));
	case (256 << 16) + 32: // decimal, float
	    float ff = ((Float)n2).floatValue();
	    if (Float.isInfinite(ff) || Float.isNaN(ff)){
		return n2;
	    }
	    return op2_bdec(operator, (BigDecimal)n1, doubleToDecimal(ff));
	case (256 << 16) + 64: // decimal, double
	    double dd = ((Double)n2).doubleValue();
	    if (Double.isInfinite(dd) || Double.isNaN(dd)){
		return n2;
	    }
	    return op2_bdec(operator, (BigDecimal)n1, doubleToDecimal(dd));
	case (256 << 16) + 256: // decimal, decimal
	    return op2_bdec(operator, (BigDecimal)n1, (BigDecimal)n2);
	case (128 << 16) + 128: // string, string
	    if (operator == COMPARE){
		int c = ((String)n1).compareTo((String)n2);
		if (c < 0){
		    return RIGHT;
		} else if (c > 0){
		    return LEFT;
		} else {
		    return EQ;
		}
	    }
	default:
	    if (Runtime.isArray(n1) && Runtime.isArray(n2)){
		if (operator == ADD){
		    return append(n1, n2);
		} else if (operator == COMPARE){
		    int len = 0;
		    if ((len = Array.getLength(n2)) == Array.getLength(n1)){
			for (int i = 0; i < len; i++){
			    Object e1 = Array.get(n1, i);
			    Object e2 = Array.get(n2, i);
			    if (e1 == null){
				if (e2 != null){
				    return NE;
				}
			    } else {
				if (Runtime.isArray(e1)){
				    if (binary(COMPARE, e1, e2) == NE){
					return NE;
				    }
				} else if (!e1.equals(e2)){
				    return NE;
				}
			    }
			}
		    } else {
			return NE;
		    }
		    return EQ;
		}
	    }
	    if (operator == COMPARE){
		int compared = Pnuts.verdep.compare(n1, n2);
		switch (compared){
		case EQUAL:
		    return EQ;
		case NOT_EQUAL:
		    return NE;
		case LEFT_IS_BIGGER:
		    return LEFT;
		case RIGHT_IS_BIGGER:
		    return RIGHT;
		}
	    }
	}
	throw new IllegalArgumentException("(" + n1 + ", " + n2 + ")");
    }

    static Object op2_bdec(int operator, BigDecimal d1, BigDecimal d2){
	switch (operator){
	case ADD:
	    return d1.add(d2);
	case SUBTRACT:
	    return d1.subtract(d2);
	case MULTIPLY:
	    return d1.multiply(d2);
	case DIVIDE:
	    return d1.divide(d2, BigDecimal.ROUND_HALF_DOWN);
	case COMPARE:
	    int i = d1.compareTo(d2);
	    if (i > 0){
		return LEFT;
	    } else if (i < 0){
		return RIGHT;
	    } else {
		return EQ;
	    }
	case POW:
	    return new Double(Math.pow(d1.doubleValue(), d2.doubleValue()));
	default:
	    throw new IllegalArgumentException("" + operator);
        }
    }

    static Object op2_bint(int operator, BigInteger i1, BigInteger i2){
	switch (operator){
	case ADD:
	    return compress(i1.add(i2));
	case SUBTRACT:
	    return compress(i1.subtract(i2));
	case MULTIPLY:
	    return compress(i1.multiply(i2));
	case DIVIDE:
	    return compress(i1.divide(i2));
	case COMPARE:
	    return new Integer(i1.compareTo(i2));
	case REMAINDER:
	    return compress(i1.remainder(i2));
	case GCD:
	    return compress(i1.gcd(i2));
	case MOD:
	    return compress(i1.mod(i2));
	case AND:
	    return compress(i1.and(i2));
	case OR:
	    return compress(i1.or(i2));
	case XOR:
	    return compress(i1.xor(i2));
	case SHIFT_L:
	    return compress(i1.shiftLeft(i2.intValue()));
	case SHIFT_R:
	    return compress(i1.shiftRight(i2.intValue()));
	case POW:
	    Number n = compress(i2);
	    if (n instanceof Integer){
		return compress(i1.pow(n.intValue()));
	    } else {
		return new Double(Math.pow(i1.doubleValue(), i2.doubleValue()));
	    }
	default:
	    throw new IllegalArgumentException("" + operator);
	}
    }

    static Object op2_double(int operator, double d1, double d2){
	switch (operator){
	case ADD:
	    return new Double(d1 + d2);
	case SUBTRACT:
	    return new Double(d1 - d2);
	case MULTIPLY:
	    return new Double(d1 * d2);
	case DIVIDE:
	    return new Double(d1 / d2);
	case COMPARE:
	    if (d1 > d2){
		return LEFT;
	    } else if (d1 < d2){
		return RIGHT;
	    } else if (d1 == d2){
		return EQ;
	    } else {
		return NE;
	    }
	case POW:
	    return new Double(Math.pow(d1, d2));
	default:
	    throw new IllegalArgumentException("" + operator);
	}
    }

    static Object op2_float(int operator, float d1, float d2){
	switch (operator){
	case ADD:
	    return new Float(d1 + d2);
	case SUBTRACT:
	    return new Float(d1 - d2);
	case MULTIPLY:
	    return new Float(d1 * d2);
	case DIVIDE:
	    return new Float(d1 / d2);
	case COMPARE:
	    if (d1 > d2){
		return LEFT;
	    } else if (d1 < d2){
		return RIGHT;
	    } else if (d1 == d2){
		return EQ;
	    } else {
		return NE;
	    }
	case POW:
	    return new Float((float)Math.pow(d1, d2));
	default:
	    throw new IllegalArgumentException("" + operator);
	}
    }

    static Object op2_long(int operator, long d1, long d2){
	long d3;
	switch (operator){
	case ADD:
	    d3 = d1 + d2;
	    if (d3 > d1 && d2 < 0 || d3 < d1 && d2 > 0){
		return op2_bint(ADD, BigInteger.valueOf(d1), BigInteger.valueOf(d2));
	    }
	    return longOrInt(d3);
	case SUBTRACT:
	    d3 = d1 - d2;
	    if (d3 > d1 && d2 > 0 || d3 < d1 && d2 < 0){
		return op2_bint(SUBTRACT, BigInteger.valueOf(d1), BigInteger.valueOf(d2));
	    }
	    return longOrInt(d3);
	case MULTIPLY:
	    d3 = d1 * d2;
	    if (d2 == 0){
		return zero;
	    } else if (d3 / d2 == d1){
		return longOrInt(d3);
	    } else {
		return op2_bint(MULTIPLY, BigInteger.valueOf(d1), BigInteger.valueOf(d2));
	    }
	case DIVIDE:
	    d3 = d1 / d2;
	    return longOrInt(d3);
	case MOD:
	    d3 = d1 % d2;
	    return longOrInt(d3);
	case COMPARE:
	    if (d1 > d2){
		return LEFT;
	    } else if (d1 < d2){
		return RIGHT;
	    } else if (d1 == d2){
	        return EQ;
	    } else {
		return NE;
	    }
	case AND:
	    return longOrInt(d1 & d2);
	case OR:
	    return longOrInt(d1 | d2);
	case XOR:
	    return longOrInt(d1 ^ d2);
	case SHIFT_L:
	    int m = 0;
	    long dd = d1;
	    while (dd > 0){
		dd >>= 1;
		m++;
	    }
	    if (d2 + m > 63){
		return op2_bint(SHIFT_L, BigInteger.valueOf(d1), BigInteger.valueOf(d2));
	    }
	    return longOrInt(d1 << d2);
	case SHIFT_R:
	    return longOrInt(d1 >> d2);
	case SHIFT_A:
	    return longOrInt(d1 >>> d2);
	case GCD:
	    return longOrInt(gcd(d1, d2));
	case POW:
	      return op2_bint(POW, BigInteger.valueOf(d1), BigInteger.valueOf(d2));
	default:
	    throw new IllegalArgumentException("" + operator);
	}
    }

    final static Object op2_int(int operator, int d1, int d2){
	switch (operator){
	  case ADD:
	    return longOrInt((long)d1 + d2);
	  case SUBTRACT:
	    return longOrInt((long)d1 - d2);
	  case MULTIPLY:
	    return longOrInt((long)d1 * d2);
	  case DIVIDE:
	    return longOrInt(d1 / d2);
	  case MOD:
	    return longOrInt(d1 % d2);
	  case COMPARE:
	    if (d1 > d2){
		return LEFT;
	    } else if (d1 < d2){
		return RIGHT;
	    } else if (d1 == d2){
	        return EQ;
	    } else {
	        return NE;
	    }
	  case AND:
	    return longOrInt(d1 & d2);
	  case OR:
	    return longOrInt(d1 | d2);
	  case XOR:
	    return longOrInt(d1 ^ d2);
	  case SHIFT_L:
	    int m = 0;
	    long dd = (long)d1;
	    while (dd > 0){
		dd >>= 1;
		m++;
	    }
	    if (d2 + m > 63){
		return op2_bint(SHIFT_L, BigInteger.valueOf(d1), BigInteger.valueOf(d2));
	    }
	    return longOrInt((long)d1 << d2);
	  case SHIFT_R:
	    return longOrInt(d1 >> d2);
	  case SHIFT_A:
	    return longOrInt(d1 >>> d2);
	  case GCD:
	    return new Integer((int)gcd(d1, d2));
	  case POW:
	      return op2_bint(POW, BigInteger.valueOf(d1), BigInteger.valueOf(d2));
	  default:
	    throw new IllegalArgumentException("" + operator);
	}
    }

    static Number compress(Number n){
	if (n instanceof BigInteger){
	    BigInteger b = (BigInteger)n;
	    if (b.compareTo(maxLong) > 0 ||
		b.compareTo(minLong) < 0){
		return n;
	    } else {
		long l = b.longValue();
		return longOrInt(l);
	    }
	} else if (n instanceof BigDecimal){
	    int j = 0;
	    String f = n.toString();
	    int p = f.indexOf((int)'.');
	    if (p > 0){
	        char ca[] = f.toCharArray();
	        for (int i = 0; i < ca.length - 2 && i < ca.length - p - 2; i++){
	            if (ca[ca.length - i - 1] == '0'){
	                j++;
	            } else {
	                break;
	            }
                }
	        f = f.substring(0, ca.length - j);
	    }
	    if (f.length() < 16){
	        return new Double(n.doubleValue());
	    }
	    if (j > 0){
	        return new BigDecimal(f);
	    } else {
	        return n;
	    }
	} else {
	    return n;
	}
    }

    final static Number longOrInt(long l){
	if (l >= 0 && l < 32){
	    return smallInt[(int)l];
	} else if (l <= Integer.MAX_VALUE && l >= Integer.MIN_VALUE){
	    return new Integer((int)l);
	} else {
	    return new Long(l);
    	}
    }

    public final static Object add(Object n1, Object n2){
	return binary(ADD, n1, n2);
    }

    public final static Object subtract(Object n1, Object n2){
	return binary(SUBTRACT, n1, n2);
    }

    public final static Object multiply(Object n1, Object n2){
	return binary(MULTIPLY, n1, n2);
    }

    public final static Object divide(Object n1, Object n2){
	return binary(DIVIDE, n1, n2);
    }

    public final static int compareTo(Object n1, Object n2){
	Object compared = binary(COMPARE, n1, n2);
	if (compared == EQ){
	    return EQUAL;
	} else if (compared == LEFT){
	    return LEFT_IS_BIGGER;
	} else if (compared == RIGHT){
	    return RIGHT_IS_BIGGER;
	} else {
	    return NOT_EQUAL;
	}
    }

    public final static Object remainder(Object n1, Object n2){
	return binary(REMAINDER, n1, n2);
    }

    public final static Object gcd(Object n1, Object n2){
	return binary(GCD, n1, n2);
    }

    public final static Object pow(Object n1, Object n2){
	return binary(POW, n1, n2);
    }

    public final static Object mod(Object n1, Object n2){
	return binary(MOD, n1, n2);
    }

    public final static Object and(Object n1, Object n2){
	return binary(AND, n1, n2);
    }

    public final static Object or(Object n1, Object n2){
	return binary(OR, n1, n2);
    }

    public final static Object xor(Object n1, Object n2){
	return binary(XOR, n1, n2);
    }

    public final static Object shiftLeft(Object n1, Object n2){
	return binary(SHIFT_L, n1, n2);
    }

    public final static Object shiftRight(Object n1, Object n2){
	return binary(SHIFT_R, n1, n2);
    }

    public final static Object shiftArithmetic(Object n1, Object n2){
	return binary(SHIFT_A, n1, n2);
    }

    public final static Object abs(Object n1){
	return unary(ABS, n1);
    }

    public static Object negate(Object n1){
	if (n1 instanceof Numeric){
	    return ((Numeric)n1).negate();
	} else {
	    return unary(NEGATE, n1);
	}
    }

    public static Object inverse(Object n1){
	if (n1 instanceof Numeric){
	    return ((Numeric)n1).inverse();
	} else {
	    return binary(DIVIDE, new Double(1.0), (Number)n1);
	}
    }

    public final static Object not(Object n1){
	return unary(NOT, n1);
    }

    static BigDecimal doubleToDecimal(double d){
	return (BigDecimal)decimal("" + d, 10);
    }

    static Number decimalNumber(String s, int radix){
	return decimal(s, radix);
    }

    static Number decimal(String s, int radix){
	if (radix == 16){
	    return new BigInteger(s, radix);
	} else {
	    return new BigDecimal(deNormalize(s));
	}
    }

    private static String deNormalize(String fmt){
	int pt = fmt.indexOf((int)'.');
	int exp = fmt.indexOf((int)'E');
	if (exp < 0){
	    return fmt;
	}
	int e = Integer.parseInt(fmt.substring(exp + 1));
	int i = 0;
	int k = 0;
	char b[];
	if (e < 0){
	    b = new char[exp - e];
	    i = 0;
	    if (fmt.charAt(k) == '-'){
	        b[i++] = '-';
		k++;
	    }
	    b[i++] = '0';
	    b[i++] = '.';
	    for (int j = 0; j < -e - 1; j++){
	        b[i++] = '0';
	    }
	    b[i++] = fmt.charAt(k++);
	    k++;
	    for (int j = 0; j < exp - pt - 1; j++){
	        b[i++] = fmt.charAt(k++);
	    }
	} else {
	    if (exp - pt - 1 > e){
		b = new char[exp];
	        i = 0;
	        if (fmt.charAt(k) == '-'){
	            b[i++] = '-';
		    k++;
	        }
	        b[i++] = fmt.charAt(k++);
	        k++;
		for (int j = 0; j < e; j++){
	            b[i++] = fmt.charAt(k++);
	        }
		b[i++] = '.';
		for (int j = 0; j < exp - pt - 1 - e; j++){
	            b[i++] = fmt.charAt(k++);
	        }
	    } else {
		b = new char[exp -1 + (e - (exp - pt - 1))];
	        i = 0;
	        if (fmt.charAt(k) == '-'){
	            b[i++] = '-';
		    k++;
	        }
	        b[i++] = fmt.charAt(k++);
	        k++;
		for (int j = 0; j < exp - pt - 1; j++){
	            b[i++] = fmt.charAt(k++);
	        }
		for (int j = 0; j < e - (exp - pt - 1); j++){
	            b[i++] = '0';
	        }
	    }
	}
	return new String(b);
    }

    static long gcd(long a, long b){ // Euclid's algorithm
	if (b > a){
	    long tmp = a;
	    a = b;
	    b = tmp;
	}
	for(;;){
	    if (b == 0){
		return a;
	    } else if (b == 1){
		return b;
	    } else {
		long tmp = b;
		b = a % b;
		a = tmp;
	    }
	}
    }

    final static String NULL = "null";

    static String concat(Object o1, Object o2){
	String s2;
	if (o2 == null){
	    s2 = NULL;
	} else if (Runtime.isArray(o2)){
	    s2 = Pnuts.format(o2);
	} else {
	    s2 = o2.toString();
	}
	String s1;
	if (o1 == null){
	    s1 = NULL;
	} else if (Runtime.isArray(o1)){
	    s1 = Pnuts.format(o1);
	} else {
	    s1 = o1.toString();
	}
	return s1 + s2;
    }

    static Object append(Object p1, Object p2){
	int l1 = Array.getLength(p1);
	int l2 = Array.getLength(p2);
	Object p3 = Array.newInstance(p1.getClass().getComponentType(), l1 + l2);
	try {
	    for (int i = 0; i < l1; i++){
		Array.set(p3, i, Array.get(p1, i));
	    }
	    for (int i = 0; i < l2; i++){
		Array.set(p3, i + l1, Array.get(p2, i));
	    }
	    return p3;
	} catch (Exception e){
	    throw new IllegalArgumentException(e.getMessage());
	}
    }
}
