/*
 * @(#) Cache.java	1.3 00/01/24
 *
 * Copyright (c) 1997-1999 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.lang;

/**
 *  An implementation of LRU cache
 */
public class Cache {

    private Cell[] cells;
    private int size;
    private boolean compareID;
    private Cell head;
    private Cell tail;
    private int count = 0;

    static class Cell {
	Cell next;
	Cell prev;
	Cell chain;
	Object key;
	Object value;
    }

    protected Cache(){
    }

    /**
     * @param size	cache size
     */
    public Cache(int size){
	this(size, true);
    }

    /**
     * @param size	cache size
     * @param compareID	If true keys are compared by their object IDs,
     * 	                 otherwise by Object.equals().
     */
    public Cache(int size, boolean compareID){
	cells = new Cell[size];
	this.size = size;
	this.compareID = compareID;
    }

    /**
     * If key is in the cache it returns value, otherwise null.
     */
    public synchronized Object get(Object key){
	boolean compareID = this.compareID;
	int index = ((compareID ? System.identityHashCode(key) : key.hashCode()) & 0x7FFFFFFF) % size;
	Cell b = cells[index];
	while (b != null){
	    if (compareID ? (b.key == key) : b.key.equals(key)){
		if (b != tail){
		    if (b.prev != null){
			b.prev.next = b.next;
		    }
		    if (b.next != null){
			b.next.prev = b.prev;
		    }
		    tail.next = b;
		    b.prev = tail;
		    tail = b;
		}
		if (b == head){
		    head = head.next;
		}
		tail.next = null;
		return b.value;
	    }
	    b = b.chain;
	}
	return null;
    }

    /**
     * Register key and its value into the cache.
     */
    public synchronized void put(Object key, Object value){
	boolean compareID = this.compareID;
	int index = ((compareID ? System.identityHashCode(key) : key.hashCode()) & 0x7FFFFFFF) % size;
	Cell b = cells[index];
	Cell n = null;
	if (b != null){
	    while (b.chain != null){
		if (compareID ? (b.key == key) : b.key.equals(key)){
		    b.value = value;
		    return;
		}
		b = b.chain;
	    }
	    if (compareID ? (b.key == key) : b.key.equals(key)){
		b.value = value;
		return;
	    }
	    if (head != null){
		if (count >= size){
		    Cell second = head.next;
		    if (b == head){
			n = b;
		    } else {
			remove(head);
			n = head;
			n.chain = null;
			b.chain = n;
		    }
		    n.next = null;
		    head = second;
		} else {
		    count++;
		    n = new Cell();
		    b.chain = n;
		}
	    } else {
		n = new Cell();
		count++;
		head = n;
		tail = n;
		b.chain = n;
	    }
	} else {
	    if (head != null){
		if (count >= size){
		    Cell second = head.next;
		    n = head;
		    remove(head);
		    n.chain = null;
		    n.next = null;
		    head = second;
		} else {
		    count++;
		    n = new Cell();
		}
	    } else {
		n = new Cell();
		count++;
		head = n;
		tail = n;
	    }
	    b = cells[index] = n;
	}
	n.key = key;
	n.value = value;
	if (tail != n){
	    tail.next = n;
	    tail = n;
	}
    }

    private void remove(Cell old){
	int index = ((compareID ? System.identityHashCode(old) : old.hashCode()) & 0x7FFFFFFF) % size;
	Cell b = cells[index];
	if (b == old){
	    cells[index] = b.chain;
	} else if (b != null){
	    while (b.chain != null){
		if (b.chain == old){
		    b.chain = b.chain.chain;
		    return;
		}
		b = b.chain;
	    }
	}
    }
}
