/*
 * @(#) Context.java	1.44 00/08/08
 *
 * Copyright (c) 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.lang;

import java.io.OutputStream;
import java.io.Writer;
import java.io.PrintWriter;
import java.io.IOException;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;

/**
 * <em>Context</em> represents an internal state in Pnuts runtime environment.
 *  A <a href="/pnuts/apidoc/pnuts.lang.Context.html">Context</a>  object contains
 *  the following information.
 * <ul>
 * <li>Which Pnuts-package being used
 * <li>Imported Java-package list
 * <li>Writer to which print() write data
 * <li>Writer to which error() write message
 * <li>Stack frame (with the pure interpreter)
 * </ul>
 *
 * @version	1.44
 */
public class Context implements Cloneable {

    public final static PrintWriter defaultOutputStream = new PrintWriter(System.out, false);
    public final static PrintWriter defaultTerminalStream = defaultOutputStream;
    public final static PrintWriter defaultErrorStream = new PrintWriter(System.err, true);

    /**
     * internal name for undefined state.
     */
    private final static Object UNDEF = new Object[0];

    static String exceptionHandlerTableSymbol = "!exception".intern();

    static ImportEnv defaultImports = new ImportEnv();
    static {
	defaultImports.addPackage("");
	defaultImports.addPackage("java.lang");
    }

    static PnutsImpl defaultPnutsImpl = new PnutsImpl();

    PnutsImpl pnutsImpl = defaultPnutsImpl;

    /*
     * A flag to tell if this context is created with eval() builtin function
     */
    boolean eval = false;

    // parent context from which this context is created
    Context parent;

    // eval depth
    protected int depth = 0;

    // caller
    Function frame;

    // function which throws an exception
    Function errorFrame;

    // the stack frame
    StackFrame stackFrame;

    // the stack which tracks load/loadFile chain
    protected Stack loadingResource = new Stack();

    // line information
    protected int beginLine = -1;

    // line information
    protected int endLine = -1;

    // import() state
    protected ImportEnv importEnv;

    private PrintWriter outputWriter;
    private PrintWriter errorWriter;
    private PrintWriter terminalWriter;
    private String name ;

    private SymbolTable environment;

    boolean namespaceRefreshed[] = {false}; // shared by all clones

    ClassLoader classLoader[] = new ClassLoader[]{null};

    Hashtable usedPackage;

    Runtime runtime = null; // used by bytecode compiler

    Package currentPackage = Package.getGlobalPackage();
    {
	currentPackage.init(this);
    }

    /**
     * Create a new context
     */
    public Context(){
	this(Package.getGlobalPackage());
    }

    /**
     * Creates a context.
     *
     * @param pkg the name of the package.
     */
    public Context(String pkg){
	this(Package.getPackage(pkg));
    }

    /**
     * Creates a context.
     *
     * @param pkg the initial package of the context.
     */
    public Context(Package pkg){
	currentPackage = pkg;
	importEnv = defaultImports;

	this.outputWriter = defaultOutputStream;
	this.terminalWriter = defaultTerminalStream;
	this.errorWriter = defaultErrorStream;
    }

    /**
     * Make a clone of the context
     */
    public Object clone(){
	return clone(true);
    }

    /**
     * Make a clone of the context
     *
     * @param clear If true, import() state and current package are reset
     *              to the default values.
     */
    public Object clone(boolean clear){
	try {
	    if (usedPackage == null){
		usedPackage = new Hashtable(8);
	    }
	    Context ret = (Context)super.clone();
	    ret.reset();
	    ret.parent = this;
	    if (clear){
		ret.eval = false;
		ret.frame = null;
		ret.importEnv = defaultImports;
		ret.currentPackage = Package.globalPackage;
		ret.beginLine = -1;
	    } else {
		if (importEnv != defaultImports){
		    ret.importEnv = (ImportEnv)importEnv.clone();
		}
	    }
	    return ret;
	} catch (Exception e){
	    throw new InternalError("clone error");
	}
    }

    /**
     * Sets the name of the context
     */
    public void setName(String name){
	this.name = name;
    }

    /**
     * Gets the name of the context
     */
    public String getName(){
	return name;
    }

    /**
     * Changes the PnutsImpl object associated with this context
     */
    public void setPnutsImpl(PnutsImpl impl){
	if (impl != null){
	    pnutsImpl = impl;
	}
    }

    /**
     * Gets the PnutsImpl object associated with this context
     */
    public PnutsImpl getPnutsImpl(){
	return pnutsImpl;
    }

    /**
     * Get an environemnt variable associated with this context
     *
     * @param key the name of the variable
     * @return the value of the variable
     * @since 1.0beta8
     */
    public Object get(String key){
	if (environment == null){
	    return null;
	}
	return environment.get(key.intern());
    }

    /**
     * Define an environemnt variable associated with this context
     *
     * @param key the name of the variable
     * @param value the value of the variable
     * @since 1.0beta8
     */
    public void set(String key, Object value){
	if (environment == null){
	    environment = new SymbolTable();
	}
	environment.put(key.intern(), value);
    }

    /**
     * Returns an enumeration of the keys in the environment of this context.
     */
    public Enumeration keys(){
	if (environment == null){
	    environment = new SymbolTable();
	}
	return environment.keys();
    }

    /**
     * set output stream of the context
     */
    public void setOutputStream(Object out, boolean autoFlush){
	if (out == null){
	    this.outputWriter = null;
	} else if (out instanceof OutputStream){
	    this.outputWriter = new PrintWriter((OutputStream)out, autoFlush);
	} else if (out instanceof Writer){
	    this.outputWriter = new PrintWriter((Writer)out, autoFlush);
	} else {
	    throw new IllegalArgumentException(PnutsException.getMessage("illegal.streamType", PnutsException.NO_PARAM));
	}
    }

    public void setOutputStream(Object outputStream){
	if (outputStream instanceof PrintWriter){
	    this.outputWriter = (PrintWriter)outputStream;
	} else {
	    setOutputStream(outputStream, false);
	}
    }

    /**
     * get output stream of the context
     */
    public PrintWriter getOutputStream(){
	return outputWriter;
    }

    /**
     * set terminal-output-stream of the context
     */
    public void setTerminalStream(Object str, boolean autoFlush){
	if (str == null){
	    this.terminalWriter = null;
	} else if (str instanceof OutputStream){
	    this.terminalWriter = new PrintWriter((OutputStream)str, autoFlush);
	} else if (str instanceof Writer){
	    this.terminalWriter = new PrintWriter((Writer)str, autoFlush);
	} else {
	    throw new IllegalArgumentException(PnutsException.getMessage("illegal.streamType", PnutsException.NO_PARAM));
	}
    }

    public void setTerminalStream(Object stream){
	if (stream instanceof PrintWriter){
	    this.terminalWriter = (PrintWriter)stream;
	} else {
	    setTerminalStream(stream, false);
	}
    }

    /**
     * get terminal-output-stream of the context
     */
    public PrintWriter getTerminalStream(){
	return terminalWriter;
    }

    /**
     * Set an OutputStream or a Writer to which error() write message
     * If errorStream is null, exception is thrown out of eval loop.
     */
    public void setErrorStream(Object errorStream, boolean autoFlush){
	if (errorStream == null){
	    this.errorWriter = null;
 	} else if (errorStream instanceof OutputStream){
	    this.errorWriter = new PrintWriter((OutputStream)errorStream, autoFlush);
 	} else if (errorStream instanceof Writer){
	    this.errorWriter = new PrintWriter((Writer)errorStream, autoFlush);
	} else {
	    throw new IllegalArgumentException(PnutsException.getMessage("illegal.streamType", PnutsException.NO_PARAM));
	}
    }

    /**
     * set an OutputStream or a Writer to which error() write message
     * If errorStream is null, exception is thrown out of eval loop.
     */
    public void setErrorStream(Object errorStream){
	if (errorStream instanceof PrintWriter){
	    this.errorWriter = (PrintWriter)errorStream;
	} else {
	    setErrorStream(errorStream, false);
	}
    }

    /**
     * Get an OutputStream or a Writer to which error() write message
     */
    public PrintWriter getErrorStream(){
	return errorWriter;
    }

    /**
     * get the current package
     */
    public Package getCurrentPackage(){
	return currentPackage;
    }

    /**
     * set the current package
     */
    public void setCurrentPackage(Package pkg){
	pkg.init(this);
	currentPackage = pkg;
    }

    /*
    void clear(){
	reset();
	eval = false;
	frame = null;
	importEnv = defaultImports;
	currentPackage = Package.globalPackage;
	beginLine = -1;
    }
    */

    void addClassToImport(String className){
	if (importEnv == defaultImports){
	    importEnv = (ImportEnv)importEnv.clone();
	}
	importEnv.addClass(className);
    }

    void addPackageToImport(String pkgName){
	if (importEnv == defaultImports){
	    importEnv = (ImportEnv)importEnv.clone();
	}
	importEnv.addPackage(pkgName);
    }

    final void pushFile(Object file){
	loadingResource.push(file);
    }

    final void popFile(){
	if (loadingResource.size() > 0){
	    loadingResource.pop();
	}
    }

    final boolean isLoading(){
	return loadingResource.size() > 0;
    }
    
    final Object peekFile(){
	return loadingResource.peek();
    }

    final void clearFile(){
	loadingResource.removeAllElements();
    }

    /**
     * Returns the list of use()'d packages
     */
    public String[] usedPackages(){
	if (usedPackage == null){
	    return new String[]{};
	}
	int size = usedPackage.size();
	String[] names = new String[size];
	int i = 0;
	for (Enumeration e = usedPackage.keys(); e.hasMoreElements(); ){
	    Package p = (Package)e.nextElement();
	    names[i++] = p.getName();
	}
	return names;
    }

    /**
     * Add a package to the use()'d package list.
     *
     * @param name the package name
     */
    public void usePackage(String name){
	Package pkg = Package.find(name);
	if (!"".equals(name)){
	    if (pkg != null){ 
		for (Enumeration e = pkg.keys(); e.hasMoreElements(); ){
		    String symbol = (String)e.nextElement();
		    Object value = pkg.get(symbol, this);
		    if (value instanceof PnutsFunction){
			PnutsFunction func = (PnutsFunction)value;
			if (symbol.equals(func.getName())){
			    Package.globalPackage.set(symbol, func, this);
			}
		    }
		}
	    }
	    if (pkg == null){
		pkg = Package.getPackage(name);
	    }
	    if (usedPackage == null){
		usedPackage = new Hashtable(8);
	    }
	    usedPackage.put(pkg, pkg);
	}
	try {
	    Pnuts.require(pkg.getInitScript(), this);
	} catch (IOException e){
	}
    }

    /**
     * Get the source of the script. 
     *
     * @return <pre>
     * java.net.URL object, when the script is not precompiled
     * pnuts.lang.Runtime object, when the script is precompiled
     * </pre>
     */
    protected Object getScriptSource(){
	if (errorFrame != null){
	    return errorFrame.file;
	} else if (frame != null && frame.file != null){
	    return frame.file;
	} else if (isLoading()){
	    return peekFile();
	} else {
	    return null;
	}
    }

    protected void updateLine(SimpleNode node){
	updateLine(node, node.beginLine, node.endLine);
    }

    protected void updateLine(SimpleNode node, int beginLine, int endLine){
	if (beginLine > 0 && getScriptSource() != null){
	    this.beginLine = beginLine;
	    this.endLine = endLine;
	}
    }

    /**
     * This method is called when the excecution is terminated normally.
     */
    protected void onExit(Object arg){
    }

    /**
     * This method is called when an exception is thrown.
     */
    protected void onError(Throwable t){
    }

    public Object getId(String symbol){
	Object v = _getId(symbol.intern());
	if (v == UNDEF){
	    return undefined(symbol);
	} else {
	    return v;
	}
    }

    Object _getId(String symbol){
	Value v = currentPackage.lookup(symbol, this);
	if (v != null){
	    return v.get();
	}

	if (symbol.charAt(0) == '!'){
	    return null;
	}

	synchronized(namespaceRefreshed){
	    if (namespaceRefreshed[0]){
		importEnv.reset();
	    }
	    namespaceRefreshed[0] = false;
	}

	Object c = importEnv.get(symbol, this);
	if (c != null){
	    return c;
	}

	return undefinedHook(symbol);
    }

    protected Object undefinedHook(String symbol){
	Package global = Package.globalPackage;
	SymbolTable tab = global.autoloadTable;

	if (tab != null){
	    AutoloadHook hook = (AutoloadHook)tab.get(symbol);
	    if (hook != null){
		hook.load(symbol, this);
		tab.remove(symbol);
		if (global.defined(symbol, this)){
		    return global.get(symbol, this);
		}
	    }
	}
	return UNDEF;
    }

    Object undefined(String sym){
	throw new PnutsException("not.defined", new Object[]{sym}, this);
    }

    public boolean defined(String name){
	return _getId(name.intern()) != UNDEF;
    }

    Object getValue(String symbol){
	Object val;

	if (stackFrame != null){
	    val = _getValue(symbol);
	} else {
	    val = _getId(symbol);
	}
	if (val == UNDEF){
	    return undefined(symbol);
	} else {
	    return val;
	}
    }

    ////// Methods below are used by the pure interpreter

    /*
     * open new scope with the pure interpreter
     */
    protected void open(String locals[]){
	errorFrame = null;
	stackFrame = new StackFrame(locals, stackFrame);
    }

    /*
     * open for/foreach scope with the pure interpreter
     */
    void openLocal(String locals[]){
	stackFrame.openLocal(locals);
    }

    /*
     * close the scope with the pure interpreter
     */
    protected void close(){
	stackFrame = stackFrame.parent;
    }

    /*
     * close for/foreach scope with the pure interpreter
     */
    void closeLocal(){
	stackFrame.closeLocal();
    }

    void reset(){
	if (stackFrame != null){
	    stackFrame = new StackFrame();
	}
    }

    protected void setValue(String symbol, Object obj){
	Binding b = stackFrame.lookup(symbol);
	if (b != null){
	    b.value = obj;
	    return;
	}

	Function ff = frame;

	while (ff != null){
	    SymbolTable ls = ff.lexicalScope;
	    if (ls != null){
		Binding bb = ls.lookup0(symbol);
		if (bb != null){
		    ((Binding)bb.value).value = obj;
		    return;
		}
	    }
	    ff = ff.outer;
	}

	if (stackFrame.parent != null){
	    stackFrame.put1(symbol, obj);
	} else {
	    currentPackage.set(symbol, obj, this);
	}
    }

    final void bind(String symbol, Object obj){
	stackFrame.bind(symbol, obj);
    }

    private Object _getValue(String symbol){
	Binding b = stackFrame.lookup(symbol);

	if (b != null){
	    return b.value;
	}

	Function ff = frame;

	while (ff != null){
	    SymbolTable ls = ff.lexicalScope;
	    if (ls != null){
		Binding bb = ls.lookup0(symbol);
		if (bb != null){
		    return ((Binding)bb.value).value;
		}
	    }
	    ff = ff.outer;
	}
	return _getId(symbol);
    }

    void catchException(Class t, PnutsFunction f){
	if (stackFrame.parent != null){
	    Hashtable table = null;
	    Value b = stackFrame.lookup(exceptionHandlerTableSymbol);
	    if (b != null){
		table = (Hashtable)b.get();
	    }
	    if (table == null){
		table = new Hashtable(10);
		stackFrame.put1(exceptionHandlerTableSymbol, table);
	    }
	    if (f == null){
		table.remove(t);
	    } else {
		table.put(t, f);
	    }
	} else {
	    Runtime.catchException(t, f, this);
	}
    }
}
