/*
 * @(#) Package.java	1.25 00/08/04
 *
 * Copyright (c) 1997-1999 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.lang;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.io.Serializable;
import java.io.OutputStream;
import java.io.InputStream;
import java.io.ByteArrayOutputStream;
import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.io.FileNotFoundException;

/**
 * This class represents a Pnuts package (not Java's).
 *
 * @version	1.25
 * @author	Toyokazu Tomatsu
 */
public class Package implements Property, Serializable, Cloneable {

    static final long serialVersionUID = 663339062422920949L;

    /**
     * All packages with a non-null name
     */
    protected static Hashtable packages = new Hashtable(10);

    /**
     * The package with name "".
     */
    public static Package globalPackage = null;

    transient protected SymbolTable table;
    transient protected Vector children;
    transient protected Package parent;

    /**
     * The name of the package.
     * @serial
     */
    protected String name;

    /**
     * This represents a set of names which are registered as autoloaded.
     * This is only for global package.
     * @serial
     */
    public SymbolTable autoloadTable; // only for global package

    /**
     * This represents a set of registered QuantityFactory.
     * This is only for global package.
     * @serial
     */
    protected SymbolTable unitTable; // only for global package

    /**
     * This represents a set of loaded script names.
     * This is only for global package.
     * @serial
     */
    protected SymbolTable requireTable; // only for global package

    static void initGlobal(){
	packages.put("", getGlobalPackage());
    }

    static PackageFactory factory;

    static {
	try {
	    String prop = System.getProperty("pnuts.package.factory");
	    if (prop != null){
		Object obj = Class.forName(prop).newInstance();
		if (obj instanceof PackageFactory){
		    factory = (PackageFactory)obj;
		}
	    }
	} catch(Throwable e){/* skip */}
    }

    static {
	initGlobal();
    }

    /**
     * @return the global package
     */
    public static Package getGlobalPackage(){
	if (globalPackage == null){
	    globalPackage = getInstance("", null);
	}
	return globalPackage;
    }

    /**
     * When an instance of PackageFactory is registered by this
     * method, the package() builtin function calls its
     * PackageFactory.createPackage() method.
     */
    public static void setPackageFactory(PackageFactory factory){
	Package.factory = factory;
    }

    static Package getInstance(String name, Package parent){
	if (name == null){
	    if (factory != null){
		return factory.createPackage(null, parent);
	    } else {
		return new Package(null, parent);
	    }
	}
	Package p = (Package)packages.get(name);
	if (p == null){
	    if (factory != null){
		p = factory.createPackage(name, parent);
	    } else {
		p = new Package(name, parent);
	    }
	    Pnuts.verdep.checkPackageAccess(p, "add");
	    packages.put(name, p);
	}
	return p;
    }

    /**
     * If package "pkg" exists returns the package, otherwise
     * creates and returns it.
     */
    public static Package getPackage(String pkg){
        Package g = globalPackage;
	Package p = g;
	if ("".equals(pkg)){
	    return g;
	}
	int index = pkg.indexOf("::");
	String rest = pkg;
	String s = "";
	while (index > 0){
	    String t = rest.substring(0, index);
	    if (t.indexOf('.') >= 0 || t.indexOf(':') >= 0){
	        new PnutsException("illegal package name " + pkg);
            }
	    s += t;
	    if (g.get(s) instanceof Class){
		new PnutsException("package name and Class conflicted: " + pkg);
	    }
	    p = getInstance(s, p);
	    rest = rest.substring(index + 2);
	    index = rest.indexOf("::");
	    s += "::";
        }
	if (rest.indexOf('.') >= 0 || rest.indexOf(':') >= 0){
	    throw new PnutsException("illegal package name " + pkg);
        }
	s += rest;
	return getInstance(s, p);
    }

    String getInitScript(){
	if ("".equals(name)){
	    return "init";
	}
	int index = name.indexOf("::");
	int start = 0;
	String s = "";
	while (index > 0){
	    String t = name.substring(start, index);
	    s += t;
	    s += "/";
	    start = index + 2;
	    index = name.indexOf("::", start);
        }
	return s + name.substring(start) + "/init";
    }

    /**
     * @return true if name is defined in the package.
     *
     * @deprecated
     */
    public boolean defined(String name){
	return defined(name, null);
    }

    /**
     * @return true if name is defined in the package.
     */
    public boolean defined(String name, Context context){
	return table.lookup0(name.intern()) != null;
    }

    /**
     * @return the value of specified variable in the package.
     *
     * @deprecated
     */
    public Object get(String symbol){
	return get(symbol, null);
    }

    /**
     * Get the value of a symbol in the package.  When the symbol is not defined in the package
     * and this.parent is not null, get the value in the parent package.
     *
     * @param symbol an interned name in the package
     * @param context the context in which the symbol is referenced.  null means "not specified".
     * @return the value of specified variable in the package.
     */
    public Object get(String symbol, Context context){
	Value b = lookup(symbol, context);
	if (b == null){
	    return null;
	} else {
	    return b.get();
	}
    }

    /**
     * Set a value of a symbol in the package.
     * @param sym a name of variable
     * @param obj the value of the variable
     *
     * @deprecated
     */
    void set(String symbol, Object obj){
	set(symbol, obj, null);
    }

    /**
     * Set a value of a symbol in the package.
     *
     * @param	symbol an interned name of variable
     * @param	obj the value of the variable
     */
    public void set(String symbol, Object obj, Context context){
	if (context != null){
	    Pnuts.verdep.checkPackageAccess(this, "write");
	}
	table.put0(symbol, obj);
	if (this != globalPackage && obj instanceof PnutsFunction){
	    if (context.usedPackage.get(this) != null){
		PnutsFunction func = (PnutsFunction)obj;
		if (symbol.equals(func.getName())){
		    globalPackage.set(symbol, obj, context);
		}
	    }
	}
    }

    /**
     * Deletes a symbol from the package.
     * @param symbol a name of variable to be deleted
     *
     * @deprecated
     */
    public void clear(String symbol){
	clear(symbol, null);
    }

    /**
     * Deletes a symbol from the package.
     *
     * @param symbol a name of variable to be deleted
     */
    public void clear(String symbol, Context context){
	Pnuts.verdep.checkPackageAccess(this, "write");
	table.remove(symbol);
    }

    /**
     * Removes the specified package.
     */
    public static void remove(String name){
	Package p = (Package)packages.get(name);
	if (p == null){
	    return;
	}
	Package p2 = p.parent;
	if (p2 == null){
	    return;
	}
	p2.children.removeElement(p);
	Pnuts.verdep.checkPackageAccess(p, "remove");
        packages.remove(name);
    }

    /**
     * @param pkg a name of package to look.
     * @return a package with name "pkg" if it exits.
     */
    public static Package find(String pkg){
        return (Package)packages.get(pkg);
    }

    /**
     * Remove all packages
     */
    public static void reset(){
	packages = new Hashtable(10);
	initGlobal();
    }

    /**
     * Creates a package that is not visible from other packages.
     */
    public Package(){
	this(null);
    }

    /**
     * Creates a package and register it in a static hashtable.
     *
     * @param name the name of the package
     */
    public Package(String name){
	this(name, globalPackage);
    }

    /**
     * Creates a package and register it in a static hashtable.
     * The method <tt>get()</tt> tries to find a symbol in this package
     * and then consult the parent package.
     * Other instance methods, such as <tt>set(), defined()</tt>, operates on this package only.
     * Other constructors implicitly specify the global package as the parent package.
     *
     * @param name the name of the package
     * @param parent the parent package.
     */
    public Package(String name, Package parent){
	if (name != null){
	    this.name = name.intern();
	}
	this.parent = parent;
	if (parent != null){
	    if (!parent.children.contains(this)){
		parent.children.addElement(this);
	    }
	}
	init();
    }

    /**
     * This method is called by the constructors.
     */
    protected void init(){
	this.table = new SymbolTable();
	this.children = new Vector();
	if (name != null){
	    if ("".equals(name)){
		autoloadTable = new SymbolTable();
		unitTable = new SymbolTable();
		requireTable = new SymbolTable();

		table.put("pnuts_version".intern(), Pnuts.pnuts_version);
		table.put("int".intern(), Integer.TYPE);
		table.put("short".intern(), Short.TYPE);
		table.put("long".intern(), Long.TYPE);
		table.put("byte".intern(), Byte.TYPE);
		table.put("char".intern(), Character.TYPE);
		table.put("float".intern(), Float.TYPE);
		table.put("double".intern(), Double.TYPE);
		table.put("boolean".intern(), Boolean.TYPE);

		PnutsFunction[] prim = PnutsFunction.primitives;
		for (int i = 0; i < prim.length; i++){
		    table.put(prim[i].getName().intern(), prim[i]);
		}
	    }
	}
    }

    /**
     * This method is called when the package become the current package with package() function.
     * This method in a subclass must call super.init(context) first.
     */
    protected void init(Context context){
	if ("".equals(name)){
	    globalPackage = this;
	}
	if (name != null){
	    if (packages.get(name) != this){
		Pnuts.verdep.checkPackageAccess(this, "add");
		packages.put(name, this);
	    }
	}
	context.currentPackage = this;
    }

    /**
     * @return the name of the package.
     */
    public String getName(){
	return name;
    }

    /**
     * Returns an enumeration of the names in the package.
     */
    public Enumeration keys(){
	return table.keys();
    }

    /**
     * Returns the parent package.
     */
    public Package getParent(){
	return parent;
    }

    /**
     * @return the number of symbols.
     */
    public int size(){
	return table.size();
    }

    /**
     * Lookup the symbol in the package.  When the symbol is not defined in the package and this.parent
     * is not null, lookup the symbol in the parent package.
     *
     * @param symbol intern'ed string
     */
    protected Value lookup(String symbol, Context context){
	Value v = table.lookup0(symbol);
	if (v == null && parent != null){
	    return parent.lookup(symbol, context);
	} else {
	    return v;
	}
    }

    transient SymbolTable notToSerialize;
    transient ByteArrayOutputStream bout;
    transient ObjectOutputStream out;

    /**
     * Saves the package into specified output stream.
     */
    void prepareToSave(){
	int size = table.size();
	for (int i = 0; i < size; i++){
	    try {
		bout = new ByteArrayOutputStream();
		out = new ObjectOutputStream(bout);
		out.writeObject(this);
		break;
	    } catch (IOException e){
		/* ignore */
		System.out.println(e);
	    }
	}
    }

    /**
     * Saves the package to an outputStream.
     * If a value of a symbol can not be serialized, the symbol is
     * ignored.
     */
    public void save(OutputStream s){
	prepareToSave();
	try {
	    bout.writeTo(s);
	} catch (IOException e){
	    throw new PnutsException(e);
	} finally {
	    out = null;
	    bout = null;
	}
    }

    private synchronized void writeObject(ObjectOutputStream s)
        throws IOException
    {
	s.defaultWriteObject();

	if (notToSerialize == null){
	    notToSerialize = new SymbolTable();
	}
	Runtime.writeSymbolTable(table, notToSerialize, s);
    }

    private synchronized void readObject(ObjectInputStream s)
         throws IOException, ClassNotFoundException
    {

	s.defaultReadObject();
	table = new SymbolTable();
	children = new Vector();
	Runtime.readSymbolTable(table, s);
	notToSerialize = null;
    }

    /**
     * Enumerates sub-packages.
     */
    public Enumeration elements(){
	return children.elements();
    }

    public static void registerQuantityFactory(String unit, QuantityFactory fac){
	Pnuts.verdep.checkPackageAccess(globalPackage, "write");
	if (fac != null){
	    globalPackage.unitTable.put(unit.intern(), fac);
	} else {
	    globalPackage.unitTable.remove(unit.intern());
	}
    }

    public Object clone(){
	try {
	    Package p = (Package)super.clone();
	    for (Enumeration e = keys(); e.hasMoreElements(); ){
		String key = (String)e.nextElement();
		p.set(key, get(key));
	    }
	    return p;
	} catch (Throwable t){
	    throw new InternalError();
	}
    }

    public String toString(){
	if (name != null){
	    return "package \"" + name + "\"";
	} else {
	    return "package null";
	}
    }
}
