/*
 * @(#) Pnuts.java	1.49 00/08/08
 *
 * Copyright (c) 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.lang;

import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.PrintWriter;
import java.io.File;
import java.io.FilePermission;
import java.io.FileInputStream;
import java.io.BufferedInputStream;
import java.io.StringReader;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.io.Serializable;
import java.net.URL;
import java.net.MalformedURLException;
import java.net.SocketPermission;
import java.util.StringTokenizer;
import java.util.Hashtable;
import java.applet.Applet;
import java.security.*;
import pnuts.lang.Runtime;
import pnuts.lang.Package;

/**
 * "Manager" class for Pnuts interpreter.<br>
 * "pnuts" command starts from main method of this class.<br>
 * Annotated reference manual is <a href="/pnuts/doc/api.html">here</a>.
 *
 * @version	1.49
 * @author	Toyokazu Tomatsu
 */
public class Pnuts implements Serializable {
    final static boolean DEBUG = false;

    static final long serialVersionUID = 7444339744293294263L;

    public final static String pnuts_version = "1.0beta8_p4";

    /**
     * "prompt" string for the command shell
     */
    public static String prompt = "> ";

    private static String java_vendor = System.getProperty("java.vendor");
    static {
	if (java_vendor == null){
	    java_vendor = "";
	}
    }

    final static String load_path_sym = "pnuts_load_path".intern();

    final static VersionDependent verdep = VersionDependent.getInstance();

    static boolean verbose = false;
    static {
	try {
	    String vb = System.getProperty("pnuts.verbose");
	    if (vb != null){
		verbose = true;
	    }
	} catch (Throwable t){}
    }

    static String compiledClassPrefix;

    static String getCompiledClassPrefix(){
	if (compiledClassPrefix == null){
	    try {
		compiledClassPrefix = System.getProperty("pnuts.compiled.script.prefix");
		if (!"".equals(compiledClassPrefix) && !compiledClassPrefix.endsWith(".")){
		    compiledClassPrefix = compiledClassPrefix + ".";
		}
	    } catch (Throwable t){}
	}
	return compiledClassPrefix;
    }

    static ClassLoader defaultClassLoader;

    private static Stack freeParsers = new Stack();
    static {
	freeParsers.push(new PnutsParser((Reader)null));
    }

    private final static Hashtable contextTable = new Hashtable(8);

    private static int java2 = 0;

    final public static boolean isJava2(){
	if (java2 == 0){
	    try {
		Class.forName("java.lang.Package");
		java2 = 1;
	    } catch (Throwable t){
		java2 = -1;
	    }
	}
	if (java2 == 1){
	    return true;
	} else {
	    return false;
	}
    }

    public static Visitor getDefaultVisitor(){
	return PnutsInterpreter.instance;
    }

    public static void registerQuantityFactory(String unit, QuantityFactory factory){
	Package.registerQuantityFactory(unit, factory);
    }

    /**
     * set ClassLoader for class access
     * @param	loader the ClassLoader with which classes are accessed
     */
    public synchronized static void setClassLoader(ClassLoader loader, Context context){
	if (context != null){
	    context.classLoader[0] = loader;
	    synchronized(context.namespaceRefreshed){
		context.namespaceRefreshed[0] = true;
	    }
	} else {
	    defaultClassLoader = loader;
	}
    }

    /**
     * 
     */
    public static void setClassLoader(ClassLoader loader){
	setClassLoader(loader, null);
    }

    /**
     * get ClassLoader for class access
     */
    public static ClassLoader getClassLoader(Context context){
	if (context == null){
	    return defaultClassLoader;
	}
	ClassLoader cl = context.classLoader[0];
	if (cl != null){
	    return cl;
	} else {
	    return defaultClassLoader;
	}
    }

    /**
     */
    public static ClassLoader getClassLoader(){
	return defaultClassLoader;
    }

    /**
     * @deprecated
     */
    public final static Class loadClass(String name) throws ClassNotFoundException {
	return loadClass(name, null);
    }

    /**
     * Loads the class with the specified name.
     *
     * @param name the class name to be loaded
     * @param context the context in which the class is loaded
     * @return the loaded class.
     */
    public final static Class loadClass(String name, Context context)
	throws ClassNotFoundException
    {
	ClassLoader classLoader = getClassLoader(context);
	if (classLoader != null){
	    if (isJava2()){
		return Class.forName(name, true, classLoader);
	    } else {
		Class c = classLoader.loadClass(name);
		return c;
	    }
	} else {
	    if (isJava2()){
		ClassLoader cl = null;
		try {
		    cl = ClassLoader.getSystemClassLoader();
		} catch (SecurityException t){/* skip */}
		if (cl != null){
		    return Class.forName(name, true, cl);
		}
	    }
	    return Class.forName(name);
	}
    }

    /**
     * set a "prompt" string for the command shell
     */
    public static void setPrompt(String str){
	prompt = str;
    }

    /**
     * set verbose mode
     */
    public static void setVerbose(boolean b) {
	verbose = b;
    }

    /**
     * @return the current verbose mode
     */
    public static boolean isVerbose() {
	return verbose;
    }

    /**
     * set a search path for scripts as an object array
     * Elements of the array should one of the followings.
     * <ul>
     *    <li>String:  base name of resource
     *    <li>URL:     base name of URL
     *    <li>File:    base directory
     * </ul>
     *
     * @deprecated 
     */
    public static void setLoadPath(Object paths[]){
	set(load_path_sym, paths);
    }

    /**
     * returns a formatted string for obj
     */
    public static String format(Object obj){
	return Runtime.format(obj, 64);
    }

    /**
     * @return the value of a global variable "str"
     */
    public static Object get(String str){
	return get(str, "");
    }

    /**
     * @return the value of a variable "str" in the package "pkg"
     */
    public static Object get(String str, String pkg){
	Package p = Package.find(pkg);
	if (p != null){
	    return p.get(str.intern(), null);
	} else {
	    return null;
	}
    }

    /**
     * set a value "val" to a global variable "str"
     */
    public static void set(String str, Object val){
	set(str, val, "");
    }

    /**
     * set a value "val" to a variable "str" in package "pkg"
     */
    public static void set(String str, Object val, String pkg){
	if (str.length() > 0 && Character.isJavaIdentifierStart(str.charAt(0))){
	    Package p = Package.find(pkg);
	    if (p != null){
		p.set(str.intern(), val, null);
	    } else {
		throw new IllegalArgumentException(PnutsException.getMessage("package.notFound", new Object[]{pkg}));
	    }
	} else {
	    throw new IllegalArgumentException(PnutsException.getMessage("illegal.symbolForId", PnutsException.NO_PARAM));
	}
    }

    /**
     * evaluate "str" in package "pkg"
     */
    public static Object eval(String str, String pkg){
	return eval(str, new Context(pkg));
    }

    /**
     * evaluate "str" in the global package
     *
     * @deprecated
     *    Replaced by eval(String, Context)
     */
    public static Object eval(String str){
	return eval(str, new Context());
    }

    /**
     * evaluate "str" in "context"
     */
    public static Object eval(String str, Context context){
	return context.pnutsImpl.eval(str, context);
    }

    /**
     * load a local script "file"
     *
     * @deprecated
     *   Replaced by loadFile(String, Context)
     */
    public static Object loadFile(String file) throws FileNotFoundException {
	return loadFile(file, new Context());
    }

    /**
     * load a local script "file" in "context"
     */
    public static Object loadFile(String file, Context context) throws FileNotFoundException {
	return context.pnutsImpl.loadFile(file, context);
    }

    /**
     * load a script "file"
     *
     * @deprecated
     *    Replaced by load(String, Context)
     */
    public static Object load(String file) throws FileNotFoundException {
	return load(file, new Context());
    }

    /**
     * load a script "file" in "context"
     */
    public static Object load(String name, final Context context)
	throws FileNotFoundException
    {
	String file = name;
	if (!name.endsWith(".pnut")){
	    file = name + ".pnut";

	    final Runtime rt = Runtime.getCompiledScript(name, context);

	    if (rt != null){
		if (Pnuts.isVerbose()){
		    System.out.println("[loading " + Pnuts.format(rt.getClass()) + "]");
		}
		context.pushFile(rt);
		try {
		    Object ret = Pnuts.verdep.execute((Runtime)rt, context);
		    provide(file);
		    return ret;
		} catch (Jump jump){
		    provide(file);
		    return jump.getValue();
		} finally {
		    context.popFile();
		}
	    }
	}

	return context.pnutsImpl.load(file, context);
    }

    public static Object load(URL url, Context context) {
	return context.pnutsImpl.load(url, context);
    }

    /**
     * load a script from InputStream "in"
     *
     * @deprecated
     *   Replaced by load(InputStream, Context)
     */
    public static Object load(InputStream in) {
	return load(in, new Context());
    }

    /**
     * load a script from InputStream "in" in "context"
     */
    public static Object load(InputStream in, Context context) {
	return load(new InputStreamReader(in), context);
    }

    /**
     * Load a script from an InputStream in the specified Context.
     *
     * @param in an InputStream from which the interpreter reads an input
     * @param interactive
     * <ul>
     * <li>When "interactive" is true, the greeting message, the prompt, and the results of
     * evaluations are displayed.  When an exception is thrown and not caught by any exception
     * handler, it is caught at the top level of the interpreter, display an error message, 
     * and resume the interactive session.  If the exception is caught by a handler that is
     * registered at the top level, the result of the handler becomes the return value of the
     * last expression.
     *
     * <li>When "interactive" is false, exceptions are caught at the top level of the interpreter
     * and exits this function.  If the exception thrown is caught by a handler that
     * is registered at the top level,  the result of the handler becomes the return value of
     * this method.
     * </ul>
     * @param context a Context in which the interpretation is taken place.
     * @return the result of the last expression
     */
    public static Object load(InputStream in,
			      boolean interactive,
			      Context context) {
	return load(new InputStreamReader(in), interactive, context);
    }

    /**
     * This method loads a script
     *
     * @param reader the Reader from which the script is loaded
     *
     * @deprecated
     *   Replaced by load(Reader, Context)
     */
    public static Object load(Reader reader) {
	return load(reader, new Context());
    }

    /**
     * This method loads a script
     *
     * @param reader the Reader from which the script is loaded
     * @param context the context in which the script is loaded
     * @return the result of the last expression
     */
    public static Object load(Reader reader, Context context) {
	return context.pnutsImpl.load(reader, context);
    }

    /**
     * This method loads a script
     *
     * @param reader the Reader from which the script is loaded
     * @param interactive specifies if the execution is in interactive mode.
     * @return the result of the last expression
     *
     * @deprecated
     *   Replaced by load(Reader, boolean, Context)
     */
    public static Object load(Reader reader, boolean interactive){
	return load(reader, interactive, new Context());
    }

    /**
     * This method loads a script
     *
     * @param reader the Reader from which the script is loaded
     * @param interactive specifies if the execution is in interactive mode.
     * @param context the context in which the script is loaded
     * @return the result of the last expression
     */
    public static Object load(Reader reader,
			      boolean interactive,
			      Context context) {
	return new Pnuts(reader, interactive, context).value;
    }

    /**
     * parse a script from InputStream and return a Pnuts object
     *
     * @return the Pnuts object including a parsed syntax tree
     * @param in the InputStream
     * @since Pnuts 1.0beta3
     */
    public static Pnuts parse(InputStream in) throws ParseException {
	return parse(new InputStreamReader(in));
    }

    /**
     * parse a script from Reader and return a Pnuts object
     *
     * @return the Pnuts object including a parsed syntax tree
     * @param reader the Reader
     * @since Pnuts 1.0beta3
     */
    public static Pnuts parse(Reader reader) throws ParseException {
	return new Pnuts(reader);
    }

    /**
     * parse a script and return a Pnuts object
     *
     * @return the Pnuts object including a parsed syntax tree
     * @param expr the script
     * @since Pnuts 1.0beta3
     */
    public static Pnuts parse(String expr) throws ParseException {
	return new Pnuts(new StringReader(expr));
    }

    /**
     * load a script "file" only when the script has not been read.
     */
    public static void require(String file, Context context)
	throws FileNotFoundException
    {
	if (file.endsWith(".pnut")){
	    file = file.substring(0, file.length() - 5);
	}
	file = file.intern();
	if (Package.globalPackage.requireTable.get(file) == null){
	    load(file, context);
	}
    }

    static void provide(String file){
	if (file.endsWith(".pnut")){
	    file = file.substring(0, file.length() - 5);
	}
	file = file.intern();
	Package.globalPackage.requireTable.put(file, file);
    }

    private void loop(PnutsParser parser, boolean interactive, Context context){
	while (true) {
	    try {
		PrintWriter term = null;
		if (interactive){
		    flush(context);

		    term = context.getTerminalStream();
		    if (term != null){
			term.print(prompt);
			term.flush();
		    }
		    System.gc();
		}
		
		SimpleNode start = parser.Start();
//		start.dump("");

		switch (getCode(start)){
		case -1:
		    ((int[])start.info)[0] = 0;
		    return;
		case 0:
		    reset(parser);
		    break;
		case 1:
		    if (interactive){
			try { 
                            value = context.pnutsImpl.accept(start, context);
			} finally {
			    context.onExit(value);
			}
		    } else {
                        value = context.pnutsImpl.accept(start, context);
		    }
		    if (getCode(start) == -1){
			value = start.value;
			return;
		    }
		    if (interactive && term != null){
			try {
			    term.println(format(value));
			    term.flush();
			} catch (ThreadDeath td){
			    throw td;
			} catch (Throwable t){
			    Runtime.checkException(context, t);
			}
		    }
		    reset(parser);
		    break;
		}
		    
	    } catch (ParseException p){
		context.beginLine = p.getErrorLine();
		error_skipto(parser, PnutsParserConstants.EOL);
		if (!interactive || context.depth > 1){
 		    Runtime.checkException(context, p);
		} else {
		    try {
			context.beginLine = p.currentToken.next.beginLine;
			Runtime.checkException(context, p);
		    } catch (Jump j){
			value = j.getValue();
			PrintWriter term = context.getTerminalStream();
			if (interactive && term != null){
			    term.println(value);
			    term.flush();
			}
		    } catch (PnutsException t){
			Runtime.printError(t, context);
		    }
		    reset(parser);
		    recover(context);
		    value = null;
		}
	    } catch (Escape esc){
		reset(parser);
		throw esc;
	    } catch (Throwable t){
		if (!interactive || context.depth > 1){
		    if (t instanceof PnutsException){
			throw (PnutsException)t;
		    } else if (t instanceof ThreadDeath){
			throw (ThreadDeath)t;
		    } else {
			throw new PnutsException(t, context);
		    }
		}
		if (t instanceof ThreadDeath){
		    if (DEBUG){
			System.err.println(">>> ThreadDeath is caught");
		    }
		    return;
		}
		Runtime.printError(t, context);
		reset(parser);
		recover(context);
		value = null;
	    }
	}
    }


    private static void error_skipto(PnutsParser parser, int kind){
	Token t;
	do {
	    t = parser.getNextToken();
	} while (t != null && t.image.length() > 0 && t.kind != kind);
    }


    /**
     * parsed scripts
     *
     * @serial
     */
    protected SimpleNode startNodes = null;

    transient protected Object value;

    protected Pnuts(){
    }

    protected Pnuts(Reader reader) throws ParseException {
	PnutsParser parser;
	synchronized (freeParsers){
	    if (freeParsers.size() > 0){
		parser = (PnutsParser)freeParsers.pop();
		parser.ReInit(reader);
	    } else {
		parser = new PnutsParser(reader);
	    }
	}
	try {
	    startNodes = parser.StartSet();
	} catch (ParseException e){
	    error_skipto(parser, PnutsParserConstants.EOL);
	    reset(parser);
	    throw e;
	} finally {
	    synchronized (freeParsers){
		if (freeParsers.size() < 3){
		    freeParsers.push(parser);
		}
	    }
	}
    }

    /**
     * executes a Pnuts object with a new Context
     *
     * @return the result
     *
     * @deprecated
     *    Replaced by run(Context)
     */
    public Object run(){
	return run(new Context());
    }

    /**
     * executes a Pnuts object with the specified Context
     *
     * @param context the Context
     * @return the result
     * @since Pnuts 1.0beta3
     */
    public Object run(Context context){
	int depth = enter(context);
	try {
	    Object value = null;
	    value = context.pnutsImpl.accept(startNodes, context);
	    context.onExit(value);
	    return value;
	} catch (Escape esc){
	    return esc.getValue();
	} finally {
	    context.depth = depth;
	}
    }

    private static void flush(Context context){
	PrintWriter out = context.getOutputStream();
	if (out != null){
	    out.flush();
	}
    }

    /**
     * traverse the parsed tree with the specified Visitor and Context
     *
     * @param context the Context
     * @return the result
     * @since Pnuts 1.0beta3
     */
    public Object accept(Visitor visitor, Context context) {
	int depth = enter(context);
	Object value = null;
	try {
	    value = startNodes.accept(visitor, context);
	    context.onExit(value);
	    return value;
	} catch (Escape esc){
	    Object val = esc.getValue();
	    flush(context);
	    return val;
	} catch (PnutsException pe){
	    context.onError(pe);
	    throw pe;
	} catch (Throwable t){
	    PnutsException e = new PnutsException(t, context);
	    context.onError(e);
	    throw e;
	} finally {
	    context.depth = depth;
	}
    }

    /**
     * Starts Pnuts interpreter. If <em>interactive</em> is true
     * prompt message and the results of evaluation are printed
     * on the screen.
     */
    protected Pnuts(Reader reader, boolean interactive, Context context){
	PnutsParser parser;
	synchronized (freeParsers){
	    if (freeParsers.size() > 0){
		parser = (PnutsParser)freeParsers.pop();
		parser.ReInit(reader);
	    } else {
		parser = new PnutsParser(reader);
	    }
	}
	synchronized(context.namespaceRefreshed){
	    context.namespaceRefreshed[0] = false;
	}

	int depth = enter(context);
	try {
	    loop(parser, interactive, context);
	    flush(context);
	} catch (Escape esc){
	    value = esc.getValue();
	    flush(context);
	} catch (PnutsException t){
	    if (!interactive || context.depth > 1){
		throw t;
	    }
	    Runtime.printError(t, context);
	} finally {
	    if (interactive) context.onExit(value);
	    context.depth = depth;
	    synchronized (freeParsers){
		if (freeParsers.size() < 3){
		    freeParsers.push(parser);
		}
	    }
	}
    }

    /**
     * all public entry points must go through this to ensure
     * the correct behavior of evalDepth method.
     *
     * @param context the context in which the execution is taken place.
     * @return the last value of context.depth.
     */
    static int enter(Context context){
	int depth = context.depth;
	if (depth < 0x7fffffff){
	    context.depth++;
	}
	contextTable.put(Thread.currentThread(), context);
	return depth;
    }

    /**
     * Get the depth of evaluation.
     *
     * This value increases when load(), loadFile(), or eval() is called.
     */
    public static int evalDepth(){
	Context context = (Context)contextTable.get(Thread.currentThread());
	if (context == null){
	    return 0;
	}
	return context.depth;
    }

    private static void reset(PnutsParser parser){
	parser.jjtree.reset();
    }

    private static void recover(Context context){
	context.reset();
	context.clearFile();
    }

    private final static int getCode(SimpleNode node){
	return ((int[])node.info)[0];
    }

    /**
     * @deprecated
     */
    public static void main(String args[]) throws Throwable {
	pnuts.tools.Main.main(args);
    }
}
