/*
 * @(#) PnutsFunction.java	1.34 00/08/08
 *
 * Copyright (c) 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.lang;

import java.io.StringReader;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.io.Serializable;
import java.io.FileNotFoundException;
import java.util.Enumeration;
import java.util.NoSuchElementException;
import java.util.Vector;

/**
 * A PnutsFunction represents a group of Pnuts functions with a same name.
 *
 * @version	1.34
 * @author	Toyokazu Tomatsu
 */
public class PnutsFunction implements Serializable {

    static final long serialVersionUID = -315478430086905925L;

    /**
     * @serial
     */
    Function[] functions = new Function[4];

    /**
     * the name
     * @serial
     */
    String name;

    final static String _getContext = "getContext".intern();
    final static String _package = "package".intern();
    final static String _import = "import".intern();
    final static String _catch = "catch".intern();
    final static String _throw = "throw".intern();
    final static String _eval = "eval".intern();
    final static String _loadFile = "loadFile".intern();
    final static String _load = "load".intern();
    final static String _quit = "quit".intern();
    final static String _defined = "defined".intern();
    final static String _use = "use".intern();
    final static String _class = "class".intern();
    final static String _require = "require".intern();
    public final static PnutsFunction GET_CONTEXT = new Builtin(_getContext);
    public final static PnutsFunction PACKAGE= new Builtin(_package);
    public final static PnutsFunction IMPORT = new Builtin(_import);
    public final static PnutsFunction CATCH = new Builtin(_catch);
    public final static PnutsFunction THROW = new Builtin(_throw);
    public final static PnutsFunction EVAL = new Builtin(_eval);
    public final static PnutsFunction LOAD_FILE = new Builtin(_loadFile);
    public final static PnutsFunction LOAD = new Builtin(_load);
    public final static PnutsFunction QUIT = new Builtin(_quit);
    public final static PnutsFunction DEFINED =  new Builtin(_defined);
    public final static PnutsFunction USE =  new Builtin(_use);
    public final static PnutsFunction CLASS =  new Builtin(_class);
    public final static PnutsFunction REQUIRE =  new Builtin(_require);
    public final static PnutsFunction primitives[] = {GET_CONTEXT, PACKAGE, IMPORT, THROW,
						       EVAL, LOAD_FILE, LOAD, QUIT, DEFINED, USE, CLASS, REQUIRE};

    protected PnutsFunction(){
	this.name = "";
    }

    protected PnutsFunction(String name){
	this.name = name.intern();
    }

    static class Auto implements AutoloadHook, Serializable {
	String file;
	
	Auto(String file){
	    this.file = file;
	}

	public void load(String name, Context context){
	    try {
		Context ct = (Context)context.clone();
		Pnuts.load(file, ct);
	    } catch (Throwable e){
		throw new PnutsException("autoload.failed", new Object[]{file + "," + e}, context);
	    }
	}
    }

    /**
     * register "name" as an autoloaded function.
     * Later "file" is automatically loaded if the function is not defined 
     */
    public static void autoload(String name, String file){
	if (file == null){
	    autoload(name, (AutoloadHook)null);
	} else {
	    autoload(name, new Auto(file));
	}
    }

    public static void autoload(String name, AutoloadHook loadHook){
	Package global = Package.globalPackage;
	Pnuts.verdep.checkPackageAccess(global, "write");
	if (loadHook == null){
	    global.autoloadTable.remove(name.intern());
	} else {
	    global.autoloadTable.put(name.intern(), loadHook);
	}
    }

    final protected void put(int narg, Function f){
	int n = narg + 1;
	if (functions.length < n + 1){
	    Function func[] = new Function[(n + 1) * 2];
	    System.arraycopy(functions, 0, func, 0, functions.length);
	    functions = func;
	}
	functions[n] = f;
	f.function = this;
	added(narg);
    }

    /**
     * This method is called when a Function object is registered to this object.
     * This method just returns.  Subclass can override this as a hook method.
     *
     * @param narg the number of parameters
     */
    protected void added(int narg){
    }

    final protected Function get(int narg){
	int n = narg + 1;
	if (n >= functions.length){
	    return null;
	}
	return functions[n];
    }

    /**
     * Check if the function with narg parameter is defined
     *
     * @param narg the number of paramters.  -1 means a arbitrary length parameter.
     * @return true if a function with narg, otherwise false
     */
    public boolean defined(int narg){
	return get(narg) instanceof Function;
    }

    /**
     * @return the name of functions
     */
    public String getName(){
	return name;
    }

    /**
     * Call a function in "context" with arguments "args".
     * Increments the counter for Pnuts.evalDepth() during
     * the execution.
     *
     * @return the result of the call
     */
    public final Object call(Object[] args, Context context){
	int depth = Pnuts.enter(context);
	try {
	    return exec(args, context);
	} finally {
	    context.depth = depth;
	}
    }

    /**
     * Call a function in "context" with arguments "args".
     * Subclasses of this class should override this method.
     *
     * @return the result of the call
     */
    protected Object exec(Object[] args, Context context){
	Function f = get(args.length);
	if (f == null){
	    f = get(-1);
	    if (f == null){
		throw new PnutsException("function.notDefined",
					 new Object[]{name, new Integer(args.length)},
					 context);
	    }
	    args = new Object[]{args};
	}

	ImportEnv saved = context.importEnv;
	Package outerPackage = context.currentPackage;
	boolean eval = context.eval;
	Function caller = context.frame;
	context.importEnv = f.importEnv;
	context.currentPackage = f.pkg;
	context.eval = false;
	context.frame = f;

	try {
	    return f.exec(args, context);
	} catch (Jump jump){
	    if (_throw == name && args.length == 1){
		throw jump;
	    }
	    return jump.getValue();
	} catch (Escape esc){
	    throw esc;
	} catch (ThreadDeath td){
	    throw td;
	} catch (Throwable t){
	    if (context.errorFrame == null){
		context.errorFrame = f;
	    }
	    String msg = null;
	    try {
		String param = Pnuts.format(args);
		msg = name + "(" + param.substring(1, param.length() - 1) + ")";
	    } catch (Throwable tt){
		msg = name + "(?)";
	    }
	    
	    PnutsException p = null;
	    if (t instanceof PnutsException){
		p = (PnutsException)t;
	    } else {
		p = new PnutsException(t, context);
	    }
	    if (p.trace == null){
		p.trace = new Vector();
	    } 
	    p.trace.addElement(msg);
	    
	    throw p;

	} finally {
	    context.frame = caller;
	    context.importEnv = saved;
	    context.eval = eval;
	    context.currentPackage = outerPackage;
	}
    }

    /**
     * call a function with arguments "args"
     *
     * @return the result of the call
     */
    public final Object call(Object[] args){
	Context context = new Context();
	Object ret = call(args, context);
	context.getOutputStream().flush();
	return ret;
    }

    public String toString(){
	StringBuffer buf = new StringBuffer();

	int i = 0;
	while (i < functions.length){
	    Function f = functions[i++];
	    if (f != null){
		buf.append(f.toString());
		break;
	    }
	}
	while (i < functions.length){
	    Function f = functions[i++];
	    if (f != null){
		buf.append(",");
		buf.append(f.paramString());
	    }
	}
	return buf.toString();
    }

    /**
     * call a function "pkg::name" with arguments "args"
     */
    public static Object call(String name, Object args[], String pkg){
	return call(name, args, new Context(pkg));
    }

    /**
     * call a function "name" in "context" with arguments "args"
     */
    public static Object call(String name, Object args[], Context context){
	Object o = context.getId(name);
	if (o instanceof PnutsFunction){
	    PnutsFunction func = (PnutsFunction)o;
	    Object ret = func.call(args, context);
	    context.getOutputStream().flush();
	    return ret;
	} else {
	    throw new PnutsException("function.notDefined", new Object[]{name, new Integer(args.length)}, context);
	}
    }

    /**
     * call a function "name" with arguments "args"
     */
    public static Object call(String name, Object args[]){
	return call(name, args, "");
    }

    /**
     * Retrieve the symbolic definition of the function.
     *
     * @param narg  the number of paramters.  -1 means a arbitrary length parameter.
     * @return the function definition
     */
    public String unparse(int narg){
	Function f = get(narg);
	if (f != null){
	    return f.unparse(null);
	}
	return null;
    }

    /**
     * @param narg  the number of paramters.  -1 means a arbitrary length parameter.
     * @return Package in which the function is defined
     * @deprecated
     */
    public Package getPackage(int narg){
	Function f = get(narg);
	if (f != null){
	    return f.pkg;
	}
	return null;
    }

    protected transient Package pkg = Package.globalPackage;

    /**
     * @return Package in which the function is defined
     */
    public Package getPackage(){
	return pkg;
    }

    /**
     * @param narg  the number of paramters.  -1 means a arbitrary length parameter.
     * @return imports of the function (array of Class or String)
     */
    public String[] getImportEnv(int narg){
	Function f = get(narg);
	if (f != null){
	    return f.importEnv.list();
	}
	return null;
    }

    public boolean isBuiltin(){
	return false;
    }

    /**
     * @param narg  the number of paramters.  -1 means a arbitrary length parameter.
     */
    public Object accept(int narg, Visitor visitor, Context context){
	Function f = get(narg);
	if (f != null){
	    return f.accept(visitor, context);
	}
	return null;
    }

    class Enum implements Enumeration {
	int idx = 0;
	int size = functions.length;

	public boolean hasMoreElements(){
	    while (!defined(idx - 1) && idx < size){
		idx++;
	    }
	    return idx < size;
	}

	public Object nextElement(){
	    if (idx < size){
		while (!defined(idx - 1)){
		    idx++;
		}
		return get(idx++ - 1);
	    } else {
		throw new NoSuchElementException(PnutsFunction.this.toString());
	    }
	}
    }

    protected Enumeration elements(){
	return new Enum();
    }

    private synchronized void readObject(ObjectInputStream s)
	throws IOException, ClassNotFoundException {
	s.defaultReadObject();
	if (name != null){
	    name = name.intern();
	}
	if (functions == null){
	    functions = new Function[4];
	}
    }
}
