/*
 * @(#) PnutsImpl.java	1.3 00/08/07
 *
 * Copyright (c) 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.lang;

import java.io.InputStream;
import java.io.Reader;
import java.io.File;
import java.io.IOException;
import java.io.FileNotFoundException;
import java.net.URL;
import java.net.MalformedURLException;
import java.security.CodeSource;
import pnuts.lang.Runtime;

/**
 * This class defines an abstract interface of script interpreter's implementation,
 * It also gives the default implementation, which is a pure interpreter.
 *
 * @see Context.setPnutsImpl()
 * @see Context.getPnutsImpl()
 */
public class PnutsImpl {

    private Visitor interpreter = new PnutsInterpreter();

    /**
     * Evaluate an expreesion
     *
     * @param expr the expression to be evaluated
     * @param context the context in which the expression is evaluated
     * @return the result of the evaluation
     */
    public Object eval(String str, Context context){
	try {
	    return Pnuts.parse(str).run(context);
	} catch (ParseException e){
	    context.beginLine = e.getErrorLine();
	    checkException(context, e);
	    return null;
	}
    }

    /**
     * Load a script file from local file system
     *
     * @param filename the file name of the script
     * @param context the context in which the expression is evaluated
     * @return the result of the evaluation
     */
    public Object loadFile(String filename, Context context)
	throws FileNotFoundException
    {
	URL scriptURL = null;
	try {
	    File f = new File(new File(filename).getCanonicalPath());
	    if (!f.exists()){
		throw new FileNotFoundException(filename);
	    }
	    scriptURL = Runtime.fileToURL(f);
	} catch (IOException e1){
	    throw new FileNotFoundException(filename);
	}
	try {
	    pushFile(scriptURL, context);
	    return load(scriptURL, context);
	} catch (Throwable t){
	    checkException(context, t);
	    return null;
	} finally {
	    popFile(context);
	}
    }

    /**
     * Load a script file using classloader
     *
     * @param file the name of the script
     * @param context the context in which the script is executed
     * @return the result of the evaluation
     */
    public Object load(String file, Context context) throws FileNotFoundException {
	URL url = Runtime.getScriptURL(file, context);
	if (url == null){
	    throw new FileNotFoundException(file);
	}
	try {
	    pushFile(url, context);
	    Object ret = load(url, context);
	    Pnuts.provide(file);
	    return ret;
	} catch (Throwable t){
	    checkException(context, t);
	    return null;
	} finally {
	    popFile(context);
	}
    }

    /**
     * Load a script file from a URL
     *
     * @param scriptURL the URL of the script
     * @param context the context in which the script is executed
     * @return the result of the evaluation
     */
    public Object load(URL scriptURL, Context context) {
	InputStream in = null;
	try {
	    in = scriptURL.openStream();
	    return Pnuts.parse(in).run(context);
	} catch (Escape e){
	    return e.getValue();
	} catch (ParseException e){
	    context.beginLine = e.getErrorLine();
	    checkException(context, e);
	    return null;
	} catch (IOException e){
	    checkException(context, e);
	    return null;
	} finally {
	    if (in != null){
		try {
		    in.close();
		} catch (IOException io){}
	    }
	}
    }

    public Object load(Reader reader, Context context){
	try {
	    return Pnuts.parse(reader).run(context);
	} catch (ParseException e){
	    context.beginLine = e.getErrorLine();
	    checkException(context, e);
	    return null;
	}
    }
    
    public Object accept(SimpleNode node, Context context){
	return node.accept(interpreter, context);
    }

    /**
     * Tell the context that it's started processing the script file.
     */
    protected void pushFile(Object file, Context context){
	context.pushFile(file);
    }

    /**
     * Tell the context that the current script file has been
     * completed.
     */
    protected void popFile(Context context){
	context.popFile();
    }

    /**
     * Check if any exception handler for the specified exception is defined.
     */
    protected void checkException(Context context, Throwable t){
	Runtime.checkException(context, t);
    }
    
}
