/*
 * @(#) PnutsInterpreter.java	1.35 00/08/01
 *
 * Copyright (c) 1997-1999 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.lang;

import java.lang.reflect.Array;
import java.io.Reader;
import java.io.PrintWriter;
import java.util.Hashtable;
import java.util.Enumeration;
import pnuts.lang.Runtime;

/**
 * The pure interpreter
 */
public class PnutsInterpreter extends Runtime implements Visitor, PnutsParserTreeConstants {

    private final static Integer one = new Integer(1);
    private final static Object[] NO_PARAM = PnutsException.NO_PARAM;

    static final PnutsInterpreter instance = new PnutsInterpreter();

    final static int CODE = 0;
    final static int VALUE = 1;
    final static int BROKEN = 0;
    final static int CONT = 1;
    final static int PASSED = 2;

    // node.info = int[]{0 or 1 or -1}; 0:EOL, 1:EXPR, -1:EOF

    public Object start(SimpleNode node, Context context){
	if (node.jjtGetNumChildren() > 0){
	    if (context.stackFrame == null){
		context.stackFrame = new StackFrame();
	    }
	    try {
		return accept(node, 0, context);
	    } catch (Throwable t){
		checkException(context, t);
	    }
	}
	return null;
    }

    public Object startSet(SimpleNode node, Context context){
	if (context.stackFrame == null){
	    context.stackFrame = new StackFrame();
	}
	int n = node.jjtGetNumChildren();
	Object ret = null;
	try {
	    for (int i = 0; i < n; i++) {
		ret = accept(node, i, context);
	    }
	} catch (Throwable t){
	    checkException(context, t);
	} finally {
	    PrintWriter pw = context.getOutputStream();
	    if (pw != null){
		pw.flush();
	    }
	}
	return ret;
    }

    public Object expressionList(SimpleNode node, Context context){
	Object ret = null;
	int n = node.jjtGetNumChildren();
	for (int i = 0; i < n; i++) {
	    ret = accept(node, i, context);
	}
	return ret;
    }

    public Object global(SimpleNode node, Context context){
	String symbol = node.str;
	if (Package.globalPackage.defined(symbol, context)){
	    return Package.globalPackage.get(symbol, context);
	} else {
	    return context.undefined(symbol);
	}
    }

    public Object idNode(SimpleNode node, Context context){
	return context.getValue(node.str);
    }

    public Object arrayType(SimpleNode node, Context context){
	SimpleNode parent = node.parent;
	int count = 1;
	while (parent.id == JJTARRAYTYPE){
	    count++;
	    parent = parent.parent;
	}
	node.info = new int[]{count}; // dimension
	Object target = accept(node, 0, context);
	SimpleNode child = node.children[0];
	if (target instanceof Class){
	    return Runtime.arrayType((Class)target, 1);
	} else if (Runtime.isArray(target)){
	    return target;
	} else {
	    throw new PnutsException("classOrArray.expected", NO_PARAM, context);
	}
    }

    public Object listElements(SimpleNode node, Context context){
	int n = node.jjtGetNumChildren();
	Object ret[] = new Object[n];

	for (int i = 0; i < n; i++){
	    ret[i] = accept(node, i, context);
	}
	return ret;
    }

    public Object castExpression(SimpleNode node, Context context){
	Class type = (Class)accept(node, 0, context);
	node.info = type;
	Object obj = accept(node, 1, context);
	return Runtime.cast(context, type, obj, true);
    }

    public Object classNode(SimpleNode node, Context context){

	String name = node.str;
	int n = node.jjtGetNumChildren();
	if (name == null){
	    return PnutsFunction.CLASS;
	}

	for (int i = 0; i < n; i++){
	    SimpleNode ch = node.children[i];
	    if (ch.id == JJTPACKAGE){
		name += ".";
		name += ch.str;
	    }
	}
    
	try {
	    return Pnuts.loadClass(name, context);
	} catch (ClassNotFoundException e){
	    throw new PnutsException(e, context);
	}
    }

    public Object rangeNode(SimpleNode node, Context context){
	Object target = accept(node, 0, context);
	Object idx1 = accept(node, 1, context);
	Object idx2 = null;
	if (node.jjtGetNumChildren() >= 3){
	    idx2 = accept(node, 2, context);
	}
	if (idx1 instanceof Number && (idx2 == null || idx2 instanceof Number)){
	    return Runtime.range(target, (Number)idx1, (Number)idx2);
	} else {
	    throw new PnutsException("number.expected", NO_PARAM, context);
	}
    }

    public Object indexNode(SimpleNode node, Context context){
	Object top = accept(node, 0, context);
	Object idx = accept(node, 1, context);

	// node.info =  [Class componentType, size_of_the_dimension, total_dimension]
	node.info = null;

	if (top == null){
	    throw new PnutsException(new NullPointerException(), context);
	}
	if (idx instanceof Number){
	    int ix = ((Number)idx).intValue();
	    SimpleNode c = node.children[0];
	    int dim = 0;
	    if (node.parent.id == JJTARRAYTYPE){
		dim = ((int[])node.parent.info)[0];
	    }
	    if (c.id == JJTINDEXNODE && c.info != null){
		Class type = (Class)((Object[])c.info)[0];
		int count = ((Integer)((Object[])c.info)[2]).intValue() + 1;
		node.info = new Object[]{type, idx, new Integer(count)};
		int sizes[] = new int[count];
		SimpleNode n = node;
		int i = sizes.length;
		while (n.info != null && i > 0){
		    sizes[--i] = ((Integer)((Object[])n.info)[1]).intValue();
		    n = n.children[0];
		}
		return Array.newInstance(Runtime.arrayType(type, dim), sizes);
	    } else if (top instanceof Indexed){
		return ((Indexed)top).get(ix);
	    } else if (top instanceof String){
		return new Character(((String)top).charAt(ix));
	    } else if (Runtime.isArray(top)){
		if (ix >= 0 && ix < Array.getLength(top)){
		    return Array.get(top, ix);
		} else {
		    throw new PnutsException(new IndexOutOfBoundsException("" + ix), context);
		}
	    } else if (c.id == JJTCLASS || c.id == JJTIDNODE && top instanceof Class){
		node.info = new Object[]{top, idx, one};
		return Array.newInstance(Runtime.arrayType((Class)top, dim), ix);
	    } else {
		throw new PnutsException("illegal.subscript", NO_PARAM, context);
	    }
	} else {
	    throw new PnutsException("number.expected", NO_PARAM, context);
	}
    }

    public Object methodNode(SimpleNode node, Context context){
	Object target = accept(node, 0, context);
	Object[] arg = (Object[])accept(node, 1, context);

	SimpleNode argNode = node.children[1];
	Class types[] = null;
	for (int i = 0; i < arg.length; i++){
	    SimpleNode n = argNode.children[i];
	    if (n.id == JJTCASTEXPRESSION){
		if (types == null){
		    types = new Class[arg.length];
		}
		types[i] = (Class)n.info;
	    }
	}

	if (target == null){
	    PnutsException pe = new PnutsException(new NullPointerException(), context);
	    pe.line = node.beginLine;
	    throw pe;
	}

	return Runtime.callMethod(context, target.getClass(), node.str, arg, types, target);
    }

    public Object staticMethodNode(SimpleNode node, Context context){
	SimpleNode c1 = node.children[0];
	String pkgName = getPackageName(c1);
			     
	SimpleNode argNode = node.children[1];
	Class types[] = null;
	Object[] arg = null;

	Package pkg = Package.find(pkgName);
	if (pkg != null){
	    Object fun = pkg.get(node.str, context);

	    if (fun instanceof PnutsFunction){
		arg = (Object[])accept(node, 1, context);
		PnutsFunction ft = (PnutsFunction)fun;
		return ft.exec(arg, context);
	    } else if (fun instanceof Class){
		arg = (Object[])accept(node, 1, context);
		for (int i = 0; i < arg.length; i++){
		    SimpleNode n = argNode.children[i];
		    if (n.id == JJTCASTEXPRESSION){
			if (types == null){
			    types = new Class[arg.length];
			}
			types[i] = (Class)n.info;
		    }
		}
		return Runtime.callConstructor(context, (Class)fun, arg, types);
	    } else {
		throw new PnutsException("illegal.staticCall", 
					new Object[]{
						pkgName,
						node.str,
						new Integer(argNode.jjtGetNumChildren())},
					context);
	    }
	}
	Object target = accept(node, 0, context);
	arg = (Object[])accept(node, 1, context);
	if (target instanceof Class){
	    for (int i = 0; i < arg.length; i++){
		SimpleNode n = argNode.children[i];
		if (n.id == JJTCASTEXPRESSION){
		    if (types == null){
			types = new Class[arg.length];
		    }
		    types[i] = (Class)n.info;
		}
	    }
	    return Runtime.callMethod(context, (Class)target, node.str, arg, types, null);
	} else {
	    throw new PnutsException("illegal.staticCall", 
					 new Object[]{pkgName, node.str, new Integer(arg.length)},
					 context);
	}
    }

    public Object memberNode(SimpleNode node, Context context){
	Object target = accept(node, 0, context);

	if (target == null){
	    throw new PnutsException(new NullPointerException(), context);
	} else if (Runtime.isArray(target) && node.str.equals("length")){
	    return new Integer(Array.getLength(target));
	}
	return Runtime.getField(context, target, node.str);
    }

    static String getPackageName(SimpleNode node){
	if (node.jjtGetNumChildren() > 0){
	    SimpleNode c1 = node.children[0];
	    return getPackageName(c1) + "::" + node.str;
	} else {
	    return node.str;
	}
    }

    public Object staticMemberNode(SimpleNode node, Context context){

	SimpleNode c1 = node.children[0];
	String pkgName = getPackageName(c1);

	Package pkg = Package.find(pkgName);
	if (pkg != null){
	    if (pkg.defined(node.str, context)){
		return pkg.get(node.str, context);
	    } else {
		return context.undefined(node.str);
	    }
	}
	Object target = c1.accept(this, context);
	if (target instanceof Class){
	    return Runtime.getStaticField(context, (Class)target, node.str);
	} else {
	    throw new PnutsException("packageOrClass.expected", NO_PARAM, context);
	}
    }

    public Object applicationNode(SimpleNode node, Context context){

	Object[] arg = (Object[])accept(node, 1, context);
	Object target = accept(node, 0, context);
	
	if (target instanceof PnutsFunction){
	    PnutsFunction ftable = (PnutsFunction)target;
	    return ftable.exec(arg, context);

	} else if (target instanceof Class){  // constructor
	    if (((Class)target).isPrimitive()){
		return Runtime.primitive(context, (Class)target, arg[0], true);
	    } else if (((Class)target).isArray()){
		return Runtime.cast(context, (Class)target, arg[0], true);
	    } else {
		SimpleNode argNode = node.children[1];
		Class types[] = null;
		for (int i = 0; i < arg.length; i++){
		    SimpleNode n = argNode.children[i];
		    if (n.id == JJTCASTEXPRESSION){
			if (types == null){
			    types = new Class[arg.length];
			}
			types[i] = (Class)n.info;
		    }
		}
		return Runtime.callConstructor(context, (Class)target, arg, types);
	    }
	}
	throw new PnutsException("funcOrType.expected", new Object[]{Pnuts.format(target)}, context);
    }

    // node.info = [int[] unitOffset]

    public Object integerNode(SimpleNode node, Context context){

	String str = node.str;
	if (node.value != null){
	    if (node.info == null){
		return node.value;
	    } else {
		int offset = ((int[])node.info)[0];
		return Runtime.quantity((Number)node.value,
				      str.substring(0, offset),
				      str.substring(offset),
				      context);
	    }
	}
	Object[] p = Runtime.parseInt(str); // [Number value, int[] unitOffset]
	Number n = (Number)p[0];
	node.value = n;
	if (p[1] != null){
	    node.info = p[1];
	    int offset = ((int[])p[1])[0];
	    return Runtime.quantity(n, str.substring(0, offset), str.substring(offset), context);
	} else {
	    return n;
	}
    }

    public Object floatingNode(SimpleNode node, Context context){
	
	String str = node.str;
	if (node.value != null){
	    if (node.info == null){ // no unit
		return node.value;
	    } else {
		int offset = ((int[])node.info)[0];
		return Runtime.quantity((Number)node.value,
				      str.substring(0, offset),
				      str.substring(offset), context);
	    }
	}
	Object[] p = Runtime.parseFloat(str); // [Number value, int[] unitOffset]
	Number n = (Number)p[0];
	node.value = n;
	if (p[1] != null){
	    node.info = p[1];
	    int offset = ((int[])p[1])[0];
	    return Runtime.quantity(n, str.substring(0, offset), str.substring(offset), context);
	} else {
	    return n;
	}
    }

    public Object characterNode(SimpleNode node, Context context){
	if (node.value != null){
	    return node.value;
	}
	try {
	    return node.value = Runtime.parseChar(node.str);
	} catch (IllegalArgumentException e){
	    throw new PnutsException("illegal.character", new Object[]{node.str}, context);
	}
    }

    public Object stringNode(SimpleNode node, Context context){
	try {
	    if (node.info != null){
		return node.info;
	    }
	    return node.info = Runtime.parseString(node.str);
	} catch (PnutsException pe){
	    throw new PnutsException(pe.getThrowable(), context);
	}
    }

    public Object trueNode(SimpleNode node, Context context){
	return Boolean.TRUE;
    }

    public Object falseNode(SimpleNode node, Context context){
	return Boolean.FALSE;
    }

    public Object nullNode(SimpleNode node, Context context){
	return null;
    }

    static Object compute(Object o1, Object o2, Context context, int id){
	switch (id){
	case JJTASSIGNMENTTA:
	    return Runtime.binary(o1, o2, Arithmetic.MULTIPLY, context);
	case JJTASSIGNMENTMA:
	    return Runtime.binary(o1, o2, Arithmetic.MOD, context);
	case JJTASSIGNMENTDA:
	    return Runtime.binary(o1, o2, Arithmetic.DIVIDE, context);
	case JJTASSIGNMENTPA:
	    return Runtime.binary(o1, o2, Arithmetic.ADD, context);
	case JJTASSIGNMENTSA:
	    return Runtime.binary(o1, o2, Arithmetic.SUBTRACT, context);
	case JJTASSIGNMENTLA:
	    return Runtime.binary(o1, o2, Arithmetic.SHIFT_L, context);
	case JJTASSIGNMENTRA:
	    return Runtime.binary(o1, o2, Arithmetic.SHIFT_R, context);
	case JJTASSIGNMENTRAA:
	    return Runtime.binary(o1, o2, Arithmetic.SHIFT_A, context);
	case JJTASSIGNMENTAA:
	    return Runtime.binary(o1, o2, Arithmetic.AND, context);
	case JJTASSIGNMENTEA:
	    return Runtime.binary(o1, o2, Arithmetic.XOR, context);
	case JJTASSIGNMENTOA:
	    return Runtime.binary(o1, o2, Arithmetic.OR, context);
	default:
	    return null; // never happen
	}
    }

    private static Object putArray(Object array, int idx, Object expr, Context context, int id){
	if (id != JJTASSIGNMENT){
	    Object obj = Array.get(array, idx);
	    expr = compute(obj, expr, context, id);
	}
	Array.set(array, idx, expr);
	return expr;
    }

    private Object doAssign(SimpleNode node, Object expr, Context context, int id){
	if (node.id == JJTIDNODE){
	    if (id != JJTASSIGNMENT){
		expr = compute(context.getValue(node.str), expr, context, id);
	    }
	    context.setValue(node.str, expr);
	    return expr;
	} else if (node.id == JJTGLOBAL){
	    if (id != JJTASSIGNMENT){
		expr = compute(Package.globalPackage.get(node.str, context), expr, context, id);
	    }
	    Package.globalPackage.set(node.str, expr, context);
	    return expr;
	} else if (node.id == JJTINDEXNODE){
	    Object o = accept(node, 0, context);
	    Object idx = accept(node, 1, context);
	    if (!(idx instanceof Number)){
		throw new PnutsException("number.expected", NO_PARAM, context);
	    }
	    int i = ((Number)idx).intValue();
	    if (o instanceof Indexed){
		Indexed indexed = (Indexed)o;
		if (id != JJTASSIGNMENT){
		    expr = compute(indexed.get(i), expr, context, id);
		}
		indexed.set(i, expr);
		return expr;
	    } else if (o instanceof String){
		if (id != JJTASSIGNMENT || !(expr instanceof Character)){
	            throw new PnutsException("illegal.assign", NO_PARAM, context);
		}
		Object value = assignRange(o, (Number)idx, (Number)idx, expr);
		SimpleNode cld = node.children[0];

		if (cld.id == JJTIDNODE ||
		    cld.id == JJTGLOBAL ||
		    cld.id == JJTINDEXNODE ||
		    cld.id == JJTSTATICMEMBERNODE ||
		    cld.id == JJTMEMBERNODE){
		    return doAssign(cld, value, context, id);
		} else {
		    return value;
		}
	    } else if (Runtime.isArray(o)){
		if (id != JJTASSIGNMENT){
		    expr = compute(Array.get(o, i), expr, context, id);
		}
		Array.set(o, i, expr);
		return expr;
	    }
	    throw new PnutsException("illegal.assign", NO_PARAM, context);
	} else if (node.id == JJTSTATICMEMBERNODE){
	    SimpleNode c1 = node.children[0];
	    String pkgName = getPackageName(c1);
	    Package pkg = Package.find(pkgName);
	    if (pkg != null){
		pkg.set(node.str, expr, context);
		return expr;
	    }
	    Object target = c1.accept(this, context);
	    if (target instanceof Class){
		Class clazz = (Class)target;
		if (id != JJTASSIGNMENT){
		    Object f = Runtime.getField(context, clazz, node.str);
		    expr = compute(f, expr, context, id);
		}
		Runtime.putStaticField(context, clazz, node.str, expr);
		return expr;
	    } else {
		throw new PnutsException("package.notFound", NO_PARAM, context);
	    }
	} else if (node.id == JJTMEMBERNODE){
	    Object target = accept(node, 0, context);

	    if (id != JJTASSIGNMENT){
		Object f = Runtime.getField(context, target, node.str);
		expr = compute(f, expr, context, id);
	    }
	    Runtime.putField(context, target, node.str, expr);
	    return expr;
	} else if (node.id == JJTRANGENODE){
	    if (id != JJTASSIGNMENT){
	        throw new PnutsException("illegal.assign", NO_PARAM, context);
	    }
	    Object target = accept(node, 0, context);
	    Object idx1 = accept(node, 1, context);
	    Object idx2 = null;
	    if (node.jjtGetNumChildren() >= 3){
		idx2 = accept(node, 2, context);
	    }
	    if (!(idx1 instanceof Number) || (idx2 != null && !(idx2 instanceof Number))){
		throw new PnutsException("number.expected", NO_PARAM, context);
	    }
	    Object value = assignRange(target, (Number)idx1, (Number)idx2, expr);
	    if (target instanceof String){
		SimpleNode cld = node.children[0];

		if (cld.id == JJTIDNODE ||
		    cld.id == JJTGLOBAL ||
		    cld.id == JJTINDEXNODE ||
		    cld.id == JJTSTATICMEMBERNODE ||
		    cld.id == JJTMEMBERNODE){
		    return doAssign(cld, value, context, id);
		} else {
		    return value;
		}
	    }
	    return value;
	} else {
	    throw new PnutsException("illegal.assign", NO_PARAM, context);
	}
    }

    private Object assign(SimpleNode node, Context context, int id){

	Object expr = accept(node, 1, context);
	try {
	    return doAssign(node.children[0], expr, context, id);
	} catch (PnutsException p){
	    return p;
	} catch (Throwable t){
	    throw new PnutsException(t, context);
	}
    }


    public Object assignment(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENT);
    }

    public Object assignmentTA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTTA);
    }

    public Object assignmentMA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTMA);
    }

    public Object assignmentDA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTDA);
    }

    public Object assignmentPA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTPA);
    }

    public Object assignmentSA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTSA);
    }

    public Object assignmentLA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTLA);
    }

    public Object assignmentRA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTRA);
    }

    public Object assignmentRAA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTRAA);
    }

    public Object assignmentAA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTAA);
    }

    public Object assignmentEA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTEA);
    }

    public Object assignmentOA(SimpleNode node, Context context){
	return assign(node, context, JJTASSIGNMENTOA);
    }

    public Object orNode(SimpleNode node, Context context){
	Object o1 = accept(node, 0, context);
	if (!(o1 instanceof Boolean)){
	    throw new PnutsException("boolean.expected", NO_PARAM, context);
	}
	if (((Boolean)o1).booleanValue()){
	    return Boolean.TRUE;
	} else {
	    Object o2 = accept(node, 1, context);
	    if (!(o2 instanceof Boolean)){
		throw new PnutsException("boolean.expected", NO_PARAM, context);
	    }
	    return o2;
	}
    }

    public Object andNode(SimpleNode node, Context context){
	Object o1 = accept(node, 0, context);
	if (!(o1 instanceof Boolean)){
	    throw new PnutsException("boolean.expected", NO_PARAM, context);
	}
	if (((Boolean)o1).booleanValue()){
	    Object o2 = accept(node, 1, context);
	    if (!(o2 instanceof Boolean)){
		throw new PnutsException("boolean.expected", NO_PARAM, context);
	    }
	    return o2;
	} else {
	    return Boolean.FALSE;
	}
    }

    public Object logOrNode(SimpleNode node, Context context){
	return Runtime.binary(accept(node, 0, context),
			      accept(node, 1, context),
			      Arithmetic.OR,
			      context);
    }

    public Object logExorNode(SimpleNode node, Context context){
	return Runtime.binary(accept(node, 0, context),
			      accept(node, 1, context),
			      Arithmetic.XOR,
			      context);
    }

    public Object logAndNode(SimpleNode node, Context context){
	return Runtime.binary(accept(node, 0, context),
			      accept(node, 1, context),
			      Arithmetic.AND,
			      context);
    }

    public Object equalNode(SimpleNode node, Context context){
	return eq(node, context) ? Boolean.TRUE : Boolean.FALSE;
    }

    public Object notEqNode(SimpleNode node, Context context){
	return eq(node, context) ? Boolean.FALSE : Boolean.TRUE;
    }

    boolean eq(SimpleNode node, Context context){
	return eq(accept(node, 0, context), accept(node, 1, context), context);
    }

    static boolean eq(Object o1, Object o2, Context context){
	return Runtime.compareTo(o1, o2, context) == Arithmetic.EQUAL;
    }

    public Object instanceofExpression(SimpleNode node, Context context){
	Object o1 = accept(node, 0, context);
	Object o2 = accept(node, 1, context);

	if (o2 instanceof Class){
	    Class c = (Class)o2;
	    return (c.isInstance(o1) ? Boolean.TRUE : Boolean.FALSE);
	} else if (o2 == null){
	    return Boolean.FALSE;
	} else {
	    throw new PnutsException("class.expected", NO_PARAM, context);
	}
    }

    public Object ltNode(SimpleNode node, Context context){
	int i = Runtime.compareTo(accept(node, 0, context),
				  accept(node, 1, context),
				  context);
	if (i == Arithmetic.NOT_EQUAL){
	    throw new PnutsException("illegal.operandType", NO_PARAM, context);
	}
	return (i == Arithmetic.RIGHT_IS_BIGGER) ? Boolean.TRUE : Boolean.FALSE;
    }

    public Object gtNode(SimpleNode node, Context context){
	int i = Runtime.compareTo(accept(node, 0, context),
				  accept(node, 1, context),
				  context);
	if (i == Arithmetic.NOT_EQUAL){
	    throw new PnutsException("illegal.operandType", NO_PARAM, context);
	}
	return (i == Arithmetic.LEFT_IS_BIGGER) ? Boolean.TRUE : Boolean.FALSE;
    }

    public Object leNode(SimpleNode node, Context context){
	int i = Runtime.compareTo(accept(node, 0, context),
				  accept(node, 1, context),
				  context);
	if (i == Arithmetic.NOT_EQUAL){
	    throw new PnutsException("illegal.operandType", NO_PARAM, context);
	}
	if (i == Arithmetic.EQUAL || i == Arithmetic.RIGHT_IS_BIGGER){
	    return Boolean.TRUE;
	} else {
	    return Boolean.FALSE;
	}
    }

    public Object geNode(SimpleNode node, Context context){
	int i = Runtime.compareTo(accept(node, 0, context),
				  accept(node, 1, context),
				  context);
	if (i == Arithmetic.NOT_EQUAL){
	    throw new PnutsException("illegal.operandType", NO_PARAM, context);
	}
	if (i == Arithmetic.EQUAL || i == Arithmetic.LEFT_IS_BIGGER){
	    return Boolean.TRUE;
	} else {
	    return Boolean.FALSE;
	}
    }

    public Object shiftLeftNode(SimpleNode node, Context context){
	return Runtime.binary(accept(node, 0, context),
			      accept(node, 1, context),
			      Arithmetic.SHIFT_L,
			      context);
    }

    public Object shiftRightNode(SimpleNode node, Context context){
	return Runtime.binary(accept(node, 0, context),
			      accept(node, 1, context),
			      Arithmetic.SHIFT_R,
			      context);
    }

    public Object shiftArithmeticNode(SimpleNode node, Context context){
	return Runtime.binary(accept(node, 0, context),
			      accept(node, 1, context),
			      Arithmetic.SHIFT_A,
			      context);
    }

    public Object addNode(SimpleNode node, Context context){
	return Runtime.binary(accept(node, 0, context),
			      accept(node, 1, context),
			      Arithmetic.ADD,
			      context);
    }

    public Object subtractNode(SimpleNode node, Context context){
	return Runtime.binary(accept(node, 0, context),
			      accept(node, 1, context),
			      Arithmetic.SUBTRACT,
			      context);
    }

    public Object multNode(SimpleNode node, Context context){
	return Runtime.binary(accept(node, 0, context),
			      accept(node, 1, context),
			      Arithmetic.MULTIPLY,
			      context);
    }

    public Object divideNode(SimpleNode node, Context context){
	return Runtime.binary(accept(node, 0, context),
			      accept(node, 1, context),
			      Arithmetic.DIVIDE,
			      context);
    }

    public Object modNode(SimpleNode node, Context context){
	return Runtime.binary(accept(node, 0, context),
			      accept(node, 1, context),
			      Arithmetic.MOD,
			      context);
    }

    public Object negativeNode(SimpleNode node, Context context){
	return Runtime.unary(accept(node, 0, context),
			     Arithmetic.NEGATE,
			     context);
    }

    public Object preIncrNode(SimpleNode node, Context context){
	return doAssign(node.children[0], one, context, JJTASSIGNMENTPA);
    }

    public Object preDecrNode(SimpleNode node, Context context){
	return doAssign(node.children[0], one, context, JJTASSIGNMENTSA);
    }

    public Object postIncrNode(SimpleNode node, Context context){
	SimpleNode n = node.children[0];
	Object ret = n.accept(this, context);
	Object dummy = doAssign(n, one, context, JJTASSIGNMENTPA);
	return ret;
    }

    public Object postDecrNode(SimpleNode node, Context context){
	SimpleNode n = node.children[0];
	Object ret = n.accept(this, context);
	Object dummy = doAssign(n, one, context, JJTASSIGNMENTSA);
	return ret;
    }

    public Object logNotNode(SimpleNode node, Context context){
	return Runtime.unary(accept(node, 0, context),
			     Arithmetic.NOT,
			     context);
    }

    public Object notNode(SimpleNode node, Context context){
	Object o = accept(node, 0, context);
	if (o instanceof Boolean){
	    return ((Boolean)o).booleanValue() ? Boolean.FALSE : Boolean.TRUE;
	} else {
	    throw new PnutsException("boolean.expected", NO_PARAM, context);
	}
    }

    public Object continueNode(SimpleNode node, Context context){
	SimpleNode parent = node.parent;
	while (parent != null){
	    if (parent.id == JJTFOREACHSTATEMENT){
 		setContinued(parent.children[1]);
		break;
	    } else if (parent.id == JJTFORSTATEMENT){
		break;
	    } else if (parent.id == JJTWHILESTATEMENT){
 		setContinued(parent.children[1]);
		break;
	    } else if (parent.id == JJTBLOCK){
 		setContinued(parent);
	    }
	    parent = parent.parent;
	}
	return null;
    }

    public Object breakNode(SimpleNode node, Context context){
	Object o = null;
	if (node.jjtGetNumChildren() > 0){
	    o = accept(node, 0, context);
	}

	SimpleNode parent = node.parent;
	while (parent != null){
	    if (parent.id == JJTFORSTATEMENT ||
		parent.id == JJTWHILESTATEMENT ||
		parent.id == JJTFOREACHSTATEMENT ||
		parent.id == JJTSWITCHSTATEMENT){
		setBroken(parent, o);
		break;
	    } else if (parent.id == JJTBLOCK ||
		       parent.id == JJTSWITCHBLOCK){
		setBroken(parent, o);
	    }
	    parent = parent.parent;
	}
	return o;
    }

    public Object returnNode(SimpleNode node, Context context){
	Object o = null;
	if (node.jjtGetNumChildren() > 0){
	    o = accept(node, 0, context);
	}
	if (context.frame != null){
	    throw new Jump(o);
	} else { 
	    throw new Escape(o);
	}
    }

    public Object catchNode(SimpleNode node, Context context){
	if (node.jjtGetNumChildren() == 0){
	    return null;
	}
	Object[] args = (Object[])accept(node, 0, context);
	if (args.length != 2){
	    throw new PnutsException("wrong.numberOfParameters", new Object[]{"catch"}, context);
	}
	context.catchException((Class)args[0], (PnutsFunction)args[1]);
	return null;
    }

    public Object blockNode(SimpleNode node, Context context){
	int n = node.jjtGetNumChildren();
	Object last = null;
	Object attr[] = (Object[])node.info;
	attr[CONT] = Boolean.FALSE;
	for (int i = 0; i < n; i++){
	    try {
		last = accept(node, i, context);
	    } catch (PnutsException e){
		Throwable t = e.getThrowable();
		if (t instanceof ThreadDeath){
		    setBroken(node, last);
		}
		throw e;
	    }
	    if (attr[BROKEN] == Boolean.TRUE){
		attr[BROKEN] = Boolean.FALSE;
		last = attr[PASSED];
		break;
	    }
	    if (attr[CONT] == Boolean.TRUE){
		break;
	    }
	}
	return last;
    }

    public Object ifStatement(SimpleNode node, Context context){

	Object con = accept(node, 0, context);
	if (condition(con, context)){
	    return accept(node, 1, context);
	}
	int n = node.jjtGetNumChildren();
	for (int i = 2; i < n; i++){
	    SimpleNode _node = node.children[i];
	    if (_node.id == JJTELSEIFNODE){
		if (condition(accept(_node, 0, context), context)){
		    return accept(_node, 1, context);
		}
	    } else if (_node.id == JJTELSENODE){
		return accept(_node, 0, context);
	    }
	}
	return null;
    }

    static boolean condition(Object cond, Context context){
	if (cond instanceof Boolean){
	    return ((Boolean)cond).booleanValue();
	} else {
	    throw new PnutsException("boolean.expected", NO_PARAM, context);
	}
    }

    public Object whileStatement(SimpleNode node, Context context){

	Object[] attr = (Object[])node.info;
	while (true){
	    Object cond = accept(node, 0, context);
	    if (cond instanceof Boolean){
		Object last = null;
		boolean b = ((Boolean)cond).booleanValue();
		if (b){
		    last = accept(node, 1, context);

		    if (attr[BROKEN] == Boolean.TRUE){
			attr[BROKEN] = Boolean.FALSE;
			return attr[PASSED];
		    }
		} else {
		    attr[BROKEN] = Boolean.FALSE;
		    return last;
		}
	    } else {
		throw new PnutsException("boolean.expected", NO_PARAM, context);
	    }
	}
    }

    public Object forStatement(SimpleNode node, Context context){
	SimpleNode initNode = null;
	SimpleNode condNode = null;
	SimpleNode updateNode = null;
	SimpleNode blockNode = null;

	int j = 0;
	SimpleNode n = node.children[j];
	if (n.id == JJTFORINIT){
	    initNode = n;
	    j++;
	}
	n = node.children[j];
	if (n.id != JJTFORUPDATE && n.id != JJTBLOCK){
	    condNode = n;
	    j++;
	}
	n = node.children[j];
	if (n.id == JJTFORUPDATE){
	    updateNode = n;
	    j++;
	}

	blockNode = node.children[j];

	Object last = null;

	if (initNode != null){
	    int num = initNode.jjtGetNumChildren();
	    String[] env = new String[num];
	    for (int i = 0; i < env.length; i++){
		SimpleNode sn = initNode.children[i];
		env[i] = sn.str;
	    }
	    context.openLocal(env);
	    for (int i = 0; i < env.length; i++){
		SimpleNode sn = initNode.children[i];
		context.bind(env[i], accept(sn, 0, context));
	    }
	} else {
	    context.openLocal(new String[]{});
	}

	Object c = null;
	if (condNode != null){
	    c = condNode.accept(this, context);
	} else {
	    c = Boolean.TRUE;
	}
	Object[] attr = (Object[])node.info;
	try {
	    while (((Boolean)c).booleanValue()){
		last = blockNode.accept(this, context);
		if (attr[BROKEN] == Boolean.TRUE){
		    attr[BROKEN] = Boolean.FALSE;
		    last = attr[PASSED];
		    break;
		}
		if (updateNode != null){
		    expressionList(updateNode, context);
		}
		if (condNode != null){
		    c = condNode.accept(this, context);
		}
	    }
	} catch (ClassCastException e){
	    throw new PnutsException("boolean.expected", NO_PARAM, context);
	}
	context.closeLocal();
	return last;
    }

    public Object foreachStatement(SimpleNode node, Context context){
	SimpleNode n2 = node.children[1];

	Object o = accept(node, 0, context);
	if (o == null){
	    return null;
	}
	Object[] attr = (Object[])node.info;
	if (Runtime.isArray(o)){
	    Object last = null;
	    context.openLocal(new String[]{node.str});
	    for (int i = 0; i < Array.getLength(o); i++){
		context.bind(node.str, Array.get(o, i));
		last = n2.accept(this, context);
		if (attr[BROKEN] == Boolean.TRUE){
		    attr[BROKEN] = Boolean.FALSE;
		    last = attr[PASSED];
		    break;
		}
	    }
	    context.closeLocal();
	    return last;
	} else if (o instanceof Enumeration){
	    Enumeration e = (Enumeration)o;
	    Object last = null;
	    context.openLocal(new String[]{node.str});
	    while (e.hasMoreElements()){
		context.bind(node.str, e.nextElement());
		last = n2.accept(this, context);
		if (attr[BROKEN] == Boolean.TRUE){
		    attr[BROKEN] = Boolean.FALSE;
		    last = attr[PASSED];
		    break;
		}
	    }
	    context.closeLocal();
	    return last;
	} else {
	    return Pnuts.verdep.doIterator(o, node, n2, this, context);
	}
    }

    public Object switchStatement(SimpleNode node, Context context){
	Object[] attr = (Object[])node.info;
	int n = node.jjtGetNumChildren();
	Object target = accept(node, 0, context);
	Object last = null;
	boolean matched = false;
	for (int i = 1; i < n; i++){
	    if (attr[BROKEN] == Boolean.TRUE){
		last = attr[PASSED];
		break;
	    }
	    SimpleNode _node = node.children[i];
	    if (_node.jjtGetNumChildren() == 1){ // case
		Object o = accept(_node, 0, context);
		i++;
		if (matched || eq(target, o, context)){
		    last = accept(node, i, context);
		    matched = true;
		}
	    } else { // default
		attr[BROKEN] = Boolean.FALSE;
		matched = true;
		last = accept(node, ++i, context);
	    }
	}
	attr[BROKEN] = Boolean.FALSE;
	return last;
    }

    public Object switchBlock(SimpleNode node, Context context){
	Object[] attr = (Object[])node.info;
	int n = node.jjtGetNumChildren();
	Object last = null;
	for (int i = 0; i < n; i++){
	    try {
		last = accept(node, i, context);
	    } catch (PnutsException e){
		Throwable t = e.getThrowable();
		if (t instanceof ThreadDeath){
		    setBroken(node, last);
		}
		throw e;
	    }
	    if (attr[BROKEN] == Boolean.TRUE){
		attr[BROKEN] = Boolean.FALSE;
		last = attr[PASSED];
		break;
	    }
	}
	return last;
    }

    public Object functionStatement(SimpleNode node, Context context){
	int n = node.jjtGetNumChildren();

	SimpleNode block = null;
	int nargs = 0;

	String locals[] = null;

	for (int i = 0; i < n; i++){
	    SimpleNode _node = node.children[i];
	    if (_node.id == JJTPARAMLIST){
		nargs = _node.jjtGetNumChildren();
		locals = new String[nargs];

		SimpleNode n0 = null;
		if (nargs == 1 &&
		    (n0 = _node.children[0]).id == JJTINDEXNODE){
		    nargs = -1;
		    locals[0] = n0.children[0].str;
		} else {
		    for (int j = 0; j < nargs; j++){
			locals[j] = _node.children[j].str;
		    }
		}
	    } else if (_node.id == JJTBLOCK){
		block = _node;
	    }
	}

	Package pkg = context.getCurrentPackage();
	Function f = new Function(node.str, locals, nargs, block, pkg, context);

	Object o = null;
	if (node.str != null && context.stackFrame != null){
	    o = context.stackFrame.get(node.str);
	}
	String name = f.getName();
	if (o == null){
	    if (pkg.defined(name, context)){
		o = pkg.get(name, context);
	    }
	}
	if (!(o instanceof PnutsFunction)){
	    o = null;
	} else if (!((PnutsFunction)o).getName().equals(name)){
	    o = null;
	}
	PnutsFunction ht = f.register((PnutsFunction)o);

	if (node.str != null){
	    context.setValue(node.str, ht);
	}
	return ht;
    }

    protected static void setBroken(SimpleNode node, Object value){
	Object[] attr = (Object[])node.info;
	attr[BROKEN] = Boolean.TRUE;
	attr[PASSED] = value;
    }

    protected static void setContinued(SimpleNode node){
	Object[] attr = (Object[])node.info;
	attr[CONT] = Boolean.TRUE;
    }

    protected Object accept(SimpleNode node, int idx, Context context){
	return node.children[idx].accept(this, context);
    }
}
