/*
 * @(#) Runtime.java	1.17 00/08/01
 *
 * Copyright (c) 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.lang;

import java.util.Hashtable;
import java.util.Enumeration;
import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;
import java.io.CharArrayWriter;
import java.io.PrintWriter;
import java.io.File;
import java.io.Serializable;
import java.io.IOException;
import java.net.URL;
import java.net.URLConnection;
import java.net.MalformedURLException;
import java.applet.Applet;
import java.lang.reflect.Method;
import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.lang.reflect.Modifier;
import java.lang.reflect.InvocationTargetException;

/**
 * This class provides runtime supports for Pnuts compiler/interpreter.
 * Most of the methods are protected static, so that only subclasses can access them.
 *
 * @version	1.17
 */
public class Runtime implements PnutsParserTreeConstants {
    private static String java_vendor = System.getProperty("java.vendor");
    static {
	if (java_vendor == null){
	    java_vendor = "";
	}
    }
    private final static Object[] defaultLoadPath = new Object[]{"/", ""};

    private final static boolean DEBUG = false;

    protected Runtime(){}

    /**
     * Executes a compiled script.  Exceptions are checked if
     * an exception handler is registered with the catch() function.
     * Output stream of the specified context is flushed after the
     * script is executed or an exception is thrown.
     *
     * @param context the context in which this object is executed
     * @return the result of the execution
     */
    public Object execute(Context context){
	context.runtime = this;
	try {
	    return exec(context);
	} catch (Throwable t){
	    checkException(context, t);
	    return null;
	} finally {
	    PrintWriter pw = context.getOutputStream();
	    if (pw != null){
		pw.flush();
	    }
	}
    }

    /**
     * This method is overrided by classes generated by the compiler.
     *
     * @param context the context in which this object is executed
     * @return the result of the execution
     */
    protected Object exec(Context context){
	return null;
    }

    /**
     * Call a method
     *
     * @param context the context in which the method is called
     * @param c the class of method
     * @param name the method name
     * @param args the paramters
     * @param types the types of the paramters
     * @param target the target object
     * @return the return value of the call
     */
    public static Object callMethod(Context context, Class c, String name, Object args[], Class types[], Object target){
	if (target instanceof AbstractData){
	    return ((AbstractData)target).invoke(name, args, context);
	}
	try {
	    return callMethod(c, name, args, types, target);
	} catch (InvocationTargetException e1){
	    Throwable t = e1.getTargetException();
	    if (t instanceof PnutsException){
		throw (PnutsException)t;
	    } else {
	        throw new PnutsException(t, context);
	    }
	} catch (IllegalArgumentException e2){
	    throw new PnutsException("method.notFound",
				     new Object[]{name, ""+(target==null?c:target), ""+Pnuts.format(args)}, context);
	} catch (PnutsException e3){
	    throw e3;
	} catch (Throwable e4){
	    throw new PnutsException(e4, context);
	}
    }

    /**
     * Call a method
     *
     * @param c the class of method
     * @param name the method name
     * @param args the paramters
     * @param target the target object
     * @return the return value
     */
    public static Object callMethod(Class c, String name, Object args[], Object target)
	throws IllegalAccessException, IllegalArgumentException, InvocationTargetException {
	    return callMethod(c, name, args, null, target);
    }

    /**
     * Call a method
     *
     * @param c the class of method
     * @param name the method name
     * @param args the paramters
     * @param types the types of the formal arguments
     * @param target the target object
     * @return the return value
     */
    public static Object callMethod(Class c, String name, Object args[], Class types[], Object target)
	throws IllegalAccessException, IllegalArgumentException, InvocationTargetException
    {
        Method method = null;
	boolean _static = (target == null);
	try {
	    if (c.isArray()){
		c = Object.class;  // This line is needed for JavaStation
	    }
	    Method m[] = getMethods(c, name);

	    int count = 0;
	    int min = Integer.MAX_VALUE;
	    Stack methods = new Stack();
	cand:
	    for (int i = 0; i < m.length; i++){
		Method mi = m[i];
		Class p[] = mi.getParameterTypes();
		if (p.length != args.length){
		    continue;
		}
		count = 0;
		for (int j = 0; j < p.length; j++){
		    Class pj = p[j];
		    if (types != null){
			Class tj = types[j];
			if (tj != null && pj != tj && !pj.isAssignableFrom(tj)){
			    continue cand;
			}
		    }
		    int t = matchType(pj, args[j]);
		    if (t < 0){
			continue cand;
		    }
		    count += t;
		}
		if (count > min){
		    continue;
		}
		
		boolean st = Modifier.isStatic(mi.getModifiers());
		if (st != _static){
		    continue;
		}
		    
		if (count < min){
		    methods.removeAllElements();
		    methods.push(mi);
		    min = count;
		} else if (count == min){
		    methods.push(mi);
		}
	    }
	    Class clazz = c;
	out:
	    while (clazz != null){
		int size = methods.size();
		for (int i = 0; i < size; i++){
		    method = (Method)methods.pop();
		    if (method.getDeclaringClass() == clazz){
			break out;
		    }
		}
		clazz = clazz.getSuperclass();
	    }

	    if (method != null){
		if (args.length > 0){
		    Class p[] = method.getParameterTypes();
		    for (int j = 0; j < p.length; j++){
			Class pj = p[j];
			if (pj.isArray() && args[j] != null && isArray(args[j])){
			    if (!pj.isInstance(args[j])){
				Class compType = getBottomType(pj);
				args[j] = transform(compType, args[j]);
			    }
			}
		    }
		}
		return method.invoke(target, args);
	    } else {
		throw new IllegalArgumentException();
	    }
	} catch (IllegalAccessException pe){
	    return Pnuts.verdep.reInvoke(pe, method, target, args);
	}
    }

    /**
     * Call a constructor
     *
     * @param context the context in which the constructor is called
     * @param c the class of method
     * @param args the paramters
     * @param types the types of the formal arguments
     * @return the created instance
     */
    public static Object callConstructor(Context context, Class c, Object args[], Class types[]){
	try {
	    return callConstructor(c, args, types);
	} catch (InvocationTargetException e1){
	    Throwable t = ((InvocationTargetException)e1).getTargetException();
	    if (t instanceof PnutsException){
		throw (PnutsException)t;
	    } else {
		throw new PnutsException(t, context);
	    }
	} catch (IllegalArgumentException e2){
	    throw new PnutsException("constructor.notFound", new Object[]{c, Pnuts.format(args)}, context);
	} catch (PnutsException e3){
	    throw e3;
	} catch (Throwable e4){
	    throw new PnutsException(e4, context);
	}
    }

    /**
     * Call a constructor
     *
     * @param c the class of constructor
     * @param args the paramters
     * @return the created instance
     */
    public static Object callConstructor(Class c, Object args[])
	throws IllegalAccessException, IllegalArgumentException, InvocationTargetException, InstantiationException
    {
	return callConstructor(c, args, null);
    }

    /**
     * Call a constructor
     *
     * @param c the class of constructor
     * @param args the paramters
     * @param types the types of the formal arguments
     * @return the created instance
     */
    public static Object callConstructor(Class c, Object args[], Class types[])
	throws IllegalAccessException, IllegalArgumentException, InvocationTargetException, InstantiationException
    {
	Constructor cs[] = getConstructors(c);

	Constructor cons = null;
	int count = 0;
	int min = Integer.MAX_VALUE;
    cand:
	for (int i = 0; i < cs.length; i++){
	    Class p[] = cs[i].getParameterTypes();
	    if (p.length != args.length){
		continue;
	    }
	    count = 0;
	    for (int j = 0; j < p.length; j++){
		Class pj = p[j];
		if (types != null){
		    Class tj = types[j];
		    if (tj != null && pj != tj && !pj.isAssignableFrom(tj)){
			continue cand;
		    }
		}
		int t = matchType(pj, args[j]);
		if (t < 0){
		    continue cand;
		}
		count += t;
	    }
	    if (count < min){
		min = count;
		cons = cs[i];
	    }
	}
	if (cons != null){
	    Class p[] = cons.getParameterTypes();
	    for (int j = 0; j < p.length; j++){
		Class pj = p[j];
		if (pj.isArray() && args[j] != null && isArray(args[j])){
		    if (!pj.isInstance(args[j])){
			Class compType = getBottomType(pj);
			args[j] = transform(compType, args[j]);
		    }
		}
	    }
	    return cons.newInstance(args);
	} else {
	    throw new IllegalArgumentException();
	}
    }

    /**
     * Assign an object to a static field.
     *
     * @param context the context in which the field is accessed
     * @param clazz the class in which the static field is defined
     * @param name the name of the static field
     * @param expr the value to be assigned
     */
    public static void putStaticField(Context context, Class clazz, String name, Object expr) {
	try {
	    clazz.getField(name).set(null, expr);
	} catch (PnutsException e0){
	    throw e0;
	} catch (NoSuchFieldException f){
	    throw new PnutsException("field.notFound", new Object[]{name, clazz}, context);
	} catch (Throwable e){
	    throw new PnutsException(e, context);
	}
    }

    /**
     * Get the value of a static field.
     *
     * @param context the context in which the field is accessed
     * @param clazz the class in which the static field is defined
     * @param name the name of the static field
     * @return the value
     */
    public static Object getStaticField(Context context, Class clazz, String name){
	try {
	    return getStaticField(clazz, name);
	} catch (PnutsException p){
	    throw p;
	} catch (NoSuchFieldException f){
	    throw new PnutsException("field.notFound", new Object[]{name, clazz}, context);
	} catch (Throwable t){
	    throw new PnutsException(t, context);
	}
    }

    static Object getStaticField(Class clazz, String name) throws IllegalArgumentException, IllegalAccessException, NoSuchFieldException {
	return clazz.getField(name).get(null);
    }

    /**
     * Assign an object to a instance field.
     *
     * @param context the context in which the field is accessed
     * @param target the target object of the field
     * @param name the name of the field
     * @param expr the value to be assigned
     */
    public static void putField(Context context, Object target, String name, Object expr) {
	try {
	    if (target instanceof Property){
		((Property)target).set(name, expr, context);
	    } else {
		target.getClass().getField(name).set(target, expr);
	    }
	} catch (PnutsException p){
	    throw p;
	} catch (NoSuchFieldException f){
	    throw new PnutsException("field.notFound", new Object[]{name, target.getClass()}, context);
	} catch (Throwable t){
	    throw new PnutsException(t, context);
	}
    }

    /**
     * Get the value of a instance field.
     *
     * @param context the context in which the field is accessed
     * @param target the target object of the field
     * @param name the name of the field
     * @return the value
     */
    public static Object getField(Context context, Object target, String name){
	try {
	    if (target instanceof Property){
		return ((Property)target).get(name, context);
	    } else {
		return target.getClass().getField(name).get(target);
	    }
	} catch (PnutsException p){
	    throw p;
	} catch (NoSuchFieldException f){
	    throw new PnutsException("field.notFound", new Object[]{name, target.getClass()}, context);
	} catch (Throwable t){
	    throw new PnutsException(t, context);
	}
    }

    /**
     * Get true component type from an array type.
     *  e.g. int[][] ==> int,  String[] ==> String
     *
     * @param clazz	An array type to be examined
     * @return		The component type of the array type.
     */
    protected static Class getBottomType(Class clazz){
	while (clazz.isArray()){
	    clazz = clazz.getComponentType();
	}
	return clazz;
    }

    /**
     * Creates an array type
     *
     * @param c		the component type
     * @param dim	the number of dimensions
     */
    public static Class arrayType(Class c, int dim){
	if (dim == 0){
	    return c;
	} else {
	    return arrayType(Array.newInstance(c, 0).getClass(), dim - 1);
	}
    }

    /**
     * Transform an array to a different array type.
     *
     * @param baseType	The component type of which the array to create
     * @param array	The source array of the transformation
     * @return		An array of the <code>clazz</code> type that has same values as <code>obj</code>
     */
    protected static Object transform(Class baseType, Object array){
	Stack dim = new Stack();
	Object b = array;
	while (isArray(b)){
	    int len = Array.getLength(b);
	    dim.push(new Integer(len));
	    if (len == 0){
		break;
	    }
	    b = Array.get(b, 0);
	}
	int dimensions[] = new int[dim.size()];
	for (int i = dimensions.length - 1; i >= 0; i--){
	    dimensions[i] = ((Integer)dim.pop()).intValue();
	}
	Object ret = Array.newInstance(baseType, dimensions);
	copy(array, ret);
	return ret;
    }

    /*
     * Copy elements of an array to another array
     *
     * @param src should be an array
     * @param dst should be an array
     */
    private static void copy(Object src, Object dst){
	for (int i = 0; i < Array.getLength(src); i++){
	    Object elem = Array.get(src, i);
	    if (isArray(elem)){
		copy(elem, Array.get(dst, i));
	    } else {
		Array.set(dst, i, Array.get(src, i));
	    }
	}
    }

    /**
     * Write a hashtable to the objectOutputStream.
     * If a value of the hashtable is a
     * Class object, it is encapsulated to avoid NotSerializableException.
     * But if a value contains a Class object in its structure
     * e.g. "new Class[]{ Object.class }",  NotSerializableException
     * is thrown, even though the Class object is Serializable.
     * If notToSerialize is not null, falied keys are accumulated in
     * notToSerialize.
     *
     * This is a workaround for bug:4075221.
     */
    public static int writeSymbolTable(SymbolTable tab, SymbolTable notToSerialize, ObjectOutputStream s) throws IOException {
	int count = 0;
	boolean verbose = Pnuts.isVerbose();
	for (Enumeration e = tab.keys(); e.hasMoreElements(); ){
	    String key = (String)e.nextElement();
	    if (verbose){
		System.out.print(key);
	    }
	    if (notToSerialize != null && notToSerialize.get(key) != null){
		continue;
	    }
	    Object value = tab.get(key);
	    if (verbose){
		System.out.println("=" + value);
	    }
	    try {
		if (!(value instanceof Serializable)){
		    continue;
		}
		s.writeObject(key);
		s.writeObject(value);
		count++;
	    } catch (IOException ex){
		if (notToSerialize != null){
		    notToSerialize.put(key, key);
		}
		throw ex;
	    }
	}
	s.writeObject(null);
	
	return count;
    }

    /**
     * Read a hashtable from objectInputStream.
     * If Class objects are encapsulated,
     * it retrieves the Class objects from them.
     */
    public static void readSymbolTable(SymbolTable table, ObjectInputStream s)
	throws IOException, ClassNotFoundException {
	    for (;;){
		String k = (String)s.readObject();
		if (k == null){
		    break;
		}
		try {
		    Object v = s.readObject();
		    if (Pnuts.verbose){
			System.out.println(k + "=" + v);
		    }
		    table.put(k.intern(), v);
		} catch (Exception e){
		}
	    }
    }

    protected static int matchType(Class type, Object obj){
	if (obj == null){
	    if (type.isPrimitive()){
		return -1;
	    }
	    return 0;
	}
	Class clazz = obj.getClass();
	if (clazz == type){
	    return 0;
	}
	if (type == boolean.class){
	    return (clazz == Boolean.class) ? 0 : -1;
	}
	if (type == byte.class){
	    return distance(0, clazz);
	}
	if (type == char.class){
	    return distance(1, clazz);
	}
	if (type == short.class){
	    return distance(1, clazz);
	}
	if (type == int.class){
	    return distance(2, clazz);
	}
	if (type == long.class){
	    return distance(3, clazz);
	}
	if (type == float.class){
	    return distance(4, clazz);
	}
	if (type == double.class){
	    return distance(5, clazz);
	}
	if (type.isAssignableFrom(clazz)){
	    return 1;
	} else if (type.isArray() && clazz.isArray()){
	    int j = 1;
	    for (int i = 0; i < Array.getLength(obj); i++){
		int s = matchType(type.getComponentType(), Array.get(obj, i));
		if (s < 0){
		    return -1;
		} else {
		    j += s;
		}
	    }
	    return j;
	} else {
	    return -1;
	}
    }

    private static int distance(int pos, Class clazz){
	if (clazz == Double.class){
	    return pos - 5;
	} else if (clazz == Float.class){
	    return pos - 4;
	} else if (clazz == Long.class){
	    return pos - 3;
	} else if (clazz == Integer.class){
	    return pos - 2;
	} else if (clazz == Short.class){
	    return pos - 1;
	} else if (clazz == Character.class){
	    return pos - 1;
	} else if (clazz == Byte.class){
	    return pos;
	} else {
	    return -1;
	}
    }

    /*
     * class <-> public methods
     */
    private static Cache mtab = new Cache(512);

    private static Method[] getMethods(Class cls, String name){
	Cache cache = (Cache)mtab.get(cls);
	if (cache == null){
	    cache = new Cache(512);
	    mtab.put(cls, cache);
	}
	Object v = cache.get(name);
	if (v instanceof Method[]){
	    return (Method[])v;
	} else {
	    Method m[] = Pnuts.verdep.getMethods(cls);
	    if (m == null){ // for Bug:4137722
		throw new NoClassDefFoundError("" + cls);
	    }
	    int j = 0;
	    for (int i = 0; i < m.length; i++){
		String m_name = m[i].getName();
		if (m_name.equals(name) && i >= j){
		    m[j] = m[i];
		    j++;
		}
	    }
	    Method m2[] = new Method[j];
	    System.arraycopy(m, 0, m2, 0, j);
	    cache.put(name, m2);
	    return m2;
	}
    }

    /*
     * class <-> Constructors
     */
    private static Cache ctab = new Cache(512);

    private static Constructor[] getConstructors(Class cls){
	Object v = ctab.get(cls);
	if (v instanceof Constructor[]){
	    return (Constructor[])v;
	} else {
	    Constructor con[] = cls.getConstructors();
	    ctab.put(cls, con);
	    return con;
	}
    }

    /*
     * method <-> parameter types
     */
    private static Cache ptab = new Cache(16);

    private static Class[] getParamTypes(Method method){
	Object obj = ptab.get(method);
	if (obj instanceof Class[]){
	    return (Class[])obj;
	} else {
	    Class[] paramTypes = method.getParameterTypes();
	    ptab.put(method, paramTypes);
	    return paramTypes;
	}
    }

    /**
     * Parse an integer.
     *
     * @return an array [Number number, int offset_of_unit_symbol]
     */
    public static Object[] parseInt(String str){
	char c1 = str.charAt(0);
	if (c1 == '#'){
	    return parseInt(str.substring(1), 16, true);
	} else if (c1 == '0'){
	    if (str.length() > 1){
		char c2 = str.charAt(1);
		if (c2 == 'x' || c2 == 'X'){
		    return parseInt(str.substring(2), 16, false);
		} else {
		    return parseInt(str, 10, false);
		}
	    } else {
		return new Object[]{new Integer(0), null};
	    }
	} else {
	    return parseInt(str, 10, false);
	}
    }

    static Object[] parseInt(String str, int radix, boolean shrink){
	boolean overflow = false;
	long value = 0;
	char chars[] = str.toCharArray();
	int i = 0;
    out:
	for (i = 0; i < chars.length; i++) {
	    int ch = chars[i];
	    switch (ch) {
	    case '0': case '1': case '2': case '3': case '4':
	    case '5': case '6': case '7': case '8': case '9':
		if (radix == 10){
		    int c = ch - '0';
		    value = (value * 10) + c;
		    if (value < 0){ 
			overflow = true;
			break out;
		    }
		    break;
		} else if (radix == 16){
		    value = (value << 4) + 
			Character.toLowerCase((char)ch) - '0';
		    if (value < 0){ 
			overflow = true;
			break out;
		    }
		    break;
		}
	    case 'a': case 'A': case 'b': case 'B': case 'c': case 'C':
	    case 'd': case 'D': case 'e': case 'E': case 'f': case 'F':
		if (radix == 10){
		    break out;
		}
		value = (value << 4) + 10 + 
		    Character.toLowerCase((char)ch) - 'a';
		if (value < 0){ 
		    overflow = true;
		    break out;
		}
		break;
	    default:
		break out;
	    }
	}

    out2:
	for (; i < chars.length; i++) {
	    int ch = chars[i];
	    switch (ch) {
	    case '0': case '1': case '2': case '3': case '4':
	    case '5': case '6': case '7': case '8': case '9':
		break;
	    case 'a': case 'A': case 'b': case 'B': case 'c': case 'C':
	    case 'd': case 'D': case 'e': case 'E': case 'f': case 'F':
		if (radix == 10){
		    break out2;
		}
		break;
	    default:
		break out2;
	    }
	}

	Number number = null;

	if (overflow){
	    String s = str;
	    if (str.length() > i){
		s = str.substring(0, i);
	    }
	    number = Arithmetic.decimalNumber(s, radix);
	} else {
	    if (value >= 0 && value <= 255 && radix == 16){
		if (shrink){
		    number = new Byte((byte)value);
		} else {
		    number = new Integer((int)value);
		}
	    } else if (value <= Integer.MIN_VALUE ||
		       value >= Integer.MAX_VALUE){
		number = new Long(value);
	    } else {
		number = new Integer((int)value);
	    }
	}
	if (str.length() > i){
	    return new Object[]{number, new int[]{i}};
	} else {
	    return new Object[]{number, null};
	}
    }

    /**
     * Parse a floating point number.
     *
     * @return an array [Number number, int offset_of_unit_symbol]
     */
    public static Object[] parseFloat(String str){
	char chars[] = str.toCharArray();
	int i = 0;
    out1:
	for (i = 0; i < chars.length; i++){
	    int ch = chars[i];
	    switch (ch){
	    case '0': case '1': case '2': case '3': case '4':
	    case '5': case '6': case '7': case '8': case '9':
	    case '.':
		break;
	    default:
		break out1;
	    }
	}
	if (i < chars.length && (chars[i] == 'e' || chars[i] == 'E')){
	    i++;
	out2:
	    for (; i < chars.length; i++){
		int ch = chars[i];
		switch (ch){
		case '+': case '-':
		case '0': case '1': case '2': case '3': case '4':
		case '5': case '6': case '7': case '8': case '9':
		    break;
		default:
		    break out2;
		}
	    }
	}
	String s = str;
	if (i < chars.length){
	    s = str.substring(0, i);
	}
	Number n = null;
	if (s.length() > 16){
	    n = Arithmetic.decimalNumber(s, 10);
	} else {
	    n = Double.valueOf(s);
	}
	if (i < chars.length){
	    return new Object[]{n, new int[]{i}};
	} else {
	    return new Object[]{n, null};
	}
    }

    /**
     * Parse a string literal.
     *
     * @return the value
     */
    public static String parseString(String str){
	try {
	    CharArrayWriter cout = new CharArrayWriter();
	    boolean backquote = (str.charAt(0) == '`');
	    int length = str.length();
	    for (int i = 1; i < length - 1; i++){
		char ch = str.charAt(i);
		if (ch == '\\' && !backquote){
		    i++;
		    switch (str.charAt(i)){
		    case '"':
			cout.write((int)'"');
			break;
		    case 'b':
			cout.write((int)'\b');
			break;
		    case 'f':
			cout.write((int)'\f');
			break;
		    case 't':
			cout.write((int)'\t');
			break;
		    case 'r':
			cout.write((int)'\r');
			break;
		    case 'n':
			cout.write((int)'\n');
			break;
		    case '0':
			cout.write((int)'\0');
			break;
		    case '\\':
			cout.write((int)'\\');
			break;
		    case 'u':
		    case 'U':
			{
			    if (i + 6 > length){
				throw new IllegalArgumentException();
			    }
			    int value = 0;
			    for (int j = 0; j < 4; j++){
				int c = str.charAt(++i);
				switch (c){
				case '0': case '1': case '2': case '3': case '4':
				case '5': case '6': case '7': case '8': case '9':
				    value = (value << 4) + (c - '0');
				    break;
				case 'a': case 'A': case 'b': case 'B': case 'c': case 'C':
				case 'd': case 'D': case 'e': case 'E': case 'f': case 'F':
				    value = (value << 4) + 10 + Character.toLowerCase((char)c) - 'a';
				    break;
				default:
				    throw new IllegalArgumentException();
				}
			    }
			    cout.write(value);
			    break;
			}
		    default:
			throw new IllegalArgumentException();
		    }
		} else {
		    cout.write((int)ch);
		}
	    }
	    return new String(cout.toCharArray());
	} catch(Throwable e){
	    throw new PnutsException(e);
	}
    }

    /**
     * Parse a character literal.
     *
     * @return the value
     */
    public static Character parseChar(String str){
	char c[] = str.toCharArray();
	if (c[1] == '\\'){
	    switch (c[2]){
	    case '\'':
		return new Character('\'');
	    case 'b':
		return new Character('\b');
	    case 'f':
		return new Character('\f');
	    case 't':
		return new Character('\t');
	    case 'r':
		return new Character('\r');
	    case 'n':
		return new Character('\n');
	    case '0':
		return new Character('\0');
	    case '\\':
		return new Character('\\');
	    case 'u':
	    case 'U':
		{
		    int len = c.length;
		    if (len != 8){
			throw new IllegalArgumentException();
		    }
		    int value = 0;
		    for (int p = 3; p < 7; p++){
			int ch = c[p];
			switch (ch){
			case '0': case '1': case '2': case '3': case '4':
			case '5': case '6': case '7': case '8': case '9':
			    value = (value << 4) + (ch - '0');
			    break;
			case 'a': case 'A': case 'b': case 'B': case 'c': case 'C':
			case 'd': case 'D': case 'e': case 'E': case 'f': case 'F':
			    value = (value << 4) + 10 + Character.toLowerCase((char)ch) - 'a';
			    break;
			default:
			    throw new IllegalArgumentException();
			}
		    }
		    return new Character((char)value);
		}
	    default:
		throw new IllegalArgumentException();
	    }
	} else {
	    if (c.length > 3){
		throw new IllegalArgumentException();
	    }
	    return new Character(c[1]);
	}
    }

    /**
     * Creates an object from a number literal and a unit symbol
     *
     * @param number		a number object
     * @param numberString	a symbol of the number literal
     * @param unit		a unit symbol
     * @param context		a context in which the quantity is created
     */
    public static Object quantity(Number number, String numberString, String unit, Context context){
	SymbolTable units = Package.globalPackage.unitTable;

	QuantityFactory factory = (QuantityFactory)units.get(unit.intern());
	if (factory != null){
	    return factory.make(number, unit);
	}
	if ("f".equals(unit) || "F".equals(unit)){
	    return new Float(number.floatValue());
	} else if ("d".equals(unit) || "D".equals(unit)){
	    return new Double(number.doubleValue());
	} else if ("b".equals(unit) || "B".equals(unit)){
	    return Arithmetic.decimalNumber(numberString, 10);
	}
	throw new PnutsException("unitName.notDefined", new Object[]{unit}, context);
    }

    /**
     * This method is called by the syntax
     * "primitiveType(object)" and "(primitiveType)object"
     *
     * @param context	the context
     * @param primitiveType	a primitive type
     * @param param	the parameter
     * @param flag	string<->number conversion
     */
    public static Object primitive(Context context, Class primitiveType, Object param, boolean flag){
	if (primitiveType == int.class){
	    if (param instanceof Byte){
		byte b = ((Byte)param).byteValue();
		return new Integer((int)b);
	    } else if (param instanceof Character){
		return new Integer((int)((Character)param).charValue());
	    } else if (param instanceof Number){
		return new Integer(((Number)param).intValue());
	    } else if (flag && param instanceof String){
		return new Integer(Integer.parseInt(((String)param).trim()));
	    }
	} else if (primitiveType == byte.class){
	    if (param instanceof Character){
		return new Byte((byte)((Character)param).charValue());
	    } else if (param instanceof Number){
		return new Byte(((Number)param).byteValue());
	    } else if (flag && param instanceof String){
		return new Byte((byte)Integer.parseInt(((String)param).trim()));
	    }
	} else if (primitiveType == char.class){
	    if (param instanceof Number){
		int i = ((Number)param).intValue();
		if (i >= 0 && i <= 0xffff){
		    return new Character((char)i);
		} else {
		    throw new PnutsException("char.outOfBound", PnutsException.NO_PARAM, context);
		}
	    } else if (param instanceof Character){
		return param;
	    }
	} else if (primitiveType == long.class){
	    if (param instanceof Number){
		return new Long(((Number)param).longValue());
	    } else if (flag && param instanceof String){
		return new Long(Long.parseLong(((String)param).trim()));
	    }
	} else if (primitiveType == short.class){
	    if (param instanceof Character){
		return new Short((short)((Character)param).charValue());
	    } else if (param instanceof Number){
		return new Short(((Number)param).shortValue());
	    } else if (flag && param instanceof String){
		return new Short(Short.parseShort(((String)param).trim()));
	    }
	} else if (primitiveType == double.class){
	    if (param instanceof Number){
		return new Double(((Number)param).doubleValue());
	    } else if (flag && param instanceof String){
		return Double.valueOf((String)param);
	    }
	} else if (primitiveType == float.class){
	    if (param instanceof Number){
		return new Float(((Number)param).floatValue());
	    } else if (flag && param instanceof String){
		return Float.valueOf((String)param);
	    }
	} else if (primitiveType == boolean.class){
	    if (param instanceof Boolean){
		return param;
	    } else if (flag && param instanceof String){
		return Boolean.valueOf((String)param);
	    }
	}
	throw new ClassCastException("(" + primitiveType.getName() + ")" + Pnuts.format(param));
    }

    /**
     * This method is called by the syntax "(Class)object"
     *
     * @param context	the context
     * @param type	the type
     * @param flag	object_array<->primitive_array conversion
     */
    public static Object cast(Context context, Class type, Object object, boolean flag){
	if (type.isPrimitive()){
	    return primitive(context, type, object, false);
	}
	if (object == null){
	    return null;
	} else if (type.isInstance(object)){
	    return object;
	} else if (type.isArray() && isArray(object)){
	    for (int i = 0; i < Array.getLength(object); i++){
		int s = matchType(type.getComponentType(), Array.get(object, i));
		if (s < 0){
		    throw new ClassCastException("(" + getClassName(type) + ")" + Pnuts.format(object));
		}
	    }
	    if (flag){
		return transform(getBottomType(type), object);
	    } else {
		return object;
	    }
	}
	throw new ClassCastException("(" + getClassName(type) + ")" + Pnuts.format(object));
    }

    /**
     * Check if the parameter is an array
     *
     * @param obj the object to be checked
     * @return true if the obj is an array, false otherwise.
     */
    public final static boolean isArray(Object obj){
	return (obj instanceof Object[] ||
		obj instanceof int[] ||
		obj instanceof char[] ||
		obj instanceof boolean[] ||
		obj instanceof byte[] ||
		obj instanceof double[] ||
		obj instanceof long[] ||
		obj instanceof short[] ||
		obj instanceof float[]);
    }

    /**
     * Range expression 'target[idx1..idx2]'
     */
    protected static Object range(Object target, Number idx1, Number idx2){
	int from = idx1.intValue();
	int to = 0;
	if (idx2 != null){
	    to = idx2.intValue();
	    if (to < 0){
		throw new ArrayIndexOutOfBoundsException("" + to);
	    }
	}
	if (target instanceof String){
	    String s = (String)target;
	    if (idx2 != null){
		return s.substring(from, to + 1);
	    } else {
		return s.substring(from);
	    }
	} else {
	    int len = Array.getLength(target);
	    if (idx2 == null){
		to = len - 1;
	    } else if (to > len - 1){
		throw new ArrayIndexOutOfBoundsException("" + to);
	    }
	    int size = to - from + 1;
	    if (from + size > len){
		size = len - from;
	    }
	    Class c = target.getClass().getComponentType();
	    Object ret = Array.newInstance(c, size);
	    if (size > 0){
		System.arraycopy(target, from, ret, 0, size);
	    }
	    return ret;
	}
    }

    /**
     * This method is called by the syntax "id[from..to] = sth"
     */
    protected static Object assignRange(Object target, Number idx1, Number idx2, Object expr){
	int from = idx1.intValue();
	int to = 0;
	if (idx2 != null){
	    to = idx2.intValue();
	    if (to < 0){
		throw new ArrayIndexOutOfBoundsException("" + to);
	    }
	}
	if (target instanceof String){
	    StringBuffer s = new StringBuffer((String)target);
	    if (idx2 == null){
		to = s.length() - 1;
	    }
	    if (expr instanceof Character){
		char c = ((Character)expr).charValue();
		for (int i = from; i < to + 1; i++){
		    s.setCharAt(i, c);
		}
	    } else {
		Pnuts.verdep.replace(s, from, to + 1, String.valueOf(expr));
	    }
	    String val = s.toString();
	    return val;
	} else if (isArray(target)){
	    int len = Array.getLength(target);
	    if (idx2 == null){
		to = len - 1;
	    } else if (to > len - 1){
		throw new ArrayIndexOutOfBoundsException("" + to);
	    }
	    int size = to - from + 1;
	    if (from + size > len){
		size = len - from;
	    }
	    if (target instanceof char[] && expr instanceof String){
		String str = (String)expr;
		int end = str.length();
		if (end > to - from + 1){
		    end = to - from + 1;
		}
		str.getChars(0, end, (char[])target, from);
	    } else {
		for (int i = from; i < to + 1; i++){
		    Array.set(target, i, expr);
		}
	    }
	    return target;
	} else {
	    throw new IllegalArgumentException();
	}
    }

    protected static void checkException(Context context, Throwable throwable){
	if (context.defined(Context.exceptionHandlerTableSymbol)){
	    checkException(context, throwable, (Hashtable)context.getValue(Context.exceptionHandlerTableSymbol));
	} else if (throwable instanceof PnutsException){
	    throw (PnutsException)throwable;
	} else {
	    throw new PnutsException(throwable, context);
	}
    }

    /**
     * Check if any exception handler is registered to the specified exception.
     * If any an exception handler is executed.  If not, the exception is thrown.
     *
     * @param context	the Context in which the exception is checked
     * @param throwable	the exception
     * @param tab	the exception handler table
     */
    protected static void checkException(Context context, Throwable throwable, Hashtable tab){
	if (DEBUG){
	    System.out.println("checkException " + throwable + ", " + tab);
	}
	Throwable e = throwable;
	if (e instanceof PnutsException){
	    throwable = ((PnutsException)e).getThrowable();
	}
	if (throwable instanceof Escape){
	    if (DEBUG){
		System.out.println(throwable);
	    }
	    throw (Escape)throwable;
	}
	if (throwable instanceof ThreadDeath){
	    throw (ThreadDeath)throwable;
	}
	if (tab == null){
	    if (e instanceof PnutsException){
		throw (PnutsException)e;
	    } else {
		throw new PnutsException(throwable, context);
	    }
	} else {
	    Class clazz = throwable.getClass();
	    PnutsFunction f = null;
	    while (clazz != null){
		f = (PnutsFunction)tab.get(clazz);
		if (f != null){
		    break;
		}
		clazz = clazz.getSuperclass();
	    }
	    if (f != null){
		Object ret = f.exec(new Object[]{e}, context);
		context.getOutputStream().flush();
		throw new Jump(ret);
	    } else {
		if (e instanceof PnutsException){
		    throw (PnutsException)e;
		} else {
		    throw new PnutsException(throwable, context);
		}
	    }
	}
    }

    /**
     * This method is called when catch() function is called in a
     * package(non-local) scope 
     *
     * @param type the exception type of which an exception handler is registered
     * @param func the function to be registered as an exception handler
     * @param context the context in which the exception handler is registered
     */
    protected static void catchException(Class type,
					 PnutsFunction func,
					 Context context){
	if (DEBUG){
	    System.out.println("catchException " +
			       type.getName() + ": " +
			       func.unparse(1));
	}
	Package pkg = context.getCurrentPackage();
	Hashtable table = (Hashtable)pkg.get(Context.exceptionHandlerTableSymbol,
					     context);
	if (table == null){
	    table = new Hashtable(10);
	    pkg.set(Context.exceptionHandlerTableSymbol, table, context);
	}
	if (func == null){
	    table.remove(type);
	} else {
	    table.put(type, func);
	}
    }

    /**
     * This method is called by the syntax "target[index]"
     */
    protected static Object getIndex(Object target, int index, Context context){
	if (target instanceof Object[]){
	    return ((Object[])target)[index];
	} else if (target instanceof String){
	    return new Character(((String)target).charAt(index));
	} else if (target instanceof int[]){
	    return new Integer(((int[])target)[index]);
	} else if (target instanceof char[]){
	    return new Character(((char[])target)[index]);
	} else if (target instanceof byte[]){
	    return new Byte(((byte[])target)[index]);
	} else if (target instanceof boolean[]){
	    return new Boolean(((boolean[])target)[index]);
	} else if (target instanceof double[]){
	    return new Double(((double[])target)[index]);
	} else if (target instanceof float[]){
	    return new Float(((float[])target)[index]);
	} else if (target instanceof long[]){
	    return new Long(((long[])target)[index]);
	} else if (target instanceof short[]){
	    return new Short(((short[])target)[index]);
	} else if (target instanceof Indexed){
	    return ((Indexed)target).get(index);
	} else {
	    throw new PnutsException("illegal.subscript", new Object[]{}, context);
	}
    }

    /**
     * This method is called by the syntax "clazz(args...)"
     */
    protected static Object newInstance(Context context, Class clazz, Object[] args, Class[] casts){
	if (clazz.isPrimitive()){
	    return primitive(context, clazz, args[0], true);
	} else if (clazz.isArray()){
	    return cast(context, clazz, args[0], true);
	} else {
	    return callConstructor(context, clazz, args, casts);
	}
    }

    /**
     * Call a function
     *
     * @param context the context in which the function is called
     * @param func the function to be called
     * @param args the arguments
     */
    protected final static Object callFunction(Context context, PnutsFunction func, Object[] args){
	return func.exec(args, context);
    }

    /**
     * This method is called by the syntax "funcOrClass(args...)"
     */
    protected final static Object call(Context context, Object funcOrClass, Object[] args, Class[] casts){
	if (funcOrClass instanceof PnutsFunction){
	    return ((PnutsFunction)funcOrClass).exec(args, context);
	} else if (funcOrClass instanceof Class){
	    return newInstance(context, (Class)funcOrClass, args, casts);
	} else {
	    throw new PnutsException("funcOrType.expected", new Object[]{Pnuts.format(funcOrClass)}, context);
	}
    }

    protected static void jump(Object v){
	throw new Jump(v);
    }

    protected static void escape(Object v){
	throw new Escape(v);
    }

    /**
     * Set line number information for error reporting
     */
    protected static void setLine(Context context, int beginLine, int endLine){
	context.updateLine(null, beginLine, endLine);
    }

    protected static void setLine(Context context, int line){
	context.updateLine(null, line, line);
    }

    protected static int getBeginLine(Context context){
	return context.beginLine;
    }

    protected static int getEndLine(Context context){
	return context.endLine;
    }

    protected static Function getFunction(PnutsFunction pf, int nargs){
	return pf.get(nargs);
    }

    protected static Enumeration getFunctions(PnutsFunction pf){
	return pf.elements();
    }

    protected static Runtime getRuntime(Context context){
	return context.runtime;
    }

    protected static Object getScriptSource(Context context){
	return context.getScriptSource();
    }

    protected static Function getFunction(Context context){
	return context.frame;
    }

    protected static boolean namespaceRefreshed(Context context){
	return context.namespaceRefreshed[0];
    }

    protected static void setPackage(Package pkg, Context context){
	pkg.init(context);
    }

    private static char[] hexDigit = {
	'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'
    };

    /**
     * Get the String representation of the specified object.
     *
     * @param object the target object.
     * @param maxArrayLength When the target object is an array and maxArrayLength is greater than zero, only the first maxArrayLength elements are printed and the rest of the elements are omitted as "...". 
     */
    public static String format(Object object, int maxArrayLength){
	if (object == null){
	    return "null";
	} else if (!(isArray(object))){
	    if (object instanceof String){
		return "\"" + object.toString() + "\"";
	    } else if (object instanceof Character){
		char ch = ((Character)object).charValue();
		switch (ch) {
		  case '\\': return "'\\'";
		  case '\b': return "'\\b'";
		  case '\f': return "'\\f'";
		  case '\t': return "'\\t'";
		  case '\n': return "'\\n'";
		  case '\r': return "'\\r'";
		  case '\'': return "'\\''";
		  case '?': return "'?'";
		  default:
		      char[] buf = new char[1];
		      buf[0] = ch;
		      byte[] conv = new String(buf).getBytes();
		      if (!Character.isISOControl(ch) &&
			  !(conv.length == 1 && conv[0] == (byte)'?')){
			  return "'" + ch + "'";
		      }
		      return "'\\u" +
			  hexDigit[(ch >> 12) & 0xF] +
			  hexDigit[(ch >> 8) & 0xF] +
			  hexDigit[(ch >> 4) & 0xF] +
			  hexDigit[(ch >> 0) & 0xF] + "'";
		}
	    } else if (object instanceof Class){
		Class c = (Class)object;
		return getClassName(c) + getSuffix(c);
	    } else if (object instanceof Byte){
		byte value = ((Byte)object).byteValue();
		return "#" + hexDigit[(value >> 4) & 0x0f] + hexDigit[value & 0x0f];
	    } else {
		return object.toString();
	    }
	}
	String s = "";
	int length = Array.getLength(object);
	int last = length;
	s += "[";
	if (length > 0){
	    s += format(Array.get(object, 0), maxArrayLength);
	}
	if (maxArrayLength > 0 && maxArrayLength < length){
	    last = maxArrayLength;
	}
	for (int j = 1; j < last; j++){
	    s += ", ";
	    s += format(Array.get(object, j), maxArrayLength);
	}
	if (maxArrayLength > 0 && maxArrayLength < length){
	    s += ", ...";
	}
	s += "]";
	return s;
    }

    static String getClassName(Class cls){
	if (cls.isArray()){
	    StringBuffer buf = new StringBuffer();
	    for (; cls.isArray(); cls = cls.getComponentType()){
		buf.append("[]");
	    }
	    buf.insert(0, cls.getName());
	    return buf.toString();
	} else {
	    return cls.getName();
	}
    }

    private static String getSuffix(Class cls){
	if (cls.isInterface()){
	    return " interface";
	} else if (cls.isPrimitive()){
	    return " type";
	} else {
	    return " class";
	}
    }

    /**
     * Binary operation
     *
     * @param n1 the first argument
     * @param n2 the second argument
     * @param operator an integer which is corresponds to the operator
     * @param context the context in which the operation is taken place
     * @return the result of the operation
     */
    protected static Object binary(Object n1, Object n2, int operator, Context context){
	try {
	    return Arithmetic.binary(operator, n1, n2);
	} catch (IllegalArgumentException e){
	    throw new PnutsException("illegal.operandType",
				     new Object[]{e.getMessage()},
				     context);
	}
    }

    /**
     * Comparison operation
     *
     * @param n1 the first argument
     * @param n2 the second argument
     * @param context the context in which the operation is taken place
     * @return the result of the operation; one of: <ul>
     *  <li>Arithmetic.EQUAL
     *  <li>Arithmetic.NOT_EQUAL
     *  <li>Arithmetic.LEFT_IS_BIGGER
     *  <li>Arithmetic.RIGHT_IS_BIGGER
     *  </ul>
     */
    protected static int compareTo(Object n1, Object n2, Context context){
	try {
	    return Arithmetic.compareTo(n1, n2);
	} catch (IllegalArgumentException e){
	    throw new PnutsException("illegal.operandType",
				     new Object[]{e.getMessage()},
				     context);
	}
    }

    /**
     * Unary operation
     *
     * @param n the argument
     * @param operator an integer which corresponds to the operator
     * @param context the context in which the operation is taken place
     * @return the result of the operation
     */
    protected static Object unary(Object n, int operator, Context context){
	try {
	    return Arithmetic.unary(operator, n);
	} catch (IllegalArgumentException e){
	    throw new PnutsException("illegal.operandType",
				     new Object[]{e.getMessage()},
				     context);
	}
    }

    private static URL getResource(String s, Context context){
	ClassLoader classLoader = Pnuts.getClassLoader(context);
	if (classLoader != null){
	    if (Pnuts.isJava2()){
		URL url = classLoader.getResource(s);
		if (url == null){
		    return Pnuts.verdep.getResource(s);
		} else {
		    return url;
		}
	    } else {
		return classLoader.getResource(s);
	    }
	} else {
	    return Pnuts.verdep.getResource(s);
	}
    }

    public static URL getScriptURL(String name, Context context){
	Object loadPath[] = null;
	Object o = Package.getGlobalPackage().get(Pnuts.load_path_sym, context);
	if (o instanceof Object[]){
	    loadPath = (Object[])o;
	} else {
	    loadPath = defaultLoadPath;
	}
	Applet ns_applet = null;
	if (java_vendor.startsWith("Netscape")){
	    Object pnuts_applet = Pnuts.get("pnuts_applet");
	    if (pnuts_applet instanceof Applet){ 
		ns_applet = (Applet)pnuts_applet;
	    }
	}

 	URLConnection conn = null;
	URL url = null;

	for (int i = 0; i < loadPath.length; i++){
	    Object obj = loadPath[i];
	    if (obj instanceof String){
		String s = (String)obj;
		if (s != "" && !s.endsWith("/")){
		    s += "/";
		}
		s += name;
		if (ns_applet != null){	    // workaround for resource bug
		    try {
			url = new URL(ns_applet.getCodeBase(), s.substring(1));
		    } catch (IOException e){}
		} else {
		    url = getResource(s, context);
		    if (DEBUG){
			System.out.println("getResource(" + s + ") = " + url);
		    }
		}
		if (url == null){
		    continue;
		}
		try {
		    conn = url.openConnection();
		    conn.connect();
		    if (Pnuts.isVerbose()){
			PrintWriter out = context.getOutputStream();
			out.println("[loading " + url + "]");
			out.flush();
		    }
		    break;
		} catch (IOException ioe){
		    if (Pnuts.isVerbose()){
			System.err.println(ioe);
		    }
		}
	    } else if (obj instanceof File){
		File dir = (File)obj;
		if (!dir.isDirectory()){
		    return null;
		}
		File f = new File(dir, name);
		if (f.exists()){
		    try {
			url = Runtime.fileToURL(f);

			conn = url.openConnection();
			conn.connect();
			if (Pnuts.isVerbose()){
			    PrintWriter out = context.getOutputStream();
			    out.println("[loading " + f + "]");
			    out.flush();
			}
			break;
		    } catch (IOException e){
		    }
		} else {
		    continue;
		}
	    } else if (obj instanceof URL){
		try {
		    url = new URL((URL)obj, name);
		    conn = url.openConnection();
		    conn.connect();
		    if (Pnuts.isVerbose()){
			PrintWriter out = context.getOutputStream();
			out.println("[loading " + url + "]");
			out.flush();
		    }
		    break;
		} catch (IOException e){ } // ignore
	    }
	}
	return url;
    }

    /**
     * This method is called by Pnuts.load() when the property "pnuts.compiled.script.prefix"
     * is defined, to load pre-compiled scripts.
     *
     * @param name	the script name
     * @param context 	the context in which the class is loaded.
     * @return		pnuts.lang.Runtime object if the class is found, otherwise null.
     */
    public static Runtime getCompiledScript(String name, Context context){
	String prefix = Pnuts.getCompiledClassPrefix();
	if (prefix != null){
	    Object rt = null;
	    try {
		Class cls = Pnuts.loadClass(prefix + name.replace('/', '.'), context);
		rt = cls.newInstance();
		if (rt instanceof Runtime){
		    return (Runtime)rt;
		}
	    } catch (Exception t){ /* ignore */ }
	}
	return null;
    }

    public static URL fileToURL(File file) throws MalformedURLException {
	String path = file.getAbsolutePath();
	if (File.separatorChar != '/') {
	    path = path.replace(File.separatorChar, '/');
	}
	if (!path.startsWith("/")) {
	    path = "/" + path;
	}
	if (!path.endsWith("/") && file.isDirectory()) {
	    path = path + "/";
	}
	return new URL("file", "", path);
    }

    public static void printError(Throwable t, Context context){
	context.onError(t);
	PrintWriter err = context.getErrorStream();
	if (err != null){
	    if (Pnuts.isVerbose()){
		t.printStackTrace(err);
	    } else {
		err.println(t);
	    }
	    err.flush();
	} else {
	    t.printStackTrace();
	}
    }
}
