/*
 * @(#) DebugContext.java	1.7 00/08/07
 *
 * Copyright (c) 1997-1999 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.tools;

import pnuts.lang.*;
import pnuts.lang.Package;
import java.util.*;

/**
 * This class is a Context used in debug mode.
 */
public class DebugContext extends Context {
    private Vector listeners = null;
    private int callDepth = 0;

    public DebugContext(){
	setPnutsImpl(new PnutsImpl());  // pure interpreter
    }

    public DebugContext(Package pkg){
	super(pkg);
	setPnutsImpl(new PnutsImpl()); // pure interpreter
    }

    public DebugContext(Context context){
	super(context.getCurrentPackage());
	Pnuts.setClassLoader(Pnuts.getClassLoader(context), this);
	String[] used = context.usedPackages();
	for (int i = 0; i < used.length; i++){
	    usePackage(used[i]);
	}
	setOutputStream(context.getOutputStream());
	setErrorStream(context.getErrorStream());
	setTerminalStream(context.getTerminalStream());
	setPnutsImpl(new PnutsImpl()); // pure interpreter
    }

    /**
     * Tracks stack depth (function call) in pure interpreter
     */
    protected void open(String[] locals){
	super.open(locals);
	++callDepth;
    }

    /**
     * Tracks stack depth (function return) in pure interpreter
     */
    protected void close(){
	super.close();
	--callDepth;
    }

    /**
     * Get the depth of evaluation.
     *
     * This value increases when load(), loadFile(), or eval() is called.
     */
    public int getEvalDepth(){
	return depth;
    }

    /**
     * Get the stack depth (in pure interpreter)
     */
    public int getCallDepth(){
	return callDepth;
    }

    /**
     * This method is called when line number is changed.
     */
    protected void updateLine(SimpleNode node, int beginLine, int endLine){
	int line = beginLine;
	if (line > 0 && this.beginLine != line){
	    this.beginLine = line;
	    this.endLine = endLine;
	    if (listeners != null){
		Vector vec = (Vector)listeners.clone();
		CommandEvent event = new CommandEvent(this, CommandEvent.LINE_UPDATED, node);
		for (Enumeration e = vec.elements(); e.hasMoreElements(); ){
		    ((CommandListener)e.nextElement()).signal(event);
		}
	    }
	}
    }

    /**
     * This method is called when some exception is thrown.
     */
    protected void onError(Throwable t){
	if (listeners != null){
	    Vector vec = (Vector)listeners.clone();
	    CommandEvent event = new CommandEvent(this, CommandEvent.EXCEPTION, t);
	    for (Enumeration e = vec.elements(); e.hasMoreElements(); ){
		((CommandListener)e.nextElement()).signal(event);
	    }
	}
    }

    /**
     * This method is called when an evaluation is terminated normally.
     */
    protected void onExit(Object arg){
	if (listeners != null){
	    Vector vec = (Vector)listeners.clone();
	    CommandEvent event = new CommandEvent(this, CommandEvent.EXITED, arg);
	    for (Enumeration e = vec.elements(); e.hasMoreElements(); ){
		((CommandListener)e.nextElement()).signal(event);
	    }
	}
    }

    public synchronized void addCommandListener(CommandListener listener){
	if (listeners == null){
	    listeners = new Vector();
	}
	listeners.addElement(listener);
    }

    public synchronized void removeCommandListener(CommandListener listener){
	listeners.removeElement(listener);
	if (listeners.size() == 0){
	    listeners = null;
	}
    }
}
