/*
 * @(#) Main.java	1.10 00/08/07
 *
 * Copyright (c) 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.tools;

import pnuts.lang.Pnuts;
import pnuts.lang.PnutsImpl;
import pnuts.lang.Context;
import pnuts.lang.Runtime;
import pnuts.compiler.CompilerPnutsImpl;
import pnuts.ext.SecurePnutsImpl;
import java.util.Properties;
import java.io.PrintWriter;
import java.io.InputStream;
import java.io.FileReader;
import java.io.Reader;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.IOException;

/**
 * The main class of "pnuts" command
 */
public class Main {

    private final static String copyrightMessage= "Copyright (c) 1997-2000 Sun Microsystems, Inc. All rights reserved.";
    private final static String pnutsDebuggerProperty = "pnuts.debugger";

    static String java_vendor = System.getProperty("java.vendor");
    static String java_version = System.getProperty("java.version");

    static void greeting(Context context){
	PrintWriter w = context.getTerminalStream();
	w.println(copyrightMessage);
	if (java_vendor == null){
	    java_vendor = System.getProperty("java.vendor");
	}
	w.print("Pnuts version " + Pnuts.pnuts_version + ", " + java_version + " (" + java_vendor + ")");
	if (Pnuts.isVerbose()){
	    String vm_name = System.getProperty("java.vm.name");
	    String vm_info = System.getProperty("java.vm.info");
	    if (vm_name != null){
		w.println("\n" + vm_name + " (" + vm_info + ")");
	    }
	} else {
	    w.println();
	}
    }

    static void setProperty(String name, String value){
	if (Pnuts.isJava2()){
	    System.setProperty(name, value);
	} else {
	    Properties prop = System.getProperties();
	    prop.put(name, value);
	    System.setProperties(prop);
	}
    }

    /**
     * starts the command shell interpreter
     */
    public static void main(String args[]) throws Throwable {
	String param = null;

	Context context = null;

	boolean evaluated = false;
	boolean interactive = false;
	boolean quiet = false;
	ContextFactory contextFactory = null;

	String iparam = System.getProperty("pnuts.interactive");
	if (iparam != null){
	    if ("false".equals(iparam)){
		quiet = true;
	    } else if ("true".equals(iparam)){
		interactive = true;
	    }
	}

	boolean use_compiler = true;
	boolean debug_mode = false;

	for (int i = 0; i < args.length; i++) {
	    if (args[i].startsWith("-d")){

		String pnutsContextFactory = System.getProperty(pnutsDebuggerProperty);
		if (pnutsContextFactory != null){
		    Class cls = Class.forName(pnutsContextFactory);
		    contextFactory = (ContextFactory)cls.newInstance();
		} else {
		    if (args[i].length() > 3 && args[i].charAt(2) == ':'){ // -d:file
			String debug_script = args[i].substring(3);
			System.out.println("reading " + debug_script);
			Reader reader = new FileReader(debug_script);
			contextFactory = new TerminalDebugger(reader);
		    } else {
			contextFactory = new TerminalDebugger();
		    }
		}
		use_compiler = false;
		debug_mode = true;
		break;
	    } else if ("-pure".equals(args[i])){
		use_compiler = false;
		break;
	    }
	}

	if (use_compiler){
	    setProperty("pnuts.compiled.script.prefix", "pnuts.precompiled.");
	}

	if (debug_mode){
	    context = contextFactory.createContext();
	} else {
	    context = new Context();
	}

	if (use_compiler){
	    PnutsImpl impl = new CompilerPnutsImpl();
	    if (Pnuts.isJava2() && System.getSecurityManager() != null){
		impl = new SecurePnutsImpl(impl);
	    }
	    context.setPnutsImpl(impl);
	}
	
	for (int i = 0; i < args.length; i++) {
	    Context ctx = (Context)context.clone();

	    try {
		if ("-r".equals(args[i])){
		    if (++i < args.length){
			Pnuts.load(args[i], ctx);
			continue;
		    } else {
			break;
		    }
		} else if ("-f".equals(args[i])){
		    if (++i < args.length){
			Pnuts.loadFile(args[i], ctx);
			continue;
		    } else {
			break;
		    }
		} else if ("-e".equals(args[i])){
		    if (++i < args.length){
			Pnuts.eval(args[i], ctx);
			evaluated = true;
			continue;
		    } else {
			break;
		    }
		} else if ("-m".equals(args[i])){
		    if (++i < args.length){
			ctx.usePackage(args[i]);
			continue;
		    } else {
			break;
		    }
		} else if (args[i].startsWith("-pure")){
		    
		} else if (args[i].startsWith("-d")){
		    
		} else if ("-v".equals(args[i])){
		    Pnuts.setVerbose(true);
		} else {
		    evaluated = true;
		    String[] scriptArgs = new String[args.length - i];
		    System.arraycopy(args, i, scriptArgs, 0, scriptArgs.length);
		    Pnuts.set("$args", scriptArgs);
		    Pnuts.loadFile(args[i], ctx);
		    break;
		}
	    } catch (Throwable t){
		Runtime.printError(t, ctx);
		return;
	    }
	}

	if (evaluated){
	    return;
	}

	if (quiet){
	    interactive = false;
	} else if (!interactive){
	    int av = 0;
	    try {
		av = System.in.available();
	    } catch (IOException e){
	    }
	    if (av < 1){
		interactive = true;
	    } else if (av == 1){ // Some implementations of JDK1.1-win32 has a bug in System.in.available()
		if (!Pnuts.isJava2()){
		    String osname = System.getProperty("os.name");
		    if (osname != null && osname.startsWith("Win") &&
			java_vendor != null && 
			(java_vendor.startsWith("Sun") || java_vendor.startsWith("IBM"))){
			interactive = true;
		    }
		}
	    }
	}
	    
	if (interactive){
	    greeting(context);
	    InputStream in = new TerminalInputStream(System.in);
	    if (contextFactory instanceof TerminalDebugger){
		((TerminalDebugger)contextFactory).setInput(new BufferedReader(new InputStreamReader(in)));
	    }
	    Pnuts.load(in, true, context);
	} else {
	    try {
		Pnuts.load(System.in, false, context);
	    } catch (Throwable t){
		Runtime.printError(t, context);
	    }
	}
    }
}
