/*
 * @(#) PersonalServer.java	1.3 00/07/06
 *
 * Copyright (c) 1997-1999 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.tools;

import java.net.*;
import java.io.*;
import java.util.*;
import pnuts.lang.*;
import pnuts.lang.Package;

/**
 * This program is a network server which receives scripts sent by
 * a local user program and executes the scripts. <p>
 *
 * When a ServerSocket is created, this program writes 10 bytes data including
 * local port number (2 bytes) and a random number (8 bytes), to file
 * ${HOME}/.pnuts/cookie.<p>
 *
 * This server receives any Pnuts script after 9 bytes control data.
 * The first 8 bytes of the control data must be same as one generated when the
 * server started.
 * If the other 1 byte is non-zero, echo is enabled. In this case, Pnuts.load() method
 * is called twice.  The first time is executed without echoes.
 *
 * In a multi-user environment, directory "${HOME}/.pnuts" should not be
 * readable from other users.<p>
 *
 * <pre>
 * Usage:
 *   java PersonalServer [ -port NNNN ] [ -home PATH ]
 * </pre>
 */
public class PersonalServer {
    private final static boolean DEBUG = false;
    ServerSocket ss;
    public final static int DEFAULT_PORT = 0;
    int port;
    String home;
    long cookie;
    boolean shutdown = false;
    RequestQueue req = new RequestQueue();

    public PersonalServer() throws IOException {
	this(0, null);
    }

    public PersonalServer(int port, String home) throws IOException {
	this.port = port;
	if (home == null){
	    home = System.getProperty("user.home");
	}
	this.home = home;
    }

    static class RequestQueue {
	class Cell {
	    Object value;
	    Cell prev;
	    Cell next;
	}
	Cell head;
	Cell tail;
	int count = 0;

	public boolean isEmpty(){
	    return (head == null);
	}

	public synchronized Object pull(){
	    while (head == null){
		try {
		    wait();
		} catch (InterruptedException e){
		}
	    }
	    Cell c = head;
	    head = head.next;
	    return c.value;
	}

	public synchronized void push(Object object){
	    Cell c = new Cell();
	    c.value = object;
	    if (head == null){
		head = c;
		notifyAll();
	    } else {
		c.prev = tail;
		tail.next = c;
	    }
	    tail = c;
	}
    }

    private final static int NTHREAD = 3;
    private int busy = 0;

    class Worker implements Runnable {
	int id;

	Worker(int id){
	    this.id = id;
	}

	public void run(){
	    while (true){
		handle();
	    }
	}
    }

    void refuse(Socket sock){
	try {
	    Thread.sleep(1000);
	} catch (InterruptedException e){
	    /* ignore */
	}
	try {
	    sock.close();
	} catch (IOException e2){
	    /* ignore */
	}
    }

    public void handle(){
	Socket s = (Socket)req.pull();
	busy++;
	PrintWriter pw = null;
	try {
	    InetAddress localAddr = s.getLocalAddress();
	    InetAddress addr = s.getInetAddress();
	    if (!localAddr.equals(addr)){
		refuse(s);
		return;
	    }
	    InputStream in = s.getInputStream();
	    long cc = readLong(in);
	    if (cc != cookie){
		refuse(s);
		System.err.println("authentication failed");
		return;
	    }
	    int echo = in.read();

	    if (DEBUG){
		System.out.println("echo = " + echo);
	    }

	    Context context = new Context(new Package());
	    OutputStream out = s.getOutputStream();
	    context.setOutputStream(out);
	    context.setErrorStream(out);
	    pw = context.getErrorStream();
	    context.setTerminalStream(null);

	    if (echo != 0){
		Pnuts.load(in, context); // initialization
		context.setTerminalStream(out);
	    }
	    Pnuts.load(in, (echo != 0), context);
	} catch (Exception e){
	    if (Pnuts.isVerbose()){
		e.printStackTrace(pw);
	    } else {
		pw.println(e);
	    }
	    pw.flush();
	} finally {
	    busy--;
	    try {
		s.close();
		if (DEBUG){
		    System.out.println("closed");
		}
	    } catch (IOException e2){
	    }
	    System.gc();
	}
    }

    void handleOnNewThread() {
	Thread th = new Thread(new Runnable(){
	    public void run(){
		handle();
	    }
	});
	th.setPriority(4);
	th.start();
    }

    public void start() throws IOException {
	ss = new ServerSocket(port);
	initialize(ss.getLocalPort());

	for (int i = 0; i < NTHREAD; i++){
	    Thread th = new Thread(new Worker(i));
	    th.setPriority(4);
	    th.start();
	}

	while (true){
	    Socket s = ss.accept();
	    if (s != null){
		req.push(s);
		if (busy >= NTHREAD){
		    handleOnNewThread();
		}
	    } else {
		throw new RuntimeException("connection refused");
	    }
	}
    }

    void initialize(int port) throws IOException {
	System.out.println("port = " + port);

	File pref = new File(home, ".pnuts");
	if (!pref.exists()){
	    if (!pref.mkdir()){
		return;
	    }
	}
	FileOutputStream out = new FileOutputStream(new File(pref, "cookie"));
	out.write((port >>> 8) & 0xff);
	out.write((port >>> 0) & 0xff);
	long r = new Random(new Random().nextLong()).nextLong();
	out.write((int)(r >>> 56) & 0xff);
	out.write((int)(r >>> 48) & 0xff);
	out.write((int)(r >>> 40) & 0xff);
	out.write((int)(r >>> 32) & 0xff);
	out.write((int)(r >>> 24) & 0xff);
	out.write((int)(r >>> 16) & 0xff);
	out.write((int)(r >>> 8) & 0xff);
	out.write((int)(r >>> 0) & 0xff);
	this.cookie = r;
	out.close();
    }

    long readLong(InputStream in) throws IOException {
	int c1 = in.read();
	int c2 = in.read();
	int c3 = in.read();
	int c4 = in.read();
	int c5 = in.read();
	int c6 = in.read();
	int c7 = in.read();
	int c8 = in.read();
	return
	    ((long)((c1 << 24) + (c2 << 16) + (c3 << 8) + (c4 << 0)) << 32) +
	    (((c5 << 24) + (c6 << 16) + (c7 << 8) + (c8 << 0)) & 0xFFFFFFFFL);
    }

    public static void main(String args[]) throws Throwable {
	Pnuts.load("init");
	Pnuts.eval("shutdown=exit;exit=quit");

	int port = DEFAULT_PORT;
	String home = null;
	for (int i = 0; i < args.length; i++){
	    if (args[i].startsWith("-port")){
		i++;
		port = Integer.parseInt(args[i]);
	    } else if (args[i].startsWith("-home")){
		i++;
		home = args[i];
	    }
	}
	PersonalServer s = new PersonalServer(port, home);
	s.start();
    }
}
