/*
 * @(#) PnutsCompiler.java	1.4 99/12/21
 *
 * Copyright (c) 1997-1999 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.tools;

import pnuts.compiler.Compiler;
import pnuts.compiler.*;
import pnuts.lang.*;
import java.util.*;
import java.util.zip.*;
import java.io.*;

/**
 * A batch compiler for Pnuts.
 * It reads Pnuts scripts and generates class files or a JAR file.
 *
 * <pre>
 * Usage:
 *    pnutsc [ -d destination_directory ] [ -o jar_file ] [ -v ] [ -no_proxy ] [ -main ] [ -C dir ] [ -prefix name ] script_file ...
 * </pre>
 */
public class PnutsCompiler {
    
    Compiler pc;
    boolean verbose = false;
    boolean dynamic = false;
    boolean includeMain = false;
    int count = 0;
    static String file_sep = System.getProperty("file.separator");
    static String prefix;

    public PnutsCompiler(){
    }

    public void setVerbose(boolean flag){
	this.verbose = flag;
    }

    public void includeMainMethod(boolean flag){
	this.includeMain = flag;
    }

    Compiler getCompiler(String name){
	Compiler compiler = new Compiler(name, false, dynamic);
	compiler.includeMainMethod(includeMain);
	return compiler;
    }

    public void compileToZip(Pnuts p, String name, ZipOutputStream zout)
    {
	ZipWriterHandler handler = new ZipWriterHandler(zout);
	if (verbose){
	    handler.setVerbose(true);
	}
	getCompiler(name).compile(p, handler);
    }

    public void compileToFile(Pnuts p, String name, File dir)
    {
	FileWriterHandler handler = new FileWriterHandler(dir);
	if (verbose){
	    handler.setVerbose(true);
	}
	getCompiler(name).compile(p, handler);
    }

    static void printUsage(){
	System.out.println("pnutsc [ -d destination_directory ] [ -o jar_file ] [ -v ] [ -no_proxy ] [ -main ] [ -C dir ] [ -prefix name ] script_file ...");
    }

    public static String relativeName(String path, String base){
	if (base == null){
	    return path;
	}
	int idx = path.indexOf(base);
	String name = path;
	if (idx >= 0){
	    int offset = base.length();
	    if (!base.endsWith(file_sep)){
		offset++;
	    }
	    name = path.substring(offset);
	}
	return name;
    }
	
    public static String className(String name){
	int pos = name.lastIndexOf('.');
	String cname = name;
	if (pos > 0){
	    cname = name.substring(0, pos).replace(file_sep.charAt(0), '.');
	}
	if (prefix != null){
	    if (!"".equals(prefix) && !prefix.endsWith(".")){
		prefix = prefix + ".";
	    }
	    cname = prefix + cname;
	}
	return cname;
    }
    
    public static void main(String args[]) throws Throwable {
	int nargs = args.length;

	if (nargs == 0){
	    printUsage();
	    return;
	}

	PnutsCompiler compiler = new PnutsCompiler();
	compiler.dynamic = true;

	String dest = null;
	String jar = null;
	Vector files = new Vector();
	String base = null;

	for (int i = 0; i < nargs; i++){
	    if ("-d".equals(args[i])){
		dest = args[++i];
	    } else if ("-no_proxy".equals(args[i])){
		compiler.dynamic = false;
	    } else if ("-o".equals(args[i])){
		jar = args[++i];
	    } else if ("-v".equals(args[i])){
		compiler.setVerbose(true);
	    } else if ("-C".equals(args[i])){
		base = args[++i];
	    } else if ("-main".equals(args[i])){
		compiler.includeMainMethod(true);
	    } else if ("-prefix".equals(args[i])){
		compiler.prefix = args[++i];
	    } else {
		files.addElement(args[i]);
	    }
	}
	if (dest == null){
	    dest = System.getProperty("user.dir");
	}
	File dir = new File(dest);

	if (!dir.exists()){
	    System.err.println("Directory " + dest + " does not exist");
	}

	if (files.size() == 0){
	    printUsage();
	    return;
	}

	ZipOutputStream zout = null;

	if (jar != null){
	    zout = new ZipOutputStream(new FileOutputStream(jar));
	}

	for (Enumeration e = files.elements(); e.hasMoreElements(); ){
	    String fileName = (String)e.nextElement();

	    Pnuts pn = Pnuts.parse(new FileInputStream(fileName));
	    String className = className(relativeName(fileName, base));

	    if (jar != null){
		compiler.compileToZip(pn, className, zout);
	    } else {
		compiler.compileToFile(pn, className, dir);
	    }
	}
	if (jar != null){
	    zout.close();
	}
    }
}
