/*
 * @(#) TerminalDebugger.java	1.7 00/08/04
 *
 * Copyright (c) 1997-1999 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.tools;

import java.io.Reader;
import java.io.File;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.IOException;
import java.net.URL;
import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.ResourceBundle;
import java.util.MissingResourceException;
import pnuts.lang.Pnuts;
import pnuts.lang.Context;
import pnuts.lang.SimpleNode;
import pnuts.lang.Runtime;
import pnuts.lang.PnutsParserTreeConstants;
import pnuts.ext.UnparseVisitor;

/**
 * This class implements a debugger for Pnuts interpreter.
 * It is used through <tt>pnuts -d</tt> command.
 * <pre>
 * Commands:
 *   stop at [FILE:]LINENO
 *       Stop execution at the LINENO
 *   stop in FUNC[:NARGS]
 *       Stop execution when FUNC is called.
 *       When NARGS is specified, stop when FUNC with NARGS is called.
 *   clear
 *       Clear all breakpoints
 *   cont
 *       Continue execution
 *   trace
 *       Toggle trace mode
 *   step [NUM]
 *       Single step NUM lines.  The default number is 1.
 *   step up
 *       Step out of the current function
 *   next [NUM]
 *       Step NUM line (step OVER calls).  The default number is 1.
 *   help
 *       Print a summary of commands
 *   ?
 *       Same as help.
 *  </pre>
 */
public class TerminalDebugger extends Runtime implements ContextFactory, CommandListener, PnutsParserTreeConstants {
    private static final boolean DEBUG = false;

    Hashtable functions;
    Hashtable files;
    BufferedReader reader;
    boolean initialized = false;

    boolean step = false;
    boolean step_up = false;
    boolean next = false;
    int nsteps;
    int nnexts;
    int c_depth;
    int initialEvalDepth;
    int initialCallDepth;
    int e_depth;
    Object file;
    int line;
    boolean trace;
    boolean interactive;

    boolean session = false;

    static String indent = " >>> ";

    public TerminalDebugger(){
	this(new InputStreamReader(System.in));
	interactive = true;
    }

    /**
     * @param reader	debug script to read in
     */
    public TerminalDebugger(Reader reader){
	initialized = false;
	interactive = false;
	if (reader instanceof BufferedReader){
	    this.reader = (BufferedReader)reader;
	} else {
	    this.reader = new BufferedReader(reader);
	}
    }

    public Context createContext(){
	DebugContext dc = new DebugContext();
	dc.addCommandListener(this);
	return dc;
    }

    void init(){
	initialEvalDepth = Pnuts.evalDepth();
	nsteps = 1;
	nnexts = 1;
	c_depth = 0;
	e_depth = 0;
	file = null;
	line = 0;
	trace = false;
	functions = new Hashtable();
	files = new Hashtable();
    }

    /**
     * @param reader	The target script to be tested
     */
    void setInput(BufferedReader reader){
	this.reader = reader;
    }

    /**
     * @see	pnuts.tools.CommandListener
     */
    void exit(CommandEvent event){
	if (session){
	    return;
	}
	DebugContext dc = (DebugContext)event.getSource();
	if (DEBUG){
	    System.out.println("depth = " + dc.getEvalDepth() + ", initial depth = " + initialEvalDepth);
	}
	if (dc.getEvalDepth() <= initialEvalDepth &&
	    dc.getCallDepth() <= initialCallDepth)
	{
	    if (DEBUG){
		System.out.println("initialCallDepth = " + initialCallDepth);
		System.out.println("callDepth = " + dc.getCallDepth());
	    }
	    PrintWriter term = dc.getTerminalStream();
	    term.println("# Returns " + Pnuts.format(event.getArg()));
	    term.flush();
	    initialized = false;
	}
    }

    /**
     * set a breakpoint at the specified position
     */
    void setBreakPoint(String file, int lineno){
	if (DEBUG){
	    System.out.println("setBreakPoint(" + file + ", " + lineno + ")");
	}
	if (file == null){
	    return;
	}
	Vector lines = (Vector)files.get(file);
	if (lines == null){
	    lines = new Vector();
	    files.put(file, lines);
	}
	Integer i = new Integer(lineno);
	if (!lines.contains(i)){
	    lines.addElement(i);
	}
    }

    Vector getBreakPoints(Object file){
	if (file instanceof URL){
	    URL url = (URL)file;
	    String f = url.getFile();
	    for (Enumeration e = files.keys(); e.hasMoreElements(); ){
		String key = (String)e.nextElement();
		if (f.endsWith(key)){
		    return (Vector)files.get(key);
		}
	    }
	} else if (file instanceof File){
	    String f = ((File)file).getPath();
	    for (Enumeration e = files.keys(); e.hasMoreElements(); ){
		String key = (String)e.nextElement();
		if (f.endsWith(key)){
		    return (Vector)files.get(key);
		}
	    }
	}
	return null;
    }

    void setBreakPointInFunction(String func_name){
	if (functions.get(func_name) == null){
	    functions.put(func_name, func_name);
	}
    }

    void setBreakPointInFunction(String func_name, int nargs){
	String key = func_name + ":" + nargs;
	if (functions.get(key) == null){
	    functions.put(key, key);
	}
    }

    void clearBreakPoints(){
	functions.clear();
	files.clear();
    }

    SimpleNode getTopNode(SimpleNode node){
	while (node != null && node.id != JJTEXPRESSIONLIST){
	    SimpleNode parent = node.jjtGetParent();
	    if (parent == null || parent.id == JJTBLOCK){
		break;
	    }
	    node = parent;
	}
	return node;
    }

    public void signal(CommandEvent event){
	int eventType = event.getType();
	if (eventType == CommandEvent.EXITED){
	    exit(event);
	    return;
	} else if (eventType == CommandEvent.EXCEPTION){
	    initialized = false;
	    return;
	} else {
	    DebugContext dc = (DebugContext)event.getSource();
	    SimpleNode node = (SimpleNode)event.getArg();
	    lineUpdated(dc, node);
	}
    }

    void lineUpdated(DebugContext dc, SimpleNode node){
	if (session){
	    return;
	}

	PrintWriter term = dc.getTerminalStream();
	int beginLine = getBeginLine(dc);
	int endLine = getEndLine(dc);

	if (!initialized){
	    init();
	    initialCallDepth = dc.getCallDepth();

	    Object f = getScriptSource(dc);
	    if (f == null){
		f = "?";
	    }
	    term.println("# Stopped at " + f + ":" + beginLine);
	    SimpleNode n = getTopNode(node);
	    if (n != null){
		term.print(indent);
		term.println(n.accept(new UnparseVisitor(), dc));
	    }
	    c_depth = dc.getCallDepth();
	    e_depth = dc.getEvalDepth();
	    this.file = getScriptSource(dc);
	    this.line = getBeginLine(dc);
	    initialized = true;
	    session(dc);
	    return;
	}
	if (step){
	    if (this.file != getScriptSource(dc) ||
		this.line != beginLine ||
		c_depth != dc.getCallDepth())
	    {

		this.line = beginLine;
		this.file = getScriptSource(dc);
		c_depth = dc.getCallDepth();

		if (--nsteps < 1){
		    Object f = file;
		    if (f == null){
			f = "?";
		    }
		    term.println("# Stopped at " + f + ":" + beginLine);
		    SimpleNode n = getTopNode(node);
		    if (n != null){
			term.print(indent);
			term.println(n.accept(new UnparseVisitor(), dc));
		    }
		    session(dc);
		} else if (trace){
		    term.print(file + ":" + line + indent);
		    term.println(node.accept(new UnparseVisitor(), dc));
		    term.flush();
		}
	    }
	} else if (step_up){
	    if (e_depth > dc.getEvalDepth() ||
		e_depth >= dc.getEvalDepth() && c_depth > dc.getCallDepth()){

		this.line = beginLine;
		this.file = getScriptSource(dc);
		this.c_depth = dc.getCallDepth();

		Object f = this.file;
		if (f == null){
		    f = "?";
		}
		term.println("# Stopped at " + f + ":" + beginLine);

		SimpleNode n = getTopNode(node);
		if (n != null){
		    term.print(indent);
		    term.println(n.accept(new UnparseVisitor(), dc));
		}
		session(dc);
	    } else if (trace){
		term.print(file + ":" + line + indent);
		term.println(node.accept(new UnparseVisitor(), dc));
		term.flush();
	    }
	} else if (next && e_depth >= dc.getEvalDepth() && c_depth >= dc.getCallDepth()){
	    if (this.file != getScriptSource(dc) ||
		this.line != beginLine ||
		c_depth != dc.getCallDepth()){

		this.line = beginLine;
		this.file = getScriptSource(dc);
		this.c_depth = dc.getCallDepth();

		if (--nnexts < 1){
		    Object f = this.file;
		    if (f == null){
			f = "?";
		    }
		    term.println("# Stopped at " + f + ":" + beginLine);
		    SimpleNode n = getTopNode(node);
		    if (n != null){
			term.print(indent);
			term.println(n.accept(new UnparseVisitor(), dc));
		    }
		    session(dc);
		}
	    } else if (trace){
		term.print(file + ":" + line + indent);
		term.println(node.accept(new UnparseVisitor(), dc));
		term.flush();
	    }
	} else {
	    Object file = getScriptSource(dc);
	    if (trace){
		term.print(file + ":" + beginLine + indent);
		term.println(node.accept(new UnparseVisitor(), dc));
		term.flush();
	    }
	    if (file != null){
		Vector lines = getBreakPoints(file);

		if (lines != null &&
		    checkBreakPoint(lines, node.id, beginLine, endLine) &&
		    (line != beginLine || c_depth != dc.getCallDepth()))
		{
		    this.line = beginLine;
		    this.file = getScriptSource(dc);
		    this.c_depth = dc.getCallDepth();
		    Object f = this.file;
		    if (f == null){
			f = "?";
		    }
		    term.println("# Stopped at " + f + ":" + beginLine);
		    SimpleNode n = getTopNode(node);
		    if (n != null){
			term.print(indent);
			term.println(n.accept(new UnparseVisitor(), dc));
		    }
		    session(dc);
		    return;
		}
	    }
	    if (node.id == JJTAPPLICATIONNODE){
		String name = node.jjtGetChild(0).str;
		if (name != null && functions.get(name) != null){
		    Object f = this.file;
		    if (f == null){
			f = "?";
		    }
		    term.println("# Stopped at " + f + ":" + beginLine);
		    SimpleNode n = getTopNode(node);
		    if (n != null){
			term.print(indent);
			term.println(n.accept(new UnparseVisitor(), dc));
		    }
		    session(dc);
		}
	    }
	}
    }

    boolean checkBreakPoint(Vector lines, int nodeID, int begin, int end){
	if (DEBUG){
	    System.out.println("checkBreakPoint(" + begin + ", " + end + ")");
	}
	if (nodeID == JJTBLOCK){
	    return false;
	}
	for (Enumeration e = lines.elements(); e.hasMoreElements(); ){
	    int bp = ((Integer)e.nextElement()).intValue();
	    if (bp >= begin && bp <= end){
		return true;
	    }
	}
	return false;
    }

    String getCommand(){
	String cmd = null;
	try {
	    cmd = reader.readLine();
	} catch (IOException e){}
	return cmd;
    }

    void session(DebugContext context){
	step = false;
	step_up = false;
	next = false;

	while (true){

	    file = getScriptSource(context);
	    c_depth = context.getCallDepth();
	    e_depth = context.getEvalDepth();

	    PrintWriter term = context.getTerminalStream();

	    if (interactive){
		term.print("debug> ");
		term.flush();
	    }

	    String cmd = getCommand();
	    if (!interactive){
		term.println("debug> " + cmd);
	    }

	    if (cmd == null){
		break;
	    }
	    int offset = 0;;
	    if ((offset = cmd.indexOf("stop at ")) >= 0){
		String arg = cmd.substring(offset + "stop at ".length()).trim();
		int idx = arg.indexOf(':');
		if (idx < 0){
		    Object f = getScriptSource(context);
		    String s = null;
		    if (f instanceof URL){
			s = ((URL)f).getFile();
		    } else if (f instanceof File){
			s = ((File)f).getPath();
		    } else if (f instanceof Runtime){
			s = f.getClass().getName();
		    }
		    setBreakPoint(s, Integer.parseInt(arg));
		} else {
		    setBreakPoint(arg.substring(0, idx), Integer.parseInt(arg.substring(idx + 1)));
		}
	    } else if ((offset = cmd.indexOf("stop in ")) >= 0){
		String arg = cmd.substring(offset + "stop at ".length()).trim();
		int idx = arg.indexOf(':');
		if (idx < 0){
		    setBreakPointInFunction(arg);
		} else {
		    setBreakPointInFunction(arg.substring(0, idx), Integer.parseInt(arg.substring(idx + 1)));
		}
	    } else if ("help".equals(cmd) || "?".equals(cmd)){
		try {
		    ResourceBundle rb = ResourceBundle.getBundle("pnuts.tools.debug");
		    String help = rb.getString("pnuts.debug.help");
		    term.println(help);
		} catch (MissingResourceException mis){
		    mis.printStackTrace(term);
		}
	    } else if ((offset = cmd.indexOf("step")) >= 0){
		String arg = cmd.substring(offset + "step".length()).trim();
		if (arg.indexOf("up") >= 0){
		    file = getScriptSource(context);
		    c_depth = context.getCallDepth();
		    e_depth = context.getEvalDepth();
		    step_up = true;
		} else if (arg.length() > 0 && Character.isDigit(arg.charAt(0))){
		    step = true;
		    nsteps = Integer.parseInt(arg);
		} else {
		    step = true;
		    nsteps = 1;
		}
		break;
	    } else if ((offset = cmd.indexOf("next")) >= 0){
		String arg = cmd.substring(offset + "step".length()).trim();
		if (arg.length() > 0 && Character.isDigit(arg.charAt(0))){
		    nnexts = Integer.parseInt(arg);
		} else {
		    nnexts = 1;
		}
		next = true;
		c_depth = context.getCallDepth();
		e_depth = context.getEvalDepth();
		file = getScriptSource(context);
		break;
	    } else if ("trace".equals(cmd)){
		trace = !trace;
		if (trace){
		    term.println("on");
		} else {
		    term.println("off");
		}
	    } else if ("cont".equals(cmd)){
		break;
	    } else if ("clear".equals(cmd)){
		clearBreakPoints();
	    } else {
		try {
		    session = true;
		    term.println(Pnuts.format(Pnuts.eval(cmd, context)));
		} catch (Throwable t){
		    term.println(t);
		} finally {
		    session = false;
		}
	    }
	}
    }
}
