/*
 * @(#) VisualDebugger.java	1.2 00/08/01
 *
 * Copyright (c) 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.tools;

import java.io.*;
import java.util.*;
import java.net.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.text.*;
import pnuts.lang.*;
import pnuts.lang.Runtime;
import pnuts.awt.*;
import pnuts.ext.*;
import pnuts.tools.*;

/**
 * A visual debugger for Pnuts
 * 
 * <pre>
 *   ctx = VisualDebugger().createContext(getContext())
 *   load(script, ctx)
 * </pre>
 *
 * @version	1.2
 */
public class VisualDebugger extends Runtime implements ContextFactory, CommandListener, PnutsParserTreeConstants {
    private final static boolean DEBUG = false;

    JFrame jfr;
    JTextArea jta;
    JMenuBar jmb;
    JDialog ins;

    DebugContext dc;

    private Object currentSource;
    private int currentLine;
    private int n_nexts;
    private int n_steps;
    private int command;
    private Element lineMap;

    private Hashtable files;		// source -> lines (break points)
    private Hashtable highlights;	// source -> Highlighter
    private Hashtable tags;		// source:line -> highlight tag
    private boolean initialized = false;
    private int initialEvalDepth;
    private int initialCallDepth;
    private int e_depth;
    private int c_depth;
    private boolean guiStarted = false;
    private PrintWriter errorStream;
    private Object traceTag;
    private Context parent;

    static final int OPEN = 1;
    static final int STEP = 2;
    static final int STEP_UP = 3;
    static final int NEXT = 4;
    static final int CONT = 5;
    static final int EXIT = 6;
    static final int CLEAR_BP = 7;
    static final int INSPECT = 8;

    public Context createContext(Context context){
	dc = new DebugContext(context);
	dc.addCommandListener(this);
	parent = context;
	return dc;
    }

    public Context createContext(){
	dc = new DebugContext();
	dc.addCommandListener(this);
	return dc;
    }

    void init(){
	files = new Hashtable();
	highlights = new Hashtable();
	tags = new Hashtable();
	initialEvalDepth = Pnuts.evalDepth();
	e_depth = 0;
	c_depth = 0;
	n_steps = 1;
	n_nexts = 1;
	if (DEBUG){
	    System.out.println("initialized");
	    System.out.println("initialEvalDepth = " + initialEvalDepth);
	}
    }

    void startGUI(){
	jfr = new JFrame();
	jta = new JTextArea();
	jta.addMouseListener(new MouseHandler());
	jta.setFont(new Font("monospaced", Font.PLAIN, 11));
	jta.setEditable(false);

	jmb = new JMenuBar();
	JMenu fileMenu = new JMenu("File");
	JMenu editMenu = new JMenu("Edit");
	jmb.add(fileMenu);
	jmb.add(editMenu);
	
	JMenuItem open = new JMenuItem("Open");
	open.addActionListener(new MenuHandler(OPEN));
	open.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_O, Event.CTRL_MASK));
	fileMenu.add(open);

	JMenuItem step = new JMenuItem("Step");
	step.addActionListener(new MenuHandler(STEP));
	step.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_S, Event.CTRL_MASK));
	fileMenu.add(step);

	JMenuItem step_up = new JMenuItem("Step Up");
	step_up.addActionListener(new MenuHandler(STEP_UP));
	step_up.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_U, Event.CTRL_MASK));
	fileMenu.add(step_up);

	JMenuItem next = new JMenuItem("Next");
	next.addActionListener(new MenuHandler(NEXT));
	next.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_N, Event.CTRL_MASK));
	fileMenu.add(next);

	JMenuItem cont = new JMenuItem("Continue");
	cont.addActionListener(new MenuHandler(CONT));
	cont.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_T, Event.CTRL_MASK));
	fileMenu.add(cont);

	JMenuItem exit = new JMenuItem("Exit");
	exit.addActionListener(new MenuHandler(EXIT));
	exit.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Q, Event.CTRL_MASK));
	fileMenu.add(exit);

	JMenuItem inspect = new JMenuItem("Show Inspector");
	inspect.addActionListener(new MenuHandler(INSPECT));
	inspect.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_H, Event.CTRL_MASK));
	editMenu.add(inspect);

	jfr.getContentPane().add(new JScrollPane(jta));
	jfr.setJMenuBar(jmb);
	jfr.setSize(400, 400);
	jfr.show();

	ins = new JDialog(jfr);
	final TextField t1 = new TextField();
	final TextArea t2 = new TextArea("", 0, 0, TextArea.SCROLLBARS_VERTICAL_ONLY);
	t2.setEditable(false);
	t2.setBackground(ins.getBackground());
	JPanel p = new JPanel();
	JButton btn = new JButton("Evaluate");
	ActionListener actionListener = new ActionListener(){
		public void actionPerformed(ActionEvent e){
		    String txt = t1.getText();
		    String result = Pnuts.format(evaluate(txt, dc));
		    t2.setText(result);
		    t1.select(0, txt.length());
		}
	    };
	t1.addActionListener(actionListener);
	btn.addActionListener(actionListener);
	p.add(btn);
	ins.getContentPane().add(t1, BorderLayout.NORTH);
	ins.getContentPane().add(t2, BorderLayout.CENTER);
	ins.getContentPane().add(p, BorderLayout.SOUTH);
	ins.setSize(400, 400);

	errorStream = new PrintWriter(new DialogOutputStream(jfr), true);
	if (parent != null){
	    parent.setErrorStream(errorStream);
	}
	guiStarted = true;
    }

    void open(){
	synchronized (this){
	    notifyAll();
	}
	JFileChooser chooser = new JFileChooser(System.getProperty("user.dir"));
	int ret = chooser.showOpenDialog(jfr);
	if (ret == JFileChooser.APPROVE_OPTION) {
	    final File file = chooser.getSelectedFile();

	    Thread t = new Thread(new Runnable(){
		    public void run(){
			open(file.getPath());
		    }
		});
	    t.start();
	}
    }

    public void open(String filename){
	command = OPEN;
	File file = new File(filename);
	try {
	    if (file.exists()){
		showScript(fileToURL(file), 1);
	    } else {
		errorStream.println(filename + " is not found");
	    }
	} catch (MalformedURLException e){
	}
    }

    void exitGUI(){
	jfr.dispose();
	jfr = null;
	jta = null;
    }

    synchronized void do_step(int n){
	command = STEP;
	n_steps = n;
	notifyAll();
    }

    synchronized void do_stepup(){
	command = STEP_UP;
	notifyAll();
    }

    synchronized void do_next(int n){
	command = NEXT;
	n_nexts = n;
	notifyAll();
    }

    synchronized void do_cont(){
	command = CONT;
	n_steps = 0;
	n_nexts = 0;
	notifyAll();
    }

    void setBreakPoint(Object source, int lineno){
	if (DEBUG){
	    System.out.println("setBreakPoint(" + source + ", " + lineno + ")");
	}
	if (source == null){
	    return;
	}
	Vector lines = (Vector)files.get(source.toString());
	if (lines == null){
	    lines = new Vector();
	    files.put(source.toString(), lines);
	}
	Integer i = new Integer(lineno + 1);
	if (!lines.contains(i)){
	    lines.addElement(i);
	}
	if (DEBUG){
	    System.out.println("files = " + files);
	    System.out.println("lines = " + lines);
	}
    }

    void removeBreakPoint(Object source, int lineno){
	if (source == null){
	    return;
	}
	Vector lines = (Vector)files.get(source.toString());
	if (lines == null){
	    return;
	}
	lines.removeElement(new Integer(lineno + 1));
    }

    Vector getBreakPoints(Object file){
	if (file == null){
	    return null;
	}
	String f = file.toString();
	for (Enumeration e = files.keys(); e.hasMoreElements(); ){
	    String key = (String)e.nextElement();
	    if (f.endsWith(key)){
		return (Vector)files.get(key);
	    }
	}
	return null;
    }

    void clearBreakPoints(){
	files.clear();
    }

    void showScript(Object source, int beginLine){
	if (DEBUG){
	    System.out.println("showScript(" + source + ", " + beginLine + ")");
	}
	try {
	    Highlighter highlighter = null;
	    if (source != null){
		highlighter = (Highlighter)highlights.get(source.toString());
	    } else {
		jfr.setTitle("?");
	    }
	    
	    if (source != null && !source.equals(currentSource)){
		currentSource = source;
		jfr.setTitle(source.toString());
		CharArrayWriter caw = new CharArrayWriter();
		if (source instanceof URL){
		    URL url = (URL)source;
		    InputStream in = url.openStream();
		    BufferedReader reader =
			new BufferedReader(new InputStreamReader(in));
		    char[] buf = new char[512];
		    int n = 0;
		    while ((n = reader.read(buf)) != -1){
			caw.write(buf, 0, n);
		    }
		    jta.setText(new String(caw.toCharArray()));

		} else if (source instanceof Runtime){
		    jta.setText("");
		    return;
		} else {
		    throw new RuntimeException("invalid source:" + source);
		}

		try {
		    if (highlighter == null){
			highlighter = new DefaultHighlighter();
			highlights.put(source.toString(), highlighter);
		    }
		    jta.setHighlighter(highlighter);

		    lineMap = jta.getDocument().getDefaultRootElement();

		    Vector lines = (Vector)files.get(source.toString());
		    if (lines != null){
			for (Enumeration e = lines.elements(); e.hasMoreElements(); ){
			    int line = ((Integer)e.nextElement()).intValue();
			    Element pos = lineMap.getElement(line - 1);
			    if (pos != null){
				int start = pos.getStartOffset();
				int end = pos.getEndOffset();
				highlighter.addHighlight(start,
							 end,
							 new DefaultHighlighter.DefaultHighlightPainter(Color.orange));
			    }
			}
		    }

		    Element pos = lineMap.getElement(beginLine - 1);
		    if (pos != null){
			int start = pos.getStartOffset();
			int end = pos.getEndOffset();
			jta.select(start, start);
			traceTag =
			    highlighter.addHighlight(start,
						     end,
						     new DefaultHighlighter.DefaultHighlightPainter(Color.cyan));
		    }
		
		} catch (BadLocationException ex){
		    ex.printStackTrace();
		    return;
		}
	    } else {

		if (lineMap != null){
		    Element pos = lineMap.getElement(beginLine - 1);
		    if (pos != null){
			int start = pos.getStartOffset();
			int end = pos.getEndOffset();
			try {
			    jta.select(start, end);
			    if (highlighter != null){
				highlighter.changeHighlight(traceTag, start, end);
			    }
			} catch (BadLocationException badLocation){
			}
		    }
		}
	    }

	    jta.repaint();
	    jfr.setVisible(true);

	} catch (IOException e){
	    e.printStackTrace();
	    System.exit(-1);
	}
    }

    public void signal(CommandEvent event){
	dc = (DebugContext)event.getSource();
	dc.setErrorStream(errorStream);
	if (!guiStarted){
	    if (DEBUG){
		System.out.println("starting GUI");
	    }
	    startGUI();
	    command = OPEN;
	}
	int eventType = event.getType();

	if (DEBUG){
	    System.out.println(eventType);
	}

	if (!initialized){
	    init();
	    initialized = true;
	    initialCallDepth = dc.getCallDepth();
	}

	if (eventType == CommandEvent.EXITED){
	    if (DEBUG){
		System.out.println("exit: src=" + getScriptSource(dc) + ", " + Pnuts.evalDepth());
	    }
	    if (initialEvalDepth == Pnuts.evalDepth()){
		showScript(getScriptSource(dc), getBeginLine(dc));
	    }
	    return;
	}

	if (eventType == CommandEvent.LINE_UPDATED){
	    SimpleNode node = (SimpleNode)event.getArg();
	    int beginLine = getBeginLine(dc);
	    int endLine = getEndLine(dc);
	    Object source = getScriptSource(dc);
	    if (DEBUG){
		System.out.println("signal " + source + ":" + beginLine);
	    }

	    if (hasToStop(source, node, beginLine, endLine)){
		c_depth = dc.getCallDepth();
		e_depth = Pnuts.evalDepth();
		if (source == null && node != null){
		    jta.setText((String)node.accept(new UnparseVisitor(), null));
		}

		showScript(source, beginLine);

		currentSource = source;
		waitForCommands();
		currentLine = beginLine;
	    }
	} else if (eventType == CommandEvent.EXCEPTION){
	    Throwable t = (Throwable)event.getArg();
	    if (DEBUG){
		System.out.println(t);
	    }
	    if (t instanceof PnutsException){
		PnutsException pe = (PnutsException)t;
		Object source = pe.getScriptSource();
		int beginLine = pe.getLine();
		showScript(source, beginLine);
	    }
	}
    }

    boolean checkBreakPoint(Vector lines, SimpleNode node, int begin, int end){
	if (DEBUG){
	    System.out.print("lines = " + lines + ", " + begin + ", " + end);
	}
	if (node != null && node.id == JJTBLOCK){
	    return false;
	}
	for (Enumeration e = lines.elements(); e.hasMoreElements(); ){
	    int bp = ((Integer)e.nextElement()).intValue();
	    if (bp >= begin && bp <= end){
		return true;
	    }
	}
	return false;
    }

    boolean hasToStop(Object source, SimpleNode node, int beginLine, int endLine){
	if (source instanceof Runtime){
	    return false;
	}
	if (source == null && node != null){
	    return true;
	}
	if (command == OPEN){
	    currentLine = beginLine;
	    command = 0;
	    return true; 
	} else if (command == STEP){
	    if (--n_steps < 1){
		return true;
	    }
	} else if (command == NEXT){
	    int d1 = Pnuts.evalDepth();
	    int d2 = dc.getCallDepth();
	    if (DEBUG){
		System.out.println(e_depth + " <-> " + d1 + ", " +
				   c_depth + " <-> " + d2);
	    }
	    if (e_depth >= d1 && c_depth >= d2 &&
		(currentSource != source || currentLine != beginLine))
	    {
		if (--n_nexts < 1){
		    return true;
		}
	    }
	} else if (command == STEP_UP){
	    int d1 = Pnuts.evalDepth();
	    int d2 = dc.getCallDepth();
	    if (e_depth > d1 || c_depth > d2){
		return true;
	    }
	}

	Vector lines = getBreakPoints(source);

	if (DEBUG){
	    System.out.println("lines = " + lines + ", " + beginLine + "<->" + endLine);
	}

	if (lines != null && checkBreakPoint(lines, node, beginLine, endLine)){
	    currentLine = beginLine;
	    return true;
	}

	return false;
    }

    synchronized void waitForCommands(){
	command = 0;
	try {
	    while (command == 0){
		wait();
	    }
	} catch (InterruptedException e){
	}
    }

    public Object evaluate(String expr, Context ctx){
	try {
	    return Pnuts.eval(expr, ctx);
	} catch (Throwable t){
	    Runtime.printError(t, ctx);
	    return null;
	}
    }

    class MenuHandler implements ActionListener {
	private int id;

	MenuHandler(int id){
	    this.id = id;
	}

	public void actionPerformed(ActionEvent e){
	    switch (id){
	    case OPEN:
		open();
		break;
	    case STEP:
		do_step(1);
		break;
	    case STEP_UP:
		do_stepup();
		break;
	    case NEXT:
		do_next(1);
		break;
	    case CONT:
		do_cont();
		break;
	    case EXIT:
		System.exit(0);
		break;
	    case CLEAR_BP:
		break;
	    case INSPECT:
		showInspector();
		break;
	    }
	}
    }

    int getElementIndex(int pos){
	Element lineMap = jta.getDocument().getDefaultRootElement();
	return lineMap.getElementIndex(pos);
    }

    Element getElement(int pos){
	Element lineMap = jta.getDocument().getDefaultRootElement();
	return lineMap.getElement(pos);
    }

    class MouseHandler extends MouseAdapter {
	public void mousePressed(MouseEvent e) {
	    maybeShowPopup(e);
	}

	public void mouseReleased(MouseEvent e) {
	    maybeShowPopup(e);
	}

	private void maybeShowPopup(MouseEvent e) {
	    if (!e.isPopupTrigger()) {
		return;
	    }
	    int x = e.getX();
	    int y = e.getY();
	    Point pt = new Point(x, y);
	    int pos = jta.viewToModel(pt);
	    if (pos > 0){
		int line = getElementIndex(pos);
		if (DEBUG){
		    System.out.println("line = " + line);
		}
		Element elem = getElement(line);

		try {
		    String key = currentSource + ":" + line;
		    Object tag = tags.get(key);
		    Highlighter highlighter = jta.getHighlighter();
		    if (tag == null){
			tag = highlighter.addHighlight(elem.getStartOffset(),
						       elem.getEndOffset(),
						       new DefaultHighlighter.DefaultHighlightPainter(Color.orange));
			tags.put(key, tag);
			setBreakPoint(currentSource, line);
		    } else {
			highlighter.removeHighlight(tag);
			tags.remove(key);
			removeBreakPoint(currentSource, line);
		    }
		} catch (BadLocationException ex){
		    ex.printStackTrace();
		}
	    }
	}
    }

    void showInspector(){
	Point pt = jfr.getLocation();
	ins.setLocation(pt.x + jfr.getSize().width, pt.y);
	ins.setVisible(true);
    }
}
