/*
 * @(#) psh.c	1.2 99/11/07
 *
 * Copyright (c) 1997-1999 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <malloc.h>
#include <fcntl.h>
#include <pwd.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/filio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>

/*
 * This program is the client side of the pserv/psh environment.
 * Psh reads Pnuts scripts from files or command-line expression and
 * sends them to the pserv program through a socket connection.
 * Pserv receives scripts sent by psh and execute them. 
 *
 * The first line that starts with '#' is ignored.
 * The variable $args is defined as an array of command arguments.
 * The variable $cwd is defined as the current working directory.
 *
 * Usage:
 *   psh [ -help | [ -port:NNNN ] [ -e expr | script_file, arg1, ...]]
 *
 */

#define INADDR_NONE 0xffffffff
#define LOOPBACK "127.0.0.1"

static int sock;

/*
 * Connect to host:port
 */
int do_connect(host, port)
     char *host;
     int port;
{
  int fd;
  in_addr_t inaddr;
  struct hostent *hp;
  struct sockaddr_in srv_addr;
  
  memset((char*)&srv_addr, 0, sizeof(srv_addr));
  srv_addr.sin_family = AF_INET;

  srv_addr.sin_port = htons(port);

  if ((inaddr = inet_addr(host)) != INADDR_NONE){
    memcpy((char*)&srv_addr.sin_addr, (char*)&inaddr, sizeof(inaddr));
  } else {
    if ((hp = gethostbyname(host)) == NULL){
      return -1;
    }
    memcpy((char*)&srv_addr.sin_addr, hp->h_addr, hp->h_length);
  }

  if (port < 0){
    return -1;
  }
  if ((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0){
    return -1;
  }
  if (connect(fd, (struct sockaddr*)&srv_addr, sizeof(srv_addr)) < 0){
    close(fd);
    return -1;
  }
  return fd;
}

/*
 * Get the home directory
 */
char* get_home()
{
  struct passwd *pass;
  pass = getpwuid(getuid());
  return pass->pw_dir;
}

/*
 * Check if the file exists
 */
int file_exists(char* file){
  struct stat st;
  if (stat(file, &st) != 0){
    return 0;
  }
  return (st.st_mode & S_IFMT) == S_IFREG;
}

/*
 * Read the cookie file into buf.
 */
int read_cookie(char *buf)
{
  char* home;
  char* path;

  home = get_home();
  path = (char*)malloc(strlen(home) + 15);
  if (path == NULL){
    return -1;
  }
  sprintf(path, "%s/.pnuts/cookie", home);
  if (file_exists(path)){
    int fd = open(path, O_RDONLY);
    if (read(fd, buf, 10) != 10){
      return -1;
    }
    return 0;
  }
  return -1;
}

int available(int fd, long *pbytes)
{
    long cur, end;
    struct stat buf;

    if (fstat(fd, &buf) >= 0) {
        int mode = buf.st_mode;
        if (S_ISCHR(mode) || S_ISFIFO(mode) || S_ISSOCK(mode)) {
	    int n;
            if (ioctl(fd, FIONREAD, &n) >= 0) {
		*pbytes = n;
                return 1;
            }
        }
    }
    if ((cur = lseek(fd, 0L, SEEK_CUR)) == -1) {
	return 0;
    } else if ((end = lseek(fd, 0L, SEEK_END)) == -1) {
	return 0;
    } else if (lseek(fd, cur, SEEK_SET) == -1) {
	return 0;
    }
    *pbytes = end - cur;
    return 1;
}

int stdin_available()
{
  long bytes;
  return available(0, &bytes) && bytes > 0;
}

void send_quit(int sock)
{
  do_write(sock, "quit()\n", 7);
}

void terminate()
{
  printf("terminate\n");
  send_quit(sock);
  close(sock);
}

#ifdef DEBUG
void dump(char* pt, int len)
{
  int i;
  for (i = 0; i < len; i++){
    printf("%x ", *(pt + i));
  }
  printf("\n");
}
#endif

int do_write(int fd, char* pt, int len)
{
  int n;

#ifdef DEBUG
  dump(pt, len);
#endif

  if ((n = write(fd, pt, len)) < 0){
    perror("write");
    exit(-1);
  }
  return n;
}

void read_stdin(int sock)
{
  int maxfdp1, nfound, nread;
  char buf[512];
  fd_set readmask;
  int first = 1;

  FD_ZERO(&readmask);

  for (;;){
    FD_SET(0, &readmask);
    FD_SET(sock, &readmask);
    maxfdp1 = sock + 1;

    nfound = select(maxfdp1, &readmask, (fd_set*)0, (fd_set*)0, (struct timeval*)0);

    if (nfound < 0){
      perror("select");
      exit(-1);
    }
    if (FD_ISSET(0, &readmask)){
      int offset = 0;
      nread = read(0, buf, sizeof(buf));
      if (nread < 0){
	perror("read");
	exit(-1);
      } else if (nread == 0){
	break;
      }

      if (first && buf[0] == '#' && buf[1] == '!'){
	char* p1 = strchr(buf, '\n');
	char* p2 = strchr(buf, '\r');
	if (p1 != NULL){
	  offset = p1 - buf + 1;
	}
	if (p2 != NULL){
	  if (p2 - buf + 1 > offset){
	    offset = p2 - buf + 1;
	  }
	}
      }

      if (write(sock, buf + offset, nread - offset) != nread - offset){
	perror("write");
	exit(-1);
      }
    }
    if (FD_ISSET(sock, &readmask)){
      nread = read(sock, buf, sizeof(buf));
      if (nread <= 0){
	break;
      }
      if (write(1, buf, nread) != nread){
	perror("write");
	exit(-1);
      }
    }
  }
}

/*
 * Read all data from fp and write to sock.
 * The first line which starts with '#!' is discarded.
 */
void read_input(FILE* fp, int sock)
{
  char line[512];
  int len;

  if (fgets(line, sizeof(line), fp) != NULL){
    if (line[0] != '#' || line[1] != '!'){
      len = strlen(line);
      line[len] = '\n';
      do_write(sock, line, len + 1);
    }
  }
  while (fgets(line, sizeof(line), fp) != NULL){
    len = strlen(line);
    line[len] = '\n';
    do_write(sock, line, len + 1);
  }
}

/*
 * Receive output from server
 */
int receive_output(int sock)
{
  int n;
  char line[512];

  while ((n = read(sock, line, sizeof(line))) > 0){
    do_write(1, line, n);
  }
}

char* append(char* buf, int* size, int* capacity, char* str)
{
  char* ret = buf;
  int len = strlen(str);
  if (*size + len + 1 > *capacity){
    ret = (char*)malloc(*capacity = (*capacity + *size + len + 1));
    if (ret == NULL){
      return NULL;
    }
    memcpy(ret, buf, *size);
  }
  memcpy(ret + *size, str, len);
  ret[*size + len] = '\0';
  *size += len;
  return ret;
}

char* escape(char* str)
{
  int i = 0;
  char* s = str;

  while (*s){
    switch (*s){
    case '\\':
    case '"':
      i++;
      break;
    }
    s++;
  }
  if (i > 0){
    char *ret;
    char *ss = ret = (char*)malloc(strlen(str) + i + 1);
    if (ss == NULL){
      return NULL;
    }
    s = str;
    while (*s){
      switch (*s){
      case '\\':
      case '"':
	*ss++ = '\\';
	break;
      }
      *ss++ = *s++;
    }
    *ss = *s;
    return ret;
  } else {
    return str;
  }
}

/*
 * Define $args=[argv[0],argv[1],..]
 */
int set_args(int sock, int argc, char** argv)
{
  int i;
  int count = 0;
  int cap = 64;
  char *buf = (char*)malloc(cap);

  buf = append(buf, &count, &cap, "$args=[\"");
  buf = append(buf, &count, &cap, escape(argv[1]));
  buf = append(buf, &count, &cap, "\"");
  for (i = 2; i < argc; i++){
    buf = append(buf, &count, &cap, ",\"");
    buf = append(buf, &count, &cap, escape(argv[i]));
    buf = append(buf, &count, &cap, "\"");
  }
  buf = append(buf, &count, &cap, "]\n");

  do_write(sock, buf, strlen(buf));

  return 0;
}

void send_cwd(int sock){
  char *cwd = getcwd((char*)NULL, 64);
  int len = strlen(cwd);
  char* cmd = (char*)malloc(8 + len + 1);
  if (!cmd){
    perror("malloc");
    return;
  }
  sprintf(cmd, "$cwd=\"%s\"\n", cwd);
  do_write(sock, cmd, strlen(cmd));
}

int main(int argc, char** argv){
  int i, n;
  int port = 0;
  char *pname;
  char *command = NULL;
  char pref[10];
  char* cookie;
  int zero = 0;
  int nonzero = -1;
  int from_pipe = 0;
  struct sigaction act;
  sigset_t sa_mask;

#ifdef DEBUG
  printf("psh starts\n");
#endif

  act.sa_handler = terminate;
  act.sa_mask = sa_mask;
  act.sa_flags = 0;
  sigaction(SIGINT, &act, NULL);

  pname = argv[0];

  i = 1;
  while (argc > 1){
    char* a = argv[i];
    if (memcmp(a, "-port:", 6) == 0){
      port = atoi(a + 6);
      argc--;
      argv++;
      i++;
    } else if (memcmp(a, "-e", 2) == 0){
      i++;
      command = strdup(argv[i]);
      i++;
      argc -= 2;
      argv += 2;
    } else if (memcmp(a, "-help", 5) == 0){
      printf("Usage:  %s [ -port:NNNN ] [ -e expr | script_file ]\n", pname);
      exit(1);
    } else {
      break;
    }
  }

  if (read_cookie(pref) < 0){
    fprintf(stderr, "cookie read failed\n");
    exit(-1);
  }
  if (port == 0){
    port = ((pref[0] & 0xff) << 8) + (pref[1] & 0xff);
  }

  cookie = pref + 2;

  sock = do_connect(LOOPBACK, port);
  if (sock < 0){
    fprintf(stderr, "connection failed\n");
    exit(-1);
  }

  from_pipe = stdin_available();

  do_write(sock, cookie, 8); /* send cookie */

  if (argc == 1 && command == NULL && from_pipe == NULL){
    int i = -1;
    do_write(sock, (char*)&i, 1);
    send_cwd(sock);
    send_quit(sock);
  } else {
    int i = 0;
    do_write(sock, (char*)&i, 1);
    send_cwd(sock);
  }

  if (command != NULL){
    do_write(sock, command, strlen(command));
    do_write(sock, "\n", 1);
  } else if (argc == 1){ /* read from stdin */
    read_stdin(sock); 
  } else {
    FILE *file;

    if (set_args(sock, argc, argv) < 0){
      /* ignore */
    }

    if ((file = fopen(argv[1], "r")) == NULL){
      fprintf(stderr, "open failed:%s\n", argv[i]);
      exit(-1);
    }
    read_input(file, sock);
    fclose(file);
  }

  send_quit(sock);
  receive_output(sock);
  close(sock);
  exit(0);
}
