/*
 * @(#) Function.java	1.23 00/07/26
 *
 * Copyright (c) 1997-2000 Sun Microsystems, Inc. All Rights Reserved.
 *
 * See the file "license.txt" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 */
package pnuts.lang;

import java.util.Vector;
import java.io.Serializable;
import java.io.StringReader;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import pnuts.lang.Runtime;
import pnuts.ext.UnparseVisitor;

/**
 * This class represents a function with a certain number of parameters.
 * In Pnuts, functions should be accessed through PnutsFunction.
 * This class is used mainly by compiler implementors.  Note that
 * there is no way to create a Function object through public API.
 *
 * @version 	1.23
 */
public class Function extends Runtime implements Serializable {

    static final long serialVersionUID = -493498203738404960L;

    /**
     * name of the function including the scope information
     *
     * @serial
     */
    protected String funcName;

    /**
     * name of the function without the scope information
     *
     * @serial
     */
    protected String name;

    /**
     * the number of arguments
     *
     * @serial
     */
    protected int nargs;

    /**
     * local parameters
     *
     * @serial
     */
    protected String locals[];

    /**
     * file name in which this function is defined
     *
     * @serial
     */
    protected Object file;

    /**
     * reference to the definition
     *
     */
    protected SimpleNode node;

    /**
     * @serial
     */
    SymbolTable lexicalScope;

    /**
     * "import" environment
     *
     * @serial
     */
    protected ImportEnv importEnv;

    /**
     * reference to the outer function
     *
     * @serial
     */
    protected Function outer;

    /**
     * the package name in which this function is defined
     *
     * @serial
     */
    protected String pkgName;

    /**
     * reference to PnutsFunction
     *
     * @serial
     */
    protected PnutsFunction function;

    transient protected Package pkg;

    protected Function(){
    }

    protected Function(String func, String[] locals, int nargs, SimpleNode node, Package pkg, Context context){
	if (context.frame != null){
	    outer = context.frame;
	    this.file = outer.file;
	    this.importEnv = outer.importEnv;
	} else {
	    if (context.isLoading()){
	        this.file = context.peekFile();
	    }
	    this.importEnv = context.importEnv;
	}
	Function of = outer;

	String scopeName = null;

	if (of != null){
	    scopeName = of.getName() + ".";
	}

	if (func == null){
	    func = "";
	}

	this.name = func;
	this.funcName = func;
	if (scopeName != null){
	    this.funcName = scopeName + func;
	}
	this.funcName = funcName.intern();
	this.nargs = nargs;
	this.locals = locals;
	this.node = node;
	this.pkg = pkg;
	this.pkgName = pkg.getName();

	StackFrame lex = context.stackFrame;
	if (lex != null){
	    SymbolTable tab = new SymbolTable();
	    lex.makeLexicalScope(tab);
	    this.lexicalScope = tab;
	}
    }

    public String getName(){
	return funcName;
    }

    public int getNumberOfParameter(){
	return nargs;
    }

    public Object getScriptSource(){
	return file;
    }

    protected Object exec(Object[] args, Context context){
	try {
	    context.open(locals);
	    for (int i = 0; i < args.length; i++){
		context.bind(locals[i], args[i]);
	    }
	    if (outer != null && name != null){
		context.bind(name, function);
	    }
	    return node.accept(PnutsInterpreter.instance, context);
	} catch (ThreadDeath td){
	    throw td;
	} catch (Throwable t){
	    Runtime.checkException(context, t);
	    return null;
	} finally {
	    context.close();
	}
    }

    public Package getPackage(){
	return pkg;
    }

    public String[] getImportEnv(){
	return importEnv.list();
    }

    String paramString(){
        String s = "";
	s += "(";
	if (nargs != 0){
	    s += locals[0];
	}
	if (nargs == -1){
	    s += "[]";
	} else {
	    for (int i = 1; i < nargs; i++){
		s += ",";
		s += locals[i];
	    }
	}
	s += ")";
	return s;
    }

    protected PnutsFunction register(PnutsFunction pf){
	if (pf == null){
	    if (name != null){
		pf = new PnutsFunction(funcName);
	    } else {
		pf = new PnutsFunction();
	    }
	    pf.pkg = pkg;
	}
	pf.put(nargs, this);
	return pf;
    }

    protected Object accept(Visitor visitor, Context context){
	return node.jjtGetParent().accept(visitor, context);
    }

    protected String unparse(Context context){
	return (String)accept(new UnparseVisitor(false), context);
    }

    public String toString(){
	String s = "function ";
	s += funcName;
	s += paramString();
	return s;
    }

    private synchronized void readObject(ObjectInputStream s) throws IOException, ClassNotFoundException {
	s.defaultReadObject();
	if (pkgName != null){
	    this.pkg = Package.getPackage(pkgName);
	} else {
	    this.pkg = new Package();
	}
	for (int i = 0; i < locals.length; i++){
	    locals[i] = locals[i].intern();
	}
    }
}
